/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.image.ImageObserver;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.SwingConstants;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.dtypes.editor.icons.IconEditor;
import at.lcm.bu21.general.dtypes.preview.TPeriodicPreview;
import at.lcm.bu21.general.plot.HOLD;
import at.lcm.bu21.general.plot.Plot2Line;
import at.lcm.bu21.general.plot.Plot2XY;
import at.lcm.bu21.general.plot.PlotException;
import at.lcm.bu21.general.utils.CloneException;

/**
 * TPeriodic Generator
 * 
 * @author Christoph Halbmayr
 */
class TPeriodicGenerator implements ActionListener, TableModelListener {

	/** Buttons Vector for Navigation Bar */
	private Vector<JButton> Buttons;

	/** 2D Overview Plot */
	private Plot2XY overviewPlot;

	/** preview plot */
	private Plot2XY plot;

	/** Navigation Ba */
	private JToolBar tb;

	/** JPanel for ToolBars */
	private final JPanel toolBarPanel;

	/** choosen line */
	int choosenline;

	/** console */
	JTextField console;

	/** TPeriodic */
	TPeriodic data = null;

	/** if down key pressed */
	boolean downPressed = false;

	/** GeneratorTable */
	TPeriodicGeneratorTable gt;

	/** TextField for Max Orders */
	JTextField maxOrdersField;

	/** Lable for Max Orders */
	JLabel maxOrdersLable;

	/** frame for displaying a signal overview */
	JFrame overview;

	/** TPeriodic Generator JPanel */
	JPanel panel;

	/** preview frame */
	JFrame preview;

	/** spectrum frame */
	JFrame spectrum;

	/** typed lines */
	Vector<String> typedLines;

	/** if up key pressed */
	boolean upPressed = false;

	/**
	 * Creats a new TPeriodic Generator
	 */
	TPeriodicGenerator() {
		this.typedLines = new Vector<String>();
		this.gt = new TPeriodicGeneratorTable();
		this.gt.addTML(this);
		this.panel = new JPanel();
		this.panel.setLayout(new BorderLayout());

		this.panel.repaint();
		this.panel.revalidate();
		this.console = new JTextField("0");
		this.maxOrdersField = new JTextField("20");

		final ActionListener alis = new ActionListener() {

			@Override
			public void actionPerformed(final ActionEvent e) {

				int maxOrder = 0;
				try {
					maxOrder = Integer.parseInt(TPeriodicGenerator.this.maxOrdersField.getText());
				} catch (final NumberFormatException ex) {
					return;
				}

				TPeriodicGenerator.this.gt.maxOrderChanged(maxOrder);

				TPeriodicGenerator.this.updateOverview();

				if (TPeriodicGenerator.this.console.getText().equals("") == false) {

					TPeriodicGenerator.this.consoleEnter();
					TPeriodicGenerator.this.updateDataPlots();
				}
			}

		};

		this.maxOrdersField.addActionListener(alis);

		this.maxOrdersLable = new JLabel("max Orders:");

		this.initKeyListener();

		this.toolBarPanel = new JPanel(new BorderLayout());
		this.panel.add(this.toolBarPanel, BorderLayout.NORTH);

		final JPanel consolePanel = new JPanel(new BorderLayout());

		consolePanel.add(this.console);

		this.maxOrdersLable.setHorizontalAlignment(SwingConstants.RIGHT);

		final JPanel pan = new JPanel(new BorderLayout());

		pan.add(this.maxOrdersLable, BorderLayout.WEST);
		pan.add(this.maxOrdersField, BorderLayout.EAST);

		consolePanel.add(pan, BorderLayout.EAST);

		this.createToolBar();
		this.toolBarPanel.add(consolePanel, BorderLayout.SOUTH);
		this.panel.add(this.gt.getScrollPane());
	}

	@Override
	public void actionPerformed(final ActionEvent e) {
		if (e.getActionCommand().equals("clear Signals")) {
			this.clear();
		}
		if (e.getActionCommand().equals("x/y plot")) {
			if (this.data != null) {
				this.preview();
			}
		}
		if (e.getActionCommand().equals("spectrum")) {
			if (this.data != null) {
				this.spectrum();
			}
		}

		if (e.getActionCommand().equals("clear Console")) {
			this.console.setText("");
		}

		if (e.getActionCommand().equals("overview")) {
			this.overview();
		}

	}

	/**
	 * Retruns the generated TPeriodic
	 * 
	 * @return TPeriodic
	 */
	public TData getData() {
		TPeriodic per = null;

		if (this.data != null) {
			per = this.data;
		} else if (this.gt.data.size() == 1) {
			per = this.gt.data.get(0);
		}

		return per;

	}

	/**
	 * Returns the GernatorPanel
	 * 
	 * @return panel
	 */
	public JPanel getPanel() {

		return this.panel;
	}

	@Override
	public void tableChanged(final TableModelEvent arg0) {
		int maxOrder = 20;
		try {
			maxOrder = Integer.parseInt(this.maxOrdersField.getText());
		} catch (final NumberFormatException e) {
			return;
		}

		this.gt.tableChanged(maxOrder);
		this.updateOverview();

		if (this.console.getText().equals("") == false) {

			this.consoleEnter();
			this.updateDataPlots();
		}

	}

	/**
	 * Clears the plot and the Table
	 */
	private void clear() {
		this.gt.clear();
		this.data = null;

		if (this.preview != null) {
			this.preview.setVisible(false);
		}

		if (this.overview != null) {
			this.overview.setVisible(false);
		}
		if (this.spectrum != null) {
			this.spectrum.setVisible(false);
		}

	}

	/**
	 * handles an console enter
	 */
	private void consoleEnter() {

		final String[] read = this.console.getText().split("\n");

		final String intput = read[read.length - 1];

		if (intput.equals("") == false) {
			this.choosenline++;
			this.typedLines.add(intput);
		}

		try {
			final int number = Integer.parseInt(intput.toString());
			this.data = this.gt.data.get(number);
		} catch (final Exception e) {

			try {
				this.data = this.sum(intput);
			} catch (final StackOverflowError e1) {
			}

		}

	}

	/**
	 * Creats the ToolBar
	 */

	private void createToolBar() {
		this.tb = new JToolBar();
		this.tb.setPreferredSize(new Dimension(ImageObserver.WIDTH, 25));
		this.Buttons = new Vector<JButton>();

		final JButton cs = new JButton("Signals", IconEditor.ICON_REMOVE);
		cs.setActionCommand("clear Signals");
		cs.setToolTipText("clear all signals");
		this.Buttons.add(cs);

		final JButton cc = new JButton("Console", IconEditor.ICON_REMOVE);
		cc.setActionCommand("clear Console");
		cc.setToolTipText("clear console text");
		this.Buttons.add(cc);

		final JButton prv = new JButton(IconEditor.ICON_XYPLOT);
		prv.setActionCommand("x/y plot");
		prv.setToolTipText("show x/y plot");
		this.Buttons.add(prv);

		final JButton spec = new JButton(IconEditor.ICON_SPECTRUM);
		spec.setActionCommand("spectrum");
		spec.setToolTipText("show spectrum plot");
		this.Buttons.add(spec);

		final JButton ove = new JButton("ALL", IconEditor.ICON_XYPLOT);
		ove.setActionCommand("overview");
		ove.setToolTipText("show all signals");
		this.Buttons.add(ove);

		for (final JButton i : this.Buttons) {
			this.tb.add(i);
			i.addActionListener(this);
		}
		this.tb.setVisible(true);
		this.toolBarPanel.add(this.tb, BorderLayout.NORTH);

	}

	/**
	 * initializes the key listner
	 */
	private void initKeyListener() {
		final KeyListener l = new KeyListener() {

			/**
			 * Key Press Key event
			 * 
			 * @param arg0
			 *            KeyEvent
			 */
			@Override
			public void keyPressed(final KeyEvent arg0) {
				final int keycode = arg0.getKeyCode();

				if (keycode == KeyEvent.VK_ENTER) {

					TPeriodicGenerator.this.consoleEnter();
					TPeriodicGenerator.this.upPressed = false;
					TPeriodicGenerator.this.downPressed = false;
					TPeriodicGenerator.this.updateOverview();
					TPeriodicGenerator.this.updateDataPlots();
				} else if (keycode == KeyEvent.VK_UP) {

					if (TPeriodicGenerator.this.choosenline > 1) {
						TPeriodicGenerator.this.choosenline--;
					}

					if (TPeriodicGenerator.this.upPressed == false) {
						TPeriodicGenerator.this.upPressed = true;
					}
					TPeriodicGenerator.this.downPressed = false;

					TPeriodicGenerator.this.removeConsoleLastLine();
					TPeriodicGenerator.this.console.setText(
							TPeriodicGenerator.this.typedLines.get(TPeriodicGenerator.this.choosenline - 1) + "\n");

				}

				else if (keycode == KeyEvent.VK_DOWN) {

					if (TPeriodicGenerator.this.choosenline < TPeriodicGenerator.this.typedLines.size()) {
						TPeriodicGenerator.this.choosenline++;
					}

					if (TPeriodicGenerator.this.downPressed == true) {
						TPeriodicGenerator.this.upPressed = false;
					} else if ((TPeriodicGenerator.this.downPressed == false)
							&& (TPeriodicGenerator.this.upPressed == false)) {

					}

					TPeriodicGenerator.this.removeConsoleLastLine();
					TPeriodicGenerator.this.downPressed = true;

					TPeriodicGenerator.this.console
							.setText(TPeriodicGenerator.this.typedLines.get(TPeriodicGenerator.this.choosenline - 1));
				}

				else if (keycode == KeyEvent.VK_BACK_SPACE) {

				}

			}

			/**
			 * Not in use
			 */
			@Override
			public void keyReleased(final KeyEvent arg0) {
			}

			/**
			 * Not in use
			 */
			@Override
			public void keyTyped(final KeyEvent arg0) {

			}

		};
		this.console.addKeyListener(l);
	}

	/**
	 * plots an 2D overview
	 */
	private void overview() {
		if (this.gt.data.size() > 0) {
			this.overview = new JFrame("ALL Signals");
			this.overviewPlot = new Plot2Line();
			this.overviewPlot.setGrid(true);
			this.overviewPlot.setLegend(true);
			int nmbr = 0;
			for (final TPeriodic s : this.gt.data) {
				try {
					this.overviewPlot.addSignal("" + nmbr, s, 500);
				} catch (PlotException e) {
				}
				nmbr++;
			}

			this.overview.add(this.overviewPlot.getPanel());
			this.overview.setSize(800, 600);
			this.overview.setVisible(true);
			this.overview.setIconImage(IconEditor.ICON_XYPLOT.getImage());
		}
	}

	/**
	 * opens a preview window (TNumeric & TPeriodic)
	 */
	private void preview() {

		this.preview = new JFrame(this.data.toString());
		this.plot = new Plot2Line();
		this.plot.setGrid(true);
		try {
			this.plot.addSignal(this.data.toString(), this.data, 500);
		} catch (PlotException e) {
		}

		this.preview.add(this.plot.getPanel());
		this.preview.setSize(800, 600);
		this.preview.setVisible(true);
		this.preview.setIconImage(IconEditor.ICON_XYPLOT.getImage());

	}

	/**
	 * @param str
	 *            intput String
	 * @return evaluated TPeriodic
	 */
	private TPeriodic prod(final String str) {
		final Vector<String> vec = TPeriodicGeneratorParser.splitProduct(TPeriodicGeneratorParser.removeBrackets(str));
		TPeriodic prod = null;
		TPeriodic per1 = null;
		TPeriodic per2 = null;

		if ((vec.size() >= 3) && (vec.get(0).equals("*") == false) && (vec.get(0).equals("/") == false)) {
			try {
				final int index = Integer.parseInt(vec.get(0));
				per1 = this.gt.data.get(index);
			} catch (final Exception e) {

				per1 = this.sum(vec.get(0));

			}

			try {
				final int index = Integer.parseInt(vec.get(2));
				per2 = this.gt.data.get(index);
			} catch (final Exception e) {
				per2 = this.sum(vec.get(2));

				per2 = this.sum(vec.get(2));

			}

			if ((per1 != null) && (per2 != null)) {

				final String sign = vec.get(1);

				if (sign.equals("*")) {
					prod = per1.multiply(per2);
				} else if (sign.equals("/")) {
					try {
						prod = per1.dividePointWise(per2);
					} catch (final CloneException e) {
					}
				}

			} else {
				prod = this.sum(str);
			}

		} else {
			prod = this.sum(str);
		}
		return prod;

	}

	/**
	 * removes the last line from the console
	 */
	private void removeConsoleLastLine() {
		final String s = this.console.getText();

		final Vector<String> lines = new Vector<String>();

		String line = new String();

		for (int i = 0; i < s.toCharArray().length; i++) {
			if (s.toCharArray()[i] == '\n') {
				lines.add(line);
				line = "";
			} else {
				line = line + "" + s.toCharArray()[i];
			}
		}

		this.console.setText("");

		for (int i = 0; i < (lines.size() - 1); i++) {
			this.console.setText(lines.get(i));
		}

	}

	/**
	 * opens a spectrum window (TNumeric & TPeriodic)
	 */
	private void spectrum() {
		this.spectrum = new JFrame(this.data.toString());
		this.spectrum.setIconImage(IconEditor.ICON_SPECTRUM.getImage());
		this.spectrum.add(TPeriodicPreview.previewSpectrum(this.data, null).getPanel());
		this.spectrum.setSize(800, 600);
		this.spectrum.setVisible(true);
	}

	/**
	 * @param str
	 *            intput String
	 * @return evaluated TPeriodic
	 */
	private TPeriodic sum(final String str) {
		final Vector<String> vec = TPeriodicGeneratorParser.splitSum(TPeriodicGeneratorParser.removeBrackets(str));
		TPeriodic sum = null;
		TPeriodic per1 = null;
		TPeriodic per2 = null;

		if ((vec.size() >= 3) && (vec.get(0).equals("+") == false) && (vec.get(0).equals("-") == false)) {

			try {
				final int index = Integer.parseInt(vec.get(0));
				per1 = this.gt.data.get(index);
			} catch (final Exception e) {

				per1 = this.sum(vec.get(0));

			}

			try {
				final int index = Integer.parseInt(vec.get(2));
				per2 = this.gt.data.get(index);
			} catch (final Exception e) {

				per2 = this.sum(vec.get(2));

			}

			if ((per1 != null) && (per2 != null)) {

				final String sign = vec.get(1);

				if (sign.equals("+")) {
					sum = per1.add(per2);
				} else if (sign.equals("-")) {
					try {
						sum = per1.subtract(per2);
					} catch (final CloneException e) {
					}
				}

			} else {
				sum = this.prod(str);
			}

		} else {
			sum = this.prod(str);
		}

		return sum;

	}

	/**
	 * updates/redraw all open plot windows
	 */
	private void updateDataPlots() {
		if ((this.spectrum != null) && this.spectrum.isVisible()) {

			this.spectrum.getContentPane().removeAll();
			this.spectrum.setLayout(new BorderLayout());
			this.spectrum.add(TPeriodicPreview.previewSpectrum(this.data, null).getPanel(), BorderLayout.CENTER);
			this.plot.getPanel().revalidate();
			this.spectrum.repaint();

		}
		if ((this.preview != null) && this.preview.isVisible()) {

			this.plot = new Plot2Line();
			this.plot.setGrid(true);
			try {
				this.plot.addSignal(this.data.toString(), this.data, 500);
			} catch (PlotException e) {
			}

			this.preview.getContentPane().removeAll();
			this.preview.add(this.plot.getPanel());

			this.preview.setLayout(new BorderLayout());
			this.preview.add(this.plot.getPanel(), BorderLayout.CENTER);
			this.plot.getPanel().revalidate();
			this.preview.repaint();

		}
	}

	/**
	 * Updates preview and spectrum plot
	 */
	private void updateOverview() {

		if ((this.overview != null) && this.overview.isVisible()) {

			this.overviewPlot = new Plot2Line();
			this.overviewPlot.setGrid(true);
			this.overviewPlot.setLegend(true);
			this.overviewPlot.setHold(HOLD.ALL);

			int nmbr = 0;
			for (final TPeriodic s : this.gt.data) {
				try {
					this.overviewPlot.addSignal("" + nmbr, s, 500);
				} catch (PlotException e) {
				}
				nmbr++;
			}

			this.overview.getContentPane().removeAll();
			this.overview.setLayout(new BorderLayout());
			this.overview.add(this.overviewPlot.getPanel(), BorderLayout.CENTER);
			this.overviewPlot.getPanel().revalidate();
			this.overview.repaint();
		}

	}

}
