/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.utils;

import java.math.BigDecimal;

/**
 * Additional mathematics operations
 * 
 * @author Werner Koppelsttter
 */
public class MathUtils {

	/**
	 * Round limits of plot
	 * 
	 * @param values
	 *            Limits to round
	 * @return Round limits
	 */
	public static double[] roundLimits(double[] values) {
		double diff, significant, lim;
		int base = 0;
		double[] limits;

		limits = values.clone();

		if (limits[0] > limits[1]) {
			diff = limits[1];
			limits[1] = limits[0];
			limits[0] = diff;
		}

		diff = limits[1] - limits[0];
		if (diff != 0.0) {
			base = (int) Math.floor(Math.log10(diff));

			significant = MathUtils.shiftRight(limits[0], base);
			lim = Math.floor(significant);
			if ((significant - lim) > (1.0 - 1e-10)) {
				lim += 1.0;
			}
			limits[0] = MathUtils.shiftLeft(lim, base);

			significant = MathUtils.shiftRight(limits[1], base);
			lim = Math.ceil(significant);
			if ((significant - lim) < -(1.0 - 1e-10)) {
				lim -= 1.0;
			}
			limits[1] = MathUtils.shiftLeft(lim, base);
		}

		return limits;
	}

	/**
	 * Shift number to right
	 * 
	 * @param value
	 *            Value to shift
	 * @param digits
	 *            Number of digits to shift
	 * @return Shifted number
	 */
	public static double shiftRight(double value, int digits) {
		return MathUtils.shiftLeft(value, -digits);
	}

	/**
	 * Shift number to left
	 * 
	 * @param value
	 *            Value to shift
	 * @param digits
	 *            Number of Digits to shift
	 * @return Shifted number
	 */
	public static double shiftLeft(double value, int digits) {
		if (!Double.isNaN(value) && !Double.isInfinite(value)) {
			value = new BigDecimal(value).movePointLeft(-digits).doubleValue();
		}

		return value;
	}

	/**
	 * Get maximum of array
	 * 
	 * @param array
	 *            Values to check
	 * @return Maximum
	 */
	public static double max(double[] array) {
		double max = Double.NEGATIVE_INFINITY;

		for (double d : array) {
			if (d > max) {
				max = d;
			}
		}

		return max;
	}

	/**
	 * Get minimum of array
	 * 
	 * @param array
	 *            Values to check
	 * @return Minimum
	 */
	public static double min(double[] array) {
		double min = Double.POSITIVE_INFINITY;

		for (double d : array) {
			if (d < min) {
				min = d;
			}
		}

		return min;
	}

	/**
	 * Get maximum of array
	 * 
	 * @param array
	 *            Values to check
	 * @return Maximum
	 */
	public static float max(float[] array) {
		float max = Float.NEGATIVE_INFINITY;

		for (float d : array) {
			if (d > max) {
				max = d;
			}
		}

		return max;
	}

	/**
	 * Get minimum of array
	 * 
	 * @param array
	 *            Values to check
	 * @return Minimum
	 */
	public static float min(float[] array) {
		float min = Float.POSITIVE_INFINITY;

		for (float d : array) {
			if (d < min) {
				min = d;
			}
		}

		return min;
	}

	/**
	 * Get maximum of array
	 * 
	 * @param array
	 *            Values to check
	 * @return Maximum
	 */
	public static long max(long[] array) {
		long max = Long.MIN_VALUE;

		for (long d : array) {
			if (d > max) {
				max = d;
			}
		}

		return max;
	}

	/**
	 * Get minimum of array
	 * 
	 * @param array
	 *            Values to check
	 * @return Minimum
	 */
	public static long min(long[] array) {
		long min = Long.MAX_VALUE;

		for (long d : array) {
			if (d < min) {
				min = d;
			}
		}

		return min;
	}

	/**
	 * Get maximum of array
	 * 
	 * @param array
	 *            Values to check
	 * @return Maximum
	 */
	public static int max(int[] array) {
		int max = Integer.MIN_VALUE;

		for (int d : array) {
			if (d > max) {
				max = d;
			}
		}

		return max;
	}

	/**
	 * Get minimum of array
	 * 
	 * @param array
	 *            Values to check
	 * @return Minimum
	 */
	public static int min(int[] array) {
		int min = Integer.MAX_VALUE;

		for (int d : array) {
			if (d < min) {
				min = d;
			}
		}

		return min;
	}

}
