from lxml import etree


class Fmi2UnitDefinitions:
    """
    Class for the UnitDefinitions xml-tag (Optional)

    XML Schema Description
    ----------------------
    Unit definition (with respect to SI base units) and default display units
    """
    def __init__(self, root, physical_quantities):
        """
        Parameters
        ----------
        root : etree.Element
                xml root element
        physical_quantities : list
                    list of quantities used in the FMU
        """
        self.__element = etree.SubElement(root, 'UnitDefinitions')
        for quantity in physical_quantities:
            quantity = quantity.lower()
            if quantity == "torque":
                self.__add_torque_unit()
            elif quantity == "angle":
                self.__add_angle_unit()
            elif quantity == "angular_velocity":
                self.__add_angular_velocity_unit()
            elif quantity == "current":
                self.__add_current_unit()
            elif quantity == "voltage":
                self.__add_voltage_unit()
            elif quantity == "flux":
                self.__add_flux_unit()
            elif quantity == "power":
                self.__add_power_unit()
            elif quantity == "resistance":
                self.__add_resistance_unit()
            else:
                raise Exception("Units for the quantity: " + quantity + " are not implemented yet!")

    def __add_torque_unit(self):
        unit = etree.SubElement(self.__element, 'Unit', name="N.m")
        etree.SubElement(unit, 'BaseUnit', kg="1", m="2", s="-1", A="0", K="0", mol="0", cd="0", rad="-1", factor="1",
                         offset="0")

    def __add_angle_unit(self):
        unit = etree.SubElement(self.__element, 'Unit', name="rad")
        etree.SubElement(unit, 'BaseUnit', kg="0", m="0", s="0", A="0", K="0", mol="0", cd="0", rad="1", factor="1",
                         offset="0")

    def __add_angular_velocity_unit(self):
        unit = etree.SubElement(self.__element, 'Unit', name="rad/s")
        etree.SubElement(unit, 'BaseUnit', kg="0", m="0", s="-1", A="0", K="0", mol="0", cd="0", rad="1", factor="1",
                         offset="0")

    def __add_current_unit(self):
        unit = etree.SubElement(self.__element, 'Unit', name="A")
        etree.SubElement(unit, 'BaseUnit', kg="0", m="0", s="0", A="1", K="0", mol="0", cd="0", rad="0", factor="1",
                         offset="0")

    def __add_voltage_unit(self):
        unit = etree.SubElement(self.__element, 'Unit', name="V")
        etree.SubElement(unit, 'BaseUnit', kg="1", m="2", s="-3", A="-1", K="0", mol="0", cd="0", rad="0", factor="1",
                         offset="0")
        etree.SubElement(unit, 'DisplayUnit', name="Wb/s")

    def __add_flux_unit(self):
        unit = etree.SubElement(self.__element, 'Unit', name="Wb")
        etree.SubElement(unit, 'BaseUnit', kg="1", m="2", s="-2", A="-1", K="0", mol="0", cd="0", rad="0", factor="1",
                         offset="0")

    def __add_power_unit(self):
        unit = etree.SubElement(self.__element, 'Unit', name="W")
        etree.SubElement(unit, 'BaseUnit', kg="1", m="2", s="-3", A="0", K="0", mol="0", cd="0", rad="0", factor="1",
                         offset="0")

    def __add_resistance_unit(self):
        unit = etree.SubElement(self.__element, 'Unit', name="Ohm")
        etree.SubElement(unit, 'BaseUnit', kg="1", m="2", s="-3", A="-2", K="0", mol="0", cd="0", rad="0", factor="1",
                         offset="0")

