"""
Created on 11.10.2016

@author: gweidenholzer

Radial basis function interpolations
ref. https://en.wikipedia.org/wiki/Polyharmonic_spline

"""

import numpy as np
from lcm.utils.pytperiodic import PyTPeriodicMatrix


def calc_weigths(xc, t, zc, kernel, gamma, max_cond):
    """

    Parameters
    ----------
    xc : (xdim, ncenters, nsamples) ndarray
        RBF center coordinates
    t : (x_dim, x_dim, n_samples) ndarray of float
        Transformation matrix for input x
    zc : (z_dim, n_centers, nsamples) ndarray
        RBF center values.
    kernel : lcm.rbf.kernel.AbstractKernel
        Function to calculate the RBF kernel.
        It expects an array_like argument (n_centers, n_samples) of radii
        and returns the RBF values in the same shape
    gamma : float
        smoothing parameter
    max_cond : float
        maximum condition number

    Returns
    -------
    xc_t : (xdim, ncenters, nsamples) ndarray
        RBF center coordinates in normalized basis
    c : (zdim, ncenters, nsamples) ndarray
        RBF weights
    d : (zdim, xdim+1, nsamples) ndarray
        polynomial coefficients
    """

    ''' get shape '''
    x_dim, n_centers, n_samples = xc.shape
    y_dim = zc.shape[0]

    ''' transform xc
            --> xc_t.shape = (x_dim, n_centers, n_samples) '''
    xc_t = np.matmul(t.transpose(2, 0, 1), xc.transpose(2, 0, 1)).transpose(1, 2, 0)

    ''' calculate matrix of center distances by double broadcasting/transposition
        --> dist.shape = (x_dim, n_centers, n_centers, n_samples)'''
    dist = xc_t[:, :, np.newaxis, :] - xc_t[:, np.newaxis, :, :]

    ''' calculate matrix of radii using L2 norm
        --> r.shape = (n_centers, n_centers, n_samples)'''
    r = np.linalg.norm(dist, axis=0)

    ''' get rbf matrix A '''
    a_rbf = kernel.eval(r)

    ''' fill in smoothing parameter gamma along diagonal'''
    smooth = np.eye(n_centers, n_centers) * gamma
    a_rbf += np.repeat(smooth[:, :, np.newaxis], n_samples, axis=2)

    ''' create polynomial submatrix P '''
    p = np.concatenate((np.ones((n_centers, 1, n_samples)), xc_t.transpose(1, 0, 2)), axis=1)

    ''' build RBF matrix a=[A,P; P', zeros] '''
    a1 = np.concatenate((a_rbf, p), axis=1)
    a2 = np.concatenate((p.transpose(1, 0, 2), np.zeros((x_dim + 1, x_dim + 1, n_samples))), axis=1)
    a = np.concatenate((a1, a2), axis=0)

    ''' build right hand side'''
    b = np.concatenate((zc, np.zeros((y_dim, x_dim + 1, n_samples))), axis=1)

    ''' check condition of matrix '''
    if max_cond > 0.:
        cond = np.max(np.linalg.cond(a.transpose(2, 1, 0)))

        if cond > max_cond:
            raise ArithmeticError(
                "RBF matrix is ill conditioned! (cond={0:7e}) Consider preconditioning!".format(cond))

    ''' solve RBF problem '''
    x = np.linalg.solve(a.transpose(2, 1, 0), b.transpose(2, 1, 0))
    x = x.transpose(2, 1, 0)

    ''' extract weights '''

    c = x[:, :n_centers, :]
    d = x[:, n_centers:, :]

    ''' change d to act on un-normalized x rather on x_t'''
    d[:, 1:, :] = np.matmul(d[:, 1:, :].transpose(2, 0, 1), t.transpose(2, 0, 1)).transpose(1, 2, 0)

    return xc_t, c, d


def eval_z(xc_t, c, d, t, x, kernel, max_dist):
    """
    Helper function: evaluates the RBF model

    Parameters
    ----------
    xc_t : (x_dim, n_centers, n_samples) ndarray of float
        RBF center coordinates in preconditioned basis
    c : (z_dim, n_centers, n_samples) ndarray of float
        RBF weigths.
    d : (z_dim, x_dim+1, n_samples) ndarray of float
        Polynomial coefficients.
    t : (x_dim, x_dim, n_samples) ndarray of float
        Transformation matrix for input x
    x : (x_dim, n_interpolations, n_samples) ndarray of float
        periodic signals representing the coordinates where to interpolate
    kernel : lcm.rbf.kernel.AbstractKernel
        Function to calculate the RBF kernel.
        It expects an array_like argument (n_centers, n_samples) of radii
        and returns the RBF values in the same shape
    max_dist : float
        Maximum distance to closest center

    Returns
    -------
    z : (z_dim, n_interpolations, n_samples) ndarray
        result array

    """

    ''' get dimensions '''
    x_dim, n_centers, n_samples = xc_t.shape
    if x.ndim != 3:
        raise ValueError("x must be of dimension 3")
    # if x.shape[0] != xc_t.shape[0]:
    #     raise ValueError("size of dimension 0 does not match")
    if x.shape[2] != xc_t.shape[2]:
        raise ValueError("size of dimension 1 does not match")

    n_interpolations = x.shape[1]

    ''' transform x
            --> x_t.shape = (x_dim, n_interpolations, n_samples) '''
    x_t = np.matmul(t.transpose(2, 0, 1), x.transpose(2, 0, 1)).transpose(1, 2, 0)

    '''
     x_t.shape = (x_dim, n_interpolations, n_samples)
     xc_t.shape = (x_dim, n_centers, n_samples)
     dist.shape = (x_dim, n_centers,  n_interpolations, n_samples)
    '''
    dist = x_t[:, np.newaxis, :, :] - xc_t[:, :, np.newaxis, :]

    ''' calculate radii using L2 norm
        --> r.shape = (n_centers, n_interpolations, n_samples)'''
    r = np.linalg.norm(dist, axis=0)

    ''' check distance threshold'''
    if max_dist and np.min(r) > max_dist:
        raise ValueError("x is located outside of confidence region. {:.2f} > {:.2f}".format(np.min(r), max_dist))

    ''' calculate rbf values
        -->rbf_values.shape = (n_centers, n_interpolations, n_samples)'''
    rbf_values = kernel.eval(r)

    ''' calculate weighted rbf values
        c.shape = (z_dim, n_centers, n_samples)
        rbf_values.shape = (n_centers, n_interpolations, n_samples)
        --> z.shape = (z_dim, n_interpolations, n_samples)'''
    z = np.matmul(c.transpose(2, 0, 1), rbf_values.transpose(2, 0, 1)).transpose(1, 2, 0)

    '''calculate polynomial part and add them to result:
        x_t.shape = (x_dim, n_interpolations, n_samples)
        --> x_ext.shape=(x_dim + 1, n_interpolations, n_samples)'''
    x_ext = np.concatenate((np.ones((1, n_interpolations, n_samples)), x), axis=0)

    '''
    x_ext.shape=(x_dim + 1, n_interpolations, n_samples)
    d.shape=(z_dim, x_dim+1, n_samples)
    --> z.shape = (z_dim, n_interpolations, n_samples)
    '''
    z += np.matmul(d.transpose(2, 0, 1), x_ext.transpose(2, 0, 1)).transpose(1, 2, 0)

    return z


def eval_dz_dx(xc_t, c, d, t, x, kernel, max_dist):
    """
    Helper function: evaluates the jacobian matrix

    Parameters
    ----------
    xc_t : (x_dim, n_centers, n_samples) ndarray of float
        RBF center coordinates in preconditioned basis
    c : (z_dim, n_centers, n_samples) ndarray of float
        RBF weigths.
    d : (z_dim, x_dim+1, n_samples) ndarray of float
        Polynomial coefficients.
    t : (x_dim, x_dim, n_samples) ndarray of float
        Transformation matrix for input x
    x : (x_dim, n_interpolations, n_samples) ndarray of float
        periodic signals representing the coordinates where to interpolate
    kernel : lcm.rbf.kernel.AbstractKernel
        Function to calculate the RBF kernel.
        It expects an array_like argument (n_centers, n_samples) of radii
        and returns the RBF values in the same shape
    max_dist : float
        Maximum distance to closest center

    Returns
    -------
    z : (z_dim, x_dim, n_interpolations, n_samples) ndarray
        result array

    """

    ''' get dimensions '''
    _, n_centers, n_samples = xc_t.shape
    x_dim = t.shape[1]

    if x.ndim != 3:
        raise ValueError("x must be of dimension 3")
    # if x.shape[0] != xc_t.shape[0]:
    #     raise ValueError("size of dimension 0 does not match")
    # if x.shape[2] != xc_t.shape[2]:
    #     raise ValueError("size of dimension 1 does not match")

    n_interpolations = x.shape[1]

    ''' transform x
            --> x_t.shape = (x_dim, n_interpolations, n_samples) '''
    x_t = np.matmul(t.transpose(2, 0, 1), x.transpose(2, 0, 1)).transpose(1, 2, 0)

    '''
     x_t.shape = (x_dim, n_interpolations, n_samples)
     xc_t.shape = (x_dim, n_centers, n_samples)
     dist.shape = (x_dim, n_centers,  n_interpolations, n_samples)
    '''
    dist = x_t[:, np.newaxis, :, :] - xc_t[:, :, np.newaxis, :]

    ''' calculate radii using L2 norm
        --> r.shape = (n_centers, n_interpolations, n_samples)'''
    r = np.linalg.norm(dist, axis=0)

    ''' check distance threshold'''
    if max_dist and np.min(r) > max_dist:
        raise ValueError("x is located outside of confidence region")

    ''' calculate rbf derivative values
        -->dphi_dr.shape = (n_centers, n_interpolations ,n_samples)'''
    dphi_dr = kernel.dkernel_dradius_dividedby_radius(r)

    '''
        t.shape=(x_dim, x_dim, n_samples)
        dist.shape=(x_dim, n_centers,  n_interpolations, n_samples)
        --> t_dist.shape=(x_dim, n_centers, n_interpolations, n_samples) '''
    t_dist = np.matmul(t[:, :, np.newaxis, :].transpose(2, 3, 1, 0), dist.transpose(2, 3, 0, 1)).transpose(2, 3, 0, 1)

    '''
            t_dist.shape=(x_dim, n_centers, n_interpolations, n_samples)
            dphi_dr.shape = (n_centers, n_interpolations ,n_samples)
            drbf_dx.shape = (n_centers, x_dim,  n_interpolations ,n_samples)'''
    # drbf_dx = np.matmul(t_dist.transpose(2, 3, 0, 1),
    #                     dphi_dr[:, np.newaxis, :, :].transpose(2, 3, 0, 1)).transpose(2, 3, 0, 1)
    # drbf_dx = np.einsum('xcis,cis->cxis', t_dist, dphi_dr)
    drbf_dx = (t_dist * dphi_dr).transpose(1, 0, 2, 3)
    '''
        drbf_dx.shape = (n_centers, x_dim,  n_interpolations ,n_samples)
        c.shape = (z_dim, n_centers, n_samples)
        jac.shape=(z_dim, x_dim, n_interpolations, n_samples) '''
    jac = np.matmul(c[:, :, np.newaxis, :].transpose(2, 3, 0, 1), drbf_dx.transpose(2, 3, 0, 1)).transpose(2, 3, 0, 1)

    '''calculate polynomial part and add them to result:
        t.shape=(x_dim, x_dim, n_samples)
        --> x_ext.shape=(x_dim + 1, x_dim , n_interpolations, n_samples)'''

    # row1 = np.zeros((1, x_dim, n_interpolations, n_samples))
    # row2 = np.eye(x_dim, x_dim)[:, :, np.newaxis, np.newaxis]
    # # row2 = np.resize(row2, (x_dim, x_dim, n_interpolations, n_samples))
    # row2 = np.repeat(row2, n_interpolations, axis=2)
    # row2 = np.repeat(row2, n_samples, axis=3)
    # x_pre_ext = np.concatenate((row1, row2), axis=0)

    zeros = np.zeros((1, x_dim, n_interpolations, n_samples))
    eye = np.broadcast_to(np.eye(x_dim, x_dim)[:, :, np.newaxis, np.newaxis],
                          (x_dim, x_dim, n_interpolations, n_samples))
    x_pre_ext = np.concatenate((zeros, eye), axis=0)

    # x_pre_ext = np.concatenate( (np.zeros((1, x_dim, n_interpolations, n_samples)), np.eye(x_dim, x_dim)[:, :,
    # np.newaxis, np.newaxis]), axis=0)

    '''
        d.shape=(z_dim, x_dim+1, n_samples)
        jac.shape=(z_dim, x_dim, n_interpolations, n_samples) '''
    jac += np.matmul(d[:, :, np.newaxis, :].transpose(2, 3, 0, 1), x_pre_ext.transpose(2, 3, 0, 1)).transpose(2, 3, 0,
                                                                                                              1)
    return jac


def eval_dz_dphi(xc_t, c, d, t, x, kernel, max_dist, omega):
    """
    Helper function: evaluates the RBF model

    Parameters
    ----------
    xc_t : (x_dim, n_centers, n_samples) ndarray of float
        RBF center coordinates in preconditioned basis
    c : (z_dim, n_centers, n_samples) ndarray of float
        RBF weigths.
    d : (z_dim, x_dim+1, n_samples) ndarray of float
        Polynomial coefficients.
    t : (x_dim, x_dim, n_samples) ndarray of float
        Transformation matrix for input x
    x : (x_dim, n_interpolations, n_samples) ndarray of float
        periodic signals representing the coordinates where to interpolate
    kernel : lcm.rbf.kernel.AbstractKernel
        Function to calculate the RBF kernel.
        It expects an array_like argument (n_centers, n_samples) of radii
        and returns the RBF values in the same shape
    max_dist : float
        Maximum distance to closest center

    Returns
    -------
    z : (z_dim, n_interpolations, n_samples) ndarray
        result array

    """

    ''' get dimensions '''
    x_dim, n_centers, n_samples = xc_t.shape
    if x.ndim != 3:
        raise ValueError("x must be of dimension 3")
    if x.shape[0] != xc_t.shape[0]:
        raise ValueError("size of dimension 0 does not match")
    if x.shape[2] != xc_t.shape[2]:
        raise ValueError("size of dimension 1 does not match")

    n_interpolations = x.shape[1]

    t_per = PyTPeriodicMatrix(t, mode="samples", nopack=True)
    c_per = PyTPeriodicMatrix(c, mode="samples", nopack=True)
    d_per = PyTPeriodicMatrix(d, mode="samples", nopack=True)
    xc_t_per = PyTPeriodicMatrix(xc_t, mode="samples", nopack=True)

    dt_dphi_per = t_per.deriv(omega)
    dc_dphi_per = c_per.deriv(omega)
    dd_dphi_per = d_per.deriv(omega)
    dxc_t_dphi_per = xc_t_per.deriv(omega)

    dt_dphi = dt_dphi_per.get_samples(n_samples)
    dc_dphi = dc_dphi_per.get_samples(n_samples)
    dd_dphi = dd_dphi_per.get_samples(n_samples)
    dxc_t_dphi = dxc_t_dphi_per.get_samples(n_samples)

    # x_per = PyTPeriodicMatrix(x, mode="samples", nopack=True)
    # dx_dphi_per = x_per.deriv(omega)
    # dx_dphi = dx_dphi_per.get_samples(n_samples)

    # x_t_per = PyTPeriodicMatrix(x_t, mode="samples", nopack=True)
    # dx_t_dphi_per = x_t_per.deriv(omega)
    # dx_t_dphi = dx_t_dphi_per.get_samples(n_samples)

    ''' transform x
            --> x_t.shape = (x_dim, n_interpolations, n_samples) '''
    x_t = np.matmul(t.transpose(2, 0, 1), x.transpose(2, 0, 1)).transpose(1, 2, 0)

    dx_t_dphi = np.matmul(dt_dphi.transpose(2, 0, 1), x.transpose(2, 0, 1)).transpose(1, 2, 0)

    '''
     x_t.shape = (x_dim, n_interpolations, n_samples)
     xc_t.shape = (x_dim, n_centers, n_samples)
     dist.shape = (x_dim, n_centers,  n_interpolations, n_samples)
    '''
    dist = x_t[:, np.newaxis, :, :] - xc_t[:, :, np.newaxis, :]

    ''' calculate radii using L2 norm
        --> r.shape = (n_centers, n_interpolations, n_samples)'''
    r = np.linalg.norm(dist, axis=0)

    ''' check distance treshold'''
    if max_dist and np.min(r) > max_dist:
        raise ValueError("x is located outside of confidence region")

    ''' calculate rbf values
        -->rbf_values.shape = (n_centers, n_interpolations, n_samples)'''
    rbf_values = kernel.eval(r)
    dPhi_dr_r = kernel.dkernel_dradius_dividedby_radius(r)
    ddist_dphi = dx_t_dphi - dxc_t_dphi
    dist = x_t - xc_t
    term = np.matmul(ddist_dphi.transpose(2, 1, 0), dist.transpose(2, 0, 1)).transpose(1, 2, 0)
    drbf_dphi_values = np.matmul(term.transpose(2, 0, 1), dPhi_dr_r.transpose(2, 0, 1)).transpose(1, 2, 0)

    ''' calculate weighted rbf values
        c.shape = (z_dim, n_centers, n_samples)
        rbf_values.shape = (n_centers, n_interpolations, n_samples)
        --> z.shape = (z_dim, n_interpolations, n_samples)'''
    z = np.matmul(dc_dphi.transpose(2, 0, 1), rbf_values.transpose(2, 0, 1)).transpose(1, 2, 0)
    z += np.matmul(c.transpose(2, 0, 1), drbf_dphi_values.transpose(2, 0, 1)).transpose(1, 2, 0)

    '''calculate polynomial part and add them to result:
        x_t.shape = (x_dim, n_interpolations, n_samples)
        --> x_ext.shape=(x_dim + 1, n_interpolations, n_samples)'''
    x_pre_ext = np.concatenate((np.ones((1, n_interpolations, n_samples)), x_t), axis=0)
    dx_pre_ext_dphi = np.concatenate((np.zeros((1, n_interpolations, n_samples)), dx_t_dphi), axis=0)

    '''
    x_ext.shape=(x_dim + 1, n_interpolations, n_samples)
    d.shape=(z_dim, x_dim+1, n_samples)
    --> z.shape = (z_dim, n_interpolations, n_samples)
    '''
    z += np.matmul(dd_dphi.transpose(2, 0, 1), x_pre_ext.transpose(2, 0, 1)).transpose(1, 2, 0)
    z += np.matmul(d.transpose(2, 0, 1), dx_pre_ext_dphi.transpose(2, 0, 1)).transpose(1, 2, 0)

    return z


def eval_r(xc_t, t, x):
    """
    Helper function: evaluates the RBF model

    Parameters
    ----------
    xc_t : (x_dim, n_centers, n_samples) ndarray of float
        RBF center coordinates in preconditioned basis
    c : (z_dim, n_centers, n_samples) ndarray of float
        RBF weigths.
    d : (z_dim, x_dim+1, n_samples) ndarray of float
        Polynomial coefficients.
    t : (x_dim, x_dim, n_samples) ndarray of float
        Transformation matrix for input x
    x : (x_dim, n_interpolations, n_samples) ndarray of float
        periodic signals representing the coordinates where to interpolate
    kernel : lcm.rbf.kernel.AbstractKernel
        Function to calculate the RBF kernel.
        It expects an array_like argument (n_centers, n_samples) of radii
        and returns the RBF values in the same shape
    max_dist : float
        Maximum distance to closest center

    Returns
    -------
    z : (z_dim, n_interpolations, n_samples) ndarray
        result array

    """

    ''' get dimensions '''
    if x.ndim != 3:
        raise ValueError("x must be of dimension 3")

    if x.shape[2] != xc_t.shape[2]:
        raise ValueError("size of dimension 1 does not match")

    ''' transform x
            --> x_t.shape = (x_dim, n_interpolations, n_samples) '''
    x_t = np.matmul(t.transpose(2, 0, 1), x.transpose(2, 0, 1)).transpose(1, 2, 0)

    '''
     x_t.shape = (x_dim, n_interpolations, n_samples)
     xc_t.shape = (x_dim, n_centers, n_samples)
     dist.shape = (x_dim, n_centers,  n_interpolations, n_samples)
    '''
    dist = x_t[:, np.newaxis, :, :] - xc_t[:, :, np.newaxis, :]

    ''' calculate radii using L2 norm
        --> r.shape = (n_centers, n_interpolations, n_samples)'''
    r = np.linalg.norm(dist, axis=0)

    return np.min(r)
