#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 977 $
$LastChangedDate:: 2019-12-11 15:24:14 +0100#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import copy
import numpy as np
from lcm.utils.pytperiodic import PyTPeriodicMatrix
from lcm.math.rbf.rbf import RBF
from lcm.math.rbf import kernels
from lcm.math.rbf.core import calc_weigths, eval_z, eval_dz_dx, eval_dz_dphi, eval_r
from lcm.utils.serialization import Serializable


class PeriodicMultiRBF:
    """
    Solves a stack of independent, multidimensional rbf problems and
    provide an interpolation function, therefor

    Attributes
    ----------
    kernel : lcm.rbf.kernel.AbstractKernel
        Instance of type AbstractKernel to calculate the RBF kernel.
    xc_t_per : (x_dim, n_centers) PyTPeriodicMatrix
        RBF center coordinates in preconditioned basis
    c_per : (z_dim, n_centers) PyTPeriodicMatrix
        RBF weigths.
    d_per : (z_dim, x_dim+1) PyTPeriodicMatrix
        Polynomial coefficients.
    t_per : (x_dim, x_dim) ndarray
        preconditioning matrix

    """

    oversampling = 2

    @classmethod
    def fromCBOR(cls, cbor_obj: dict, nopack: bool = False):
        """
        Create a deserialized instance from a CBOR serializabled object

        ref: http://cbor.io/

        Parameters
        ----------
        cbor_obj : dict
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PeriodicMultiRBF
        """
        xc_per = Serializable.fromCBOR(cbor_obj['xi_per'], nopack=nopack)
        t_per = Serializable.fromCBOR(cbor_obj['t_per'], nopack=nopack)
        c_per = Serializable.fromCBOR(cbor_obj['c_per'], nopack=nopack)
        d_per = Serializable.fromCBOR(cbor_obj['d_per'], nopack=nopack)
        kernel_id = Serializable.fromCBOR(cbor_obj['kernel'])
        kernel_params = Serializable.fromCBOR(cbor_obj['params'])
        max_dist = Serializable.fromCBOR(cbor_obj['max_dist'])
        kernel = kernels.AbstractKernel.fromID(kernel_id, kernel_params)
        return cls.fromMatrices(xc_per, c_per, d_per, kernel, t_per, max_dist=max_dist)

    @classmethod
    def fromMatrices(cls, xc_per, c_per, d_per, kernel, t_per=None, max_dist=0.):
        """
        Instantiate StackedMultiRBF from pre-calculated matrices

        Parameters
        ----------
        xc_per : PyTPeriodicMatrix
        c_per : PyTPeriodicMatrix
        d_per : PyTPeriodicMatrix
        kernel : kernels.AbstractKernel
        t_per : PyTPeriodicMatrix
        max_dist : float

        Returns
        -------
        PeriodicMultiRBF
        """
        return cls(xc_per, None, kernel, t_per, max_dist=max_dist, c_per=c_per, d_per=d_per)

    def __init__(self, xc_per, zc_per, kernel, t_per=None, gamma=0., max_dist=0., max_cond=0., **kwargs):
        """
        Initialize PeriodicMultiRBF object

        Parameters
        ----------

        xc_per : (x_dim, n_centers) PyTPeriodicMatrix
            RBF center coordinates.
        zc_per :  (z_dim, n_centers) PyTPeriodicMatrix
            RBF center values.
        kernel : lcm.rbf.kernel.AbstractKernel
            Function to calculate the RBF kernel.
            It expects an array_like argument (n_centers, n_samples) of radii
            and returns the RBF values in the same shape
        t_per : (x_dim, x_dim) PyTPeriodicMatrix
            Preconditioning matrix
            xc_per_precond = t_per * xc_per
        gamma : float
            Smoothing parameter.
        max_dist : float
            Maximum distance to closest center
        max_cond : float
            threshold for condition number. If maxcond == 0 (default),
            no condition number is calculated/checked!s
        """

        self.kernel = kernel
        self.max_dist = max_dist
        self.xc_per = xc_per

        ''' Preconditioning Matrix '''
        if t_per:
            ''' change basis '''
            self.t_per = t_per
        else:
            ''' keep basis 
                --> t_per.shape = (x_dim, x_dim)
            '''
            x_dim = xc_per.nrows()
            self.t_per = PyTPeriodicMatrix(x_dim, x_dim, mode="eye")

        # self.xc_per.expand_to_odd()
        # self.t_per.expand_to_odd()

        if zc_per is None:
            xc_t, xc_t_per, c_per, d_per = self.__initFromMatrices(kwargs['c_per'], kwargs['d_per'])
        else:
            xc_t, xc_t_per, c_per, d_per = self.__initFromSamples(zc_per, gamma, max_cond)

        # c_per.expand_to_odd()
        # d_per.expand_to_odd()
        # xc_t_per.expand_to_odd()

        self.c_per = c_per
        self.d_per = d_per
        self.xc_t = xc_t
        self.xc_t_per = xc_t_per
    
        '''pre allocate the the sample matrices 
        to speed up periodic interpolation in cases where x_bins < n_bins_init
        '''
        self.__n_bins_init = max(self.xc_t_per.nbins(),
                                 self.c_per.nbins(),
                                 self.d_per.nbins(),
                                 self.t_per.nbins())

        n_samples = self.__n_bins_init * PeriodicMultiRBF.oversampling
        self.__xc_t_samples = self.xc_t_per.get_samples(n_samples)
        self.__c_samples = self.c_per.get_samples(n_samples)
        self.__d_samples = self.d_per.get_samples(n_samples)
        self.__t_samples = self.t_per.get_samples(n_samples)

    @property
    def xdim(self):
        return self.d_per.ncols() - 1

    @property
    def zdim(self):
        return self.d_per.nrows()

    def __initFromSamples(self, zc_per, gamma, max_cond):
        """
        Initailize Xc_T, C and D matrices from Zc samples
        
        Parameters
        ----------
        zc_per : PyTPeriodicMatrix
        gamma : float
        max_cond : float

        Returns
        -------
        xc_t : np.ndarray
        xc_t_per : PyTPeriodicMatrix
        c_per : PyTPeriodicMatrix
        d_per : PyTPeriodicMatrix
        """

        ''' determine number of samples along phi'''
        # zc_per.expand_to_odd()
        n_samples = max(self.xc_per.nbins(), zc_per.nbins()) * PeriodicMultiRBF.oversampling

        ''' do sampling of rbf centers:
            --> xc.shape = (x_dim, n_centers, n_samples)
            --> zc.shape = (z_dim, n_centers, n_samples)
        '''
        xc = self.xc_per.get_samples(n_samples)
        t = self.t_per.get_samples(n_samples)

        zc = zc_per.get_samples(n_samples)
        (xc_t, c, d) = calc_weigths(xc, t, zc, self.kernel, gamma, max_cond)
        c_per = PyTPeriodicMatrix(c, mode="samples", nopack=True)
        d_per = PyTPeriodicMatrix(d, mode="samples", nopack=True)

        xc_t_per = PyTPeriodicMatrix(xc_t, mode="samples", nopack=True)
        return xc_t, xc_t_per, c_per, d_per

    def __initFromMatrices(self, c_per, d_per):
        """
        Initailize Xc_T, C and D matrices from matrices
        
        Parameters
        ----------
        c_per : PyTPeriodicMatrix
        d_per : PyTPeriodicMatrix

        Returns
        -------
        xc_t : np.ndarray
        xc_t_per : PyTPeriodicMatrix
        c_per : PyTPeriodicMatrix
        d_per : PyTPeriodicMatrix

        """
        # c_per.expand_to_odd()
        # d_per.expand_to_odd()

        n_samples = max(self.xc_per.nbins(), c_per.nbins(), d_per.nbins()) * PeriodicMultiRBF.oversampling
        xc = self.xc_per.get_samples(n_samples)

        xc_t_per = self.xc_per
        xc_t = xc
        return xc_t, xc_t_per, c_per, d_per

    def __getSampledMatrices(self, x_per : PyTPeriodicMatrix):
        """
        Return number of samples and sampled matrices
        Parameters
        ----------
        x_per: np.ndarray

        Returns
        -------
        n_samples: int
        xc_t_samples : np.ndarray
        c_samples: np.ndarray
        d_samples: np.ndarray 
        t_samples: np.ndarray
        """

        nx_bins = x_per.nbins()
        if nx_bins > self.__n_bins_init:
            n_bins_max = nx_bins
            n_samples = n_bins_max * PeriodicMultiRBF.oversampling
            xc_t_samples = self.xc_t_per.get_samples(n_samples)
            c_samples = self.c_per.get_samples(n_samples)
            d_samples = self.d_per.get_samples(n_samples)
            t_samples = self.t_per.get_samples(n_samples)
        else:
            n_samples = self.__n_bins_init * PeriodicMultiRBF.oversampling
            xc_t_samples = self.__xc_t_samples
            c_samples = self.__c_samples
            d_samples = self.__d_samples
            t_samples = self.__t_samples

        return n_samples, xc_t_samples, c_samples, d_samples, t_samples

    def __call__(self, x, *args):
        """
        Evaluate interpolation

        Parameters
        ----------
        x : (x_dim, 1) PyTPeriodicMatrix  or (x_dim, n_samples) ndarray of float
            Coordinate, where to interpolate
        phi : (n_samples,) ndarray of float, optional
            Used in in case of an interpolation at a single angular position "phi"
            In this case, x must be of shape xxx

        Returns
        -------

        """

        if len(args) == 0:
            '''periodic interpolation'''
            return self.periodic_interpolation(x)

        elif len(args) == 1:
            ''' single point interpolation '''
            return self.discrete_angle_interpolation(x, args[0])

        else:
            raise ValueError("too many arguments")

    def jacobian(self, x, *args):
        """
        Evaluate partial derivative along x

        Parameters
        ----------
        x : (x_dim, 1) PyTPeriodicMatrix  or (x_dim, n_samples) ndarray of float
            Coordinate, where to interpolate
        phi : (n_samples,) ndarray of float, optional
            Used in in case of an interpolation at a single angular position "phi"
            In this case, x must be of shape xxx

        Returns
        -------

        """

        if len(args) == 0:
            '''periodic interpolation'''
            return self.periodic_jacobian(x)

        elif len(args) == 1:
            ''' single point interpolation '''
            return self.discrete_angle_jacobian(x, args[0])

        else:
            raise ValueError("too many arguments")

    def discrete_angle_interpolation(self, x, phi):
        """
        Evaluate interpolation for discrete angles

        Parameters
        ----------
        x : (x_dim, n_interpolations, n_samples) ndarray of float
            periodic signals representing the coordinates where to interpolate
        phi : (n_samples,) ndarray of float
            angular position(s), where to interpolate

        Returns
        -------
        z : (z_dim, n_interpolations, n_samples) ndarray of float
            result of interpolation
        """

        if isinstance(phi, float) or isinstance(phi, int):
            phi = np.array([phi])

        if phi.ndim != 1:
            raise ValueError("phi must be 1d")

        n_samples = phi.shape[0]

        if x.shape[2] != n_samples:
            raise ValueError("n_samples doesn't match")

        ''' do sampling: -->c.shape=(z_dim, n_centers, n_samples)'''
        c = self.c_per.value_at(phi)

        ''' do sampling: -->c.shape=(x_dim + 1, n_centers, n_samples)'''
        d = self.d_per.value_at(phi)

        ''' do sampling: -->t.shape=(x_dim, x_dim, n_samples)'''
        t = self.t_per.value_at(phi)

        ''' do sampling: -->xc_t.shape=(x_dim, n_centers, n_samples)'''
        xc_t = self.xc_t_per.value_at(phi)

        return eval_z(xc_t, c, d, t, x, self.kernel, self.max_dist)

    def periodic_interpolation(self, x_per):
        """
        Evaluate interpolations of periodic signal
        Parameters
        ----------
        x_per : (x_dim, n_interpolations) PyTPeriodicMatrix
            periodic signals representing the coordinates where to interpolate

        Returns
        -------
        z_per : (z_dim, n_interpolations) PyTPeriodicMatrix
            periodic signals as result of interpolations
        """

        # if x_per.ncols() > 1:
        #    raise ValueError("ncols=1 required!")

        ''' do sampling '''
        samples = self.__getSampledMatrices(x_per)
        x_samples = x_per.get_samples(samples[0])

        z = eval_z(samples[1], samples[2], samples[3], samples[4], x_samples, self.kernel, self.max_dist)

        ''' periodize z and return'''
        z_per = PyTPeriodicMatrix(z, mode="samples", nopack=True)
        # z_per.expand_to_odd()
        return z_per

    def discrete_angle_jacobian(self, x, phi):
        """
        Evaluate partial derivatives along x for discrete angles

        Parameters
        ----------
        x : (x_dim, n_interpolations, n_samples) ndarray of float
            periodic signals representing the coordinates where to interpolate
        phi : (n_samples,) ndarray of float
            angular position(s), where to interpolate

        Returns
        -------
        z : (z_dim, n_interpolations, n_samples) ndarray of float
            result of interpolation
        """

        if phi.ndim != 1:
            raise ValueError("phi must be 1d")

        n_samples = phi.shape[0]

        if x.shape[2] != n_samples:
            raise ValueError("n_samples doesn't match")

        ''' do sampling: -->c.shape=(z_dim, n_centers, n_samples)'''
        c = self.c_per.value_at(phi)

        ''' do sampling: -->c.shape=(x_dim + 1, n_centers, n_samples)'''
        d = self.d_per.value_at(phi)

        ''' do sampling: -->t.shape=(x_dim, x_dim, n_samples)'''
        t = self.t_per.value_at(phi)

        ''' do sampling: -->xc_t.shape=(x_dim, n_centers, n_samples)'''
        xc_t = self.xc_t_per.value_at(phi)

        return eval_dz_dx(xc_t, c, d, t, x, self.kernel, self.max_dist)

    def periodic_jacobian(self, x_per):
        """
        Evaluate partial derivatives along x of periodic signal
        Parameters
        ----------
        x_per : (x_dim, n_interpolations) PyTPeriodicMatrix
            periodic signals representing the coordinates where to interpolate

        Returns
        -------
        # dz_dx_per : (z_dim, x_dim n_interpolations) PyTPeriodicMatrix
        #     periodic signals as result of interpolations
        """

        if x_per.ncols() > 1:
            ''' n_interpolations must be 1! '''
            raise ValueError("only n_interpolations=1 supported!")

        ''' do sampling '''
        samples = self.__getSampledMatrices(x_per)
        x_samples = x_per.get_samples(samples[0])

        ''' z: 
            (z_dim, x_dim, n_interpolations, n_samples) ndarray '''
        dz_dx = eval_dz_dx(samples[1], samples[2], samples[3], samples[4], x_samples, self.kernel, self.max_dist)

        ''' periodize dz_dx and return'''
        samples = dz_dx[:, :, 0, :]
        return PyTPeriodicMatrix(samples, mode="samples")

    def periodic_angle_derivative(self, x_per, omega):
        """
        Evaluate partial derivatives along samples of periodic signal
        Parameters
        ----------
        x_per : (x_dim, n_interpolations) PyTPeriodicMatrix
            periodic signals representing the coordinates where to interpolate

        Returns
        -------
        dz_dphi_per : (z_dim, x_dim n_interpolations) PyTPeriodicMatrix
            periodic signals as result of interpolations
        """

        if x_per.ncols() > 1:
            raise ValueError("ncols=1 required!")

        ''' do sampling '''
        samples = self.__getSampledMatrices(x_per)
        x_samples = x_per.get_samples(samples[0])

        dz_dphi = eval_dz_dphi(samples[1], samples[2], samples[3], samples[4], x_samples, 
                               self.kernel, self.max_dist, omega)

        ''' periodize dz_dphi and return'''
        return PyTPeriodicMatrix(dz_dphi[:, :, :], mode="samples")

    def transform_order(self, p=1, q=1, nopack=True):
        """
        Transform frequency. Source object remains unchanged! new_order =  order*p/q

        Parameters
        ----------

        p : factor to multiply frequency with
        q : factor to divide frequency by
        nopack : bool
            whether to perform packing or not

        Returns
        -------
        res : Transformed object

        """

        the_clone = copy.deepcopy(self)

        the_clone.xc_t_per = self.xc_t_per.transform_order(p, q, nopack)
        the_clone.c_per = self.c_per.transform_order(p, q, nopack)
        the_clone.d_per = self.d_per.transform_order(p, q, nopack)
        the_clone.t_per = self.t_per.transform_order(p, q, nopack)

        return the_clone

    def to_msgpack(self):
        """
        Create msgpack serializable dict representing an instance of this class
        ref: http://msgpack.org/

        Returns
        -------
        msgpack : dict
            msgpack representation

        """

        msgpack = {"kernel": self.kernel.to_msgpack(),
                   "max_dist": self.max_dist,
                   "xc_t_per": self.xc_t_per.to_msgpack(),
                   "t_per": self.t_per.to_msgpack(),
                   "c_per": self.c_per.to_msgpack(),
                   "d_per": self.d_per.to_msgpack()}

        return msgpack

    def write_msgpack(self, file):
        """
        Create msgpack file representing an instance of this class
        ref: http://msgpack.org/


        Parameters
        ----------
        file : String
            path to msgpack file
        """
        import msgpack
        data = self.to_msgpack()
        data["guid"] = "{8c4e810f-3df3-4a00-8276-176fa3c9f000}"

        with open(file, "wb") as fh:
            msgpack.dump(data, fh)

    def toCBOR(self, f=1, nopack=False):
        """
        Create CBOR serializable dict representing an instance of this class
        ref: http://cbor.io/

        Parameters
        ----------
        f : integer
            angular scaling factor, 1 (Default)
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        cbor_obj : dict
                cbor representation

        """

        ''' create cbor serializable dict '''
        cbor_obj = {'type': '__periodicRBF__',
                    'xi_per': self.xc_t_per.toCBOR(f, nopack),
                    't_per': self.t_per.toCBOR(f, nopack),
                    'c_per': self.c_per.toCBOR(f, nopack),
                    'd_per': self.d_per.toCBOR(f, nopack),
                    'kernel': Serializable.toCBOR(self.kernel.id),
                    'params': Serializable.toCBOR(self.kernel.params),
                    'max_dist': Serializable.toCBOR(self.max_dist)
                    }

        return cbor_obj
    
    def toRBF(self) -> RBF:
        xc = self.xc_per.mean()
        c = self.c_per.mean()
        d = self.d_per.mean()
        t = self.t_per.mean()
        rbf_mean = RBF.fromMatrices(xc, c, d, kernel=self.kernel, t=t, max_dist=self.max_dist)
        return rbf_mean
    
    def select(self, orders, nopack=True):
        """
        Select specified orders of signals.

        Parameters
        ----------
        orders : integer, list or set of integers
            only selected harmonics will be returned
            does not affect dc offset!
        nopack : bool
            don't use pack, False (Default)

        Returns
        -------

        """
        the_clone = copy.deepcopy(self)
        the_clone.xc_per = self.xc_per.select(orders, nopack)
        the_clone.xc_t_per = self.xc_t_per.select(orders, nopack)
        the_clone.t_per = self.t_per.select(orders, nopack)
        the_clone.c_per = self.c_per.select(orders, nopack)
        the_clone.d_per = self.d_per.select(orders, nopack)
        return the_clone

    def getMaxOrder(self):
        orders = [x.get_max_order() for x in [self.xc_per, self.xc_t_per, self.t_per, self.c_per, self.d_per]]
        return np.max(orders)

    def sample(self, n_xsamples=None, mode='mean', xmin=None, xmax=None):

        if n_xsamples is None:
            n_xsamples = [10] * self.xdim

        if self.xdim != len(n_xsamples):
            raise Exception("Wrong size of n_samples")

        if xmin is None:
            xmin = self.xc_per.mean().min(axis=1)[:, np.newaxis]
        else:
            xmin = np.array(xmin)[:, np.newaxis]
        if xmax is None:
            xmax = self.xc_per.mean().max(axis=1)[:, np.newaxis]
        else:
            xmax = np.array(xmax)[:, np.newaxis]
        space = [np.linspace(xmin[index], xmax[index], n_xsamples[index]) for index in range(xmin.size)]

        grid = np.meshgrid(*space)

        positions = np.vstack(map(np.ravel, grid))
        n_loadpoints = positions.shape[1]

        rbfIn_array = np.zeros((n_loadpoints, self.xdim, 1), dtype=object)
        rbfOut_array = np.zeros((n_loadpoints, self.zdim, 1), dtype=object)
        # cnt = 0
        for index in range(n_loadpoints):
            x_array = positions[:, index, np.newaxis]

            rbfIn = PyTPeriodicMatrix(x_array, mode='dc')
            # try:
            rbfOut = self.periodic_interpolation(rbfIn)
            # except:
            #     # rbfOut = np.zeros((self.zdim, 1))
            #     # rbfOut.fill(np.nan)
            #     continue
            if mode == 'mean':
                rbfIn = rbfIn.mean()
                rbfOut = rbfOut.mean()
            elif mode == 'full':
                pass
            else:
                raise Exception("Unknown mode: " + str(mode))
            rbfIn_array[index, :, :] = rbfIn
            rbfOut_array[index, :, :] = rbfOut
            # cnt += 1
        # rbfIn_array = rbfIn_array[:idx, :, :]
        # rbfOut_array = rbfOut_array[:idx, :, :]
        # n_xsamples = idx
        return rbfIn_array, rbfOut_array, n_xsamples

    def evalRadius(self, x_per) -> np.ndarray:
        """
        Evaluates the radius

        Parameters
        ----------
        x_per : (x_dim, n_interpolations) PyTPeriodicMatrix
            periodic signals representing the coordinates where to evaluate

        Returns
        -------
        r : float
            radius
        """

        # if x_per.ncols() > 1:
        #    raise ValueError("ncols=1 required!")

        samples = self.__getSampledMatrices(x_per)
        x_samples = x_per.get_samples(samples[0])
        xc_t = x_samples[1]
        t_samples = x_samples[4]
        
        r = eval_r(xc_t, t_samples, x_samples)

        return r

    # def preview(self, rbfOut_array, n_xsamples, x_axis=0, y_axis=1, z_axis=1):
    #     from symspace.wrapper.symplot import SymPlot3Surface
    #
    #     xmin = self.xc_per.mean().min(axis=1)[:, np.newaxis]
    #     xmax = self.xc_per.mean().max(axis=1)[:, np.newaxis]
    #     space = [np.linspace(xmin[index], xmax[index], n_xsamples[index]) for index in range(xmin.size)]
    #
    #     x3d = space[x_axis]
    #     y3d = space[y_axis]
    #
    #     zz = rbfOut_array[:, z_axis, 0].reshape(n_xsamples)
    #     z3d = zz.astype(float)
    #
    #     surface3D = SymPlot3Surface()
    #     surface3D.addSignal('surface3D', x3d, y3d, z3d)
    #     surface3D.setTitle('surface3D')
    #     surface3D.setViewPoint(0, 0, 0)
    #     # surface3D.show()
    #     return surface3D
    #
    # def previewAllOutputs(self, n_xsamples=None, x_axis=0, y_axis=1):
    #     from symspace.wrapper.symplot import SymSubPlot
    #
    #     if n_xsamples is None:
    #         n_xsamples = [10] * self.xdim
    #
    #     nrows, ncols = [np.ceil(np.sqrt(self.zdim))] * 2
    #     rbfIn_array, rbfOut_array, n_xsamples = self.sample(n_xsamples)
    #
    #     sym_subplot = SymSubPlot(int(nrows), int(ncols))
    #     for z_axis in range(self.zdim):
    #         row = int(np.floor(z_axis / nrows))
    #         col = int(np.mod(z_axis, nrows))
    #         sym_subplot[row, col] = self.preview(rbfOut_array, n_xsamples, x_axis, y_axis, z_axis)
    #     return sym_subplot
    #
    def plotAllOutputs(self, n_xsamples=None, x_axis=0, y_axis=1):
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        if n_xsamples is None:
            n_xsamples = [10] * self.xdim

        fig = plt.figure()
        nrows, ncols = [np.ceil(np.sqrt(self.zdim))] * 2
        rbfIn_array, rbfOut_array, n_xsamples = self.sample(n_xsamples)
        for z_axis in range(self.zdim):
            ax = fig.add_subplot(nrows, ncols, 1 + z_axis, projection=Axes3D.name)
            self.plot(rbfIn_array, rbfOut_array, n_xsamples, x_axis, y_axis, z_axis, ax)
        plt.tight_layout()
        return fig

    def plot(self, rbfIn_array, rbfOut_array, n_xsamples, x_axis=0, y_axis=1, z_axis=0, ax=None):
        from matplotlib import cm
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        if self.xdim == 1:
            x = rbfIn_array[:, 0, 0]
            z = rbfOut_array[:, z_axis, 0]

            xx, yy = np.meshgrid(x, x)
            zz, zz1 = np.meshgrid(z, z)

            # fig = plt.figure()
            # surf = plt.scatter(x, z, c=z, cmap=cm.coolwarm)
            # fig.colorbar(surf)
        else:

            if self.xdim == 2:
                xx = rbfIn_array[:, x_axis, 0].reshape(n_xsamples)
                yy = rbfIn_array[:, y_axis, 0].reshape(n_xsamples)
                zz = rbfOut_array[:, z_axis, 0].reshape(n_xsamples)
            else:
                # xx = rbfIn_array[:, x_axis, 0].reshape(n_xsamples).take(0, axis=2)
                # yy = rbfIn_array[:, y_axis, 0].reshape(n_xsamples).take(0, axis=2)
                # zz = rbfOut_array[:, z_axis, 0].reshape(n_xsamples).take(0, axis=2)

                xx = rbfIn_array[:, x_axis, 0].reshape(n_xsamples)[:, :, 0]
                yy = rbfIn_array[:, y_axis, 0].reshape(n_xsamples)[:, :, 0]
                zz = rbfOut_array[:, z_axis, 0].reshape(n_xsamples)[:, :, 0]

                # xx = rbfIn_array[:, x_axis, 0].reshape(n_xsamples)[:, 0, :]
                # yy = rbfIn_array[:, y_axis, 0].reshape(n_xsamples)[:, 0, :]
                # zz = rbfOut_array[:, z_axis, 0].reshape(n_xsamples)[:, 0, :]

        if ax is None:
            fig = plt.figure()
            ax = fig.gca(projection=Axes3D.name)
        else:
            fig = ax.get_figure()

        surf = ax.plot_surface(xx, yy, zz, cmap=cm.coolwarm, linewidth=0,
                               antialiased=False)
        ax.set_xlabel("rbfIn[" + str(x_axis) + "]")
        ax.set_ylabel("rbfIn[" + str(y_axis) + "]")
        ax.set_zlabel("rbfOut[" + str(z_axis) + "]")
        fig.colorbar(surf, shrink=0.5, aspect=5)
        return fig, ax

    def calcMinRadius2D(self, xaxis=0, yaxis=1, nx_samples=(10, 10)):
        xc = (self.t_per.inv() * self.xc_t_per).mean()

        x = xc[xaxis, :]
        y = xc[yaxis, :]

        nx, ny = nx_samples

        xi = np.linspace(np.min(x), np.max(x), nx)
        yi = np.linspace(np.min(y), np.max(y), ny)
        ri = np.zeros((nx, ny))

        for row in range(nx):
            for col in range(ny):
                x_per = PyTPeriodicMatrix.fromDC(np.array([[xi[row]], [yi[col]]]))
                ri[row, col] = self.evalRadius(x_per)

        return xi, yi, ri

    def calcMinRadius3D(self, z_values, xaxis=0, yaxis=1, nx_samples=(10, 10)):
        xc = (self.t_per.inv() * self.xc_t_per).mean()

        x = xc[xaxis, :]
        y = xc[yaxis, :]

        nx, ny = nx_samples
        nz = z_values.size

        xi = np.linspace(np.min(x), np.max(x), nx)
        yi = np.linspace(np.min(y), np.max(y), ny)
        ri = np.zeros((nx, ny, nz))

        for index, zi in enumerate(z_values):
            for row in range(nx):
                for col in range(ny):
                    x_per = PyTPeriodicMatrix.fromDC(np.array([[xi[row]], [yi[col]], [zi]]))
                    ri[row, col, index] = self.evalRadius(x_per)

        return xi, yi, ri

    def plotMinRadius2DPLY(self, xaxis=0, yaxis=1):
        import plotly.graph_objs as go

        xi, yi, ri = self.calcMinRadius2D(xaxis, yaxis)

        contour = go.Contour(z=ri, x=xi, y=yi, colorscale='Jet',
                             showscale=True,
                             zmin=0, zmax=self.max_dist,
                             contours=dict(
                                 coloring='heatmap',
                                 showlabels=True,
                             )
                             )

        fig = go.Figure(data=[contour])
        fig['layout']['xaxis'].update(title='Input[{}]'.format(xaxis))
        fig['layout']['yaxis'].update(title='Input[{}]'.format(yaxis))
        return fig

    def plotMinRadius3DPLY(self, z_values, xaxis=0, yaxis=1, nx_samples=(10, 10)):
        from plotly import tools
        import plotly.graph_objs as go

        nz = z_values.size

        xi, yi, ri = self.calcMinRadius3D(z_values, xaxis=xaxis, yaxis=yaxis, nx_samples=nx_samples)

        fig = tools.make_subplots(rows=nz, cols=1, subplot_titles=['z-axis @ {:.2f}'.format(z) for z in z_values])

        for index in range(nz):
            contour = go.Contour(z=ri[:, :, index], x=xi, y=yi, colorscale='Jet',
                                 showscale=True,
                                 zmin=0, zmax=self.max_dist,
                                 contours=dict(
                                     coloring='heatmap',
                                     showlabels=True,
                                 )
                                 )
            fig.add_trace(contour, index + 1, 1)
            l_xaxis = 'xaxis{}'.format(index + 1) if index != 0 else 'xaxis'
            l_yaxis = 'yaxis{}'.format(index + 1) if index != 0 else 'yaxis'
            fig['layout'][l_xaxis].update(title='Input[{}]'.format(xaxis))
            fig['layout'][l_yaxis].update(title='Input[{}]'.format(yaxis))

        return fig
