#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.5

$LastChangedRevision: 1026 $
$LastChangedDate:: 2020-06-30 08:34:20 +0200#$

This file is part of SyMSpace.

Copyright (c) 2017, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""


import numpy as np
from lcm.tdata_interface.pyjavadata_interface import PyJavaDataInterface


class PyTNumericInterface(PyJavaDataInterface):
    """TNumeric conversion
    Java            | Python
    -------------------------
    TNumeric        | numpy.ndarray, int, float


    """

    @staticmethod
    def convert_J2Py(jobj):
        """
        Return a Python object of type "int" , "float" or "np.ndarray, dtype != object"
        equivalent to given Java object of type "TNumeric"
        """

        if jobj.isScalar():
            '''Variable is a scalar'''
            real = jobj.getReal(0, 0)
            imag = jobj.getImag(0, 0)

            if jobj.isReal():
                '''Variable is not a complex number'''
                pyobj = real
            else:
                '''Variable is a complex number'''
                pyobj = complex(real, imag)

        else:
            '''Variable is an array'''
            shape = jobj.size()
            real = np.zeros(shape)
            imag = np.zeros(shape)

            for row in range(shape[0]):
                for column in range(shape[1]):
                    real[row, column] = jobj.getReal(row, column)
                    imag[row, column] = jobj.getImag(row, column)

            if jobj.isReal():
                '''Variable is not a complex number'''
                pyobj = real
            else:
                '''Variable is a complex number'''
                pyobj = np.zeros(shape, dtype=complex)
                pyobj.real = real
                pyobj.imag = imag
        return pyobj

    # noinspection PyUnresolvedReferences
    @staticmethod
    def convert_Py2J(pyobj):
        """
        Return a Java object of type "TNumeric"
        equivalent to given python object of type "int" , "float" or "np.ndarray, dtype != object"
        """
        import jep
        from at.lcm.bu21.general.dtypes import TNumeric
        from org.apache.commons.math3.linear import BlockRealMatrix

        if isinstance(pyobj, int) or isinstance(pyobj, float):
            """
            Real scalar
            """
            jobj = TNumeric(float(pyobj))
        elif isinstance(pyobj, complex):
            """
            Complex scalar
            """
            jobj = TNumeric(float(pyobj.real), float(pyobj.imag))
        elif isinstance(pyobj, np.ndarray):
            """
            Real and complex arrays of dimension 1 or 2
            """
            if pyobj.size == 0:
                # Note: returning TNull or TArray could be a workaround,
                #  but results in other problems, hence it is better to not allow this option at all
                raise ValueError("Empty NDArrays cannot be converted to TNumeric!")
            elif pyobj.ndim == 1:
                j_real_array = jep.jarray(pyobj.size, jep.JDOUBLE_ID, 0)
                j_imag_array = jep.jarray(pyobj.size, jep.JDOUBLE_ID, 0)
                for index in range(pyobj.size):
                    j_real_array[index] = float(pyobj[index].real)
                    j_imag_array[index] = float(pyobj[index].imag)
                jobj = TNumeric(j_real_array, j_imag_array)
            elif pyobj.ndim == 2:
                j_real_matrix = BlockRealMatrix(pyobj.shape[0], pyobj.shape[1])
                j_imag_matrix = BlockRealMatrix(pyobj.shape[0], pyobj.shape[1])
                for row in range(pyobj.shape[0]):
                    for col in range(pyobj.shape[1]):
                        j_real_matrix.setEntry(row, col, pyobj.real[row, col])
                        j_imag_matrix.setEntry(row, col, pyobj.imag[row, col])
                jobj = TNumeric(j_real_matrix, j_imag_matrix)
            else:
                raise ValueError("Maximum number of dimension for TNumeric constructor arguments is 2, "
                                 "instead pyobj.ndim is:", pyobj.ndim)
        else:
            raise TypeError("Matching TNumeric constructor for conversion from a Python class not found")

        return jobj


if __name__ == "__main__":
    from lcm.tdata_interface.pyjdata_interface import PyJDataInterface

    nd_array = np.array([1, 2, 3, 4])

    tmp_jobj = PyJDataInterface.convert_Py2J(nd_array)
    tmp_pyobj = PyJDataInterface.convert_J2Py(tmp_jobj)
    print('java:' + str(tmp_jobj))
    print('python:' + str(tmp_pyobj))
    print('python type:' + str(type(tmp_pyobj)))
