import io
import os
import matplotlib.pyplot as plt


def serializeFigure(figure, figure_format='svg', *args, **kwargs):
    if figure_format == 'base64_png' or format == 'bytes':
        serialized_figure = figureToBase64PNG(figure, *args, **kwargs)
        if figure_format == 'bytes':
            import base64
            serialized_figure = base64.b64decode(serialized_figure)
    elif figure_format == 'svg':
        serialized_figure = figureToSVG(figure, *args, **kwargs)
    elif figure_format == 'mpld3':
        serialized_figure = figureToMPLD3(figure, *args, **kwargs)
    elif figure_format == 'plotly':
        serialized_figure = figureToPlotlyHtml(figure, *args, **kwargs)
    elif figure_format == 'bokeh':
        serialized_figure = figureToBokehHtml(figure, *args, **kwargs)
    else:
        raise Exception('Figure format: "' + figure_format + '" is no valid option!')
    return serialized_figure


def figureToBase64PNG(matplotlib_figure, width=None, height=None, facecolor='lightgray', edgecolor='k',
                      transparent=False, dpi=300, *args, **kwargs):
    import base64
    matplotlib_figure = setFigureSize(matplotlib_figure, width, height, dpi)
    buf = io.BytesIO()
    matplotlib_figure.savefig(buf, format='png', facecolor=facecolor, edgecolor=edgecolor, transparent=transparent,
                              dpi=dpi, *args, **kwargs)
    buf.seek(0)
    img = base64.b64encode(buf.getvalue())
    buf.close()
    plt.close()
    return img.decode("utf-8")


def figureToSVG(matplotlib_figure, width=500, height=500, facecolor='lightgray', edgecolor='k', transparent=False,
                enable_zoom=False, *args, **kwargs):
    buf = io.BytesIO()
    matplotlib_figure.savefig(buf, dpi=matplotlib_figure.dpi, format='svg', facecolor=facecolor, edgecolor=edgecolor,
                              transparent=transparent, *args, **kwargs)
    buf.seek(0)
    img = buf.getvalue()
    svg_body = img.decode("utf-8")
    buf.close()
    plt.close()
    if enable_zoom:
        html = svgAddZoom(svg_body, width, height, facecolor)
    else:
        html = svgAutoFitSize(svg_body, facecolor)

    return html


def figureToMPLD3(matplotlib_figure, *args, **kwargs):
    import mpld3
    html = mpld3.fig_to_html(matplotlib_figure)
    plt.close()
    return html


def figureToPlotlyHtml(matplotlib_figure, *args, **kwargs):
    import plotly
    import plotly.tools as tls
    import matplotlib.figure

    if isinstance(matplotlib_figure, matplotlib.figure.Figure):
        plotly_figure = tls.mpl_to_plotly(matplotlib_figure)
    else:
        plotly_figure = matplotlib_figure

    div = plotly.offline.plot(plotly_figure, auto_open=False, output_type='div')
    html = '<html><body>' + div + '</body></html>'
    return html


def figureToBokehHtml(bokeh_figure, figure_title='Bokeh Figure', *args, **kwargs):
    import matplotlib.figure
    from bokeh.resources import INLINE
    from bokeh.embed import file_html

    if isinstance(bokeh_figure, matplotlib.figure.Figure):
        raise Exception()

    html = file_html(bokeh_figure, INLINE, figure_title)
    return html


def setFigureSize(matplotlib_figure, width, height, dpi):
    matplotlib_figure.dpi = dpi
    figure_size = matplotlib_figure.get_size_inches()

    if width is not None:
        figure_size[0] = width / matplotlib_figure.dpi
    if height is not None:
        figure_size[1] = height / matplotlib_figure.dpi

    matplotlib_figure.set_size_inches(figure_size)
    return matplotlib_figure


def showBase64Figure(base64figure):
    import base64
    from PIL import Image
    img = Image.open(io.BytesIO(base64.b64decode(base64figure)))
    img.show()


def showHTML(html):
    import webbrowser
    import tempfile
    import time

    temp_file_name = os.path.join(tempfile.gettempdir(), 'temp_' + str(time.time()) + '.html')
    with open(temp_file_name, 'w') as fh:
        print(html, file=fh)

    webbrowser.open('file://' + temp_file_name, new=2)


def saveHTML(html, filename):
    with open(filename, 'w', encoding='utf-8') as f:
        print(html, file=f)


def svgSplit(svg_body):
    svg_open = """
           <svg id="svg" """
    svg_close = svg_body.split('<svg')[1]
    svg_header = svg_body.split('<svg')[0]
    return svg_header, svg_open, svg_close


def svgAutoFitSize(svg_body, facecolor='white'):
    style = """
        <style>
            html, body {
                width: 100%;
                height: 100%;
                margin: 0;
                padding: 0;
                background-color: """ + facecolor + """
                ;
            }
            #svg {
                width: inherit;
                height: inherit;
                margin: 0;
                padding: 0;
            }
        </style>"""
    svg_header, svg_open, svg_close = svgSplit(svg_body)
    html = svg_header + style + svg_open + svg_close
    return html


def svgAddZoom(svg_body, width, height, facecolor='white'):
    svg_header, svg_open, svg_close = svgSplit(svg_body)
    style = """
        <style>
            html, body {
                width: 100%;
                height: 100%;
                margin: 0;
                padding: 0;
                background-color: """ + facecolor + """
                ;
            }
            #svg {
                width: inherit;
                height: inherit;
                margin: 0;
                padding: 0;
            }
            #container {
                widthDummy:px;
                heightDummy:px;
                border: 1px solid black;
            }
        </style>"""
    style = style.replace('widthDummy:', 'width: {w}'.format(w=width), 1)
    style = style.replace('heightDummy:', 'height: {h}'.format(h=height), 1)
    script = """
        <script src="http://ariutta.github.io/svg-pan-zoom/dist/svg-pan-zoom.js"></script>
        <script>
          // Don't use window.onLoad like this in production, because it can only listen to one function.
          window.onload = function() {
            // Expose to window namespase for testing purposes
            window.zoomSVG = svgPanZoom('#svg', {
              zoomEnabled: true,
              panEnabled: true,
              controlIconsEnabled: true,
              fit: true,
              center: true,

            });
          };
        </script>
        """
    div_open = """
        <div id="container">
        """
    div_close = """
        </div>
        """
    html = svg_header + style + script + div_open + svg_open + svg_close + div_close
    return html


# def htmlSlideshow(svg_body):
#     buttons = """<button class="w3-button w3-display-left" onclick="plusDivs(-1)">&#10094;</button>
#     <button class="w3-button w3-display-right" onclick="plusDivs(+1)">&#10095;</button>"""
#     script = """<script>
#     var slideIndex = 1;
#     showDivs(slideIndex);
#
#     function plusDivs(n) {
#         showDivs(slideIndex += n);
#     }
#
#     function showDivs(n) {
#         var i;
#         var x = document.getElementsByClassName("svg");
#         if (n > x.length) {slideIndex = 1}
#         if (n < 1) {slideIndex = x.length} ;
#         for (i = 0; i < x.length; i++) {
#             x[i].style.display = "none";
#         }
#         x[slideIndex-1].style.display = "block";
#     }
#     </script>
#     """
#     return svg_body + buttons + script

def figure_to_base64(matplotlib_figure, *args, **kwargs):
    import warnings
    warnings.warn('The function figure_to_base64 is deprecated use the function serializeFigure instead!')
    serialized_figure = serializeFigure(matplotlib_figure, figure_format='base64_png', *args, **kwargs)
    return serialized_figure
