# -*- coding: utf-8 -*-
"""
Created on Thu Sep 22 16:39:02 2016

@author: gweidenholzer, mkitzberger
"""
import copy
import numbers
import os
import sys

import math
import numpy as np
from scipy.fftpack import rfft, irfft
from scipy.optimize import brentq, newton
from scipy.signal import freqs

from lcm.utils.serialization import Serializable


# Todo: make sure scaling_factor used correctly


class PyTPeriodicMatrix:
    """
    This class represents a (N, M) matrix of time periodic signals.
    It allows operations and calculations on the signals in both time
    and frequency domain.

    Attributes
    ----------
    data : (nrows, ncols, nbins) ndarray
        RFFT representation of periodic signals according to scipy.fftpack.rfft

    shape : tuple
        number of rows and columns

    """

    minPreviewSamples = 200
    powerThreshold = 1.e-6
    resolution = 10.
    scaling_factor = 1  # type: int

    @classmethod
    def fromCBOR(cls, cbor_obj: dict, nopack: bool = False) -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix from CBOR serialized data

        Parameters
        ----------
        cbor_obj : dict
            CBOR serialized data
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PyTPeriodicMatrix
        """
        rfft_array = Serializable.fromCBOR(cbor_obj['bins'])
        f = Serializable.fromCBOR(cbor_obj['f'])
        obj = cls.fromRFFT(rfft_array, nopack=nopack)

        # ToDo: implement scaling factor
        # obj.scaling_factor = f
        obj = obj.transform_order(f, 1)
        return obj

    @classmethod
    def fromFit(cls, phi: np.ndarray, samples: np.ndarray, nopack: bool = False, maxorder: int = 30) \
            -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix from a samples fitting

        Parameters
        ----------
        phi : np.ndarray
            angle vector with interval length 2*pi
        samples : np.ndarray
            vector with equal length as angle vector
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false
        maxorder : int, optional
            Keyword argument setting the highest order. Applicable for modes "random" and "fit"
            Detault: 30

        Returns
        -------
        PyTPeriodicMatrix
        """
        obj = cls(phi, samples, mode='fit', maxorder=maxorder, nopack=nopack)
        return obj

    @classmethod
    def fromObjectArray(cls, objarray, nopack: bool = False) -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix from ndarray of 1x1 PyTPeriodicMatrix

        Parameters
        ----------
        objarray : np.ndarray
            ndarray of 1x1 PyTPeriodicMatrix objects
            ndarray of type object, containing PyTPeriodicMatrix items of size (1,1)
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------

        """
        obj = cls(objarray, mode='objarray', nopack=nopack)
        return obj

    @classmethod
    def fromSamples(cls, samples: np.ndarray, nopack: bool = False) -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix from samples

        Parameters
        ----------
        samples : np.ndarray
                Equidistant sample points over a entire period.
                samples[:,:,k] corresponds to phi=k*2*pi/n_samples
                size: (m, n, n_samples)
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PyTPeriodicMatrix
        """
        obj = cls(samples, mode='samples', nopack=nopack)
        return obj

    @classmethod
    def fromRFFT(cls, rfft_array: np.ndarray, nopack: bool = False) -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix from rfft samples

        Parameters
        ----------
        rfft_array: np.ndarray
            FFT representation of the signals according to scipy.fftpack.rfft
            size: (n, m, n_samples)
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PyTPeriodicMatrix
        """
        obj = cls(rfft_array, mode='rfft', nopack=nopack)
        return obj

    @classmethod
    def fromDC(cls, dc: np.ndarray, nopack: bool = False) -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix from dc-values

        Parameters
        ----------
        dc: np.ndarray
            Matrix containing dc-values
            size: (m, n)
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PyTPeriodicMatrix
        """
        obj = cls(dc, mode='dc', nopack=nopack)
        return obj

    @classmethod
    def fromCoeffs(cls, dc: np.ndarray, cos_coeffs: np.ndarray, sin_coeffs: np.ndarray, nopack: bool = False) \
            -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix from fourier coefficients

        y(phi) = dc + sum( cos_coeffs * cos(k*phi) + sin_coeffs * sin(k*phi), k=1..maxorder)

        Parameters
        ----------

        dc: np.ndarray
            dc offsets
            size: (n,m)
        cos_coeffs: np.ndarray
            coefficients of cosine terms
            size: (n,m,maxorder)
        sin_coeffs: np.ndarray
            coefficients of sine terms
            size: (n,m,maxorder)
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PyTPeriodicMatrix
        """
        obj = cls((dc, cos_coeffs, sin_coeffs), mode='coeffs', nopack=nopack)
        return obj

    @classmethod
    def fromOrderMagPhase(cls, order: int, mag: float, phase: float, nopack: bool = False) \
            -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix from order, magnitude, phase

        Parameters
        ----------

        order : int
        mag : float
        phase : float
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PyTPeriodicMatrix
        """
        obj = cls(order, mag, phase, nopack=nopack)
        return obj

    @classmethod
    def zeros(cls, size: tuple, nopack: bool = False) -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix zero matrix of size (ncols, nrows)

        Parameters
        ----------
        size : array_like
            (nrows, ncols)
            nrows: number of rows
            ncols: number of columns
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false


        Returns
        -------
        PyTPeriodicMatrix
        """
        obj = cls(size[0], size[1], mode='zero', nopack=nopack)
        return obj

    @classmethod
    def eye(cls, size: tuple, nopack: bool = False) -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix indentity matrix of size (ncols, nrows)

        Parameters
        ----------
        size : array_like
            (ncols, nrows)
            ncols: number of columns
            nrows: number of rows
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PyTPeriodicMatrix
        """
        obj = cls(size[0], size[1], mode='eye', nopack=nopack)
        return obj

    @classmethod
    def random(cls, size: tuple, nopack: bool = False, maxorder: int = 30) -> 'PyTPeriodicMatrix':
        """
        Initializes a PyTPeriodicMatrix of size (ncols, nrows) from random values

        Parameters
        ----------
        size : array_like
            (ncols, nrows)
            ncols: number of columns
            nrows: number of rows
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false
        maxorder : int, optional
            Keyword argument setting the highest order. Applicable for modes "random" and "fit"
            Detault: 30
        Returns
        -------
        PyTPeriodicMatrix
        """
        obj = cls(size[0], size[1], mode='random', nopack=nopack, maxorder=maxorder)
        return obj

    @classmethod
    def concatenate(cls, a_tuple, axis=0, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Join a sequence of PyTPeriodicMatrices along an existing axis.

        Parameters
        ----------
        a_tuple : tuple/list
            Tuple/List of PyTPeriodicMatrices
        axis : int, optional
            The axis along which the PyTPeriodicMatrices will be joined.
            Default is 0.

        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PyTPeriodicMatrix
        """
        if axis not in [0, 1]:
            raise Exception("PyTPeriodicMatrix.concatenate: Only axis 0 or 1 are allowed!")

        a_list = []
        dim = 'rows' if axis == 1 else 'cols'
        axis_check = 0 if axis == 1 else 1

        for element in a_tuple:
            if a_tuple[0].shape[axis_check] != element.shape[axis_check]:
                raise Exception(
                    "PyTPeriodicMatrix.concatenate: All elements must have the same number of {} !".format(dim))
            a_list.append(element.toObjArray())

        objarray = np.concatenate(a_list, axis=axis)

        return cls.fromObjectArray(objarray, nopack=nopack)

    @classmethod
    def hstack(cls, a_tuple, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Stack PyTPeriodicMatrices in sequence horizontally (column wise)

        Parameters
        ----------
        a_tuple : tuple/list
            Tuple/List of PyTPeriodicMatrices
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PyTPeriodicMatrix
        """
        return cls.concatenate(a_tuple, axis=1, nopack=nopack)

    @classmethod
    def vstack(cls, a_tuple, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Stack PyTPeriodicMatrices in sequence vertically (row wise)

        Parameters
        ----------
        a_tuple : tuple/list
            Tuple/List of PyTPeriodicMatrices
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        PyTPeriodicMatrix
        """
        return cls.concatenate(a_tuple, axis=0, nopack=nopack)

    @classmethod
    def insert(cls, main, insert_obj, index, axis, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Insert PyTPeriodicMatrix along the given axis at the given index.

        Parameters
        ----------
        main : PyTPeriodicMatrix
        insert_obj : PyTPeriodicMatrix
        index : int
        axis : int, optional
            Axis along which to insert values.
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------

        """
        if axis not in [0, 1]:
            raise Exception("PyTPeriodicMatrix.insert: Only axis 0 or 1 are allowed!")

        objarry1 = main.toObjArray()
        objarry2 = insert_obj.toObjArray()

        objarray = np.insert(objarry1, [index] * objarry2.shape[axis], objarry2, axis)

        return cls.fromObjectArray(objarray, nopack=nopack)

    def __init__(self, *args, **kwargs):
        """
        PyTPeriodicMatrix initialization routine. For initialization examples see the following examples section.

        Parameters
        ----------
         mode : {'objarray', 'samples', 'rfft', 'dc', 'coeffs', 'zero', 'eye', 'random', 'fit'}
            This keyword argument sets the mode of initialization. Depending on the selected mode
            a different number of positional arguments has to be supplied before the keyword
            argument. The following describes the parameters needed for the chosen mode.
            For initialization examples in each mode also refer to the example section below.

            fit: phi, samples phi: array_like angle vector with interval length 2*pi. samples: array_like samples
            vector with equal length as angle vector.

            objarray: (m, n) ndarray of 1x1 PyTPeriodicMatrix objects
                ndarray of type object, containing PyTPeriodicMatrix items of size (1,1)

            samples: (m, n, n_samples) ndarray
                Equidistant sample points over a entire period.
                samples[:,:,k] corresponds to phi=k*2*pi/n_samples

            rfft: (n, m, n_samples) ndarray
                FFT representation of the signals according to scipy.fftpack.rfft

            dc: (m, n) ndarray
                Matrix containing dc-values

            coeffs: tuple
                Tuple of fourier coefficients containing dc, cos_coeffs, sin_coeffs:
                'dc' is a (n,m) ndarray, containing dc offsets
                'cos_coeffs' is a (n,m,maxorder) ndarray, containing coefficients of cosine terms
                'sin_coeffs' is a (n,m,maxorder) ndarray, containing coefficients of sine terms
                y(phi) = dc + sum( cos_coeffs * cos(k*phi) + sin_coeffs * sin(k*phi), k=1..maxorder)

            zero: (nrows, ncols)
                ncols: number of columns, nrows: number of rows

            eye: (nrows, ncols)
                ncols: number of columns, nrows: number of rows

            random: (nrows, ncols)
                ncols: number of columns, nrows: number of rows

        maxorder : int, optional
            Keyword argument setting the highest order. Applicable for modes "random" and "fit"

        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Examples
        --------

        **Initialize from samples**:

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2 * np.pi, 100, endpoint=False)
           samples = np.empty((2, 3, 100), dtype=np.float)

           samples[0, 0] = np.zeros_like(phi)
           samples[0, 1] = 1.5 * np.ones_like(phi)
           samples[0, 2] = 1.5 + 2 * np.cos(phi)

           samples[1, 0] = 1.5 + 2 * np.cos(phi) - 0.5 * np.sin(2 * phi)
           samples[1, 1] = 2 * np.cos(phi) - 0.5 * np.sin(3 * phi)
           samples[1, 2] = 1.5 + 2 * np.cos(phi) + 1e-6 * np.cos(7 * phi)

           tp = pm.PyTPeriodicMatrix(samples, mode='samples')
           tp.show_preview()
           plt.show()


        **Initialize from rfft- vector**:

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           rfft_vec = np.array([[[0., 0., 0., 0., 0., 0., 0.],
                           [10.5, 0., 0., 0., 0., 0., 0.],
                           [10.5, 7., -0., 0., 0., 0., 0.]],
                          [[10.5, 7., -0., 0., 1.75, 0., 0.],
                           [-0., 7., -0., -0., -0., 0., 1.75],
                           [10.5, 7., -0., 0., 0., 0., 0.]]])

           tp = pm.PyTPeriodicMatrix(rfft_vec, mode='rfft')
           tp.show_spectrum()
           plt.show()

        **Initialize from coeffs**:

        .. plot::
           :include-source:

            import numpy as np
            import matplotlib.pyplot as plt
            import lcm.utils.pytperiodic as pm

            dc = np.zeros((2, 2))
            dc[0, 0] = 0
            dc[1, 0] = 1
            dc[0, 1] = 2
            dc[1, 1] = 3

            cos_coeffs = np.zeros((2, 2, 5))
            cos_coeffs[0, 0] = np.arange(5)
            cos_coeffs[1, 0] = np.arange(5)
            cos_coeffs[0, 1] = np.arange(5)
            cos_coeffs[1, 1] = np.arange(5)

            sin_coeffs = np.zeros((2, 2, 5))
            sin_coeffs[0, 0] = np.arange(5)
            sin_coeffs[1, 0] = np.arange(5)
            sin_coeffs[0, 1] = np.arange(5)
            sin_coeffs[1, 1] = np.arange(5)

            tp = pm.PyTPeriodicMatrix((dc, cos_coeffs, sin_coeffs), mode="coeffs")
            tp.show_spectrum()
            plt.show()


        **Initialize from dc- values**:

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           dc_vals = np.ones((2, 3), dtype=np.float)
           tp = pm.PyTPeriodicMatrix(dc_vals, mode='dc')
           tp.show_preview()
           plt.show()

        **Initialize with dc values corresponding to eye matrix**:

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           tp = pm.PyTPeriodicMatrix(3, 3, mode='eye')
           tp.show_preview()
           plt.show()

        **Initialize from angle and sample vector**:

        .. plot::
          :include-source:

          import numpy as np
          import matplotlib.pyplot as plt
          import lcm.utils.pytperiodic as pm

          phi = np.linspace(0, 2*np.pi, 1000, endpoint=False)
          tp = pm.PyTPeriodicMatrix(phi, np.sin(phi), mode='fit')

          tp.show_preview()
          plt.show()


        **Initialize with random values**:

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           tp = pm.PyTPeriodicMatrix(3, 3, mode='random', maxorder = 20)
           tp.show_preview()
           plt.show()

        **Initialize using order, mag, phase**:

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # first order cosine
           tp = pm.PyTPeriodicMatrix(1, 1, 0)
           tp.show_preview()
           plt.show()

           # first order sine
           tp = pm.PyTPeriodicMatrix(1, 1, -np.pi/2)
           tp.show_preview()
           plt.show()

           # second order cosine
           tp = pm.PyTPeriodicMatrix(2, 1, 0)
           tp.show_preview()
           plt.show()

           # second order sine
           tp = pm.PyTPeriodicMatrix(2, 1, -np.pi/2)
           tp.show_preview()
           plt.show()

        **Initialize with other PyTPeriodicMatrix objects**:

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2 * np.pi, 1000, endpoint=False)
           tp_array = np.array([[pm.PyTPeriodicMatrix(1, 1, 0), pm.PyTPeriodicMatrix(1, 1, 0)],
                                    [pm.PyTPeriodicMatrix(1, 1, 0), pm.PyTPeriodicMatrix(1, 1, 0)]], dtype=object)

           tp_array2 = np.array([[pm.PyTPeriodicMatrix(phi, np.sin(phi), mode='fit')],
                               [pm.PyTPeriodicMatrix(1, 1, 0)]], dtype=object)

           tp1 = pm.PyTPeriodicMatrix(tp_array, mode='objarray')
           tp2 = pm.PyTPeriodicMatrix(tp_array2, mode='objarray')

           tp1.show_preview()
           plt.show()

           tp2.show_preview()
           plt.show()


        """

        # get maxorder
        if kwargs.get("maxorder"):
            self.maxorder = kwargs["maxorder"]
        else:
            self.maxorder = 30

        if kwargs.get("nopack"):
            self.nopack = kwargs["nopack"]
        else:
            self.nopack = False

        if args:
            nargs = len(args)
            if nargs in [1, 2]:
                if kwargs.get("mode"):
                    mode = kwargs["mode"].lower()
                else:
                    raise ValueError("keyword argument 'mode' required!")
            else:
                mode = None

            if nargs == 1:
                if mode == "objarray":
                    self._data = self.__initObjarray(*args)
                elif mode == "samples":
                    self._data = self.__initSamples(*args)
                elif mode == "rfft":
                    self._data = self.__initRFFT(*args)
                elif mode == "dc":
                    self._data = self.__initDC(*args)
                elif mode == "coeffs":
                    self._data = self.__initCoeffs(*args[0])
                else:
                    raise ValueError("unknown mode: '" + str(mode) + "'")

            elif nargs == 2:
                if mode == "zero":
                    self._data = self.__initZero(*args)
                elif mode == "eye":
                    self._data = self.__initEye(*args)
                elif mode == "random":
                    self._data = self.__initRandom(*args)
                elif mode == "fit":
                    self._data = self.__initFit(*args)
                else:
                    raise ValueError("unknown mode: '" + str(mode) + "'")

            elif nargs == 3:
                self._data = self.__initOrderMagPhase(*args)

        else:
            self._data = np.zeros((1, 1, 1))

        self._data = np.array(self._data, dtype=np.float)

        if self._data.ndim != 3:
            raise ValueError("wrong dimension")

        self.pack()

    @property
    def interpreter(self) -> str:
        """
        Determine current interpreter (required for custom plot functionality)
        interpreter can be:   'python' for normal cpython
                              'java'   for execution in jep
                              'javaw'  for execution in jep

        Moved to property due to performance reasons

        Returns
        -------
        str
        """

        interpPath = os.path.normpath(sys.executable)
        return os.path.splitext(os.path.basename(interpPath))[0]

    @staticmethod
    def __initObjarray(objarray) -> np.ndarray:
        if not isinstance(objarray, np.ndarray):
            raise TypeError('objarray needs to be of type np.ndarray')
        if objarray.dtype != np.dtype('O'):
            raise TypeError('objarray needs to be of dtype np.object')

        m = objarray.shape[0]
        n = objarray.shape[1]
        max_order = 0

        ''' find out maximum order '''
        for tmp in objarray.flat:
            if not tmp.data.shape[0:2] == (1, 1):
                raise ValueError("objarray's item's shape must be (1,1)!")
            max_order = max(max_order, tmp.get_max_order())

        ''' create empty data array '''
        bins = 1 + max_order * 2
        data = np.zeros((m, n, bins))

        ''' fill data '''
        for row in range(0, m):
            for col in range(0, n):
                tmp = copy.deepcopy(objarray[row, col])
                tmp.expand_to_odd(bins)
                data[row, col, :] = tmp.data
        return data

    @staticmethod
    def __initSamples(samples: np.ndarray) -> np.ndarray:
        """
        PyTPeriodicMatrix(samples, mode="samples")
        """
        data = rfft(samples)
        return data

    @staticmethod
    def __initRFFT(rfft: np.ndarray) -> np.ndarray:
        """
        PyTPeriodicMatrix(rfft_array, mode="rfft")
        """
        data = copy.deepcopy(rfft)
        return data

    @staticmethod
    def __initDC(dc: np.ndarray) -> np.ndarray:
        """
        PyTPeriodicMatrix(dc_array, mode="dc")
        """
        data = dc[:, :, np.newaxis]
        return data

    @staticmethod
    def __initCoeffs(dc: np.ndarray, cos_coeffs: np.ndarray, sin_coeffs: np.ndarray) -> np.ndarray:
        """
        PyTPeriodicMatrix( (dc, cos_coeffs, sin_coeffs), mode ="coeffs" )
        """
        rows, cols, orders = cos_coeffs.shape
        n = 1 + 2 * orders

        data = np.zeros((rows, cols, n))
        data[:, :, 0] = dc * n
        data[:, :, 1::2] = cos_coeffs * n / 2.
        data[:, :, 2::2] = -sin_coeffs * n / 2.
        return data

    @staticmethod
    def __initZero(nrows: int, ncols: int) -> np.ndarray:
        """
        PyTPeriodicMatrix(n, m, mode='zero')
        """
        data = np.zeros((nrows, ncols, 1))
        return data

    @staticmethod
    def __initEye(nrows: int, ncols: int) -> np.ndarray:
        """
        PyTPeriodicMatrix(n, m, mode='eye')
        """
        data = np.zeros((nrows, ncols, 1))
        data[:, :, 0] = np.eye(nrows, ncols)
        return data

    def __initRandom(self, nrows: int, ncols: int) -> np.ndarray:
        """
        PyTPeriodicMatrix(n, m, mode='random')
        """
        data = np.random.randn(nrows, ncols, 1 + 2 * self.maxorder)
        return data

    def __initFit(self, phi: np.ndarray, ydata: np.ndarray) -> np.ndarray:
        """
        PyTPeriodicMatrix(phi, y, mode='fit')
        Take sample points at arbitrary angles phi
        """
        # convert to numpy array
        phi = np.array(phi, np.float)
        ydata = np.array(ydata, np.float)

        if ydata.ndim == 1:
            if ydata.size == phi.size:
                ydata_new = ydata
            else:
                raise ValueError("wrong number of samples")
        elif ydata.ndim == 2:
            if ydata.shape[0] == 0 or ydata.shape[1] == 0:
                ydata_new = ydata.flatten()
            elif ydata.shape[0] == phi.size:
                ydata_new = ydata.transpose()
            elif ydata.shape[1] == phi.size:
                ydata_new = ydata
            else:
                raise ValueError("wrong number of samples")
        else:
            raise ValueError("wrong number of dimensions")

        # find max. distance btw. samples
        wrapped = (phi % (2. * np.pi))
        hlp = np.sort(wrapped)
        maxdist = np.max(hlp[2:] - hlp[1:-1])
        maxdist = max([hlp[0] - (hlp[-1] - 2 * np.pi), maxdist])

        # mindist = np.min(hlp[2:] - hlp[1:-1])
        # mindist = min([hlp[0] - (hlp[-1] - 2 * np.pi), mindist])

        # even for non-equidistant sampling, it is sufficient when
        # the highest harmonic satisfies the shannon theoreme.
        # A safty factor 0.9 is applied here
        self.maxorder = min(self.maxorder, int(0.9 * np.pi / maxdist))

        if self.maxorder < 2:
            raise Exception("No periodic fitting possible!")

            # if maxdist / mindist < 1.001:
            # option: calculate dft for higher performance
            # pass

        orders = np.arange(1, self.maxorder + 1)
        a = np.ones((len(phi), 2 * self.maxorder + 1))
        b = np.outer(phi, orders)
        a[:, 1::2] = 2 * np.cos(b)
        a[:, 2::2] = -2 * np.sin(b)
        a = np.matrix(a)

        try:
            # Todo: check the correct useage since None does not seem to work
            x = np.linalg.lstsq(a, np.matrix(ydata_new).T, rcond=None)[0]
        except:
            x = np.linalg.lstsq(a, np.matrix(ydata_new).T, rcond=-1)[0]

        if x.ndim == 2 and x.shape[1] == 1:
            x = np.array(x)[:, 0]
        else:
            x = np.array(x)
        # convert vector of coefficients to rfft vector
        tmp = x * len(x)
        if ydata.ndim == 1:
            data = tmp[np.newaxis, np.newaxis, :]
        elif ydata.ndim == 2:
            tmp = np.atleast_2d(tmp.transpose())
            if ydata.shape[0] == phi.size:
                data = tmp[np.newaxis, :, :]
            else:
                data = tmp[:, np.newaxis, :]
        else:
            raise ValueError("wrong dimension")
        return data

    @staticmethod
    def __initOrderMagPhase(order: int, mag: float, phase: float) -> np.ndarray:
        """
        construct using order, magnitude, phase
        """
        n = int(order * 2 + 1)
        re = mag * math.cos(phase)
        im = mag * math.sin(phase)

        data = np.zeros((1, 1, n))
        data[0, 0, -2] = re * n / 2
        data[0, 0, -1] = im * n / 2
        return data

    @property
    def data(self) -> np.array:
        """
        3-dimensions array containing the periodic signal matrix data.

        data.shape=[rows, cols, rfft]


        """
        return self._data

    @data.setter
    def data(self, value) -> np.array:
        self._data = value

    @property
    def shape(self) -> tuple:
        """
        Shape of periodic signal matrix
        """
        return self.nrows(), self.ncols()

    def checkScalingFactor(self, other) -> None:
        if self.scaling_factor != other.scaling_factor:
            raise Exception('Scalingfactors are not the same {} != {}'.format(self.scaling_factor,
                                                                              other.scaling_factor))

    def reshape(self, rows, cols) -> 'PyTPeriodicMatrix':
        """
        Set new shape of periodic signal matrix.

        Parameters
        ----------
        rows : int
            Number of rows of new matrix

        cols : int
            Number of columns of new matrix


        Returns
        -------
        ret : PyTPeriodicMatrix
            Resulting object


        Example
        -------

        .. testcode::

            import numpy as np
            import matplotlib.pyplot as plt
            import lcm.utils.pytperiodic as pm

            tp1 = pm.PyTPeriodicMatrix(1, 10, mode='random')
            print('shape:')
            print(tp1.shape)

            print('reshaped:')
            print(tp1.reshape(5, 2).shape)

            tp2 = pm.PyTPeriodicMatrix(3, 3, mode='random')
            print('shape:')
            print(tp2.shape)

            print('reshaped:')
            print(tp2.reshape(1, 9).shape)

        .. testoutput::

           shape
           (1, 10)
           reshaped:
           (5, 2)
           shape
           (3, 3)
           reshaped:
           (1, 9)


        """
        if rows * cols != self.nrows() * self.ncols():
            raise ValueError('reshape(): reshaped matrix must contain same number of elements')

        data = self.data
        nbins = self.nbins()

        new_data = data.reshape(rows, cols, nbins)
        ret = PyTPeriodicMatrix(new_data, mode='rfft', nopack=self.nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def nrows(self) -> int:
        """
        Get number of rows.

        Returns
        -------
        int
            number of rows

        Example
        -------

        .. testcode::

            import numpy as np
            import matplotlib.pyplot as plt
            import lcm.utils.pytperiodic as pm

            dc_vals = np.ones((2, 5), dtype=np.float)
            tp = pm.PyTPeriodicMatrix(dc_vals, mode='dc')

            print('number of rows:')
            print(tp.nrows())

        .. testoutput::

           number of rows:
           2

        """
        return self._data.shape[0]

    def ncols(self) -> int:
        """
        Get number of columns.

        Returns
        -------
        int
            number of columns


        Example
        -------

        .. testcode::

            import numpy as np
            import matplotlib.pyplot as plt
            import lcm.utils.pytperiodic as pm

            dc_vals = np.ones((2, 5), dtype=np.float)
            tp = pm.PyTPeriodicMatrix(dc_vals, mode='dc')

            print('number of columns:')
            print(tp.ncols())

        .. testoutput::

           number of columns:
           5

        """
        return self._data.shape[1]

    def nbins(self) -> int:
        """
        Get length of numpy.rfft vector.

        Returns
        -------
        int
            length of vector

        Example
        -------

        .. testcode::

            import numpy as np
            import matplotlib.pyplot as plt
            import lcm.utils.pytperiodic as pm

            rfft_vec = np.array([[[0., 0., 0., 0., 0., 0., 0.],
                 [10.5, 0., 0., 0., 0., 0., 0.],
                 [10.5, 7., -0., 0., 0., 0., 0.]],
                [[10.5, 7., -0., 0., 1.75, 0., 0.],
                 [-0., 7., -0., -0., -0., 0., 1.75],
                 [10.5, 7., -0., 0., 0., 0., 0.]]])

            tp = pm.PyTPeriodicMatrix(rfft_vec, mode='rfft')

            print(tp.nbins())

        Output:

        .. testoutput::

           7



        """
        return self._data.shape[2]

    def min(self) -> np.array:
        """
        Get minimum values of signals in interval [0, 2*pi].

        Returns
        -------
        min : np.array
            minimum values

        Example
        -------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # first order cosine with magnitude 2
           tp = 2 * pm.PyTPeriodicMatrix(1, 1, 0)

           print('min:')
           print(tp.min())

        .. testoutput::

           min:
           [[ 2.]]




        """
        return np.min(self.get_samples(), axis=2)

    def max(self) -> np.array:
        """
        Get maximum values of signals in interval [0, 2*pi].

        Returns
        -------
        max : np.array
            maximum values

        Example
        -------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # first order cosine with magnitude 2
           tp = 2 * pm.PyTPeriodicMatrix(1, 1, 0)

           print('max:')
           print(tp.max())

        .. testoutput::

           max:
           [[ 2.]]



        """
        return np.max(self.get_samples(), axis=2)

    def pp(self) -> np.array:
        """
        Get peak to peak values of signals in interval [0, 2*pi].

        Returns
        -------
        pp : np.array
            peak to peak values


        Example
        -------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # first order cosine with magnitude 2
           tp = 2 * pm.PyTPeriodicMatrix(1, 1, 0)

           print('peak-peak:')
           print(tp.pp())

        .. testoutput::

           peak-peak:
           [[ 4.]]

        """
        tmp = self.get_samples()
        return np.max(tmp, axis=2) - np.min(tmp, axis=2)

    def mean(self) -> np.ndarray:
        """
        Get mean values of signals over one period.

        Returns
        -------
        mean : np.array
            mean value of signals

        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # first order sine with offset
           tp = pm.PyTPeriodicMatrix(1, 1, -np.pi/2) + 1.5
           print(tp.mean())

        Output:

        .. testoutput::

           [[ 1.5]]

        """
        return self._data[:, :, 0] / self.nbins()

    def rms(self) -> np.array:
        """
        Get root mean square values of signals.

        Returns
        -------
        rms : np.array
            root mean square values

        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # first order sine
           tp = pm.PyTPeriodicMatrix(1, 1, -np.pi/2)
           print(tp.rms())

        Output:

        .. testoutput::

           [[ 0.70710678]]

        """
        tmp = copy.deepcopy(self._data) / self.nbins() * 2.
        tmp[:, :, 0] /= np.sqrt(2.)
        return np.linalg.norm(tmp, axis=2) / np.sqrt(2.)

    def thd(self) -> np.array:
        """
        Get total harmonic distortion (THD) of signals.

        Returns
        -------
        thd : np.array
            total harmonic distortion

        Notes
        -----
        Definition of total harmonic distortion is not standardized.
        This function uses the most common :math:`\\operatorname{THD}_F` definition
        where F stands for Fundamental frequency.
        :math:`\\operatorname{THD}_F` is defined as:

        .. math::

           \\operatorname{THD}_{F}(x(\\omega t)) = \\frac{\\sqrt{\\sum\\limits_{k=2}^n X_k^2}}{X_1}

        where :math:`X_i` with :math:`i > 1` means the RMS-amplitude of the :math:`i`-th harmonic of the signal and
        :math:`X_1` means the RMS-Amplitude of the fundamental frequency.


        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # first order sine
           tp = pm.PyTPeriodicMatrix(1, 1, -np.pi/2)
           print(tp.thd())

        Output:

        .. testoutput::

           [[ 0.]]

        """
        p_base = np.linalg.norm(self.data[:, :, 1:3], axis=2)
        p_harm = np.linalg.norm(self.data[:, :, 3:], axis=2)

        return np.divide(p_harm, p_base)

    def get_samples(self, n_samples=None) -> np.ndarray:
        """
        Get samples over one period.

        Parameters
        ----------
        n_samples : int
            number of samples

        Returns
        -------
        samples : np.array
            samples for each signal

        Examples
        --------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # create third order sine with magnitude 2
           tp = pm.PyTPeriodicMatrix(3, 2, -np.pi/2)

           # get 1000 samples of signal
           samples = tp.get_samples(samples=1000).flatten()

           plt.plot(samples)
           plt.grid()
           plt.ylim(np.min(samples), np.max(samples))
           plt.show()


        """

        if n_samples is None:
            n_samples = int(self.nbins() * self.resolution)

        n = self._data.shape[2]

        return irfft(self._data, n_samples) / n * n_samples

    def value_at(self, phi: np.array) -> np.array:
        """
        Get values of signals at specific angle phi.

        Parameters
        ----------
        phi : float or numpy.ndarray
            Radian Angle to query value of signals in.

        Returns
        -------
        np.array
            Values of signals at radian angle phi.

        Example
        -------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2 * np.pi, 100, endpoint=False)
           samples = np.empty((2, 3, 100), dtype=np.float)

           samples[0, 0] = np.zeros_like(phi)
           samples[0, 1] = 1.5 * np.ones_like(phi)
           samples[0, 2] = 1.5 + 2 * np.cos(phi)

           samples[1, 0] = 1.5 + 2 * np.cos(phi) - 0.5 * np.sin(2 * phi)
           samples[1, 1] = 2 * np.cos(phi) - 0.5 * np.sin(3 * phi)
           samples[1, 2] = 1.5 + 2 * np.cos(phi) + 1e-6 * np.cos(7 * phi)

           tp = pm.PyTPeriodicMatrix(samples, mode='samples')

           print('values at angle pi:')
           print(tp.value_at(np.pi))

        .. testoutput::

           values at angle pi:
           [[[ 0. ]
             [ 1.5]
             [-0.5]]

            [[-0.5]
             [-2. ]
             [-0.5]]]


        """
        phi = np.atleast_1d(phi) * self.scaling_factor
        if phi.ndim > 1:
            raise ValueError("too many dimensions of array phi!")
        n = phi.shape[0]

        (dc, cos_coeffs, sin_coeffs) = self.get_coeffs()

        # copy dc along axis 2
        res = np.tile(dc[:, :, np.newaxis], (1, 1, n))

        orders = np.arange(1, self.get_max_order() + 1)

        # create matrix of shape (nsamples, norders)
        sin_mat = np.sin(np.outer(phi, orders))
        cos_mat = np.cos(np.outer(phi, orders))

        """
        axes:
        "s"... samples
        "o"... order
        "r"... row
        "c"... column
        """
        res += np.einsum("so,rco->rcs", sin_mat, sin_coeffs)
        res += np.einsum("so,rco->rcs", cos_mat, cos_coeffs)

        return res

    def value_at_pos(self, phi, row, col) -> float:
        """
        Get value of single signal at position 'row', 'col' in signal matrix.

        Parameters
        ----------
        phi : float
            Angle to query value of signal
        row : int
            Row of signal in signal matrix
        col : int
            Column of signal in signal matrix

        Returns
        -------
        float
            Value of signal at angle phi

        Example
        -------

        .. testcode::

            import numpy as np
            import matplotlib.pyplot as plt
            import lcm.utils.pytperiodic as pm

            phi = np.linspace(0, 2 * np.pi, 100, endpoint=False)
            samples = np.empty((2, 3, 100), dtype=np.float)

            samples[0, 0] = np.zeros_like(phi)
            samples[0, 1] = 1.5 * np.ones_like(phi)
            samples[0, 2] = 1.5 + 2 * np.cos(phi)

            samples[1, 0] = 1.5 + 2 * np.cos(phi) - 0.5 * np.sin(2 * phi)
            samples[1, 1] = 2 * np.cos(phi) - 0.5 * np.sin(3 * phi)
            samples[1, 2] = 1.5 + 2 * np.cos(phi) + 1e-6 * np.cos(7 * phi)

            tp = pm.PyTPeriodicMatrix(samples, mode='samples')

            # first row first column
            print(tp.value_at_pos(np.pi, 0, 0))

            # first row second column
            print(tp.value_at_pos(np.pi, 0, 1))

            # second row second column
            print(tp.value_at_pos(np.pi, 1, 1))

        Output:

        .. testoutput::

            0.0
            1.5
            -2.0

        """
        return self.value_at(phi)[:, :, 0][row, col]

    def get_fft(self) -> np.array:
        """
        Get complex fft vector of signals in signal matrix.

        The class PyTPeriodicMatrix uses the scipy.fftpack.rfft-representation
        (https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.fftpack.rfft.html)
        for storing frequency domain data. This function converts the scipy.fftpack.rfft entries
        to a representation according to scipy.fftpack.fft
        (https://docs.scipy.org/doc/scipy-0.18.1/reference/generated/scipy.fftpack.fft.html#scipy.fftpack.fft).

        Returns
        -------
        fft : complex np.array
            fft vector for each signal

        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # second order cosine
           tp = pm.PyTPeriodicMatrix(2, 1, 0)

           # rfft representation
           print('rfft:')
           print(tp.data)

           # fft representation
           print('fft:')
           print(tp.get_fft())

        Output:

        .. testoutput::

           rfft:
           [[[ 0.   0.   0.   2.5  0. ]]]
           fft:
           [[[ 0.0+0.j  0.0+0.j  2.5+0.j  2.5+0.j  0.0+0.j]]]


        """
        n = self.nbins()
        res = np.zeros_like(self._data, dtype=np.complex)

        res[0] = self._data[0]

        reals = self._data[:, :, 1::2]
        imags = self._data[:, :, 2::2]

        if n % 2:
            res[:, :, 1:n // 2 + 1] = reals
            res[:, :, (n + 1) // 2:] = reals[:, :, ::-1]
            res[:, :, 1:n // 2 + 1] += 1j * imags
            res[:, :, n // 2 + 1:] -= 1j * imags[:, :, ::-1]
        else:
            res[:, :, 1:n // 2 + 1] = reals
            res[:, :, n // 2:] = reals[:, :, ::-1]
            res[:, :, 1:n // 2] += 1j * imags
            res[:, :, n // 2 + 1:] -= 1j * imags[:, :, ::-1]

        return res

    def get_magnitudes(self) -> np.array:

        """
        Get magnitudes of harmonic content.
        Returned array equals single sided amplitude spectrum without dc-value.

        Returns
        -------

        magnitudes : np.array
            array of harmonic magnitudes going from first order to maxorder. Array doesn't contain
            dc-values.


        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2 * np.pi, 1000, endpoint=False)

           # fourier approximation rectangle function sig = 1.5 + 1 * np.sin(phi) + (1 / 3) * np.sin(3 * phi) + (1 /
           5) * np.sin(5 * phi) + (1 / 7) * np.sin(7 * phi) + (1 / 9) * np.sin(9 * phi) tp = pm.PyTPeriodicMatrix(
           phi, sig, mode='fit')

           print(tp.get_magnitudes())

           magnitudes = tp.get_magnitudes()
           # make 1d array
           magnitudes = magnitudes.flatten()

           first_ord_magn = magnitudes[0]
           second_ord_magn = magnitudes[1]
           third_ord_magn = magnitudes[2]

           print(third_ord_magn)

        Output:

        .. testoutput::

           0.333333333333



        """
        (dc, cos_coeffs, sin_coeffs) = self.get_coeffs()
        amplitudes = np.sqrt(cos_coeffs ** 2 + sin_coeffs ** 2)
        return amplitudes

    def get_phases(self) -> np.array:
        """
        Get phases of harmonic content.
        Returned array gives the phases as np.array starting from the base harmonic

        Returns
        -------

        phases : np.array
            array of harmonic phases going from first order to maxorder.


        Examples
        --------


        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2 * np.pi, 1000, endpoint=False)

           # fourier approximation rectangle function sig = 1.5 + 1 * np.sin(phi) + (1 / 3) * np.sin(3 * phi) + (1 /
           5) * np.sin(5 * phi) + (1 / 7) * np.sin(7 * phi) + (1 / 9) * np.sin(9 * phi) tp = pm.PyTPeriodicMatrix(
           phi, sig, mode='fit')

           print(tp.get_phases())

        Output:

        .. testoutput::

           [[[-3.141592, 0, -3.141592, 0, -3.141592, 0, -3.141592, 0, -3.141592]]] 

        """
        (dc, cos_coeffs, sin_coeffs) = self.get_coeffs()
        phases = -np.arctan2(sin_coeffs, cos_coeffs)
        return phases

    def get_max_order(self) -> int:
        """
        Get maximum harmonic order of signals in signal matrix.

        Returns
        -------
        int
            maximum order

        Example
        -------

        .. testcode::
           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2*np.pi, 1000, endpoint='false')
           sig = 1 + 2 * np.sin(8*phi) + 10 * np.cos(11*phi)

           tp = pm.PyTPeriodicMatrix(phi, sig, mode='fit')

           print('Max order:')
           print(tp.get_max_order())

        Output:

        .. testoutput::

           Max order:
           11



        """
        return self.nbins() // 2

    def get_coeffs(self, orders=None) -> tuple:
        """
        Get cosine and sine coefficients of each signal in signal matrix.

        Returns
        -------
        tuple
            dc : np.array
                dc - value of each signal
            cos_coeffs : np.array
                cosine coefficients of each signal
            sin_coeffs : np.array
                sine coefficients of each signal

        Example
        -------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2*np.pi, 1000, endpoint=False)
           sig = 1 + 5 * np.sin(2*phi) + 10 * np.sin(3*phi) + 15 * np.cos(4*phi) + 20 * np.cos(5*phi)
           tp = pm.PyTPeriodicMatrix(phi, sig, mode='fit')

           dc, cos_coeffs, sin_coeffs = tp.get_coeffs()

           print('dc:')
           print(dc)
           print('cos coeffs:')
           print(cos_coeffs)
           print('sin_coeffs:')
           print(sin_coeffs)

        .. testoutput::

           dc:
           [[ 1.]]
           cos coeffs:
           [[[ -0.  -0.   0.  15.  20.]]]
           sin_coeffs:
           [[[-0.  5. 10. -0. -0.]]]


        """

        self.expand_to_odd()
        n = self.nbins()

        dc = self._data[:, :, 0] / n
        cos_coeffs = 2. / n * self._data[:, :, 1::2]
        sin_coeffs = -2. / n * self._data[:, :, 2::2]

        if orders is not None:

            # get list of orders
            try:
                o = list(set(orders))
            except:
                o = [orders]

            # remove orders < 1
            o = [item for item in o if item > 0]

            if o:

                maxorder = max(o)

                # index array
                idx = np.array(o) - 1

                # zero pad if needed
                if maxorder > self.get_max_order():
                    npad = max(o) - self.get_max_order()  # number of zeros to pad to third axis

                    cos_coeffs = np.pad(cos_coeffs, ((0, 0), (0, 0), (0, npad)), 'constant')
                    sin_coeffs = np.pad(sin_coeffs, ((0, 0), (0, 0), (0, npad)), 'constant')

                # truncate if needed
                cos_coeffs = cos_coeffs[:, :, :maxorder]
                sin_coeffs = sin_coeffs[:, :, :maxorder]

                mask = np.zeros_like(cos_coeffs)
                mask[:, :, idx] = 1.

                cos_coeffs *= mask
                sin_coeffs *= mask

            else:
                sh = list(cos_coeffs.shape)
                sh[2] = 0
                cos_coeffs = np.empty(sh)
                sin_coeffs = np.empty(sh)

        return dc, cos_coeffs, sin_coeffs

    def select(self, orders, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Select specified orders of signals.

        Parameters
        ----------
        orders : integer, list or set of integers
            only selected harmonics will be returned
            does not affect dc offset!
        nopack : bool
            don't use pack, False (Default)

        Returns
        -------
        ret : PyTPeriodicMatrix
            Resulting object

        Example
        -------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2*np.pi, 1000, endpoint=False)
           sig = 1 + 2 * np.sin(2*phi) + 3 * np.sin(3*phi) + 4 * np.sin(4*phi) + 5 * np.sin(5*phi) + 6 * np.sin(6*phi)
           tp = pm.PyTPeriodicMatrix(phi, sig, mode='fit')

           harmonics = [5, 6]

           dc1, cos_coeffs1, sin_coeffs1 = tp.get_coeffs()
           dc2, cos_coeffs2, sin_coeffs2 = tp.select(harmonics).get_coeffs()

           print('all harmonics:')
           print(sin_coeffs1)
           print('selected harmonics:')
           print(sin_coeffs2)

        .. testoutput::

           all harmonics:
           [[[-0.  2.  3.  4.  5.  6.]]]
           selected harmonics:
           [[[-0.  0.  0.  0.  5.  6.]]]


        """

        dc, cos_coeffs, sin_coeffs = self.get_coeffs(orders)
        ret = PyTPeriodicMatrix((dc, cos_coeffs, sin_coeffs), mode="coeffs", nopack=nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def pack(self) -> None:
        """
        Remove close to zero coefficient and truncate trailing zeros.

        Example
        -------



        """

        if not self.nopack:
            rms = self.rms()

            if np.linalg.norm(rms) > 0.:

                tres = rms * np.sqrt(PyTPeriodicMatrix.powerThreshold) * self.nbins() / 2.

                mask = np.where(np.abs(self._data) > tres[:, :, np.newaxis], 1., 0.)
                self._data *= mask  # zero out very small entries

                max_idx = np.max(np.nonzero(mask)[2])  # find max index of nonzero elements along axis 2

                n = self.nbins()
                nnew = max_idx + 1

                ''' for convenience reasons, when working with fourier coefficients, make odd nr of bins  '''
                if nnew % 2 == 0:
                    nnew += 1

                self._data = self._data[:, :, :nnew] * nnew / n

            else:
                self._data = np.zeros((self.nrows(), self.ncols(), 1))

    def expand_to_odd(self, newbins=0) -> None:
        """
        Expand to odd number of rfft-entries.

        Parameters
        ----------
        newbins: int

        Example
        -------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           rfft_vec = np.array([[[0., 0., 0., 0., 0., 0.],
                [10.5, 0., 0., 0., 0., 0.],
                [10.5, 7., -0., 0., 0., 0.]],
               [[10.5, 7., -0., 0., 1.75, 0.],
                [-0., 7., -0., -0., -0., 0.],
                [10.5, 7., -0., 0., 0., 0.]]])

           tp = pm.PyTPeriodicMatrix(rfft_vec, mode='rfft')
           print('before:')
           print(tp.data)
           tp.expand_to_odd(10)
           print('after:')
           print(tp.data)

        .. testoutput::

            before:
            [[[ 0.          0.          0.          0.          0.        ]
              [ 8.75        0.          0.          0.          0.        ]
              [ 8.75        5.83333333 -0.          0.          0.        ]]

             [[ 8.75        5.83333333 -0.          0.          1.45833333]
              [-0.          5.83333333 -0.         -0.         -0.        ]
              [ 8.75        5.83333333 -0.          0.          0.        ]]]
            after:
            [[[  0.           0.           0.           0.           0.           0.
                 0.           0.           0.           0.           0.        ]
              [ 19.25         0.           0.           0.           0.           0.
                 0.           0.           0.           0.           0.        ]
              [ 19.25        12.83333333  -0.           0.           0.           0.
                 0.           0.           0.           0.           0.        ]]

             [[ 19.25        12.83333333  -0.           0.           3.20833333   0.
                 0.           0.           0.           0.           0.        ]
              [ -0.          12.83333333  -0.          -0.          -0.           0.
                 0.           0.           0.           0.           0.        ]
              [ 19.25        12.83333333  -0.           0.           0.           0.
                 0.           0.           0.           0.           0.        ]]]



        """

        n = self.nbins()

        if newbins == 0:
            newbins = n

        ''' for convenience reasons, when working with fourier coefficients, make odd nr of bins  '''
        if newbins % 2 == 0:
            newbins += 1

        if newbins < n:
            raise ValueError("newbins must be >= nbins")
        elif newbins > n:
            new_shape = list(self._data.shape)
            new_shape[2] = newbins
            new = np.zeros(new_shape)
            new[:, :, 0:n] = self._data * newbins / n
            self._data = new
        else:
            pass  # leave it unchanged!

    def transform_order(self, p=1, q=1, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Transform frequency. Source object remains unchanged! new_order =  order*p/q.

        Parameters
        ----------
        p : float
            factor to multiply frequency with
        q : float
            factor to divide frequency by
        nopack : bool
            whether to perform packing or not

        Returns
        -------
        res : Transformed signal
        """

        self.pack()

        nrows = self.nrows()
        ncols = self.ncols()
        orders = self.get_max_order()
        (dc, cos_coeffs, sin_coeffs) = self.get_coeffs()

        '''
        Frequency multiplication
        '''
        cos_coeffs2 = np.zeros((nrows, ncols, orders * p))
        sin_coeffs2 = np.zeros((nrows, ncols, orders * p))

        cos_coeffs2[:, :, (p - 1)::p] = cos_coeffs[:, :, :]
        sin_coeffs2[:, :, (p - 1)::p] = sin_coeffs[:, :, :]

        # TODO siehe Fehlermeldung:
        r'''T:\MagOpt\System\General\Python\src\lcm\utils\pytperiodic.py:700: VisibleDeprecationWarning: using a 
        non-integer number instead of an integer will result in an error in the future cos_coeffs2 = np.zeros((nrows, 
        ncols, orders * p)) T:\MagOpt\System\General\Python\src\lcm\utils\pytperiodic.py:701: 
        VisibleDeprecationWarning: using a non-integer number instead of an integer will result in an error in the 
        future sin_coeffs2 = np.zeros((nrows, ncols, orders * p)) 
        T:\MagOpt\System\General\Python\src\lcm\utils\pytperiodic.py:703: VisibleDeprecationWarning: using a 
        non-integer number instead of an integer will result in an error in the future cos_coeffs2[:, :, (p - 1)::p] 
        = cos_coeffs[:, :, :] T:\MagOpt\System\General\Python\src\lcm\utils\pytperiodic.py:704: 
        VisibleDeprecationWarning: using a non-integer number instead of an integer will result in an error in the 
        future sin_coeffs2[:, :, (p - 1)::p] = sin_coeffs[:, :, :] '''

        '''
        Frequency division
        '''
        cos_coeffs3 = cos_coeffs2[:, :, (q - 1)::q]
        sin_coeffs3 = sin_coeffs2[:, :, (q - 1)::q]

        ret = PyTPeriodicMatrix((dc, cos_coeffs3, sin_coeffs3), mode="coeffs", nopack=nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def remove_dc(self, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Remove dc-value of signals.

        Returns
        -------
        ret : PyTPeriodicMatrix

        Example
        -------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2*np.pi, 500, endpoint=True)
           tp = pm.PyTPeriodicMatrix(phi, 5 + np.sin(phi), mode='fit')
           tp.remove_dc().show_preview()
           plt.show()


        """
        d = copy.deepcopy(self._data)
        d[:, :, 0] = 0.
        ret = PyTPeriodicMatrix(d, mode="rfft", nopack=nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def remove_ac(self, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Remove ac-value of signals.

        Returns
        -------
        ret : PyTPeriodicMatrix

        Example
        -------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2*np.pi, 1000, endpoint=False)
           sig = 5 + 5 * np.sin(phi) + 5 * np.sin(2*phi) + 5 * np.sin(3*phi) + 5 * np.cos(4*phi) + 5 * np.cos(5*phi)
           tp = pm.PyTPeriodicMatrix(phi, sig, mode='fit')

           # spectrum of original signal
           tp.show_spectrum()
           plt.show()

           # spectrum after removing ac components
           tp.remove_ac().show_spectrum()
           plt.show()

        """
        d = np.zeros_like(self._data)
        d[:, :, 0] = self._data[:, :, 0]
        ret = PyTPeriodicMatrix(d, mode="rfft", nopack=nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def det(self, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Calculates the pointwise determinante of PyTPeriodicMatrix for nxn Matrices.

        Returns
        -------
        ret: PyTPeriodicMatrix with shape = (1,1)


        """

        if self.nrows() != self.ncols():
            raise ValueError('Matrix must be quadratic')

        samples = self.get_samples()
        tmp = np.linalg.det(samples.transpose(2, 1, 0))

        ret = PyTPeriodicMatrix(tmp[np.newaxis, np.newaxis, :], mode='samples', nopack=nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def inv(self, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Calculates the pointwise inverse of PyTPeriodicMatrix for nxn Matrices.

        Returns
        -------
        ret: PyTPeriodicMatrix

        """

        if self.nrows() != self.ncols():
            raise ValueError('Matrix must be quadratic')

        n = 1 + 2 * self.get_max_order()
        samples = self.get_samples(n)
        tmp = np.linalg.inv(samples.transpose(2, 1, 0))
        ret = PyTPeriodicMatrix(tmp.transpose(2, 1, 0), mode='samples', nopack=nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def find_zerocrossing(self, row=0, col=0, **kwargs):
        """
        Find zero crossings of signal during one period.

        Parameters
        ----------
        row : int
            row of signal in signal matrix
        col : int
            column of signal in signal matrix
        **kwargs : mode
            'all' :   find all zero crossings
            'guess' : find zero crossing near guess using secant method
            guess : float
            initial guess in interval [0,2*pi]

        Returns
        -------
        crossing :  float
            in mode='guess'
        crossings : np.array
            in mode='all'

        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # create first order sine shifted by 10 degree to the right
           shift_angle = 10
           tp = pm.PyTPeriodicMatrix(1, 1, -np.pi/2 - shift_angle * np.pi / 180)
           crossings = tp.find_zerocrossing(mode='all')
           print(crossings * 180 / np.pi)

        Output:

        .. testoutput::

           [  10.  190.]

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           npts = 20
           phi = np.linspace(0, 2 * np.pi, npts, endpoint=False)

           samples = np.empty((2, 3, npts), dtype=np.float)

           samples[0, 0] = np.zeros_like(phi)
           samples[0, 1] = 1.5 * np.ones_like(phi)
           samples[0, 2] = 1.5 + 2 * np.cos(phi)

           samples[1, 0] = 1.5 + 2 * np.cos(phi) - 0.5 * np.sin(2 * phi)
           samples[1, 1] = 2 * np.cos(phi) - 0.5 * np.sin(3 * phi)
           samples[1, 2] = 1.5 + 2 * np.cos(phi) + 1e-6 * np.cos(7 * phi)

           tp = pm.PyTPeriodicMatrix(samples, mode='samples')

           crossings = tp.find_zerocrossing(0, 0)
           print(crossings)
           crossings = tp.find_zerocrossing(0, 1, mode='all')
           print(crossings)
           crossings = tp.find_zerocrossing(0, 2)
           print(crossings)
           crossings = tp.find_zerocrossing(1, 0)
           print(crossings)
           crossings = tp.find_zerocrossing(1, 1)
           print(crossings)
           crossings = tp.find_zerocrossing(1, 2)
           print(crossings)

        Output:

        .. testoutput::

           []
           []
           [ 2.41885841  3.8643269 ]
           [ 2.75283098  4.15819805]
           [ 1.77623977  4.91783242]
           [ 2.41885841  3.8643269 ]

        """
        if kwargs.get("mode"):
            mode = kwargs["mode"]
        else:
            mode = "all"

        if mode == "all":
            # find all zero crossings in interval ]0 2*pi] using brent method
            samples = self.get_samples()
            vec = samples[row, col, :]
            positive = vec >= 0
            guesses = np.where(np.bitwise_xor(positive[1:], positive[:-1]))[0]
            if np.sign(vec[-1]) != np.sign(vec[0]):  # handle zero cross from first to last element
                guesses = np.append(guesses, len(vec) - 1)

            crossings = np.zeros(len(guesses))

            for i, guess in enumerate(guesses):
                k = 1
                bound_a = (guess - 1) * 2 * np.pi / len(vec)
                bound_b = (guess + 1) * 2 * np.pi / len(vec)
                val_a = self.value_at_pos(bound_a, row, col)
                val_b = self.value_at_pos(bound_b, row, col)
                while k < len(vec) and ((val_a > 0 and val_b > 0) or (val_a < 0 and val_b < 0)):
                    k += 1
                    bound_a = (guess - k) * 2 * np.pi / len(vec)
                    bound_b = (guess + k) * 2 * np.pi / len(vec)
                    val_a = self.value_at_pos(bound_a, row, col)
                    val_b = self.value_at_pos(bound_b, row, col)

                crossings[i] = brentq(self.value_at_pos, bound_a, bound_b, args=(row, col))

            return crossings

        elif mode == "guess":
            # find zero cross with initial guess using secant method
            if 'guess' not in kwargs.keys():
                raise ValueError("in mode: ' + str(mode) + ' a initial guess must be provided ")

            else:
                guess = kwargs.get('guess')

            if (guess <= 0) or (guess >= 2 * np.pi):
                raise ValueError("guess must be number in interval [0, 2*pi]")

            crossing = newton(self.value_at_pos, guess, args=(row, col))

            return crossing

        else:
            raise ValueError("unknown mode: '" + str(mode) + "'")

    def flip(self, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Mirror signals around pi.

        Returns
        -------
        ret : PyTPeriodicMatrix

        Example
        -------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2 * np.pi, 1000, endpoint=False)

           # fourier approximation sawtooth function sig = 1 * np.sin(phi) - (1 / 2) * np.sin(2 * phi) + (1 / 3) *
           np.sin(3 * phi) - (1 / 4) * np.sin(4 * phi)+ (1 / 5) * np.sin(5 * phi) - (1 / 6) * np.sin(6 * phi) + (1 /
           7) * np.sin(7 * phi) - (1 / 8) * np.sin(8 * phi) + (1 / 9) * np.sin(9 * phi) tp = pm.PyTPeriodicMatrix(
           phi, sig, mode='fit')

           # original signal
           tp.show_preview()
           plt.show()

           # flipped signal
           tp.flip().show_preview()
           plt.show()

        """
        (dc, cos_coeffs, sin_coeffs) = self.get_coeffs()
        ret = PyTPeriodicMatrix((dc, cos_coeffs, -sin_coeffs), mode="coeffs", nopack=nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def shift(self, angle, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Shift signals by angle, negative angle means shift to right.

        Parameters
        ----------
        angle : float
            angle to shift signal
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        ret: PyTPeriodicMatrix

        Example
        -------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           angle = -45

           # create first order sine
           tp = pm.PyTPeriodicMatrix(1, 1, -np.pi/2)
           tp_shifted = tp.shift(angle * np.pi / 180)

           tp.show_preview()
           plt.show()

           tp_shifted.show_preview()
           plt.show()

        """
        (dc, cos_coeffs, sin_coeffs) = self.get_coeffs()
        max_order = self.get_max_order()
        list_angle = np.arange(1, max_order + 1) * angle
        cos_coeffs2 = cos_coeffs * np.cos(-list_angle) - sin_coeffs * np.sin(-list_angle)
        sin_coeffs2 = cos_coeffs * np.sin(-list_angle) + sin_coeffs * np.cos(-list_angle)

        ret = PyTPeriodicMatrix((dc, cos_coeffs2, sin_coeffs2), mode="coeffs", nopack=nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def deriv(self, omega, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Get derivative of signals with angular frequency omega.

        Parameters
        ----------
        omega : float
            angular frequency of signal
        nopack : bool
            whether to perform packing or not, False (Default)

        Returns
        -------
        ret : PyTPeriodicMatrix

        Examples
        --------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # angular frequency
           omega = 10

           # create first order cosine
           tp = pm.PyTPeriodicMatrix(1,1,0)

           # get derivative of cosine
           tp.deriv(omega).show_preview()
           plt.title('Signal Derivative, omega = {} rad / s'.format(omega))
           plt.show()


           
        """
        if isinstance(omega, np.ndarray):
            if omega.ndim == 2:
                ''' introduce a new dimension'''
                omega = omega[:, np.newaxis, :]
            elif omega.ndim > 2:
                raise Exception("Maximum supported number of omega dimensions is 2!")

        dc, cos_coeffs, sin_coeffs = self.get_coeffs()

        omega_order = omega * np.arange(1, self.get_max_order() + 1)

        ret = PyTPeriodicMatrix((np.zeros_like(dc),
                                 sin_coeffs * omega_order,
                                 -cos_coeffs * omega_order), mode="coeffs", nopack=nopack)

        ret.scaling_factor = self.scaling_factor
        return ret

    def integ(self, omega, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Get integral of signals with angular frequency omega.
        Any Dc-offset is set to zero before integration.

        Parameters
        ----------
        omega : float
            angular frequency of signal
        nopack : bool
            whether to perform packing or not, False (Default)
        Returns
        -------
        ret : PyTPeriodicMatrix

        Examples
        --------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # angular frequency
           omega = 10

           # create first order cosine
           tp = pm.PyTPeriodicMatrix(1,1,0)

           # get integral of cosine
           tp.integ(omega).show_preview()
           plt.title('Signal Integral, omega = {} rad / s'.format(omega))
           plt.show()

        """

        dc, cos_coeffs, sin_coeffs = self.get_coeffs()

        omega_order = omega * np.arange(1, self.get_max_order() + 1)

        ret = PyTPeriodicMatrix((np.zeros_like(dc),
                                 -sin_coeffs / omega_order,
                                 cos_coeffs / omega_order), mode="coeffs", nopack=nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def tf(self, b, a, omega, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Get response on transfer function described by numerator and denominator.
        For definition of numerator and denominator see scipy.signal.freqs method:
        https://docs.scipy.org/doc/scipy-0.13.0/reference/generated/scipy.signal.freqs.html


        Parameters
        ----------
        b : np.array
            numerator
        a : np.array
            denominator
        omega : float
            radian frequency of signal
        nopack : bool
            whether to perform packing or not, False (Default)

        Returns
        -------
        ret : PyTPeriodicMatrix
            response signal on transfer function

        Example
        -------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm
           from scipy.signal import freqs

           # first order lowpass
           num = np.array([1])
           den = np.array([100, 1])

           # bode plot
           w, h = freqs(num, den)
           fig, axes = plt.subplots(2, sharex=True)
           axes[0].semilogx(w / (2 * np.pi), 20 * np.log10(abs(h)))
           axes[0].set_ylabel('magnitude / dB')
           axes[0].set_xlim(np.min(w / (2*np.pi)), np.max(w / (2*np.pi)))
           axes[0].grid()
           axes[1].semilogx(w / (2 * np.pi), 180 * np.angle(h) / np.pi)
           axes[1].set_xlabel('frequency / Hz')
           axes[1].set_ylabel('phase / degree')
           axes[1].grid()
           plt.show()

           omega = 100
           phi = np.linspace(0, 2 * np.pi, 1000, endpoint='false')
           tp = pm.PyTPeriodicMatrix(phi, np.sin(phi), mode='fit')
           tp_response = tp.tf(num, den, omega)
           tp_response.show_preview()
           plt.show()

        """
        dc, cos_coeffs, sin_coeffs = self.get_coeffs()

        omega_order = omega * np.arange(1, self.get_max_order() + 1)

        w, h = freqs(b, a, omega_order)
        mag = np.absolute(h)
        phase = np.angle(h)

        cos_coeffs1 = mag * (cos_coeffs * np.cos(-phase) - sin_coeffs * np.sin(-phase))
        sin_coeffs1 = mag * (cos_coeffs * np.sin(-phase) + sin_coeffs * np.cos(-phase))

        ret = PyTPeriodicMatrix((dc * b[-1] / a[-1], cos_coeffs1, sin_coeffs1), mode="coeffs", nopack=nopack)

        ret.scaling_factor = self.scaling_factor
        return ret

    def enclosed_area(self, other) -> np.array:
        """
        Calculate enclosed area using Leibnitz sector formula
        (https://de.wikipedia.org/wiki/Sektorformel_von_Leibniz).
        Shapes of signal matrices must match.

        Parameters
        ----------
        other : PyTPeriodicMatrix

        Returns
        -------
        ret : np.array
            enclosed area of signals

        Notes
        -----

        The formula for the enclosed area :math:`A` is given by:

        .. math:: A = \\frac{1}{2} \\int_{0}^{2\\pi} (x(\\omega t) \\dot{y}(\\omega t) - y(\\omega t) \\dot{x}(
        \\omega t))\\; \\mathrm{d}(\\omega t)

        with :math:`x(\\omega t)`, :math:`y(\\omega t)` forming a parametrized curve in the :math:`xy`- plane:

        .. math::

           \\omega t \\rightarrow \\begin{cases}x(\\omega t)&\\\\y(\\omega t)&\\end{cases}

        and :math:`\\omega t \\in [0, 2\\pi]`.

        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # one circle radius 1
           tpx = pm.PyTPeriodicMatrix(1, 1, 0)
           tpy = pm.PyTPeriodicMatrix(1, 1, -np.pi / 2)
           area = tpx.enclosed_area(tpy)
           print(area)

        Output:

        .. testoutput::

           [[ 3.14159265]]


        """

        self.checkScalingFactor(other)

        if (self.nrows() != other.nrows()) or (self.ncols() != other.ncols()):
            raise ValueError('Matrix dimensions must agree')

        dx = self.deriv(1)
        dy = other.deriv(1)

        tmp = self.elementwise_mul(dy).__sub__(other.elementwise_mul(dx))

        return (1 / 2) * tmp.mean() * 2 * np.pi

    def curve2dPerimeter(self, other) -> np.array:
        """
        Calculate perimeter of curves.
        Shapes of signal matrices must match.

        Parameters
        ----------
        other : PyTPeriodicMatrix
            Signal for perimeter calculation

        Returns
        -------
        ret : np.array
            perimeter of signals

        Notes
        -----

        Calculation of perimeter :math:`s` is based on the formula:

        .. math::

           s = \\int_{0}^{2\\pi} \\sqrt{\\dot{x}^2 (\\omega t) + \\dot{y}^2 (\\omega t)}\\; \\mathrm{d} (\\omega t)

        with :math:`x(\\omega t)`, :math:`y(\\omega t)` forming a parametrized curve in the xy-plane:

        .. math::

           \\omega t \\rightarrow \\begin{cases}x(\\omega t)&\\\\y(\\omega t)&\\end{cases}

        and :math:`\\omega t \\in [0, 2\\pi]`.

        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           # one circle radius 1
           tpx = pm.PyTPeriodicMatrix(1, 1, 0)
           tpy = pm.PyTPeriodicMatrix(1, 1, -np.pi / 2)
           perimeter = tpx.curve2dPerimeter(tpy)
           print('circumference:')
           print(perimeter)

        Output:

        .. testoutput::

            circumference:
            [[ 6.28318531]]


        """
        if (self.nrows() != other.nrows()) or (self.ncols() != other.ncols()):
            raise ValueError('Matrix dimensions must agree')
        self.checkScalingFactor(other)

        dx = self.deriv(1)
        dy = other.deriv(1)

        tmp = (dx.elementwise_power(2) + dy.elementwise_power(2)).elementwise_power(1 / 2)

        return tmp.mean() * 2 * np.pi

    def elementwise_mul(self, other, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Elementwise multiplication.

        Shapes of signal matrices must match:

        --> self.shape = (nrows, ncols)
        --> other.shape = (nrows, ncols) or (nrows, 1)
        --> return.shape = (nrows, ncols)

        Parameters
        ----------
        other : PyTPeriodicMatrix, np.ndarray,
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false


        Returns
        -------
        ret : PyTPeriodicMatrix


        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           dc_vals1 = 3 * np.ones((3, 3), dtype=np.float)
           tp1 = pm.PyTPeriodicMatrix(dc_vals1, mode='dc')

           dc_vals2 = 2 * np.ones((3, 3), dtype=np.float)
           tp2 = pm.PyTPeriodicMatrix(dc_vals2, mode='dc')

           # elementwise multiplication of signal matrices
           tp_mul = tp1.elementwise_mul(tp2)
           dc_mul, sin_coeff, cos_coeff = tp_mul.get_coeffs()

           # matrix multiplication of signal matrices
           tp_mat = tp1 * tp2
           dc_mat, sin_coeff, cos_coeff = tp_mat.get_coeffs()

           print(dc_mul)
           print(dc_mat)

        Output:

        .. testoutput::

           result elementwise multiplication:
           [[ 6.  6.  6.]
            [ 6.  6.  6.]
            [ 6.  6.  6.]]
           result matrix multiplication:
           [[ 18.  18.  18.]
            [ 18.  18.  18.]
            [ 18.  18.  18.]]


        """
        if isinstance(other, PyTPeriodicMatrix):

            self.checkScalingFactor(other)

            if (self.nrows() != other.nrows()) or (self.ncols() != other.ncols() and 1 != other.ncols()):
                raise ValueError('elementwise_mul(): Matrix dimensions must agree')
            else:
                n = 1 + 2 * (self.get_max_order() + other.get_max_order())
                a = self.get_samples(n)
                b = other.get_samples(n)
                res = np.multiply(a, b)
                ret = PyTPeriodicMatrix(res, mode="samples", nopack=nopack)
                ret.scaling_factor = self.scaling_factor
                return ret

        elif isinstance(other, np.ndarray):
            if other.ndim != 2:
                raise ValueError('elementwise_mul(): Matrix should have two dimensions')

            if self.nrows() != other.shape[0] or (self.ncols() != other.shape[1] and 1 != other.shape[1]):
                raise ValueError('elementwise_mul(): Matrix dimensions must agree')

            a = self.get_samples()
            b = np.repeat(other[:, :, np.newaxis], a.shape[2], axis=2)
            res = np.multiply(a, b)
            ret = PyTPeriodicMatrix(res, mode='samples', nopack=nopack)
            ret.scaling_factor = self.scaling_factor
            return ret

        else:
            raise ValueError('elementwise_mul(): Unsupported matrix')

    def elementwise_div(self, other, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Elementwise multiplication.
        Shapes of signal matrices must match.

        Parameters
        ----------
        other : PyTPeriodicMatrix
        nopack : bool
            whether to perform packing or not, False (Default)


        Returns
        -------
        ret : PyTPeriodicMatrix


        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           dc_vals1 = 3 * np.ones((3, 3), dtype=np.float)
           tp1 = pm.PyTPeriodicMatrix(dc_vals1, mode='dc')

           dc_vals2 = 2 * np.ones((3, 3), dtype=np.float)
           tp2 = pm.PyTPeriodicMatrix(dc_vals2, mode='dc')


        Output:

        .. testoutput::


        """

        if isinstance(other, PyTPeriodicMatrix):
            self.checkScalingFactor(other)
            # if (self.nrows() != other.nrows()) or (self.ncols() != other.ncols() and 1 != other.ncols()):
            #     raise ValueError('elementwise_div(): Matrix dimensions must agree')
            #
            # else:
            n = 1 + 2 * (self.get_max_order() + other.get_max_order())
            a = self.get_samples(n)
            b = other.get_samples(n)
            res = np.divide(a, b)
            ret = PyTPeriodicMatrix(res, mode="samples", nopack=nopack)
            ret.scaling_factor = self.scaling_factor
            return ret

        elif isinstance(other, np.ndarray):
            if other.ndim != 2:
                raise ValueError('elementwise_div(): Matrix should have two dimensions')

            if self.nrows() != other.shape[0] or (self.ncols() != other.shape[1] and 1 != other.shape[1]):
                raise ValueError('elementwise_div(): Matrix dimensions must agree')

            a = self.get_samples()
            b = np.repeat(other[:, :, np.newaxis], a.shape[2], axis=2)
            res = np.divide(a, b)
            ret = PyTPeriodicMatrix(res, mode='samples', nopack=nopack)
            ret.scaling_factor = self.scaling_factor
            return ret

        else:
            raise ValueError('elementwise_div(): Unsupported matrix')

    def elementwise_power(self, p, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Raises every signal in signal matrix to power 'p'.

        Parameters
        ----------
        p : float
            power
        nopack : bool
            whether to perform packing or not, False (Default)

        Returns
        -------
        ret : PyTPeriodicMatrix

        Examples
        --------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           dc_vals = 4 * np.ones((3, 3), dtype=np.float)
           tp = pm.PyTPeriodicMatrix(dc_vals, mode='dc')

           dc_coeffs, cos_coeffs, sin_coeffs = tp.elementwise_power(1/2).get_coeffs()

           print('sqrt:')
           print(dc_coeffs)

        .. testoutput::

           sqrt:
           [[ 2.  2.  2.]
            [ 2.  2.  2.]
            [ 2.  2.  2.]]


        """
        if isinstance(p, PyTPeriodicMatrix):
            raise ValueError('elementwise_power(): unsupported argument')

        else:
            p = np.array(p)
            if p.ndim == 0:
                if (p % 1.) == 0.:
                    # integer power
                    if p == 0.:
                        d = np.atleast_3d(np.ones_like(self._data[:, :, 0]))
                        ret = PyTPeriodicMatrix(d, mode="rfft", nopack=nopack)
                        ret.scaling_factor = self.scaling_factor
                        return ret

                    elif p == 1.:
                        ret = PyTPeriodicMatrix(self._data, mode="rfft", nopack=nopack)
                        ret.scaling_factor = self.scaling_factor
                        return ret

                    elif p > 1.:
                        tmp = copy.deepcopy(self)
                        for k in range(1, p):
                            tmp = tmp.elementwise_mul(tmp)

                        return tmp

                else:
                    # non integer power
                    # calc pointwise power
                    tres = -1e-14
                    samples = self.get_samples()
                    if np.any(samples < tres):
                        raise ValueError('elementwise_power(): non real valued power encountered')
                    samples[samples < 0] = 0
                    y = np.power(samples, p)

                    ret = PyTPeriodicMatrix(y, mode="samples", nopack=nopack)
                    ret.scaling_factor = self.scaling_factor
                    return ret

            else:
                raise ValueError("elementwise_power(): unsupported argument")

    def skew(self, angle, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Calculates infinite, symmetric, linear skewing of the signal. The angle
        will be swept between -angle/2 and angle/2.
        Source object remains unchanged!

        Parameters
        ----------
        angle : float
            Skewing angle in radians
        nopack : bool
            whether to perform packing or not, False (Default)

        Returns
        -------
        result : PyTPeriodicMatrix
            skewed object

        """

        if np.isclose(angle, 0.):
            ret = copy.deepcopy(self)

        else:

            (dc, cos_coeffs, sin_coeffs) = self.get_coeffs()

            nue = np.arange(1., self.get_max_order() + 1.)
            tmp = angle * nue / 2.
            cos_coeffs = cos_coeffs * np.sin(tmp) / tmp
            sin_coeffs = sin_coeffs * np.sin(tmp) / tmp

            ret = PyTPeriodicMatrix((dc, cos_coeffs, sin_coeffs), mode="coeffs", nopack=nopack)

        ret.scaling_factor = self.scaling_factor
        return ret

    def transpose(self, nopack=False) -> 'PyTPeriodicMatrix':
        """
        Transpose object.

        Parameters
        ----------
        nopack : bool
            whether to perform packing or not, False (Default)

        Returns
        -------
        result : PyTPeriodicMatrix
            Transposed object

        Example
        -------

        .. testcode::

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           dc_vals = np.zeros((3, 3), dtype=np.float)
           dc_vals[0, -1] = 2
           dc_vals[-1, 0] = 1
           tp = pm.PyTPeriodicMatrix(dc_vals, mode='dc')
           dc_vals_transposed, sin_coeff, cos_coeff = tp.transpose().get_coeffs()

           print('matrix:')
           print(dc_vals)

           print('transposed matrix:')
           print(dc_vals_transposed)

        Output:

        .. testoutput::

          matrix:
          [[ 0.  0.  2.]
           [ 0.  0.  0.]
           [ 1.  0.  0.]]
          transposed matrix:
          [[ 0.  0.  1.]
           [ 0.  0.  0.]
           [ 2.  0.  0.]]


        """
        ret = PyTPeriodicMatrix(self._data.transpose(1, 0, 2), mode="rfft", nopack=nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def sum(self, axis=(0, 1)) -> 'PyTPeriodicMatrix':
        if axis == 0:
            data = np.sum(self._data, axis=0)[np.newaxis, :, :]
        elif axis == 1:
            data = np.sum(self._data, axis=1)[:, np.newaxis, :]
        elif axis == (0, 1):
            data = np.sum(self._data, axis=(0, 1))[np.newaxis, np.newaxis, :]
        else:
            raise Exception('Only axis 0 and 1 are supported')

        ret = PyTPeriodicMatrix(data, mode="rfft", nopack=self.nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def toObjArray(self) -> np.array:
        """

        Returns
        -------
        np.ndarray
        """
        tmp = copy.deepcopy(self)  # type: PyTPeriodicMatrix

        objarray = np.empty((tmp.nrows(), tmp.ncols()), dtype=np.object)
        for row in range(tmp.nrows()):
            for col in range(tmp.ncols()):
                objarray[row, col] = tmp[row, col]
        return objarray

    def to_msgpack(self) -> dict:
        """
        Create msgpack serializable dict representing an instance of this class
        ref: http://msgpack.org/

        Returns
        -------
        msgpack : dict
            msgpack representation

        Example
        -------

        .. testcode::

            import numpy as np
            import matplotlib.pyplot as plt
            import lcm.utils.pytperiodic as pm

            phi = np.linspace(0, 2 * np.pi, 100, endpoint=False)
            samples = np.sin(phi)

            tp = pm.PyTPeriodicMatrix(phi, samples, mode='fit')

            print(tp.to_msgpack())

        """

        tmp = copy.deepcopy(self)
        tmp.expand_to_odd()

        msgpack = {"nrows": tmp.nrows(),
                   "ncols": tmp.ncols(),
                   "nbins": tmp.nbins(),
                   "data": tmp._data.astype('>f8', order='C').tobytes()}

        return msgpack

    def toCBOR(self, f=1, nopack=False) -> dict:
        """
        Create CBOR serializable dict representing an instance of this class
        ref: http://cbor.io/

        Parameters
        ----------
        f : integer
            angular scaling factor, 1 (Default)
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false
        Returns
        -------
        cbor_obj : dict
            cbor representation


        Example
        -------

        .. testcode::

            import numpy as np
            import matplotlib.pyplot as plt
            import lcm.utils.pytperiodic as pm

            phi = np.linspace(0, 2 * np.pi, 100, endpoint=False)
            samples = np.sin(phi)

            tp = pm.PyTPeriodicMatrix(phi, samples, mode='fit')

            print(tp.toCBOR())

        """

        tmp = copy.deepcopy(self)  # type: PyTPeriodicMatrix
        tmp = tmp.transform_order(1, f, nopack=nopack)
        tmp.expand_to_odd()

        ''' create cbor serializable dict from 3d-array z '''
        tp_per = Serializable.toCBOR(tmp._data)

        ''' create cbor serializable dict '''
        cbor_obj = {'type': '__periodicMatrix__',
                    'bins': tp_per,
                    'f': f}

        return cbor_obj

    def saveCBOR(self, file, f=1):
        """
        Create CBOR file representing an instance of this class
        ref: http://cbor.io/


        Parameters
        ----------
        file : String
            path to CBOR file
        f : integer
            angular scaling factor, 1 (Default)
        """
        Serializable.saveCBOR(self, file, f)

    def getOrders(self, orders=None) -> np.array:
        if orders is not None:

            # get list of orders
            try:
                o = list(set(orders))
            except:
                o = [orders]

            # remove orders < 1
            o = [item for item in o if item >= 0]
            o = range(0, max(o) + 1)
        else:
            o = range(0, self.maxorder + 1)
        return np.array(o, dtype=int)

    def getSpectrum(self, orders=None) -> np.array:

        (dc, cos_coeffs, sin_coeffs) = self.get_coeffs(orders)
        amplitude = np.sqrt(cos_coeffs ** 2 + sin_coeffs ** 2)
        spectrum = np.zeros((self.nrows(), self.ncols(), 1 + amplitude.shape[2]))

        for row in range(0, self.nrows()):
            for col in range(0, self.ncols()):
                spectrum[row, col, :] = np.append(dc[row, col], amplitude[row, col, :])
        return spectrum

    # def insert(self, tp_single, index, axis=0):
    #     return insert(self, tp_single, index, axis)

    def __preparePlotSamples(self, unit='deg'):
        nsamples = int(self.nbins() * self.resolution)
        nsamples = max(nsamples, self.minPreviewSamples)

        phi = np.linspace(0, 2 * np.pi, nsamples)
        phi_unscaled = phi / self.scaling_factor
        samples = self.value_at(phi_unscaled)

        if unit == 'deg':
            phi = np.rad2deg(phi)

        return phi, samples

    def plot(self, fig, mode='matrix'):
        """Plots samples in custom figure which need to be provided as argument.

        Parameters
        ----------

        fig : matplotlib.pyplot.figure
            Matplotlib Figure Object
        mode : str (optional)
            'matrix'    plot one subplot per signal(default)
            'row'       plot each columns into one diagram with different colors
            'col'       plot each row into one diagram with different colors
            'single'    plot all signals into one diagram with different colors and line types

        Example
        -------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           tp = pm.PyTPeriodicMatrix(1, 5, -np.pi/4, mode='fit')

           fig = plt.figure(figsize=(3, 2), facecolor='b', frameon=True)
           tp.plot(fig)
           plt.show()

        """

        from matplotlib.ticker import FixedLocator, FormatStrFormatter

        phi, samples = self.__preparePlotSamples()

        if mode == 'single':
            ax = fig.gca()
            for row in range(0, self.nrows()):
                for col in range(0, self.ncols()):
                    ax.plot(phi * 180. / np.pi, samples[row, col, :])

            ax.grid(True)
            ax.xaxis.grid(True, 'minor')
            ax.xaxis.set_major_locator(FixedLocator(np.arange(0, 360, 90)))
            ax.xaxis.set_major_formatter(FormatStrFormatter(u'%d°'))
            ax.xaxis.set_minor_locator(FixedLocator(np.arange(0, 360, 30)))

        elif mode == 'row':
            for col in range(0, self.ncols()):
                ax = fig.add_subplot(1, self.ncols(), col + 1)

                for row in range(0, self.nrows()):
                    ax.plot(phi, samples[row, col, :])

                ax.grid(True)
                ax.xaxis.grid(True, 'minor')
                ax.xaxis.set_major_locator(FixedLocator(np.arange(0, 360, 90)))
                ax.xaxis.set_major_formatter(FormatStrFormatter(u'%d°'))
                ax.xaxis.set_minor_locator(FixedLocator(np.arange(0, 360, 30)))

        elif mode == 'col':
            for row in range(0, self.nrows()):
                ax = fig.add_subplot(self.nrows(), 1, row + 1)

                for col in range(0, self.ncols()):
                    ax.plot(phi, samples[row, col, :])

                ax.grid(True)
                ax.xaxis.grid(True, 'minor')
                ax.xaxis.set_major_locator(FixedLocator(np.arange(0, 360, 90)))
                ax.xaxis.set_major_formatter(FormatStrFormatter(u'%d°'))
                ax.xaxis.set_minor_locator(FixedLocator(np.arange(0, 360, 30)))

        else:
            cnt = 1
            for row in range(0, self.nrows()):
                for col in range(0, self.ncols()):
                    ax = fig.add_subplot(self.nrows(), self.ncols(), cnt)
                    cnt += 1

                    ax.plot(phi, samples[row, col, :])
                    ax.grid(True)
                    ax.xaxis.grid(True, 'minor')
                    ax.xaxis.set_major_locator(FixedLocator(np.arange(0, 360, 90)))
                    ax.xaxis.set_major_formatter(FormatStrFormatter(u'%d°'))
                    ax.xaxis.set_minor_locator(FixedLocator(np.arange(0, 360, 30)))

    def plot_spectrum(self, fig, orders=None, mode='matrix'):
        """Plots spectrum in custom figure which need to be provided as argument.

        Parameters
        ----------
        fig : matplotlib.pyplot.figure
            Matplotlib Figure Object
        orders : array_like

        mode : str (optional)
            'matrix'    plot one subplot per signal(default)
            'single'    plot all signals into one diagram with different colors and line types

        Example
        -------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           tp = pm.PyTPeriodicMatrix(1, 5, -np.pi/4, mode='fit')

           fig = plt.figure(figsize=(3, 2), facecolor='b', frameon=True)
           tp.plot_spectrum(fig)
           plt.show()

        """

        from matplotlib.ticker import MultipleLocator
        cnt = 1
        (dc, cos_coeffs, sin_coeffs) = self.get_coeffs(orders)
        amplitude = np.sqrt(cos_coeffs ** 2 + sin_coeffs ** 2)
        for row in range(0, self.nrows()):
            for col in range(0, self.ncols()):
                if mode == 'single':
                    ax = fig.gca()
                else:
                    ax = fig.add_subplot(self.nrows(), self.ncols(), cnt)
                cnt += 1
                x = range(0, 1 + amplitude[row, col, :].size)
                height = np.append(dc[row, col], amplitude[row, col, :])
                # ax.stem(np.append(dc[row, col], amplitude[row, col, :]))
                ax.bar(x, height)
                ax.set_axisbelow(True)
                ax.grid(True)
                ax.xaxis.grid(True, 'minor')
                ax.xaxis.set_major_locator(MultipleLocator(1))

    def plotMPL(self, fig, mode='matrix'):
        self.plot(fig, mode)

    def plotBKH(self, mode='matrix'):
        """
        Plot with bokeh library

        Parameters
        ----------
         mode : str (optional)
            'matrix'    plot one subplot per signal(default)
            'row'       plot each columns into one diagram with different colors
            'col'       plot each row into one diagram with different colors
            'single'    plot all signals into one diagram with different colors and line types


        .. bokeh-plot::

           import numpy as np
           import lcm.utils.pytperiodic as pm
           import bokeh.io as bkio

           tp = pm.PyTPeriodicMatrix(1, 5, -np.pi/4, mode='fit')

           fig = tp.plotBKH()
           bkio.show(fig)

        Returns
        -------
        fig : bokeh.layouts.LayoutDOM
        """
        from bokeh.models import ColumnDataSource
        from bokeh.plotting import figure
        import bokeh.layouts as bkl

        phi, samples = self.__preparePlotSamples()

        data = dict(phi=phi)
        for row in range(self.nrows()):
            for col in range(self.ncols()):
                data['{}x{}'.format(row, col)] = samples[row, col, :]

        source = ColumnDataSource(data=data)

        if mode == 'single':
            fig = figure()
            for row in range(self.nrows()):
                for col in range(self.ncols()):
                    fig.line(x='phi', y='{}x{}'.format(row, col), source=source)
        else:
            p = [[figure() for _ in range(self.ncols())] for _ in range(self.nrows())]
            for row in range(self.nrows()):
                for col in range(self.ncols()):
                    p[row][col].line(x='phi', y='{}x{}'.format(row, col), source=source)
            if mode == 'row':
                flat_p = [val for sublist in p for val in sublist]
                fig = bkl.row(flat_p)
            elif mode == 'col':
                flat_p = [val for sublist in p for val in sublist]
                fig = bkl.column(flat_p)
            else:
                fig = bkl.gridplot(p)
        return fig

    def plotPLY(self, mode='matrix', titles=None):
        """
        Plot with plotly library

        Parameters
        ----------
         mode : str (optional)
            'matrix'    plot one subplot per signal(default)
            'row'       plot each columns into one diagram with different colors
            'col'       plot each row into one diagram with different colors
            'single'    plot all signals into one diagram with different colors and line types


        .. plot::
           :include-source:

           import numpy as np
           import lcm.utils.pytperiodic as pm
           import plotly.offline as ply

           tp = pm.PyTPeriodicMatrix(1, 5, -np.pi/4, mode='fit')

           fig = tp.plotPLY()
           ply.show(fig)

        Returns
        -------
        fig : plotly.graph_objs.Figure
            Plotly figure
        """
        import plotly.graph_objs as go
        from plotly import tools

        if titles is None:
            titles = []

        phi, samples = self.__preparePlotSamples()

        if mode == 'single':
            fig = tools.make_subplots(rows=1, cols=1, subplot_titles=titles)
            for row in range(0, self.nrows()):
                for col in range(0, self.ncols()):
                    line = go.Scatter(
                        x=phi,
                        y=samples[row, col, :],
                        mode='lines',
                        name="[{:d}, {:d}]".format(row, col)
                    )
                    fig.add_trace(line, 1, 1)

            fig['layout']['xaxis'].update(dtick=90.0)

        elif mode == 'row':
            fig = tools.make_subplots(rows=1, cols=self.ncols(), subplot_titles=titles)
            for col in range(0, self.ncols()):
                for row in range(0, self.nrows()):
                    line = go.Scatter(
                        x=phi,
                        y=samples[row, col, :],
                        mode='lines',
                        name="[{:d}, {:d}]".format(row, col)
                    )
                    fig.add_trace(line, 1, col + 1)

        elif mode == 'col':
            fig = tools.make_subplots(rows=self.nrows(), cols=1, subplot_titles=titles)
            for row in range(0, self.nrows()):
                for col in range(0, self.ncols()):
                    line = go.Scatter(
                        x=phi,
                        y=samples[row, col, :],
                        mode='lines',
                        name="[{:d}, {:d}]".format(row, col)
                    )
                    fig.add_trace(line, row + 1, 1)

        else:
            fig = tools.make_subplots(rows=self.nrows(), cols=self.ncols(), subplot_titles=titles)
            for row in range(0, self.nrows()):
                for col in range(0, self.ncols()):
                    line = go.Scatter(
                        x=phi,
                        y=samples[row, col, :],
                        mode='lines',
                        name="[{:d}, {:d}]".format(row, col)
                    )
                    fig.add_trace(line, row + 1, col + 1)
        return fig

    def plotSpectrumPLY(self, orders=None, mode='matrix', titles=None):
        """Plots spectrum in custom figure which need to be provided as argument.

        Parameters
        ----------

        orders : array_like

        mode : str (optional)
            'matrix'    plot one subplot per signal(default)
            'single'    plot all signals into one diagram with different colors and line types
        titles : list

        Example
        -------

        .. plot::
           :include-source:

           import numpy as np
           import lcm.utils.pytperiodic as pm
           import plotly.offline as ply

           tp = pm.PyTPeriodicMatrix(1, 5, -np.pi/4, mode='fit')
           fig = tp.plotSpectrumPLY()
           ply.show(fig)

        Returns
        -------
        fig : plotly.graph_objs.Figure
            Plotly figure

        """
        if titles is None:
            titles = []

        subplot_titles = ['Spectrum {}'.format(title) for title in titles]

        import plotly.graph_objs as go
        from plotly import tools

        (dc, cos_coeffs, sin_coeffs) = self.get_coeffs(orders)
        amplitude = np.sqrt(cos_coeffs ** 2 + sin_coeffs ** 2)

        if mode == 'single':
            fig = tools.make_subplots(rows=1, cols=1, subplot_titles=subplot_titles)
            for row in range(0, self.nrows()):
                for col in range(0, self.ncols()):
                    height = np.append(dc[row, col], amplitude[row, col, :])
                    line = go.Bar(y=height, name="[{:d}, {:d}]".format(row, col))
                    fig.add_trace(line, 1, 1)
                    fig['layout']['xaxis'].update(dtick=1.0)

        elif mode == 'row':
            fig = tools.make_subplots(rows=1, cols=self.ncols(), subplot_titles=subplot_titles)
            for col in range(0, self.ncols()):
                for row in range(0, self.nrows()):
                    height = np.append(dc[row, col], amplitude[row, col, :])
                    line = go.Bar(y=height, name="[{:d}, {:d}]".format(row, col))
                    fig.add_trace(line, 1, col + 1)

        elif mode == 'col':
            fig = tools.make_subplots(rows=self.nrows(), cols=1, subplot_titles=subplot_titles)
            for row in range(0, self.nrows()):
                for col in range(0, self.ncols()):
                    height = np.append(dc[row, col], amplitude[row, col, :])
                    line = go.Bar(y=height, name="[{:d}, {:d}]".format(row, col))
                    fig.add_trace(line, row + 1, 1)

        else:
            fig = tools.make_subplots(rows=self.nrows(), cols=self.ncols(), subplot_titles=subplot_titles)
            for row in range(0, self.nrows()):
                for col in range(0, self.ncols()):
                    height = np.append(dc[row, col], amplitude[row, col, :])
                    line = go.Bar(y=height, name="[{:d}, {:d}]".format(row, col))
                    fig.add_trace(line, row + 1, col + 1)
        return fig

    def plotFullPreviewHTML(self, mode='matrix', orders=None, titles=None):
        import plotly.offline as ply

        fig1 = self.plotPLY(mode=mode, titles=titles)
        fig2 = self.plotSpectrumPLY(orders, mode=mode, titles=titles)

        div1 = ply.plot(fig1, auto_open=False, output_type='div')
        div2 = ply.plot(fig2, auto_open=False, output_type='div')

        html = '<html><body>' \
               '<table style="width:100%">' \
               '<col width = "50%" >' \
               '<col width = "50%" >' \
               '<tr><td>{}</td><td>{}</td></tr>' \
               '</body></html>'.format(div1, div2)
        return html

    def showPreviewPLY(self, mode='matrix', titles=None):
        import uuid
        import plotly.offline as ply
        fig = self.plotPLY(mode=mode, titles=titles)
        ply.plot(fig, filename='temp-{}.html'.format(uuid.uuid4()))

    def showFullPreviewPLY(self, mode='matrix', orders=None, titles=None):
        import webbrowser
        import tempfile
        import uuid

        html = self.plotFullPreviewHTML(mode, orders=orders, titles=titles)

        temp_file_name = os.path.join(tempfile.gettempdir(), 'temp-{}.html'.format(uuid.uuid4()))
        with open(temp_file_name, 'w') as fh:
            print(html, file=fh)

        webbrowser.open('file://' + temp_file_name, new=2)
        return temp_file_name

    def show_preview(self, mode='matrix'):
        """
        Show simple preview of the signal.
        If matplotlib is used for plotting additional keyword arguments given under parameters are supported.
        If called inside SyMSpace jep console keyword arguments are ignored.

        Parameters
        ----------
        mode: string
            'matrix'    plot one subplot per signal(default)
            'row'       plot each columns into one diagram with different colors
            'col'       plot each row into one diagram with different colors
            'single'    plot all signals into one diagram with different colors and line types

        Example
        -------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2 * np.pi, 1000, endpoint=False)

           # fourier approximation sawtooth function sig = 1 * np.sin(phi) - (1 / 2) * np.sin(2 * phi) + (1 / 3) *
           np.sin(3 * phi) - (1 / 4) * np.sin(4 * phi)+ (1 / 5) * np.sin(5 * phi) - (1 / 6) * np.sin(6 * phi) + (1 /
           7) * np.sin(7 * phi) - (1 / 8) * np.sin(8 * phi) + (1 / 9) * np.sin(9 * phi) tp = pm.PyTPeriodicMatrix(
           phi, sig, mode='fit')

           # time domain
           tp.show_preview()
           plt.show()


        """
        # Get the name of the current interpreter
        # interp = 'javaw' for execution in conjunction with symspace and jep
        # interp = 'python' for execution in cpython interpreter

        if self.interpreter == 'python':
            import matplotlib.pyplot as plt
            fig = plt.figure()
            self.plot(fig, mode)
            plt.draw()
        elif self.interpreter == 'javaw' or self.interpreter == 'java':
            self.__show_java_preview()
        else:
            # assume python interpreter
            import matplotlib.pyplot as plt
            fig = plt.figure()
            self.plot(fig, mode)
            plt.draw()

    def show_full_preview(self, orders=None):
        """
        Show full preview of signal including spectrum inside the SyMSpace jep environment.
        In case matplotlib is used for plotting from normal python interpreter
        the function defaults to show_preview().

        """
        if self.interpreter == 'python:':
            # self.show_preview()
            import matplotlib.pyplot as plt
            fig = plt.figure()
            plt.subplot(211)
            self.plot(fig, 'single')
            plt.subplot(212)
            self.plot_spectrum(fig, orders, 'single')
            plt.draw()
        elif self.interpreter == 'javaw' or self.interpreter == 'java':
            if self.shape == (1, 1):
                self.__show_java_full_preview()
            else:
                print('Full preview including spectrum only supported for PyTPeriodicMatrix with shape=(1,1)')
        else:
            # assume python interpreter
            import matplotlib.pyplot as plt
            fig = plt.figure()
            plt.subplot(211)
            self.plot(fig, 'single')
            plt.subplot(212)
            self.plot_spectrum(fig, orders, 'single')
            plt.draw()

    def show_spectrum(self, orders=None):
        """
        Show single sided amplitude spectrum of the signal.

        Examples
        --------

        .. plot::
           :include-source:

           import numpy as np
           import matplotlib.pyplot as plt
           import lcm.utils.pytperiodic as pm

           phi = np.linspace(0, 2*np.pi, 1000, endpoint=False)

           # fourier approximation rectangle function sig = 1.5 + 1 * np.sin(phi) + (1/3) * np.sin(3*phi) + (1/5) *
           np.sin(5*phi) + (1/7) * np.sin(7*phi) + (1/9) * np.sin(9*phi) tp = pm.PyTPeriodicMatrix(phi, sig,
           mode='fit')

           # frequency domain
           tp.show_spectrum()
           plt.show()

           # time domain
           tp.show_preview()
           plt.show()

        """
        if self.interpreter == 'python':
            import matplotlib.pyplot as plt
            fig = plt.figure()
            self.plot_spectrum(fig, orders)
            plt.draw()
        elif self.interpreter == 'java' or self.interpreter == 'javaw':
            if self.shape == (1, 1):
                self.__show_java_spectrum()
            else:
                print('Spectrum can only be shown for PyTPeriodicMatrix with shape = (1,1)')
        else:
            # Assume python
            import matplotlib.pyplot as plt
            fig = plt.figure()
            self.plot_spectrum(fig, orders)
            plt.draw()

    def __show_java_preview(self):
        """Show preview of signal inside SyMSpace jep environment"""
        # title = 'Periodic Signal'
        # showFundamental = False
        # nPeriods = int(1)
        try:
            from lcm.tdata_interface.pyjdata_interface import PyJDataInterface
            javaTp = PyJDataInterface.convert_Py2J(self)
            javaTp.showPreview()
        except Exception as e:
            print(e)

    def __show_java_full_preview(self):
        """Show full preview of signal inside SyMSpace jep environment"""
        # title = 'Periodic Signal'
        # showFundamental = False
        # nPeriods = int(1)
        try:
            from lcm.tdata_interface.pyjdata_interface import PyJDataInterface
            javaTp = PyJDataInterface.convert_Py2J(self)
            javaTp.showFullPreview()
        except Exception as e:
            print(e)

    def __show_java_spectrum(self):
        """Show spectrum of signal inside SyMSpace jep environment"""
        # title = 'Periodic Signal'
        # showFundamental = False
        # nPeriods = int(1)
        try:
            from lcm.tdata_interface.pyjdata_interface import PyJDataInterface
            javaTp = PyJDataInterface.convert_Py2J(self)
            javaTp.showFullPreview()
        except Exception as e:
            print(e)

    def __getitem__(self, key) -> 'PyTPeriodicMatrix':

        try:
            if len(key) != 2:
                raise IndexError("array indexing requires 2 arguments!")
        except:
            raise IndexError("array indexing requires 2 arguments!")

        if isinstance(key[0], slice):
            row = key[0]
        elif np.ndim(key[0]) <= 1:
            row = np.atleast_2d(key[0])

        else:
            raise IndexError("unsupported row index")

        if isinstance(key[1], slice):
            col = key[1]
        elif np.ndim(key[1]) <= 1:
            col = np.atleast_2d(key[1])
        else:
            raise IndexError("unsupported column index")

        if type(row) != type(col):
            # Mixed Type indexing
            if isinstance(row, slice):
                col = col.flatten()

            elif isinstance(col, slice):
                row = row.flatten()

            else:
                raise IndexError("in mixed type indexing one index needs to be slice instance")
            ret_data = self._data[row, col, :]

        else:
            ret_data = self._data[row, col, :]

        ret = PyTPeriodicMatrix(ret_data, mode="rfft", nopack=True)
        # ret = PyTPeriodicMatrix(ret_data, mode="rfft", nopack=self.nopack)
        ret.scaling_factor = self.scaling_factor
        return ret

    def __setitem__(self, key, item):

        if not isinstance(item, PyTPeriodicMatrix):  # get rfft data
            raise ValueError("PyTPeriodicMatrix required")

        try:
            if len(key) != 2:
                raise IndexError("array indexing requires 2 arguments!")
        except:
            raise IndexError("array indexing requires 2 arguments!")

        if isinstance(key[0], int):
            rows = slice(key[0], key[0] + 1, None)
        else:
            rows = key[0]

        if isinstance(key[1], int):
            cols = slice(key[1], key[1] + 1, None)
        else:
            cols = key[1]

        if not (isinstance(rows, slice) and isinstance(cols, slice)):
            raise IndexError("array indexing requires integers or slices!")

        if self.nbins() > item.nbins():
            self.expand_to_odd()
            item.expand_to_odd(self.nbins())

        elif self.nbins() < item.nbins():
            item.expand_to_odd()
            self.expand_to_odd(item.nbins())

        shape1 = self._data[rows, cols, :].shape
        shape2 = item._data.shape

        if shape1 != shape2:
            raise Exception('Shapes do not match: {} != {} '.format(shape1, shape2))

        self._data[rows, cols, :] = copy.deepcopy(item._data)
        self.pack()

    def __str__(self) -> str:
        st = "["
        for r in range(0, self.nrows()):
            st += "["
            for c in range(0, self.ncols()):
                if c > 0:
                    st += ",\t"
                mean = self[r, c].mean()[0, 0]
                pp = self[r, c].pp()[0, 0]
                st += "mean={0:g} pp={1:g}".format(mean, pp)
            st += "]"
            if r + 1 < self.nrows():
                st += ",\n"
        st += "]"
        return st

    def __repr__(self) -> str:
        """

        Returns
        -------
        result: basestring
            String representation

        """
        return "PyTPeriodicMatrix( " + self._data.__str__() + " , mode='rfft')"

    def __add__(self, other) -> 'PyTPeriodicMatrix':
        """
        Overrides the "+" operator

        Parameters
        ----------
        other : PyTPeriodicMatrix or float or ndarray
            object to add

        Returns
        -------
        result : PyTPeriodicMatrix
            Result of multiplication, new instance

        """

        if isinstance(other, PyTPeriodicMatrix):

            self.checkScalingFactor(other)

            other = copy.deepcopy(other)
            n = max(self.nbins(), other.nbins())
            self.expand_to_odd(n)
            other.expand_to_odd(n)
            res = PyTPeriodicMatrix(self._data + other._data, mode="rfft", nopack=self.nopack or other.nopack)
        else:
            try:
                other = np.array(other)

                if other.ndim == 0:
                    # scalar addition

                    tmp = copy.deepcopy(self._data)
                    tmp[:, :, 0] += self.nbins() * other
                    res = PyTPeriodicMatrix(tmp, mode="rfft", nopack=self.nopack)

                elif other.ndim == 2:
                    # matrix addition

                    if self._data.shape[:-1] != other.shape:
                        raise ValueError("__add__: mismatching shapes!")

                    tmp = copy.deepcopy(self._data)
                    tmp[:, :, 0] += self.nbins() * other
                    res = PyTPeriodicMatrix(tmp, mode="rfft", nopack=self.nopack)

                else:
                    raise ValueError("__mul__: unsupported argument")

            except:
                raise ValueError("__add__: unsupported argument")
        res.scaling_factor = self.scaling_factor
        return res

    def __radd__(self, other) -> 'PyTPeriodicMatrix':
        return self.__add__(other)

    def __sub__(self, other) -> 'PyTPeriodicMatrix':
        if isinstance(other, numbers.Real):
            return self.__add__(-other)
        else:
            return self.__add__(other.__mul__(-1.0))

    def __rsub__(self, other) -> 'PyTPeriodicMatrix':
        return self.__sub__(other).__mul__(-1.)

    def __mul__(self, other) -> 'PyTPeriodicMatrix':
        """
        Overrides the "*" operator

        Parameters
        ----------
        other : PyTPeriodicMatrix or float or ndarray
            object to multiply with

        Returns
        -------
        result : PyTPeriodicMatrix
            Result of multiplication, new instance

        """

        if isinstance(other, PyTPeriodicMatrix):

            self.checkScalingFactor(other)

            if self.ncols() != other.nrows():
                raise ValueError("Matrix dimensions do not agree")
            if 0 == self.ncols():
                res = PyTPeriodicMatrix(self.nrows(), 0, mode='zero')
            else:
                n = 1 + 2 * (self.get_max_order() + other.get_max_order())

                # get samples and move axis of samples to first position
                a = self.get_samples(n)
                b = other.get_samples(n)

                # calculate matrix products (ik) = (ij) x (jk) along last axis "s"
                # res = np.einsum("ijs,jks->iks",a,b)
                res = np.matmul(a.transpose(2, 0, 1), b.transpose(2, 0, 1)).transpose(1, 2, 0)
                res = PyTPeriodicMatrix(res, mode="samples", nopack=self.nopack or other.nopack)

        else:
            try:
                other = np.array(other)

                if other.ndim == 0:
                    # scalar multiplication
                    if np.iscomplex(other):
                        c_mag = math.sqrt(other.real * other.real + other.imag * other.imag)
                        c_phase = math.atan2(other.imag, other.real)

                        res_mag = c_mag * self.get_magnitudes()
                        res_phase = c_phase + self.get_phases()
                        res_dc = c_mag * self.mean()
                        res_cos = res_mag * np.cos(res_phase)
                        res_sin = res_mag * np.sin(res_phase)

                        res = PyTPeriodicMatrix((res_dc, res_cos, res_sin), mode="coeffs")

                    else:
                        res = PyTPeriodicMatrix(self._data * other, mode="rfft", nopack=self.nopack)

                elif other.ndim == 2:
                    # matrix multiplication
                    if np.iscomplex(other).any():
                        res = PyTPeriodicMatrix(self.nrows(), len(other[0, :]), mode='zero')
                        for r in range(len(other[0, :])):
                            for k in range(self.nrows()):
                                for c in range(len(other[:, 0])):
                                    res[k, r] = self[k, c] * other[c, r]

                    else:
                        res = PyTPeriodicMatrix(np.einsum("ijs,jk->iks", self._data, other), mode="rfft",
                                                nopack=self.nopack)
                else:
                    raise ValueError("__mul__: unsupported argument")

            except:
                raise ValueError("__mul__: unsupported argument")
        res.scaling_factor = self.scaling_factor
        return res

    def __rmul__(self, other) -> 'PyTPeriodicMatrix':

        if isinstance(other, PyTPeriodicMatrix):

            self.checkScalingFactor(other)

            """ right side matrix multiplication
            other * self = (self' * other')'
            """
            res = (self.transpose().__mul__(other.transpose())).transpose()

        else:
            try:
                other = np.array(other)
                if other.ndim == 0:
                    # scalar multiplication
                    res = self.__mul__(other)
                elif other.ndim == 2:
                    # matrix multiplication
                    res = (self.transpose().__mul__(other.transpose())).transpose()
                else:
                    raise ValueError("__rmul__: unsupported argument")

            except:
                raise ValueError("__rmul__: unsupported argument")

        return res

    # def __matmul__(self, other):
    #     """
    #     Overrides the "@" operator
    #
    #     Parameters
    #     ----------
    #     other : PyTPeriodicMatrix or float or ndarray
    #         object to multiply with
    #
    #     Returns
    #     -------
    #     result : PyTPeriodicMatrix
    #         Result of multiplication, new instance
    #
    #     """
    #
    #     return self.__mul__(other)

    def __truediv__(self, other) -> 'PyTPeriodicMatrix':
        if isinstance(other, PyTPeriodicMatrix):
            raise ValueError("__truediv__: unsupported argument")

        else:
            try:

                other = np.array(other)
                ''' remove all dimensions of size 1'''
                other = other.squeeze()
                if other.ndim == 0:
                    # scalar division
                    res = self.__mul__(1 / other)
                elif other.ndim == 2:
                    # matrix division
                    raise ValueError("__truediv__: unsupported argument")
                else:
                    raise ValueError("__truediv__: unsupported argument")

            except:
                raise ValueError("__truediv__: unsupported argument")

        return res

    def __rtruediv__(self, other):
        # @TODO implement __rtruediv__!
        raise Exception("not implemented")

    def __neg__(self) -> 'PyTPeriodicMatrix':
        """
        Overrides the "-val" operator.

        Returns
        -------
        result : PyTPeriodicMatrix
        """

        return self.__mul__(-1.)

    def __pos__(self) -> 'PyTPeriodicMatrix':
        """
        Overrides the "+val" operator.

        Returns
        -------
        result : PyTPeriodicMatrix
        """

        return copy.deepcopy(self)

    def __lt__(self, other) -> bool:
        """
        Overrides the "<" operator.

        Parameters
        ----------
        other : PyTPeriodicMatrix

        Returns
        -------
        result : bool
        """

        if not isinstance(other, PyTPeriodicMatrix):
            raise ValueError('__lt__: unsupported argument')

        if (self.nrows() != other.nrows()) or (self.ncols() != other.ncols()):
            raise ValueError('__lt__: Matrix dimensions must agree')

        return False not in np.less(self.get_samples(), other.get_samples())

    def __gt__(self, other) -> bool:
        """
        Overrides the ">" operator.

        Parameters
        ----------
        other : PyTPeriodicMatrix

        Returns
        -------
        result : bool

        """

        if not isinstance(other, PyTPeriodicMatrix):
            raise ValueError('__lt__: unsupported argument')

        if (self.nrows() != other.nrows()) or (self.ncols() != other.ncols()):
            raise ValueError('__lt__: Matrix dimensions must agree')

        res_max = np.max([self.resolution, other.resolution])
        res1, res2 = self.resolution, other.resolution
        self.resolution, other.resolution = res_max, res_max

        ret = False not in np.greater(self.get_samples(), other.get_samples())

        self.resolution, other.resolution = res1, res2

        return ret

    def __eq__(self, other) -> bool:
        """
        Overrides the "==" operator.
        Checks equality of rfft- data array with relative tolerance of 1e-5

        Parameters
        ----------
        other : PyTPeriodicMatrix

        Returns
        -------
        result : bool
                Result of data equality check.

        """

        if not isinstance(other, PyTPeriodicMatrix):
            return False

        self.pack()
        other.pack()

        return np.allclose(self.data, other.data, rtol=1e-5)

    def allclose(self, other, rtol=1e-5, atol=1e-8) -> bool:

        if not isinstance(other, PyTPeriodicMatrix):
            return False

        self.pack()
        other.pack()

        n_samples = int(self.nbins() * self.resolution * 10)

        a = other.get_samples(n_samples)
        b = self.get_samples(n_samples)

        return np.allclose(a, b, rtol=rtol, atol=atol)

    def _repr_png_(self, title=None, width=400, height=300) -> str:
        from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
        from matplotlib.figure import Figure
        import io
        import base64

        fig = Figure(figsize=(width / 100., height / 100.), dpi=100.)
        canvas = FigureCanvas(fig)

        self.plot(fig)

        buf = io.BytesIO()
        canvas.print_figure(buf)

        buf.seek(0)
        png = base64.b64encode(buf.getvalue())
        buf.close()

        return png.decode("utf-8")


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    cos_orig = PyTPeriodicMatrix(2, 1., 0.)
    cos_orig.show_full_preview()

    cbor_obj = cos_orig.toCBOR(f=2)

    cos_new = PyTPeriodicMatrix.fromCBOR(cbor_obj)
    cos_new.show_full_preview()

    plt.show()
