# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 1024 $
$LastChangedDate:: 2020-06-25 18:45:41 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import unittest
import numpy as np
import numpy.testing as npt
import lcm.math.rbf as rbf
import lcm.math.rbf.kernels as kernels
from lcm.math.rbf.utils.analytic_StackedMulitRBF import AnalyticFunction1


class TestStackedMultiRBF(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.plot = False
        n_phi = 1000
        phi = np.linspace(0, 2 * np.pi, n_phi, endpoint=False)

        xi1 = np.array([[0.5 + 0.5 * np.sin(phi)], [0.2 * np.cos(3. * phi)]])
        xi2 = np.array([[0.5 * np.ones_like(phi)], [0.2 * np.ones_like(phi)]])

        md = 0.5

        cls.ref_data = \
            {'fun1-xi1-CS': dict(kernelID=1, maxdist=md, xi=xi1, fnc=AnalyticFunction1(10, 10, n_phi)),
             'fun1-xi2-CS': dict(kernelID=1, maxdist=md, xi=xi2, fnc=AnalyticFunction1(10, 10, n_phi)),
             'fun1-xi1-TPS': dict(kernelID=2, maxdist=md, xi=xi1, fnc=AnalyticFunction1(10, 10, n_phi)),
             'fun1-xi2-TPS': dict(kernelID=2, maxdist=md, xi=xi2, fnc=AnalyticFunction1(10, 10, n_phi)),
             }

        key = ''
        try:
            for key, data in cls.ref_data.items():
                fnc = data['fnc']
                xc = fnc.getXc()
                zc = fnc.getZc()
                maxdist = data['maxdist']
                kernel = kernels.AbstractKernel.fromID(data['kernelID'])
                data['rbf'] = rbf.StackedMultiRBF(xc, zc, kernel, max_dist=maxdist, t=None)
        except Exception as e:
            raise unittest.SkipTest("Constructor failed:\n{}: {}".format(key, e))

    def test__call__(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):
                xi = data['xi']
                fnc = data['fnc']
                phi = fnc.phi
                rbf_interp = data['rbf']  # type: rbf.StackedMultiRBF

                ''' rbf result'''
                zi = rbf_interp(xi)

                ''' analytic result'''
                zi_analytic = fnc.getZi(xi[0, :], xi[1, :], phi)

                '''error'''
                if key in ['fun1-xi1-CS', 'fun1-xi1-TPS']:
                    rtol = 5e-2
                    atol = 1e-3
                else:
                    rtol = 1e-2
                    atol = 1e-3

                '''plot'''
                if self.plot:
                    fnc.plotComparison(zi, zi_analytic, atol=atol, rtol=rtol)

                error = np.max(np.abs(zi - zi_analytic))
                max_error = np.linalg.norm(zi_analytic)
                npt.assert_array_less(error, rtol * max_error)

                '''compare signals'''
                npt.assert_allclose(zi, zi_analytic, atol=atol, rtol=rtol)

    def test_fromCBOR(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):
                rbf_interp = data['rbf']  # type: rbf.StackedMultiRBF

                cbor_obj = rbf_interp.toCBOR()
                obj = rbf.StackedMultiRBF.fromCBOR(cbor_obj)

                npt.assert_array_equal(obj.xc_t, rbf_interp.xc_t)
                npt.assert_array_equal(obj.t, rbf_interp.t)
                npt.assert_array_equal(obj.c, rbf_interp.c)
                npt.assert_array_equal(obj.d, rbf_interp.d)

    @unittest.skip
    def test_fromMatrices(self):
        self.fail()

    @unittest.skip
    def test_xdim(self):
        self.fail()

    @unittest.skip
    def test_zdim(self):
        self.fail()

    @unittest.skip
    def test_n_stacks(self):
        self.fail()

    @unittest.skip
    def test_initFromSamples(self):
        self.fail()

    @unittest.skip
    def test_initFromMatrices(self):
        self.fail()

    def test_eval_dz_dx(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):
                
                if key in ['fun1-xi1-CS', 'fun1-xi1-TPS']:
                    self.skipTest('subtest {} skipped'.format(key))

                xi = data['xi']
                fnc = data['fnc']
                phi = fnc.phi
                rbf_interp = data['rbf']  # type: rbf.StackedMultiRBF

                ''' get jacobian matrix ( derivation along x-dimensions)'''
                dzi_dx = rbf_interp.eval_dz_dx(xi)

                ''' analytic result'''
                dzi_dx_analytic = fnc.get_dzi_dx(xi[0, :], xi[1, :], phi)

                '''calculate error'''
                rtol = 1e-2
                atol = 1e-3

                if self.plot:
                    fnc.plotComparison(dzi_dx, dzi_dx_analytic, atol=atol, rtol=rtol)

                error = np.max(np.abs(dzi_dx - dzi_dx_analytic))
                max_error = np.linalg.norm(dzi_dx_analytic)
                npt.assert_array_less(error, rtol * max_error)

                '''compare signals'''
                npt.assert_allclose(dzi_dx, dzi_dx_analytic, atol=atol, rtol=rtol)




    @unittest.skip
    def test_to_msgpack(self):
        self.fail()

    @unittest.skip
    def test_write_msgpack(self):
        self.fail()

    def test_toCBOR(self):
        for key, data in self.ref_data.items():
            with self.subTest(msg=key):

                rbf_interp = data['rbf']  # type: rbf.StackedMultiRBF

                cbor_obj = rbf_interp.toCBOR()

                self.assertEqual(cbor_obj['type'], '__stackedRBF__')
                self.assertEqual(cbor_obj['max_dist'], rbf_interp.max_dist)
                self.assertEqual(cbor_obj['kernel'], rbf_interp.kernel.id)
                npt.assert_array_equal(cbor_obj['params'], rbf_interp.kernel.params)

                self.assertEqual(tuple(cbor_obj['xi']['shape']), rbf_interp.xc_t.shape)
                self.assertEqual(tuple(cbor_obj['t']['shape']), rbf_interp.t.shape)
                self.assertEqual(tuple(cbor_obj['c']['shape']), rbf_interp.c.shape)
                self.assertEqual(tuple(cbor_obj['d']['shape']), rbf_interp.d.shape)

