# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 927 $
$LastChangedDate:: 2019-01-30 15:22:56 +0100#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import shutil
import tempfile
import unittest

import numpy as np

import lcm.math.rbf.kernels as kernels
import lcm.math.rbf as rbf
import lcm.utils.pytperiodic as pm
from lcm.utils.serialization import Serializable


class TestSerializer(unittest.TestCase):

    def setUp(self):
        """

        Returns
        -------

        """

        ''' Create a temporary directory'''
        self.test_dir = tempfile.mkdtemp()

        ''' Create test data'''
        n_x = 2
        n_y = 2
        n_phi = 100
        phi = np.linspace(0, 2 * np.pi, n_phi, endpoint=False)
        xc_samples = np.empty((2, n_x * n_y, n_phi), dtype=np.float)
        for col, y0 in enumerate(np.linspace(-2, 2, n_y)):
            for row, x0 in enumerate(np.linspace(-1, 1, n_x)):
                idx = col * n_x + row
                xc_samples[0, idx] = x0 * np.ones_like(phi)
                xc_samples[1, idx] = y0 * np.ones_like(phi)

        zc_samples = np.ones((xc_samples.shape[0], xc_samples.shape[1], n_phi), dtype=np.float)
        xc_per = pm.PyTPeriodicMatrix(xc_samples, mode='samples')
        zc_per = pm.PyTPeriodicMatrix(zc_samples, mode='samples')
        periodic_rbf = rbf.PeriodicMultiRBF(xc_per, zc_per, kernels.KernelCS())
        stacked_rbf = rbf.StackedMultiRBF(xc_samples, zc_samples, kernels.KernelCS())

        self.test_data_toCBOR = [(np.zeros((2, 3)), '__NDRealArray__', None),
                                 (pm.PyTPeriodicMatrix(3, 3, mode='eye'), '__periodicMatrix__', 6),
                                 (periodic_rbf, '__periodicRBF__', 3),
                                 (stacked_rbf, '__stackedRBF__', None),
                                 ]

        self.test_data_dict_toCBOR = [({'__NDRealArray__': np.zeros((2, 3)),
                                        '__periodicMatrix__': pm.PyTPeriodicMatrix(3, 3, mode='eye'),
                                        '__periodicRBF__': periodic_rbf,
                                        '__stackedRBF__': stacked_rbf})
                                      ]

    def tearDown(self):
        """

        Returns
        -------

        """

        '''Remove the directory after the test'''
        shutil.rmtree(self.test_dir)

        '''assert if directory was not removed'''
        self.assertFalse(os.path.exists(self.test_dir))

    def checkInvCBOR(self, obj, inv_cbor_obj, expected_type):
        """Check function 1

        Parameters
        ----------
        obj : object, array_like, PyTPeriodicMatrix, rbf.PeriodicMultiRBF, rbf.StackedMultiRBF
            Serializable instance
        inv_cbor_obj: object, array_like, PyTPeriodicMatrix, rbf.PeriodicMultiRBF, rbf.StackedMultiRBF
            Deserialized instance
        expected_type: str
            Type string

        """

        if expected_type == '__NDRealArray__':
            '''check if actual shapes match'''
            self.assertEqual(inv_cbor_obj.shape, obj.shape)
        elif expected_type == '__periodicMatrix__':
            '''check if actual shapes match'''
            self.assertEqual(inv_cbor_obj.shape, obj.shape)
        elif expected_type == '__periodicRBF__':
            '''check if actual shapes match'''
            self.assertEqual(inv_cbor_obj.xc_per.shape, obj.xc_per.shape)
            self.assertEqual(inv_cbor_obj.t_per.shape, obj.t_per.shape)
            self.assertEqual(inv_cbor_obj.c_per.shape, obj.c_per.shape)
            self.assertEqual(inv_cbor_obj.d_per.shape, obj.d_per.shape)
        elif expected_type == '__stackedRBF__':
            '''check if actual shapes match'''
            self.assertEqual(inv_cbor_obj.xc.shape, obj.xc.shape)
            self.assertEqual(inv_cbor_obj.t.shape, obj.t.shape)
            self.assertEqual(inv_cbor_obj.c.shape, obj.c.shape)
            self.assertEqual(inv_cbor_obj.d.shape, obj.d.shape)

    def checkCBOR(self, obj, cbor_obj, expected_type):
        """Check function 2

        Parameters
        ----------
        obj : object, array_like, PyTPeriodicMatrix, rbf.PeriodicMultiRBF, rbf.StackedMultiRBF
            Serializable instance
        cbor_obj: dict
            Serialized instance
        expected_type: str
            Type string

        Returns
        -------

        """
        if expected_type == '__NDRealArray__':
            '''check if actual shapes match'''
            self.assertEqual(tuple(cbor_obj['shape']), obj.shape)
        elif expected_type == '__periodicMatrix__':
            '''check if actual shapes match'''
            self.assertEqual(tuple(cbor_obj['bins']['shape'][0:2]), obj.shape)
        elif expected_type == '__periodicRBF__':
            '''check if actual shapes match'''
            self.assertEqual(tuple(cbor_obj['xi_per']['bins']['shape'][0:2]), obj.xc_per.shape)
            self.assertEqual(tuple(cbor_obj['t_per']['bins']['shape'][0:2]), obj.t_per.shape)
            self.assertEqual(tuple(cbor_obj['c_per']['bins']['shape'][0:2]), obj.c_per.shape)
            self.assertEqual(tuple(cbor_obj['d_per']['bins']['shape'][0:2]), obj.d_per.shape)
        elif expected_type == '__stackedRBF__':
            '''check if actual shapes match'''
            self.assertEqual(tuple(cbor_obj['xi']['shape']), obj.xc.shape)
            self.assertEqual(tuple(cbor_obj['t']['shape']), obj.t.shape)
            self.assertEqual(tuple(cbor_obj['c']['shape']), obj.c.shape)
            self.assertEqual(tuple(cbor_obj['d']['shape']), obj.d.shape)

    def test_fromCBOR(self):
        for data in self.test_data_toCBOR:
            obj, expected_type, f = data

            with self.subTest(msg=type(obj)):
                cbor_obj = Serializable.toCBOR(obj, f)
                inv_cbor_obj = Serializable.fromCBOR(cbor_obj)

                '''check if actual types match'''
                self.assertEqual(type(inv_cbor_obj), type(obj))

                self.checkInvCBOR(obj, inv_cbor_obj, expected_type)

    def test_toCBOR(self):
        for data in self.test_data_toCBOR:
            obj, expected_type, f = data
            with self.subTest(msg=type(obj)):
                cbor_obj = Serializable.toCBOR(obj, f)

                '''check if type strings match'''
                self.assertEqual(cbor_obj['type'], expected_type)

                self.checkCBOR(obj, cbor_obj, expected_type)

    def test_dict_fromCBOR(self):
        for data in self.test_data_dict_toCBOR:
            obj = data
            with self.subTest(msg=type(obj)):
                cbor_obj = Serializable.toCBOR(obj)
                inv_cbor_obj = Serializable.fromCBOR(cbor_obj)
                for key, value in obj.items():
                    inv_cbor_value = inv_cbor_obj[key]

                    '''check if actual types match'''
                    self.assertEqual(type(inv_cbor_value), type(value))

    def test_dict_toCBOR(self):
        for data in self.test_data_dict_toCBOR:
            obj = data
            with self.subTest(msg=type(obj)):
                cbor_obj = Serializable.toCBOR(obj)

                '''check if actual types match'''
                self.assertTrue(isinstance(cbor_obj, dict))

                for key, value in cbor_obj.items():
                    '''check if type strings match'''
                    self.assertEqual(key, value['type'])

    def test_saveCBOR(self):
        for data in self.test_data_toCBOR:
            obj, expected_type, f = data
            file = tempfile.mktemp(suffix='.cbor', dir=self.test_dir)
            with self.subTest(msg=type(obj)):
                Serializable.saveCBOR(obj, file, f)

                '''check if file was created'''
                self.assertTrue(os.path.isfile(file))

    def test_loadCBOR(self):
        for data in self.test_data_toCBOR:
            obj, expected_type, f = data
            file = tempfile.mktemp(suffix='.cbor', dir=self.test_dir)
            with self.subTest(msg=type(obj)):
                Serializable.saveCBOR(obj, file, f)

                '''check if file was created'''
                self.assertTrue(os.path.isfile(file))
                cbor_obj = Serializable.loadCBOR(file)
                inv_cbor_obj = Serializable.fromCBOR(cbor_obj)
                '''check if actual types match'''
                self.assertEqual(type(inv_cbor_obj), type(obj))


if __name__ == '__main__':
    unittest.main()


