/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9733 $
 * $Date: 2020-04-06 09:48:37 +0200 (Mon, 06 Apr 2020) $
 */
package at.lcm.symspace.center.utils;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.io.File;

import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.WindowConstants;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.utils.MagOptConfig;
import at.lcm.symspace.center.icons.sym.SyMIcons;
import at.lcm.symspace.gui.utils.Storage;

/**
 * Select and run macro (python script)
 *
 * @author Werner Koppelsttter
 */
public class MacroDialog extends JDialog {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Is run pressed */
	boolean isRun = false;
	/** Selected macro (python script) */
	File selectedMacro = null;

	/** Name of selected macro */
	JTextField macroName;
	/** Select type of macro */
	JTabbedPane macroType;
	/** List of user-macros */
	JList<File> userMacros;
	/** List of system-macros */
	JList<File> sysMacros;

	/** user-path field */
	JTextField userPathField;

	/** Button panel */
	JPanel buttons;
	/** Run macro */
	JButton run;
	/** Close dialog */
	JButton close;

	/**
	 * Open maro-selection dialog
	 * 
	 * @param owner
	 */
	public MacroDialog(Component owner) {

		this.setIconImages(SyMIcons.getIcons());

		this.setModal(true);

		this.init();

		this.setSize(500, 500);
		this.setLocationRelativeTo(owner);

		this.setVisible(true);
	}

	/**
	 * Get Macro to execute
	 * 
	 * @return Python script
	 */
	public File getMacro() {
		if (this.isRun) {
			return this.selectedMacro;
		} else {
			return null;
		}
	}

	/**
	 * Init dialog
	 */
	private void init() {
		JPanel p;
		DefaultListModel<File> model;
		String path;
		File file;

		// init dialog
		this.setTitle("Run macro");
		this.setLayout(new BorderLayout());
		this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

		// main panel
		p = new JPanel();
		p.setLayout(new BorderLayout());
		p.add(this.initMacroSelection(), BorderLayout.CENTER);
		p.add(this.initPathSelection(), BorderLayout.SOUTH);
		this.add(p, BorderLayout.CENTER);

		// buttons
		this.initButtons();
		this.add(this.buttons, BorderLayout.EAST);

		// update system-macros
		model = new DefaultListModel<File>();
		file = FileUtils.getFile(MagOptConfig.getHome(), "System", "Python", "scripts");
		for (File f : file.listFiles()) {
			if (f.isFile() && f.getName().toLowerCase().endsWith(".py")) {
				model.addElement(f);
			}
		}

		// update user-macros
		path = new Storage().getUserMacroPath();
		if ((path != null) && !path.isEmpty()) {
			this.userPathField.setText(path);
			this.userPathChanged();
		}

		this.sysMacros.setModel(model);
		this.sysMacros.setSelectedIndex(0);
	}

	/**
	 * Initialize macro selection
	 * 
	 * @return Panel
	 */
	private JPanel initMacroSelection() {
		JPanel panel;

		panel = new JPanel();
		panel.setBorder(new TitledBorder("Macro name"));
		panel.setBorder(
				new CompoundBorder(new CompoundBorder(new EmptyBorder(10, 10, 5, 0), new TitledBorder("Macro name")),
						new EmptyBorder(10, 10, 10, 10)));
		panel.setLayout(new BorderLayout());

		// display of selected macro
		this.macroName = new JTextField("");
		this.macroName.setEnabled(false);
		panel.add(this.macroName, BorderLayout.NORTH);

		// tabbed pane for marcos
		this.macroType = new JTabbedPane();
		this.macroType.setBorder(new EmptyBorder(10, 0, 0, 0));
		panel.add(this.macroType, BorderLayout.CENTER);

		// user macro list
		this.userMacros = new JList<File>();
		this.userMacros.setBorder(new CompoundBorder(new EmptyBorder(5, 5, 5, 5), new EtchedBorder()));
		this.userMacros.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		this.userMacros.setCellRenderer(new DefaultListCellRenderer() {
			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getListCellRendererComponent(JList<? extends Object> list, Object value, int index,
					boolean isSelected, boolean cellHasFocus) {
				super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);

				File file = (File) value;
				this.setText(file.getName());
				this.setToolTipText(file.getAbsolutePath());
				return this;
			}

		});
		this.userMacros.addListSelectionListener(new ListSelectionListener() {
			@Override
			public void valueChanged(ListSelectionEvent e) {
				MacroDialog.this.updateMacro();
			}
		});
		this.macroType.addTab("User macro", new JScrollPane(this.userMacros));

		// system macro list
		this.sysMacros = new JList<File>();
		this.sysMacros.setBorder(new CompoundBorder(new EmptyBorder(5, 5, 5, 5), new EtchedBorder()));
		this.sysMacros.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

		this.sysMacros.setCellRenderer(new DefaultListCellRenderer() {
			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getListCellRendererComponent(JList<? extends Object> list, Object value, int index,
					boolean isSelected, boolean cellHasFocus) {
				super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);

				File file = (File) value;
				this.setText(file.getName());
				this.setToolTipText(file.getAbsolutePath());
				return this;
			}

		});
		this.sysMacros.addListSelectionListener(new ListSelectionListener() {
			@Override
			public void valueChanged(ListSelectionEvent e) {
				MacroDialog.this.updateMacro();
			}
		});
		this.macroType.addTab("System macro", new JScrollPane(this.sysMacros));

		// macro type selection
		this.macroType.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent e) {
				MacroDialog.this.updateMacro();
			}
		});

		return panel;
	}

	/**
	 * Initialize user-path selection
	 * 
	 * @return Panel
	 */
	private JPanel initPathSelection() {
		JPanel panel = new JPanel();
		JButton selectUserPath;

		panel.setBorder(new CompoundBorder(new EmptyBorder(5, 10, 10, 0), new TitledBorder("Location of user-macro")));

		// text field for path
		this.userPathField = new JTextField("");
		this.userPathField.addKeyListener(new KeyAdapter() {
			@Override
			public void keyReleased(KeyEvent e) {
				MacroDialog.this.userPathChanged();
			}
		});
		panel.add(this.userPathField);

		// select button
		selectUserPath = new JButton("...");
		selectUserPath.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				JFileChooser chooser;

				chooser = new JFileChooser();
				chooser.setDialogTitle("Select location for macro-user-path");
				chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				File file = new File(MacroDialog.this.userPathField.getText());
				if (file.exists()) {
					chooser.setCurrentDirectory(file);
				}
				if ((chooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION)
						&& !chooser.getSelectedFile().getPath().isEmpty()) {
					MacroDialog.this.userPathField.setText(chooser.getSelectedFile().getAbsolutePath());
					MacroDialog.this.userPathChanged();
				}
			}
		});
		panel.add(selectUserPath);

		// define layout
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setAutoCreateGaps(true);
		layout.setAutoCreateContainerGaps(true);
		layout.setHorizontalGroup(
				layout.createSequentialGroup().addComponent(this.userPathField).addComponent(selectUserPath));
		layout.setVerticalGroup(
				layout.createParallelGroup().addComponent(this.userPathField).addComponent(selectUserPath));

		return panel;
	}

	/**
	 * Change user path
	 */
	private void userPathChanged() {
		File file;
		DefaultListModel<File> model;

		file = new File(this.userPathField.getText());
		model = new DefaultListModel<File>();

		// list directory
		if (file.exists()) {
			for (File f : file.listFiles()) {
				if (f.isFile() && f.getName().toLowerCase().endsWith(".py")) {
					model.addElement(f);
				}
			}
			new Storage().setUserMacroPath(file.getAbsolutePath());
		}

		// update list
		file = this.userMacros.getSelectedValue();
		this.userMacros.setModel(model);
		if (file == null) {
			this.userMacros.setSelectedIndex(0);
		} else {
			this.userMacros.setSelectedValue(file, true);
		}
	}

	/**
	 * Update selected macro
	 */
	private void updateMacro() {

		// select script
		if (this.macroType.getSelectedIndex() == 0) {
			this.selectedMacro = this.userMacros.getSelectedValue();
		} else if (this.macroType.getSelectedIndex() == 1) {
			this.selectedMacro = this.sysMacros.getSelectedValue();
		} else {
			this.selectedMacro = null;
		}

		// display selection
		if (this.selectedMacro == null) {
			this.run.setEnabled(false);
			this.macroName.setText("");
		} else {
			this.run.setEnabled(true);
			this.macroName.setText(this.selectedMacro.getName());
		}
	}

	/**
	 * Initialize button panel
	 */
	private void initButtons() {
		this.buttons = new JPanel();

		// run button
		this.run = new JButton("Run");
		this.run.setEnabled(false);
		this.run.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				MacroDialog.this.isRun = true;
				MacroDialog.this.setVisible(false);
				MacroDialog.this.dispose();
			}
		});
		this.buttons.add(this.run);

		// close button
		this.close = new JButton("Close");
		this.close.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				MacroDialog.this.setVisible(false);
				MacroDialog.this.dispose();
			}
		});
		this.buttons.add(this.close);

		// define layout
		GroupLayout layout = new GroupLayout(this.buttons);
		this.buttons.setLayout(layout);
		layout.setAutoCreateGaps(true);
		layout.setAutoCreateContainerGaps(true);
		layout.setHorizontalGroup(layout.createParallelGroup().addComponent(this.run).addComponent(this.close));
		layout.setVerticalGroup(layout.createSequentialGroup().addComponent(this.run).addComponent(this.close));
		layout.linkSize(this.run, this.close);
	}
}
