/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.docu;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.Hashtable;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.UIManager;
import javax.swing.WindowConstants;
import javax.swing.border.EmptyBorder;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.structure.DocFile;
import at.lcm.bu21.magopt.structure.ParamContainer;
import at.lcm.bu21.magopt.structure.ParamField;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.magopt.utils.MagOptConfig;
import magopt.utils.MopFile;

/**
 * Fill report. Replace tag from project
 * 
 * @author Werner Koppelsttter
 */
public class FillTexReport extends DocTexWriter {

	/** Default project */
	Project globalProject = null;

	/** Default index */
	int globalIndex = -1;

	/** Identifier for __self__ keyword */
	String selfIdentifier = null;

	/** Logging panel */
	JTextArea log;

	/** Compile button */
	JButton compile;

	/**
	 * Program entry point
	 * 
	 * @param args
	 *            Program arguments
	 */
	public static void main(String[] args) {
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
		}

		// evaluate parameter
		if (args.length != 1) {
			System.out.println("USAGE:");
			System.out.println("\tmagopt.jar <file>");
			System.out.println();
		} else {
			new FillTexReport(args[0]);
		}

	}

	/**
	 * Create class to fill report
	 * 
	 * @param file
	 *            File for new report
	 */
	public FillTexReport(String file) {
		this(new File(file));
	}

	/**
	 * Create class to fill report
	 * 
	 * @param file
	 *            File for new report
	 */
	public FillTexReport(File file) {
		this(file, null);
	}

	/**
	 * Create class to fill report
	 * 
	 * @param file
	 *            File for new report
	 * @param project
	 *            Default project
	 */
	public FillTexReport(String file, Project project) {
		this(new File(file), project);
	}

	/**
	 * Create class to fill report
	 * 
	 * @param file
	 *            File for new report
	 * @param project
	 *            Default project
	 */
	public FillTexReport(File file, Project project) {
		this(file, project, null);
	}

	/**
	 * Create class to fill report
	 * 
	 * @param file
	 *            File for new report
	 * @param project
	 *            Default project
	 * @param selfIdentifier
	 *            Identifier for __self__ keyword
	 */
	public FillTexReport(File file, Project project, String selfIdentifier) {
		DocTexWriter texSource;
		File target;
		int index;

		this.log = new JTextArea();
		this.compile = new JButton("Compile");
		this.compile.setEnabled(false);

		if (project != null) {
			try {
				this.globalProject = project.deepClone();
				this.globalIndex = project.getIndex();
			} catch (Exception e) {
			}
		}
		this.selfIdentifier = selfIdentifier;

		if (file.exists() && file.getName().toLowerCase().endsWith(".tex")) {
			index = file.getName().toLowerCase().indexOf(".template.");
			if (index == -1) {
				target = new File(file.getParentFile(),
						file.getName().substring(0, file.getName().length() - 4) + "_filled.tex");
			} else {
				target = new File(file.getParentFile(),
						file.getName().substring(0, index) + file.getName().substring(index + 9));
			}
			this.setFile(target);

			// read source report to fill
			texSource = new DocTexWriter(file);
			try {
				texSource.readContent(file);
			} catch (IOException e) {
				this.logging("Unable to read file");
			}
			this.fillProject(texSource.toString());
			this.compile.setEnabled(true);
		} else {
			this.logging("File doesn't exists or is invalid");
		}

	}

	/**
	 * Initialize gui
	 */
	public void show() {
		JFrame frame;
		JPanel panel, buttonPanel;

		// display warnings
		frame = new JFrame("Log FillTex");
		frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		frame.setLayout(new BorderLayout());

		panel = new JPanel();
		frame.add(panel, BorderLayout.CENTER);
		panel.setLayout(new BorderLayout());
		panel.setBorder(new EmptyBorder(10, 10, 10, 10));

		panel.add(new JScrollPane(this.log), BorderLayout.CENTER);

		buttonPanel = new JPanel();
		buttonPanel.setLayout(new BorderLayout());
		panel.add(buttonPanel, BorderLayout.SOUTH);

		this.compile.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				try {
					FillTexReport.this.compile(true);
					FillTexReport.this.logging("Compiled");
				} catch (IOException e) {
					FillTexReport.this.logging("Error by compiling");
				}
			}
		});
		buttonPanel.add(this.compile, BorderLayout.CENTER);
		buttonPanel.add(new JPanel(), BorderLayout.NORTH);

		frame.setLocation(200, 200);
		frame.setSize(500, 400);
		frame.setVisible(true);

	}

	/**
	 * Fill source-report with data from project
	 * 
	 * @param source
	 *            Source of tex-file
	 */
	void fillProject(String source) {
		int startIndex, endIndex, index, index1, index2, lineIndex;
		String command, name, doc, line;
		Project project;
		Parameter param;
		String[] lines;
		int lineNumber;
		Hashtable<String, String> alias;
		String symHome = "\\newcommand{\\SyMHomePath}";
		String moHome = "\\newcommand{\\MoHomePath}";

		if (source.contains("\\begin{document}")) {
			// replace SyMHomePath
			startIndex = source.indexOf(symHome);
			if (startIndex != -1) {
				endIndex = source.indexOf("}", startIndex + symHome.length());
				source = source.substring(0, startIndex) + source.substring(endIndex + 1);
			}
			source = symHome + "{" + MagOptConfig.getHome().replace("\\", "/") + "}\n" + source;

			// replace MoHomePath
			startIndex = source.indexOf(moHome);
			if (startIndex != -1) {
				endIndex = source.indexOf("}", startIndex + moHome.length());
				source = source.substring(0, startIndex) + source.substring(endIndex + 1);
			}
			source = moHome + "{" + MagOptConfig.getHome().replace("\\", "/") + "}\n" + source;
		}

		// replace commands in template
		alias = new Hashtable<String, String>();
		lines = source.split("\n");
		source = null;
		for (lineNumber = 0; lineNumber < lines.length; lineNumber++) {
			line = lines[lineNumber];

			// search for keywords
			do {
				index1 = line.indexOf("\\SYM");
				index2 = line.indexOf("\\MO");
				if (index1 != -1) {
					if (index2 != -1) {
						if (index1 > index2) {
							lineIndex = index2;
						} else {
							lineIndex = index1;
						}
					} else {
						lineIndex = index1;
					}
				} else {
					lineIndex = index2;
				}

				if (lineIndex != -1) {
					if (lineIndex > 0) {
						this.add(line.substring(0, lineIndex));
						line = line.substring(lineIndex);
					}
					startIndex = line.indexOf("{");
					endIndex = line.indexOf("}");
					if ((startIndex > -1) && (endIndex > -1) && (startIndex < endIndex)) {
						command = line.substring(0, startIndex);
						name = line.substring(startIndex + 1, endIndex).trim();
						try {
							if ((command.compareTo("\\SYMALIAS") == 0) || (command.compareTo("\\MOALIAS") == 0)) {
								line = this.addAlias(line.substring(endIndex + 1), name, alias);
							} else if ((command.compareTo("\\SYMSETINDEX") == 0)
									|| (command.compareTo("\\MOSETINDEX") == 0)) {
								project = this.getProject(name, alias);
								this.globalProject = project;
								this.globalIndex = project.getIndex();
								line = line.substring(endIndex + 1);
							} else {
								if ((name.indexOf("[") == 0) && name.contains("]")) {
									index = name.indexOf("]");
									project = this.getProject(name.substring(1, index), alias);
									name = name.substring(index + 1).trim();
								} else {
									project = this.getProject(null, alias);
								}

								// split first parameter to name and file
								index = name.indexOf('(');
								if ((index != -1) && name.endsWith(")")) {
									doc = name.substring(index + 1, name.length() - 1).trim();
									name = name.substring(0, index).trim();
								} else {
									doc = null;
								}

								// replace __self__ keyword
								if (name.toLowerCase().startsWith("__self__")) {
									if (this.selfIdentifier != null) {
										name = this.selfIdentifier + name.substring(8);
									} else {
										throw new Exception("Self identifier is not defined!");
									}
								}

								// Get selected parameter of project
								param = project.getObject(name);
								if ((command.compareTo("\\SYMEXIST") == 0) || (command.compareTo("\\MOEXIST") == 0)) {
									if (doc != null) {
										throw new Exception("Undefined syntax (DocFile isn't allowed!)");
									}
									line = this.fillExist(param, line.substring(endIndex + 1));
								} else {

									if (param == null) {
										throw new Exception("Parameter not found!");
									}

									if ((command.compareTo("\\SYMPARAM") == 0)
											|| (command.compareTo("\\MOPARAM") == 0)) {
										if (doc != null) {
											throw new Exception("Undefined syntax (DocFile isn't allowed!)");
										}
										line = this.fillParam(param, line.substring(endIndex + 1));
									} else if ((command.compareTo("\\SYMGRAPHIC") == 0)
											|| (command.compareTo("\\MOGRAPHIC") == 0)) {
										line = this.fillGraphic(param, doc, project.getFilesDirectory(),
												line.substring(endIndex + 1));
									} else if ((command.compareTo("\\SYMGRAPHICPATH") == 0)
											|| (command.compareTo("\\MOGRAPHICPATH") == 0)) {
										line = this.fillGraphicPath(param, doc, project.getFilesDirectory(),
												line.substring(endIndex + 1));
									} else if ((command.compareTo("\\SYMTEX") == 0)
											|| (command.compareTo("\\MOTEX") == 0)) {
										line = this.fillTex(param, doc, project.getFilesDirectory(),
												line.substring(endIndex + 1));
									} else if ((command.compareTo("\\SYMTABLE") == 0)
											|| (command.compareTo("\\MOTABLE") == 0)) {
										line = this.fillTable(param, doc, project.getFilesDirectory(),
												line.substring(endIndex + 1));
									} else if ((command.compareTo("\\SYMINFO") == 0)
											|| (command.compareTo("\\MOINFO") == 0)) {
										line = this.fillInfo(param, line.substring(endIndex + 1));
									} else if ((command.compareTo("\\SYMCOMPONENTS") == 0)
											|| (command.compareTo("\\MOCOMPONENTS") == 0)) {
										line = this.fillComponents(param, project.getFilesDirectory(),
												line.substring(endIndex + 1));
									} else {
										this.logging((lineNumber + 1) + ": UNDEFINED COMMAND '" + command + "'");
										this.add(line.substring(0, 3));
										line = line.substring(3);
									}
								}
							}
						} catch (Exception ex) {
							this.logging((lineNumber + 1) + ": ERROR BY '" + line.substring(0, endIndex + 1) + "'");
							this.logging("\t" + ex.getMessage());
							this.add(line.substring(0, 3));
							line = line.substring(3);
						}
					} else {
						this.logging((lineNumber + 1) + ": UNDEFINED COMMAND IN LINE '" + line + "'");
						this.add(line.substring(0, 3));
						line = line.substring(3);
					}
				}
			} while (lineIndex != -1);

			this.add(line);
			this.add("\n");
		}

		while ((index = this.content.indexOf("__self__.")) != -1) {
			this.content.replace(index, index + 9, "");
		}
		while ((index = this.content.indexOf("__self__")) != -1) {
			this.content.replace(index, index + 8, "");
		}

		// write and compile report
		try {
			this.write();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		this.logging("Finished fill");

	}

	/**
	 * Get project to fill report
	 * 
	 * @param parameter
	 *            Parameter with project-name and index
	 * @param alias
	 *            List of alias names
	 * @return Selected Project
	 * @throws Exception
	 */
	protected Project getProject(String parameter, Hashtable<String, String> alias) throws Exception {
		Project proj = this.globalProject;
		int index = this.globalIndex;
		String fileName, aliasParameter;
		MopFile file;

		if (parameter != null) {
			aliasParameter = alias.get(parameter);
			if (aliasParameter != null) {
				parameter = aliasParameter;
			}

			// open Project
			if (parameter.contains(":")) {
				fileName = parameter.substring(0, parameter.lastIndexOf(":")).trim();
				file = new MopFile(FileUtils.trimPath(new File(this.getDirectoryFile(), fileName)));
				if (!file.exists()) {
					throw new Exception("File doesn't exist '" + fileName + "'");
				}
				if ((proj == null) || (proj.getFile().compareToIgnoreCase(file.getPath()) != 0)) {
					proj = Project.onlyLoad(file);
					if (proj.getMopFile().getName().compareTo(file.getName()) != 0) {
						throw new Exception("Error by loading Project");
					}
				}
				parameter = parameter.substring(parameter.lastIndexOf(":") + 1);
			}
			// get index
			try {
				index = Integer.parseInt(parameter.trim()) - 1;
			} catch (Exception ex) {
				throw new Exception("Undefined Project index");
			}
		}

		if (proj == null) {
			throw new Exception("Project not defined");
		}

		if ((index >= 0) && (index < proj.getLength())) {
			proj.setIndex(index);
		} else {
			throw new Exception("Project index out of bounds");
		}

		return proj;
	}

	/**
	 * Fill keyword MOPARAM
	 * 
	 * @param param
	 *            Parameter to fill
	 * @param content
	 *            String to convert
	 * @return Remain of the source string
	 * @throws Exception
	 */
	protected String fillParam(Parameter param, String content) throws Exception {
		int index;
		String opt = null, texString = "", str;
		TData data;

		// get second parameter with options
		if (!content.isEmpty() && (content.charAt(0) == '{') && content.contains("}")) {
			index = content.indexOf('}');
			opt = content.substring(1, index);
			content = content.substring(index + 1);
		}
		if (opt == null) {
			opt = "v";
		}

		// Add values from options parameter
		for (int i = 0; i < opt.length(); i++) {
			switch (Character.toLowerCase(opt.charAt(i))) {
			case 'p': // name
				texString += this.replaceSpecialSigns(param.getName());
				break;
			case 'v': // value
				if (param.getUncertain()) {
					throw new Exception("Uncertain (Is parameter up to date?)");
				}
				data = param.getData();
				if (data instanceof TString) {
					texString += this.replaceSpecialSigns(data.toString().substring(1, data.toString().length() - 1));
				} else {
					texString += this.replaceSpecialSigns(data.toString());
				}
				break;
			case 'd': // description
				texString += this.replaceSpecialSigns(param.getDescription());
				break;
			case 't': // tex symbol
				if (!(param instanceof ParamField)) {
					throw new Exception("ParamField not found!");
				}
				str = ((ParamField) param).getTexString();
				if (str.isEmpty()) {
					str = " ";
				}
				texString += "$" + str + "$";
				break;
			case 'u': // unit
				if (!(param instanceof ParamField)) {
					throw new Exception("ParamField not found!");
				}
				texString += this.replaceSpecialSigns(((ParamField) param).getUnit());
				break;
			default:
				texString += opt.charAt(i);
				break;
			}
		}

		// add tex string
		this.add(texString);
		return content;
	}

	/**
	 * Fill keyword MOGRAPHIC
	 * 
	 * @param param
	 *            Parameter to fill
	 * @param doc
	 *            DocFile to add
	 * @param path
	 *            Path to doc-files
	 * @param content
	 *            String to convert
	 * @return Remain of the source string
	 * @throws Exception
	 */
	protected String fillGraphic(Parameter param, String doc, File path, String content) throws Exception {
		int index;
		String opt = null, file, texString;
		double value;
		File relFile;

		if (doc == null) {
			throw new Exception("DocFile isn't defined");
		}

		// check if docfiles is up to date
		if (param.getUncertain()) {
			throw new Exception("Uncertain (Is parameter up to date?)");
		}
		if (param.getDocFile(doc) == null) {
			throw new Exception("DocFile not exists for parameter!");
		}
		if (!param.checkDocFile(doc)) {
			throw new Exception("File to add is not up to date!");
		}

		// get second parameter with options
		if (!content.isEmpty() && (content.charAt(0) == '{') && content.contains("}")) {
			index = content.indexOf('}');
			opt = content.substring(1, index);
			content = content.substring(index + 1);
			try {
				value = Double.parseDouble(opt);
				opt = "width=" + value + "cm";
			} catch (Exception ex) {
			}
		}

		// insert Docfile and insert
		file = FileUtils.trimPath(new File(path, param.getDocFile(doc).getFile())).getPath();
		relFile = FileUtils.getRelativePath(this.getDirectoryFile(), new File(this.getDirectoryFile(), file));
		if (new File(this.getDirectoryFile(), relFile.getPath()).exists()) {
			texString = this.includeGraphic(relFile, opt);
			if (texString == null) {
				throw new Exception("Unable to add File!");
			}
		} else {
			throw new Exception("File to add doesn't exists!");
		}

		// add tex string
		this.add(texString);
		return content;
	}

	/**
	 * Fill keyword MOTEX
	 * 
	 * @param param
	 *            Parameter to fill
	 * @param doc
	 *            DocFile to add
	 * @param path
	 *            Path to doc-files
	 * @param content
	 *            String to convert
	 * @return Remain of the source string
	 * @throws Exception
	 */
	protected String fillTex(Parameter param, String doc, File path, String content) throws Exception {
		String file;
		File relFile;

		if (doc == null) {
			throw new Exception("DocFile isn't defined");
		}

		// check if docfiles is up to date
		if (param.getUncertain()) {
			throw new Exception("Uncertain (Is parameter up to date?)");
		}
		if (!param.checkDocFile(doc)) {
			throw new Exception("File to add is not up to date!");
		}

		// get Docfile and insert
		file = FileUtils.trimPath(new File(path, param.getDocFile(doc).getFile())).getPath();
		relFile = FileUtils.getRelativePath(this.getDirectoryFile(), new File(this.getDirectoryFile(), file));
		if (new File(this.getDirectoryFile(), relFile.getPath()).exists()) {
			this.addTex(relFile);
		} else {
			throw new Exception("File to add doesn't exists!");
		}

		return content;
	}

	/**
	 * Fill keyword MOTABLE
	 * 
	 * @param param
	 *            Parameter to fill
	 * @param doc
	 *            DocFile to add
	 * @param path
	 *            Path to doc-files
	 * @param content
	 *            String to convert
	 * @return Remain of the source string
	 * @throws Exception
	 */
	protected String fillTable(Parameter param, String doc, File path, String content) throws Exception {
		int index;
		String file, caption = null, label = null;
		File relFile;

		if (doc == null) {
			throw new Exception("DocFile isn't defined");
		}

		// check if docfiles is up to date
		if (param.getUncertain()) {
			throw new Exception("Uncertain (Is parameter up to date?)");
		}
		if (!param.checkDocFile(doc)) {
			throw new Exception("File to add is not up to date!");
		}

		// get second parameter (caption)
		if (!content.isEmpty() && (content.charAt(0) == '{') && content.contains("}")) {
			index = content.indexOf('}');
			caption = content.substring(1, index);
			content = content.substring(index + 1);
		}

		// get third parameter (Label)
		if (!content.isEmpty() && (content.charAt(0) == '{') && content.contains("}")) {
			index = content.indexOf('}');
			label = content.substring(1, index);
			content = content.substring(index + 1);
		}

		// get Docfile and insert
		file = FileUtils.trimPath(new File(path, param.getDocFile(doc).getFile())).getPath();
		relFile = FileUtils.getRelativePath(this.getDirectoryFile(), new File(this.getDirectoryFile(), file));
		if (new File(this.getDirectoryFile(), relFile.getPath()).exists()) {
			this.addTable(relFile, caption, label);
		} else {
			throw new Exception("File to add doesn't exists!");
		}

		return content;
	}

	/**
	 * Fill keyword MOGRAPHIC
	 * 
	 * @param param
	 *            Parameter to fill
	 * @param doc
	 *            DocFile to add
	 * @param path
	 *            Path to doc-files
	 * @param content
	 *            String to convert
	 * @return Remain of the source string
	 * @throws Exception
	 */
	protected String fillGraphicPath(Parameter param, String doc, File path, String content) throws Exception {
		// FIXME
		String file, texString = "";
		File relFile;

		if (doc == null) {
			throw new Exception("DocFile isn't defined");
		}

		// check if docfiles is up to date
		if (param.getUncertain()) {
			throw new Exception("Uncertain (Is parameter up to date?)");
		}
		if (param.getDocFile(doc) == null) {
			throw new Exception("DocFile not exists for parameter!");
		}
		if (!param.checkDocFile(doc)) {
			throw new Exception("File to add is not up to date!");
		}

		// insert Docfile and insert
		file = FileUtils.trimPath(new File(path, param.getDocFile(doc).getFile())).getPath();
		relFile = FileUtils.getRelativePath(this.getDirectoryFile(), new File(this.getDirectoryFile(), file));
		if (new File(this.getDirectoryFile(), relFile.getPath()).exists()) {
			texString = this.GraphicPath(relFile);

		} else {
			throw new Exception("File to add doesn't exists!");
		}

		// add tex string
		this.add(texString);
		return content;
	}

	/**
	 * Add alias project (project-name and index)
	 * 
	 * @param content
	 *            Content of line
	 * @param name
	 *            Alias name
	 * @param alias
	 *            List of alias
	 * @return New content of line
	 * @throws Exception
	 */
	protected String addAlias(String content, String name, Hashtable<String, String> alias) throws Exception {
		int index;
		String proj;

		// get second parameter with options
		if (!content.isEmpty() && (content.charAt(0) == '{') && content.contains("}")) {
			index = content.indexOf('}');
			proj = content.substring(1, index);
			content = content.substring(index + 1);
			alias.put(name, proj);
		} else {
			throw new Exception("Wrong number of parmeter");
		}

		return content;
	}

	/**
	 * Fill keyword MOINFO
	 * 
	 * @param param
	 *            Parameter to fill
	 * @param content
	 *            String to convert
	 * @return Remain of the source string
	 * @throws Exception
	 */
	protected String fillInfo(Parameter param, String content) throws Exception {
		int index;
		String opt = null;

		// get second parameter with options
		if (!content.isEmpty() && (content.charAt(0) == '{') && content.contains("}")) {
			index = content.indexOf('}');
			opt = content.substring(1, index);
			content = content.substring(index + 1);
		}
		if ((opt == null) || opt.isEmpty()) {
			throw new Exception("Option is missing!");
		}

		if (opt.compareToIgnoreCase("Directory") == 0) {
			if (param instanceof Project) {
				this.addString(((Project) param).getMopFile().getParent());
			} else {
				throw new Exception("'" + opt + "' is not defined for class '" + param.getClass().getName() + "'!");
			}
		} else if (opt.compareToIgnoreCase("File") == 0) {
			if (param instanceof Project) {
				this.addString(((Project) param).getMopFile().getName());
			} else {
				throw new Exception("'" + opt + "' is not defined for class '" + param.getClass().getName() + "'!");
			}
		} else if (opt.compareToIgnoreCase("Comment") == 0) {
			if (param instanceof Project) {
				this.addString(((Project) param).getComment());
			} else {
				throw new Exception("'" + opt + "' is not defined for class '" + param.getClass().getName() + "'!");
			}
		} else if (opt.compareToIgnoreCase("Index") == 0) {
			if (param instanceof Project) {
				this.addString(Integer.toString(((Project) param).getIndex() + 1));
			} else {
				throw new Exception("'" + opt + "' is not defined for class '" + param.getClass().getName() + "'!");
			}
		} else if (opt.compareToIgnoreCase("DSComment") == 0) {
			if (param instanceof Project) {
				this.addString(((Project) param).getDSComment());
			} else {
				throw new Exception("'" + opt + "' is not defined for class '" + param.getClass().getName() + "'!");
			}
		} else if (opt.compareToIgnoreCase("Identifier") == 0) {
			this.addString(param.getIdentifier());
		} else if (opt.compareToIgnoreCase("Description") == 0) {
			this.addString(param.getDescription());
		} else {
			throw new Exception("Undefined option!");
		}

		return content;
	}

	/**
	 * Fill keyword MOEXIST
	 * 
	 * @param param
	 *            Parameter to fill
	 * @param content
	 *            String to convert
	 * @return Remain of the source string
	 * @throws Exception
	 */
	protected String fillExist(Parameter param, String content) throws Exception {

		if (param == null) {
			this.addString("01");
		} else {
			this.addString("11");
		}

		return content;
	}

	/**
	 * Fill keyword MOCOMPONENTS
	 * 
	 * @param param
	 *            Parameter to fill
	 * @param path
	 *            Path to doc-files
	 * @param content
	 *            String to convert
	 * @return Remain of the source string
	 * @throws Exception
	 */
	protected String fillComponents(Parameter param, File path, String content) throws Exception {
		ParamContainer cont;
		String file;
		File relFile;
		DocFile[] docs;

		for (Parameter p : param.getParameters(ParamContainer.class)) {
			if (param != p) {
				cont = (ParamContainer) p;
				if (cont.getComponent() != null) {
					// check if docfiles is up to date
					if (cont.getUncertain()) {
						throw new Exception("Uncertain (Is parameter up to date?)");
					}
					docs = cont.getDocFiles();
					for (DocFile doc : docs) {
						if (doc.getDescription().toLowerCase().startsWith("component tex template ")) {
							if (this.shouldInsertTex(doc, docs)) {
								// get DocFile and insert
								file = FileUtils.trimPath(new File(path, doc.getFile())).getPath();
								relFile = FileUtils.getRelativePath(this.getDirectoryFile(),
										new File(this.getDirectoryFile(), file));
								if (new File(this.getDirectoryFile(), relFile.getPath()).exists()) {
									this.addTex(relFile);
									this.add("\n");
								} else {
									throw new Exception("File to add doesn't exists!");
								}
							}
						}
					}
				}
			}
		}

		return content;
	}

	/**
	 * Should the DocFile be inserted
	 * 
	 * @param doc
	 *            DocFile to check
	 * @param docs
	 *            List of DocFiles
	 * @return Should insert
	 */
	private boolean shouldInsertTex(DocFile doc, DocFile[] docs) {
		boolean insert = false, existEqualLang = false, existNoneLang = false;
		String[] thisItems, docItems, dItems;

		if (doc.getDescription().toLowerCase().startsWith("component tex template ")) {
			thisItems = FillTexReport.splitTexName(this.getFile().getName());
			docItems = FillTexReport.splitTexName(new File(doc.getFile()).getName());

			// first choice if same language
			insert = docItems[1].equalsIgnoreCase(thisItems[1]);
			if (!insert) {
				// check other DocFiles for no language or english
				for (DocFile d : docs) {
					if (d.getDescription().toLowerCase().startsWith("component tex template ")) {
						dItems = FillTexReport.splitTexName(new File(d.getFile()).getName());
						if (dItems[0].equalsIgnoreCase(docItems[0])) {
							if (dItems[1].equalsIgnoreCase(thisItems[1])) {
								existEqualLang = true;
							}
							if (dItems[1].equals("")) {
								existNoneLang = true;
							}
						}
					}
				}

				insert = !existEqualLang && docItems[1].equals("");
				if (!insert) {
					insert = !existEqualLang && !existNoneLang && docItems[1].equalsIgnoreCase("en");
				}
			}
		}
		return insert;
	}

	/**
	 * Split name to base-name and language
	 * 
	 * @param name
	 *            Name of file
	 * @return List of base-name and language
	 */
	private static String[] splitTexName(String name) {
		String lang = "";
		Pattern pat;
		Matcher match;

		// remove .tex end
		pat = Pattern.compile("(.*)\\.tex", Pattern.CASE_INSENSITIVE);
		match = pat.matcher(name);
		if (match.matches()) {
			name = match.group(1);
		}

		// remove _filled end
		pat = Pattern.compile("(.*)_filled", Pattern.CASE_INSENSITIVE);
		match = pat.matcher(name);
		if (match.matches()) {
			name = match.group(1);
		}

		// remove string to become unique name ("(number)")
		pat = Pattern.compile("(.*)\\(\\d+\\)");
		match = pat.matcher(name);
		if (match.matches()) {
			name = match.group(1);
		}

		// split base-name and language
		pat = Pattern.compile("(.*)\\.(.+)");
		match = pat.matcher(name);
		if (match.matches()) {
			name = match.group(1);
			lang = match.group(2);
		}

		return new String[] { name, lang };
	}

	/**
	 * Write message to log
	 * 
	 * @param msg
	 *            Message
	 */
	protected void logging(String msg) {
		this.log.append(msg + "\n");
	}

	/**
	 * Get messages from log
	 * 
	 * @return Logging
	 */
	public String getLogging() {
		return this.log.getText();
	}
}
