/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9749 $
 * $Date: 2020-04-10 16:00:50 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.rmi.RemoteException;
import java.util.StringTokenizer;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamComboBox;
import at.lcm.bu21.magopt.structure.rmi.RParamComboBox;

/**
 * ComboBox parameter to select an entry only from a given list
 * 
 * @author Werner Koppelsttter
 */
public class ParamComboBox extends ParamField implements IParamComboBox {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** List of ComboBox entries */
	private String comboList = new String();

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	public ParamComboBox() throws Exception {
	}

	/**
	 * Constructor with name
	 * 
	 * @param name
	 *            Name of Parameter
	 * @throws Exception
	 */
	public ParamComboBox(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Create a ComboBox parameter with given name and description
	 * 
	 * @param name
	 *            Name of Parameter
	 * @param description
	 *            Description of Parameter
	 * @throws Exception
	 */
	public ParamComboBox(String name, String description) throws Exception {
		super(name, description);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamComboBox other = (ParamComboBox) obj;
		if (this.comboList == null) {
			if (other.comboList != null) {
				return false;
			}
		} else if (!this.comboList.equals(other.comboList)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamComboBox deepClone() throws CloneException {
		return (ParamComboBox) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException, InstantiationException, IllegalAccessException {

		super.cloneParam(clone);

		if (clone instanceof ParamComboBox) {
			ParamComboBox other = (ParamComboBox) clone;

			other.comboList = this.comboList;
		}
	}

	@Override
	public String[] getList() {
		int i = 0;
		StringTokenizer tok = new StringTokenizer(this.comboList, ";");
		String[] list = new String[tok.countTokens()];
		while (tok.hasMoreTokens()) {
			list[i++] = tok.nextToken().trim();
		}

		if (list.length == 0) {
			list = new String[] { "" };
		}
		return list;
	}

	@Override
	public String getComboList() {
		return this.comboList;
	}

	@Override
	public void setComboList(String comboList) throws Exception {
		if (comboList == null) {
			comboList = "";
		}
		if (this.comboList.compareTo(comboList) != 0) {
			this.comboList = comboList;
			if (this.getConstant()) {
				this.setComboIndex(this.getComboIndex());
			} else {
				for (int i = 0; i < this.getLength(); i++) {
					this.setComboIndex(i, this.getComboIndex(i));
				}
			}
		}
	}

	@Override
	public int getComboIndex() {
		return this.getComboIndex(this.getIndex());
	}

	/**
	 * Get index of combobox
	 * 
	 * @param index
	 *            Index of dataset
	 * @return Index of comobox
	 */
	private int getComboIndex(int index) {
		String[] list;
		int comboIndex;

		list = this.getList();
		comboIndex = list.length - 1;
		while ((comboIndex > 0) && !this.getTData(list[comboIndex]).equals(this.getData(index))) {
			comboIndex--;
		}

		return comboIndex;
	}

	@Override
	public void setComboIndex(int comboIndex) throws Exception {
		this.setComboIndex(this.getIndex(), comboIndex);
	}

	/**
	 * Set index of combobox
	 * 
	 * @param index
	 *            Index of dataset
	 * @param comboIndex
	 *            Index of combobox
	 * @throws Exception
	 */
	private void setComboIndex(int index, int comboIndex) throws Exception {
		TData data = new TNull();

		if ((comboIndex >= 0) && (comboIndex < this.getList().length)) {
			data = this.getTData(this.getList()[comboIndex]);
		}

		this.setData(index, data);
	}

	/**
	 * Get TData-object from string
	 * 
	 * @param str
	 *            string of convert
	 * @return TData-object
	 */
	private TData getTData(String str) {
		TData data;

		if ((str == null) || (str.isEmpty())) {
			data = new TNull();
		} else {
			data = new TString(str);
			try {
				data = new TNumeric(Double.parseDouble(str));
			} catch (Exception e) {
				if (str.compareToIgnoreCase(Boolean.TRUE.toString()) == 0) {
					data = new TBoolean(true);
				} else if (str.compareToIgnoreCase(Boolean.FALSE.toString()) == 0) {
					data = new TBoolean(false);
				}
			}
		}
		return data;
	}

	@Override
	public void setData(TData data) throws Exception {
		int index = -1;
		String[] values = this.getList();
		for (int i = 0; i < values.length; i++) {
			if (this.getTData(values[i]).equals(data)) {
				index = i;
			}
		}

		if (index == -1) {
			this.warning("Could not find \"" + data.toString() + "\" in combobox list");
		} else {
			this.setComboIndex(index);
		}
	}

	/**
	 * Set unit of Parameter. Default behavior is not convert values.
	 * 
	 * @param unit
	 *            unit
	 * @throws RemoteException
	 * @throws Exception
	 */
	@Override
	public void setUnit(String unit) throws Exception {
		this.setUnit(unit, false);
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamFieldList\n");
		str.append("\n\tName        : ");
		str.append(this.getName());
		str.append("\n\tDescription : ");
		str.append(this.getDescription());
		str.append("\n\tData        : ");
		str.append(this.getData());
		str.append("\n\tUnit        : ");
		str.append(this.getUnit());
		str.append("\n\tTex String  : ");
		str.append(this.getTexString());
		str.append("\n\tType        : ");
		str.append(this.getType());
		str.append("\n\tList        : ");
		str.append(this.getList());
		return str.toString();
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamField")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("ComboList")) {
				xmlReader.next();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.comboList += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamComboBox");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		// List
		xmlWriter.writeStartElement("ComboList");
		if (this.comboList != null) {
			xmlWriter.writeCharacters(this.comboList);
		}
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	void reportInfo(DocTexWriter content) {
		super.reportInfo(content);

		content.add("\tItems & ");
		content.addString(this.getComboList());
		content.add(" \\\\\n");
	}

	@Override
	public IParamComboBox getRMIImpl() throws RemoteException {
		return new RParamComboBox(this);
	}

	@Override
	public IParamComboBox getClientImpl() {
		return new CParamComboBox(this);
	}
}
