/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9868 $
 * $Date: 2020-06-04 12:11:54 +0200 (Thu, 04 Jun 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.rmi.RemoteException;
import java.util.List;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.dtypes.unit.ConversionException;
import at.lcm.bu21.general.dtypes.unit.UnitConversion;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamField;
import at.lcm.bu21.magopt.structure.rmi.RParamField;
import at.lcm.bu21.magopt.utils.MagOptConfig;

/**
 * Parameter for data-representation
 * 
 * @author Werner Koppelsttter
 */
public class ParamField extends Parameter implements IParamField {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Tex String of Parameter */
	private String texString = new String();

	/** Unit of Parameter */
	private String unit = new String();

	/** Type of Parameter */
	private String type = new String();

	/** Database identifier (-1 ... not linked with database) */
	private long dbId = -1;

	/** Lock flag */
	private boolean lock = false;

	/** Data from constant parameter */
	private TData data = null;

	@Override
	public List<String> listTabNames() {
		List<String> list = super.listTabNames();
		list.add(0, "Field Info");
		return list;
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamField() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name
	 *            Name of Parameter
	 * @throws Exception
	 */
	public ParamField(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name
	 *            Name of Parameter
	 * @param description
	 *            Description of Parameter
	 * @throws Exception
	 */
	public ParamField(String name, String description) throws Exception {
		super(name, description);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamField other = (ParamField) obj;
		if (this.data == null) {
			if (other.data != null) {
				return false;
			}
		} else if (!this.data.equals(other.data)) {
			return false;
		}
		if (this.dbId != other.dbId) {
			return false;
		}
		if (this.lock != other.lock) {
			return false;
		}
		if (this.texString == null) {
			if (other.texString != null) {
				return false;
			}
		} else if (!this.texString.equals(other.texString)) {
			return false;
		}
		if (this.type == null) {
			if (other.type != null) {
				return false;
			}
		} else if (!this.type.equals(other.type)) {
			return false;
		}
		if (this.unit == null) {
			if (other.unit != null) {
				return false;
			}
		} else if (!this.unit.equals(other.unit)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamField deepClone() throws CloneException {
		return (ParamField) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException, InstantiationException, IllegalAccessException {

		super.cloneParam(clone);

		if (clone instanceof ParamField) {
			ParamField other = (ParamField) clone;

			other.texString = this.texString;
			other.unit = this.unit;
			other.type = this.type;
			other.dbId = this.dbId;
			other.lock = this.lock;

			if (this.data != null) {
				other.data = this.data.deepClone();
			} else {
				other.data = null;
			}
		}
	}

	@Override
	void cloneParam(Parameter clone, Dataset dsThis, Dataset dsClone, boolean docFiles) throws CloneException {
		TData data;

		super.cloneParam(clone, dsThis, dsClone, docFiles);

		if (clone instanceof ParamField) {
			ParamField other = (ParamField) clone;

			data = dsThis.getValue(this.getPId(), ParamSetKey.DATA);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.DATA, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.DATA);
			}
		}

	}

	@Override
	public ParamField convertToParamField() {
		ParamField other = null;

		try {
			// clone parameter
			other = new ParamField("replace");
			this.convertParamField(other);
		} catch (Exception ex) {
			this.error("Unable to convert", ex.getStackTrace());
		}

		return other;
	}

	@Override
	public ParamField convertToParamComboBox() {
		ParamComboBox other = null;

		try {
			// clone parameter
			other = new ParamComboBox("replace");
			// FIXME COMBOBOX add current values to list
			this.convertParamField(other);
		} catch (Exception ex) {
			this.error("Unable to convert", ex.getStackTrace());
		}

		return other;
	}

	@Override
	public ParamFormula convertToParamFormula() {
		ParamFormula other = null;

		if (this.getConstant()) {
			this.warning("Cannot convert from constant Field to Formula!");
		} else {
			try {
				// clone parameter
				other = new ParamFormula("replace");
				this.convertParamField(other);
			} catch (Exception ex) {
				this.error("Unable to convert", ex.getStackTrace());
			}
		}

		return other;
	}

	@Override
	public ParamLinkedField convertToParamLinkedField() {
		ParamLinkedField other = null;

		if (this.getConstant()) {
			this.warning("Cannot convert from constant Field to LinkedField!");
		} else {
			try {
				// clone parameter
				other = new ParamLinkedField("replace");
				this.convertParamField(other);
			} catch (Exception ex) {
				this.error("Unable to convert", ex.getStackTrace());
			}
		}

		return other;
	}

	/**
	 * Convert ParamField
	 * 
	 * @param other
	 *            ParamContainer
	 * @throws Exception
	 */
	private void convertParamField(ParamField other) throws Exception {

		this.cloneParam(other);
		if (this.getParent() != null) {
			this.getParent().replaceObj(this, other);
		}
		this.refresh();
		other.refresh();
	}

	@Override
	public String getTexString() {
		return this.texString;
	}

	@Override
	public void setTexString(String texString) {
		this.texString = texString;
	}

	@Override
	public String getUnit() {
		return this.unit;
	}

	@Override
	public void setUnit(String unit) throws Exception {
		this.setUnit(unit, true);
	}

	@Override
	public void setUnit(String unit, boolean convert) throws Exception {
		String oldUnit;

		if (this.unit.compareTo(unit) != 0) {
			oldUnit = this.unit;
			this.unit = unit;
			for (int k = 0; k < this.getLength(); k++) {
				this.setModified(k, true);
				if (convert) {
					this.setData(k, this.getData(k), oldUnit);
				}
			}
		}
	}

	@Override
	public String getType() {
		return this.type;
	}

	@Override
	public void setType(String type) {
		this.type = type;
	}

	@Override
	public long getDbId() {
		return this.dbId;
	}

	/**
	 * Set the linked database identifier
	 * 
	 * @param id
	 *            Database id
	 */
	public void setDbId(long id) {
		this.dbId = id;
	}

	@Override
	public void clearDbId() {
		this.dbId = -1;
	}

	@Override
	public boolean getLock() {
		return this.lock;
	}

	/**
	 * Set lock status
	 * 
	 * @param lock
	 *            Lock status
	 */
	public void setLock(boolean lock) {
		this.lock = lock;
	}

	@Override
	public TData getData(int index) {
		Dataset ds;
		TData data = null;

		if (this.getConstant()) {
			data = this.data;
		} else {
			ds = this.getDataset(index);
			if (ds != null) {
				data = ds.getValue(this.getPId(), ParamSetKey.DATA);
			}
		}

		try {
			data = data.deepClone();
		} catch (Exception e) {
			data = new TNull();
		}

		return data;
	}

	@Override
	public TData getData(String unit) throws ConversionException {
		return this.getData(this.getIndex(), unit);
	}

	@Override
	public TData getData(int index, String unit) throws ConversionException {
		TData data;
		UnitConversion uc;
		boolean unitConversion = false;
		String tmp;
		String sUnit;

		data = this.getData(index);
		if (unit == null) {
			unit = "";
		}
		sUnit = this.getUnit();
		if (sUnit == null) {
			sUnit = "";
		}
		if (!(unit.isEmpty() && sUnit.isEmpty())) {
			if (!unit.equals(sUnit)) {
				tmp = MagOptConfig.getConfig("EnableUnitConversion");
				if (tmp != null) {
					unitConversion = Boolean.parseBoolean(tmp);
				}

				if (unitConversion) {
					uc = new UnitConversion(sUnit, unit);
					data = uc.convert(data);
				}
			}
		}

		return data;
	}

	@Override
	public TStruct getDataStruct() {
		TStruct struct;

		struct = super.getDataStruct();
		struct.put("unit__", new TString(this.getUnit()));

		return struct;
	}

	@Override
	public void setData(TData data) throws Exception {
		this.setData(this.getIndex(), data);
	}

	/**
	 * Set data of given dataset
	 * 
	 * @param index
	 *            Index of dataset
	 * @param data
	 *            data
	 * @throws Exception
	 */
	protected void setData(int index, TData data) throws Exception {
		Dataset ds;

		if (!this.getData(index).equals(data)) {
			if (this.getLock()) {
				throw new Exception("Value of locked field cannot be changed! (" + this.getIdentifier() + ")");
			} else {
				data = data.deepClone();
				if (this.getConstant()) {
					this.data = data;
					for (int i = 0; i < this.getLength(); i++) {
						this.setModified(i, true);
					}
				} else {
					ds = this.getDataset(index);
					if (ds != null) {
						ds.setValue(this.getPId(), ParamSetKey.DATA, data);
					}
					this.setModified(index, true);
				}
			}
		}
	}

	@Override
	public void setData(TData data, String unit) throws Exception {
		this.setData(this.getIndex(), data, unit);
	}

	/**
	 * Set data of field with values of unit for conversion.
	 * 
	 * @param index
	 *            Index of data
	 * @param data
	 *            Data to set
	 * @param unit
	 *            Unit of data
	 * @throws Exception
	 */
	protected void setData(int index, TData data, String unit) throws Exception {
		UnitConversion uc;
		boolean unitConversion = false;
		String tmp;
		String dUnit;

		if (unit == null) {
			unit = "";
		}
		dUnit = this.getUnit();
		if (dUnit == null) {
			dUnit = "";
		}
		if (!(unit.isEmpty() && dUnit.isEmpty())) {
			if (!unit.equals(dUnit)) {
				tmp = MagOptConfig.getConfig("EnableUnitConversion");
				if (tmp != null) {
					unitConversion = Boolean.parseBoolean(tmp);
				}

				if (unitConversion) {
					uc = new UnitConversion(unit, dUnit);
					data = uc.convert(data);
				}
			}
		}

		this.setData(index, data);
	}

	@Override
	public void setConstant(boolean constant) {
		Dataset ds;

		if (this.getConstant() != constant) {
			if (this.getLock()) {
				this.warning("Constant cannot changed of locked field! (" + this.getIdentifier() + ")");
			} else {
				if (constant) {
					this.data = this.getData();
					for (int i = 0; i < this.getLength(); i++) {
						if (!this.data.equals(this.getData(i))) {
							this.setModified(i, true);
						}
						ds = this.getDataset(i);
						if (ds != null) {
							ds.rmValue(this.getPId(), ParamSetKey.DATA);
						}
					}
				} else {
					for (int i = 0; i < this.getLength(); i++) {
						try {
							ds = this.getDataset(i);
							if (ds != null) {
								ds.setValue(this.getPId(), ParamSetKey.DATA, this.data.deepClone());
							}
						} catch (CloneException ex) {
							this.warning("Unable to clone constant data");
						}
					}
					this.data = null;
				}
				super.setConstant(constant);
			}
		}
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamField\n");
		str.append("\n\tName        : ");
		str.append(this.getName());
		str.append("\n\tDescription : ");
		str.append(this.getDescription());
		str.append("\n\tData        : ");
		str.append(this.getData());
		str.append("\n\tUnit        : ");
		str.append(this.getUnit());
		str.append("\n\tTex String  : ");
		str.append(this.getTexString());
		str.append("\n\tType        : ");
		str.append(this.getType());
		return str.toString();
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("Parameter")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("Data")) {
				xmlReader.next();
				if (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
					this.data = TData.xml2obj(xmlReader);
				}
				xmlReader.next();
			} else if (element.equals("TexString")) {
				xmlReader.next();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.texString += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Unit")) {
				xmlReader.next();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.unit += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Type")) {
				xmlReader.next();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.type += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("DbId")) {
				xmlReader.next();
				this.dbId = Long.parseLong(xmlReader.getText());
				xmlReader.next();
				xmlReader.next();
			} else if (element.equals("Lock")) {
				xmlReader.next();
				this.lock = Boolean.parseBoolean(xmlReader.getText());
				xmlReader.next();
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamField");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		// Data
		xmlWriter.writeStartElement("Data");
		if (this.data != null) {
			this.data.obj2xml(xmlWriter);
		}
		xmlWriter.writeEndElement();

		// TexString
		xmlWriter.writeStartElement("TexString");
		if (this.texString != null) {
			xmlWriter.writeCharacters(this.texString);
		}
		xmlWriter.writeEndElement();

		// Unit
		xmlWriter.writeStartElement("Unit");
		if (this.unit != null) {
			xmlWriter.writeCharacters(this.unit);
		}
		xmlWriter.writeEndElement();

		// Type
		xmlWriter.writeStartElement("Type");
		if (this.type != null) {
			xmlWriter.writeCharacters(this.type);
		}
		xmlWriter.writeEndElement();

		// Database identifier
		xmlWriter.writeStartElement("DbId");
		xmlWriter.writeCharacters(Long.toString(this.dbId));
		xmlWriter.writeEndElement();

		// Lock
		xmlWriter.writeStartElement("Lock");
		xmlWriter.writeCharacters(Boolean.toString(this.lock));
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	void reportTable(DocTexWriter content, int indent) {
		String tex;

		super.reportTable(content, indent);

		tex = this.getTexString();
		if ((tex != null) && !tex.isEmpty()) {
			if (!this.getUsed()) {
				content.add("\\textcolor{gray}{");
			}
			content.add("$" + this.getTexString() + "$");
			if (!this.getUsed()) {
				content.add("}");
			}
		}
		content.add(" & ");

		if (!this.getUsed()) {
			content.add("\\textcolor{gray}{");
		}
		if (this.getUncertain()) {
			content.add("[Uncertain]");
		} else {
			content.addString(this.getData().toString());
		}
		if (!this.getUsed()) {
			content.add("}");
		}
		content.add(" & ");

		if (!this.getUsed()) {
			content.add("\\textcolor{gray}{");
		}
		content.addString(this.getUnit());
		if (!this.getUsed()) {
			content.add("}");
		}

	}

	@Override
	void reportInfo(DocTexWriter content) {
		String tex = this.getTexString();

		super.reportInfo(content);

		content.add("\tValue & ");

		if ((tex != null) && (tex.length() != 0)) {
			content.add("$" + this.getTexString() + "$ = ");
		}
		if (this.getUncertain()) {
			content.add("[Uncertain]");
		} else {
			content.addString(this.getData().toString());
		}
		content.add(" ");
		content.addString(this.getUnit());
		content.add("\\\\\n");
	}

	@Override
	List<ParamSetKey> listParamSetKey() {
		List<ParamSetKey> list;

		list = super.listParamSetKey();
		list.add(ParamSetKey.DATA);

		return list;
	}

	@Override
	public IParamField getRMIImpl() throws RemoteException {
		return new RParamField(this);
	}

	@Override
	public IParamField getClientImpl() {
		return new CParamField(this);
	}
}
