/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.util.Hashtable;
import java.util.Map;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.magopt.docu.DocTexWriter;

/**
 * Base class of java-functions with defines interface to the structure classes
 * 
 * @author Werner Koppelsttter
 */
public abstract class ParamJavaFnc {

	/** Reference to the structure of program */
	public ParamFunction structure = null;

	/**
	 * Returns the function id
	 * 
	 * @return Function-id
	 */
	public String getFunctionID() {
		return "";
	}

	/**
	 * Returns the default description of the function
	 * 
	 * @return Default-description
	 */
	public abstract String getDefaultDescription();

	/**
	 * Abstract method definition which return list of import variables
	 * 
	 * @return list of imports
	 */
	public abstract PortDesc[] importVars();

	/**
	 * Abstract method definition which return list of export variables
	 * 
	 * @return list of exports
	 */
	public abstract PortDesc[] exportVars();

	/**
	 * Abstract method definition for the main functionality of class
	 * 
	 * @param args
	 *            Arguments of caller (Control flags)
	 * @return status code. 0 if method runs through and otherwise any other value
	 * @throws Exception
	 */
	public abstract int run(Map<String, Object> args) throws Exception;

	/**
	 * Main method which is called from structure
	 * 
	 * @param args
	 *            Arguments of caller (Control flags)
	 * @return status code
	 * @throws Exception
	 */
	public int update(Map<String, Object> args) throws Exception {
		int status = 0;
		Parameter p;

		status = this.run(args);

		p = this.structure.getRoot();
		if (p instanceof Project) {
			((Project) p).check();
		}
		if (status == 0) {
			this.structure.setUpdateRequired(false);
		}

		return status;
	}

	/**
	 * Refresh the dependencies of function
	 */
	public void refreshDependency() {
		this.setImportVars(this.importVars());
		this.setExportVars(this.exportVars());
	}

	/**
	 * Reset function
	 */
	public void reset() {
	}

	/**
	 * Function for generating documentation-files
	 * 
	 * @param directory
	 *            Directory for export-files
	 * @throws Exception
	 */
	public void generateDocFiles(File directory) throws Exception {

	}

	/**
	 * Get array of import-assignments of function
	 * 
	 * @return Array of imports
	 */
	protected DependItem[] getImports() {
		return this.structure.getDepend().getImports();
	}

	/**
	 * Set list of import-variables
	 * 
	 * @param vars
	 *            Array of import-variable
	 */
	protected void setImportVars(PortDesc[] vars) {
		this.structure.getDepend().setImportVars(vars);
	}

	/**
	 * Assign Parameters to the import-variables
	 * 
	 * @param vars
	 *            Array of assignments
	 */
	protected void setImportAssignments(String[][] vars) {
		this.structure.getDepend().setImportAssignments(vars);
	}

	/**
	 * Get array of export-assignments of function
	 * 
	 * @return Array of exports
	 */
	protected DependItem[] getExports() {
		return this.structure.getDepend().getExports();
	}

	/**
	 * Set list of export-variables
	 * 
	 * @param vars
	 *            Array of export-variable
	 */
	protected void setExportVars(PortDesc[] vars) {
		this.structure.getDepend().setExportVars(vars);
	}

	/**
	 * Assign Parameters to the export-variables
	 * 
	 * @param vars
	 *            Array of assignments
	 */
	protected void setExportAssignments(String[][] vars) {
		this.structure.getDepend().setExportAssignments(vars);
	}

	/**
	 * Add information message to event-log
	 * 
	 * @param message
	 *            Message
	 */
	public void info(String message) {
		if (this.structure != null) {
			this.structure.info(message, this.structure.getStackTrace());
		}
	}

	/**
	 * Add information message to event-log
	 * 
	 * @param message
	 *            Message
	 * @param stack
	 *            Stack-trace
	 */
	protected void info(String message, StackTraceElement[] stack) {
		if (this.structure != null) {
			this.structure.info(message, stack);
		}
	}

	/**
	 * Add warning message to event-log
	 * 
	 * @param message
	 *            Message
	 */
	public void warning(String message) {
		if (this.structure != null) {
			this.structure.warning(message, this.structure.getStackTrace());
		}
	}

	/**
	 * Add warning message to event-log
	 * 
	 * @param message
	 *            Message
	 * @param stack
	 *            Stack-trace
	 */
	protected void warning(String message, StackTraceElement[] stack) {
		if (this.structure != null) {
			this.structure.warning(message, stack);
		}
	}

	/**
	 * Add error message to event-log
	 * 
	 * @param message
	 *            Message
	 */
	public void error(String message) {
		if (this.structure != null) {
			this.structure.error(message, this.structure.getStackTrace());
		}
	}

	/**
	 * Add error message to event-log
	 * 
	 * @param message
	 *            Message
	 * @param stack
	 *            Stack-trace
	 */
	protected void error(String message, StackTraceElement[] stack) {
		if (this.structure != null) {
			this.structure.error(message, stack);
		}
	}

	/**
	 * Get import values
	 * 
	 * @return Map with import values
	 * @throws Exception
	 */
	protected Map<String, TData> importValues() throws Exception {
		Map<String, TData> map = new Hashtable<String, TData>();
		DependItem[] depends = this.getImports();
		String name;
		TData data;

		for (DependItem depend : depends) {
			name = depend.getVarName();
			data = this.structure.getImportValue(name);
			if (data != null) {
				map.put(name, data);
			}
		}
		return map;
	}

	/**
	 * Get export with the given name
	 * 
	 * @param map
	 *            Map for export values
	 * @throws Exception
	 */
	protected void revertExportValue(Map<String, TData> map) throws Exception {
		DependItem[] depends = this.getExports();
		String name;
		TData data;

		for (DependItem depend : depends) {
			name = depend.getVarName();
			data = this.structure.getExportValue(name);
			if (data != null) {
				map.put(name, data);
			}
		}
	}

	/**
	 * Set value to the export with the given name
	 * 
	 * @param map
	 *            Map with export values
	 * @throws Exception
	 */
	protected void exportValues(Map<String, TData> map) throws Exception {
		DependItem[] depends = this.getExports();
		String name;

		for (DependItem depend : depends) {
			name = depend.getVarName();
			this.structure.setExportValue(name, map.get(name));
		}
	}

	/**
	 * Get unit of import. The unit of the import description and if this is empty
	 * then the unit from the parameter
	 * 
	 * @param name
	 *            Name of import
	 * @return Unit string
	 */
	protected String getImportUnit(String name) {
		return this.structure.getImportUnit(name);
	}

	/**
	 * Get unit of export. The unit of the import description and if this is empty
	 * then the unit from the parameter
	 * 
	 * @param name
	 *            Name of export
	 * @return Unit string
	 */
	protected String getExportUnit(String name) {
		return this.structure.getExportUnit(name);
	}

	/**
	 * Register documentation-file
	 * 
	 * @param name
	 *            Name for registration
	 * @param strFile
	 *            Name of file
	 * @param description
	 *            Description to the file
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 */
	protected void registerDocFile(String name, String strFile, String description)
			throws NoSuchAlgorithmException, IOException {
		this.structure.registerDocFile(name, strFile, description);
	}

	/**
	 * Get array with registered documentation-files
	 * 
	 * @return Array with doc-files
	 */
	protected DocFile[] getDocFiles() {
		return this.structure.getDocFiles();
	}

	/**
	 * Get array with registered documentation-files
	 * 
	 * @param name
	 *            Name of doc-file
	 * 
	 * @return Array with doc-files
	 */
	protected DocFile getDocFile(String name) {
		return this.structure.getDocFile(name);
	}

	/**
	 * Get temporary data with given key
	 * 
	 * @param key
	 *            Key for data
	 * @return Value of data
	 */
	public Object getTmpData(String key) {
		return this.structure.getTmpData(key);
	}

	/**
	 * Set temporary data under given key
	 * 
	 * @param key
	 *            Key for data
	 * @param obj
	 *            Value of data
	 */
	public void setTmpData(String key, Object obj) {
		this.structure.setTmpData(key, obj);
	}

	/**
	 * Remove temporary data of given key
	 * 
	 * @param key
	 *            Key of data
	 */
	public void rmTmpData(String key) {
		this.structure.rmTmpData(key);
	}

	/**
	 * Clear temporary data
	 */
	public void clearTmpData() {
		this.structure.clearTmpData();
	}

	/**
	 * Get temporary dataset data with given key
	 * 
	 * @param key
	 *            Key for data
	 * @return Value of data
	 */
	public Object getTmpDSData(String key) {
		return this.structure.getTmpDSData(key);
	}

	/**
	 * Set temporary dataset data under given key
	 * 
	 * @param key
	 *            Key for data
	 * @param obj
	 *            Value of data
	 */
	public void setTmpDSData(String key, Object obj) {
		this.structure.setTmpDSData(key, obj);
	}

	/**
	 * Remove temporary dataset data of given key
	 * 
	 * @param key
	 *            Key for data
	 */
	public void rmTmpDSData(String key) {
		this.structure.rmTmpDSData(key);
	}

	/**
	 * Clear temporary dataset data
	 */
	public void clearTmpDSData() {
		this.structure.clearTmpDSData();
	}

	/**
	 * Report TeX-code
	 * 
	 * @param template
	 *            Content of template
	 * @return TeX-code
	 * @throws Exception
	 */
	protected DocTexWriter reportTex(DocTexWriter template) throws Exception {
		return template;
	}

	/**
	 * Get unique doc-file name in directory. Compare without extension
	 * 
	 * @param file
	 *            File to start
	 * @return Unique file
	 */
	public File getUniqueDocFileName(File file) {
		return this.structure.getUniqueDocFileName(file);
	}

	/**
	 * Get parameter-directory of dataset
	 * 
	 * @return Parameter-directroy
	 */
	public File getParamDirectory() {
		return this.structure.getParamDirectory();
	}

	/**
	 * Get parameter-identifier
	 * 
	 * @return Parameter-directroy
	 */
	public String getPId() {
		return this.structure.getPId();
	}

	/**
	 * Get Identifier
	 * 
	 * @return Parameter-directroy
	 */
	public String getIdentifier() {
		return this.structure.getIdentifier();
	}

	/**
	 * Get project
	 * 
	 * @return Project if exist
	 */
	public Project getProject() {
		Parameter param;
		Project proj = null;

		param = this.structure.getRoot();
		if (param instanceof Project) {
			proj = (Project) param;
		}

		return proj;
	}

}
