/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.client;

import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.security.NoSuchAlgorithmException;
import java.util.List;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.magopt.structure.DocFile;
import at.lcm.bu21.magopt.structure.IParamContainer;
import at.lcm.bu21.magopt.structure.IParameter;

/**
 * Client adapter for parameter interface
 * 
 * @author Werner Koppelsttter
 */
public abstract class CParameter implements IParameter {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Parameter class */
	private IParameter param;

	/**
	 * Create client adapter for parameter interface
	 * 
	 * @param param
	 *            Parameter
	 */
	public CParameter(IParameter param) {
		this.param = param;
	}

	@Override
	public String getPId() throws RemoteException {
		return this.param.getPId();
	}

	@Override
	public String getName() throws RemoteException {
		return this.param.getName();
	}

	@Override
	public void setName(String name) throws Exception {
		this.param.setName(name);
	}

	@Override
	public String getDescription() throws RemoteException {
		return this.param.getDescription();
	}

	@Override
	public void setDescription(String description) throws RemoteException {
		this.param.setDescription(description);
	}

	@Override
	public String getComment() throws RemoteException {
		return this.param.getComment();
	}

	@Override
	public void setComment(String comment) throws RemoteException {
		this.param.setComment(comment);
	}

	@Override
	public String getIdentifier() throws RemoteException {
		return this.param.getIdentifier();
	}

	@Override
	public String getShortIdentifier() throws RemoteException {
		return this.param.getShortIdentifier();
	}

	@Override
	public int getIndex() throws RemoteException {
		return this.param.getIndex();
	}

	@Override
	public void setIndex(int index) throws RemoteException {
		this.param.setIndex(index);
	}

	@Override
	public int getLength() throws RemoteException {
		return this.param.getLength();
	}

	@Override
	public TData getData() throws RemoteException {
		return this.param.getData();
	}

	@Override
	public TData getData(int index) throws RemoteException {
		return this.param.getData(index);
	}

	@Override
	public void setData(TData data) throws Exception {
		this.param.setData(data);
	}

	@Override
	public boolean getPrint() throws RemoteException {
		return this.param.getPrint();
	}

	@Override
	public void setPrint(boolean print) throws RemoteException {
		this.param.setPrint(print);
	}

	@Override
	public boolean getConstant() throws RemoteException {
		return this.param.getConstant();
	}

	@Override
	public void setConstant(boolean constant) throws RemoteException {
		this.param.setConstant(constant);
	}

	@Override
	public boolean getUsed() throws RemoteException {
		return this.param.getUsed();
	}

	@Override
	public boolean getReadOnly() throws RemoteException {
		return this.param.getReadOnly();
	}

	@Override
	public boolean getModified() throws RemoteException {
		return this.param.getModified();
	}

	@Override
	public void setModified(boolean modified) throws RemoteException {
		this.param.setModified(modified);
	}

	@Override
	public IParameter[] getUsedBy() throws RemoteException {
		IParameter[] params;
		IParameter[] rmiParams;

		params = this.param.getUsedBy();
		rmiParams = new IParameter[params.length];
		for (int k = 0; k < params.length; k++) {
			rmiParams[k] = params[k].getClientImpl();
		}

		return rmiParams;
	}

	@Override
	public IParameter[] getSetBy() throws RemoteException {
		IParameter[] params;
		IParameter[] rmiParams;

		params = this.param.getSetBy();
		rmiParams = new IParameter[params.length];
		for (int k = 0; k < params.length; k++) {
			rmiParams[k] = params[k].getClientImpl();
		}

		return rmiParams;
	}

	@Override
	public IParameter removeParam() throws RemoteException {
		return this.param.removeParam().getClientImpl();
	}

	@Override
	public void moveUp() throws RemoteException {
		this.param.moveUp();
	}

	@Override
	public void moveDown() throws RemoteException {
		this.param.moveDown();
	}

	@Override
	public int getIndexOfParameter() throws RemoteException {
		return this.param.getIndexOfParameter();
	}

	@Override
	public boolean getUncertain() throws RemoteException {
		return this.param.getUncertain();
	}

	@Override
	public boolean getError() throws RemoteException {
		return this.param.getError();
	}

	@Override
	public boolean[] getAllError() throws RemoteException {
		return this.param.getAllError();
	}

	@Override
	public void setError(boolean error) throws RemoteException {
		this.param.setError(error);
	}

	@Override
	public void refresh() throws RemoteException {
		this.param.refresh();
	}

	@Override
	public IParamContainer getParentParameter() throws RemoteException {
		IParamContainer cont;
		IParamContainer rmiCont = null;

		cont = this.param.getParentParameter();
		if (cont != null) {
			rmiCont = cont.getClientImpl();
		}

		return rmiCont;
	}

	@Override
	public IParameter getUniqueObject(String uniqueIdentifier) throws RemoteException {
		return this.param.getUniqueObject(uniqueIdentifier).getClientImpl();
	}

	@Override
	public IParameter getObject(String identifier) throws RemoteException {
		return this.param.getObject(identifier).getClientImpl();
	}

	@Override
	public IParameter getParamByPid(String pid) throws RemoteException {
		return this.param.getParamByPid(pid).getClientImpl();
	}

	@Override
	public IParameter[] getParameters(String c) throws RemoteException {
		IParameter[] params;
		IParameter[] rmiParams;

		params = this.param.getParameters(c);
		rmiParams = new IParameter[params.length];
		for (int k = 0; k < params.length; k++) {
			rmiParams[k] = params[k].getClientImpl();
		}

		return rmiParams;
	}

	@Override
	public IParameter[] getAnalysisParams() throws RemoteException {
		IParameter[] params;
		IParameter[] rmiParams;

		params = this.param.getAnalysisParams();
		rmiParams = new IParameter[params.length];
		for (int k = 0; k < params.length; k++) {
			rmiParams[k] = params[k].getClientImpl();
		}

		return rmiParams;
	}

	@Override
	public IParameter[] getDynamicParams() throws RemoteException {
		IParameter[] params;
		IParameter[] rmiParams;

		params = this.param.getDynamicParams();
		rmiParams = new IParameter[params.length];
		for (int k = 0; k < params.length; k++) {
			rmiParams[k] = params[k].getClientImpl();
		}

		return rmiParams;
	}

	@Override
	public void info(String message) throws RemoteException {
		this.param.info(message);
	}

	@Override
	public void info(String message, StackTraceElement[] stack) throws RemoteException {
		this.param.info(message, stack);
	}

	@Override
	public void info(String identifier, String message, StackTraceElement[] stack) throws RemoteException {
		this.param.info(identifier, message, stack);
	}

	@Override
	public void warning(String message) throws RemoteException {
		this.param.warning(message);
	}

	@Override
	public void warning(String message, StackTraceElement[] stack) throws RemoteException {
		this.param.warning(message, stack);
	}

	@Override
	public void warning(String identifier, String message, StackTraceElement[] stack) throws RemoteException {
		this.param.warning(identifier, message, stack);
	}

	@Override
	public void error(String message) throws RemoteException {
		this.param.error(message);
	}

	@Override
	public void error(String message, StackTraceElement[] stack) throws RemoteException {
		this.param.error(message, stack);
	}

	@Override
	public void error(String identifier, String message, StackTraceElement[] stack) throws RemoteException {
		this.param.error(identifier, message, stack);
	}

	@Override
	public DocFile[] getDocFiles() throws RemoteException {
		return this.param.getDocFiles();
	}

	@Override
	public DocFile getDocFile(String name) throws RemoteException {
		return this.param.getDocFile(name);
	}

	@Override
	public File getFileFromDocFile(DocFile doc) throws RemoteException {
		return this.param.getFileFromDocFile(doc);
	}

	@Override
	public void registerDocFile(String name, String strFile, String description)
			throws RemoteException, NoSuchAlgorithmException, IOException {
		this.param.registerDocFile(name, strFile, description);
	}

	@Override
	public void unregisterDocFiles() throws RemoteException {
		this.param.unregisterDocFiles();
	}

	@Override
	public void unregisterDocFile(String name) throws RemoteException {
		this.param.unregisterDocFile(name);
	}

	@Override
	public void resetDocFiles() throws RemoteException {
		this.param.resetDocFiles();
	}

	@Override
	public void resetDocFiles(int index) throws RemoteException {
		this.param.resetDocFiles(index);
	}

	@Override
	public void deleteDocFiles() throws RemoteException {
		this.param.deleteDocFiles();
	}

	@Override
	public void deleteDocFiles(int index) throws RemoteException {
		this.param.deleteDocFiles(index);
	}

	@Override
	public boolean checkDocFiles() throws RemoteException {
		return this.param.checkDocFiles();
	}

	@Override
	public boolean checkDocFiles(int index) throws RemoteException {
		return this.param.checkDocFiles(index);
	}

	@Override
	public boolean checkDocFile(String docfile) throws RemoteException {
		return this.param.checkDocFile(docfile);
	}

	@Override
	public boolean checkDocFile(int index, String docfile) throws RemoteException {
		return this.param.checkDocFile(index, docfile);
	}

	@Override
	public String getDsId() throws RemoteException {
		return this.param.getDsId();
	}

	@Override
	public File getParamDirectory() throws RemoteException {
		return this.param.getParamDirectory();
	}

	@Override
	public String getUniqueDocFileName(String file) throws RemoteException {
		return this.param.getUniqueDocFileName(file);
	}

	@Override
	public File getUniqueDocFileName(File file) throws RemoteException {
		return this.param.getUniqueDocFileName(file);
	}

	@Override
	public String toStringI() throws RemoteException {
		return this.param.toStringI();
	}

	@Override
	public List<String> listTabNames() throws RemoteException {
		return this.param.listTabNames();
	}

	@Override
	public IParameter getRoot() throws RemoteException {
		return this.param.getRoot().getClientImpl();
	}
}
