/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.extra;

import java.awt.Color;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import org.apache.poi.hssf.usermodel.HSSFPalette;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.hssf.util.HSSFColor;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.usermodel.XSSFCellStyle;
import org.apache.poi.xssf.usermodel.XSSFColor;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.xml.sax.SAXException;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.magopt.structure.IParamField;
import at.lcm.bu21.magopt.structure.IParameter;
import at.lcm.bu21.magopt.structure.IProject;
import at.lcm.bu21.magopt.structure.xsd.XMLValidator;

/**
 * Write list of parameter to excel-file
 *
 * @author Werner Koppelsttter
 */
public class Analysis {

	/** Project for analysis */
	private IProject project;

	/** List of unused parameters */
	private List<IParameter> unused = new LinkedList<IParameter>();

	/** List of used parameters */
	private List<IParameter> used = new LinkedList<IParameter>();

	/**
	 * Create analysis for project
	 * 
	 * @param project
	 *            Based project
	 * @throws RemoteException
	 */
	public Analysis(IProject project) throws RemoteException {

		this.project = project;
		this.unused = new LinkedList<IParameter>();
		this.used = new LinkedList<IParameter>();

		this.reset();
	}

	/**
	 * Create excel-file
	 * 
	 * @param excel
	 *            * Excel-file
	 * @return Excel-file
	 * @throws IOException
	 */
	public File create(File excel) throws IOException {
		Workbook wb;
		Font boldFont;
		CellStyle boldStyle, incompleteStyle;
		XSSFColor incompleteColor;
		HSSFPalette palette;
		HSSFColor color;
		Sheet sheet;
		Cell cell;
		IProject root;
		IParameter p;
		int bckIndex;
		TData data;
		FileOutputStream fileOut;

		// open report file
		if (excel.getName().toLowerCase().endsWith(".xls")) {
			wb = new HSSFWorkbook();

			// incomplete Style
			palette = ((HSSFWorkbook) wb).getCustomPalette();
			color = palette.findSimilarColor(255, 140, 0);
			incompleteStyle = wb.createCellStyle();
			incompleteStyle.setFillPattern(
					org.apache.poi.ss.usermodel.CellStyle.SOLID_FOREGROUND);
			incompleteStyle.setFillForegroundColor(color.getIndex());

		} else {
			if (!excel.getName().toLowerCase().endsWith(".xlsx")) {
				excel = new File(excel.getParentFile(),
						excel.getName() + ".xlsx");
			}

			wb = new XSSFWorkbook();

			// incomplete Style
			incompleteColor = new XSSFColor(new Color(255, 140, 0));
			incompleteStyle = wb.createCellStyle();
			incompleteStyle.setFillPattern(
					org.apache.poi.ss.usermodel.CellStyle.SOLID_FOREGROUND);
			((XSSFCellStyle) incompleteStyle)
					.setFillForegroundColor(incompleteColor);
		}

		// bold style
		boldFont = wb.createFont();
		boldFont.setBoldweight(Font.BOLDWEIGHT_BOLD);
		boldStyle = wb.createCellStyle();
		boldStyle.setFont(boldFont);

		sheet = wb.createSheet();

		// write head
		for (int k = 0; k < this.getUsedSize(); k++) {
			p = this.getUsed(k);
			Analysis.getCell(sheet, 0, 0).setCellValue("Description");
			cell = Analysis.getCell(sheet, 0, k + 1);
			if (p.getDescription() == null) {
				cell.setCellValue("");
			} else {
				cell.setCellValue(p.getDescription());
			}

			Analysis.getCell(sheet, 1, 0).setCellValue("Name");
			cell = Analysis.getCell(sheet, 1, k + 1);
			cell.setCellValue(p.getName());
			cell.setCellStyle(boldStyle);

			Analysis.getCell(sheet, 2, 0).setCellValue("Unit");
			cell = Analysis.getCell(sheet, 2, k + 1);
			if (((IParamField) p).getUnit() == null) {
				cell.setCellValue("");
			} else {
				cell.setCellValue(((IParamField) p).getUnit());
			}
			cell.setCellStyle(boldStyle);
		}

		Analysis.getCell(sheet, 3, 0).setCellValue("Index");

		// writing values
		root = (IProject) this.getUsed(0).getRoot();
		bckIndex = root.getIndex();
		for (int i = 0; i < root.getLength(); i++) {

			root.setIndex(i);
			root.refresh();
			System.out.println(root.getIndex());
			for (int j = 0; j < this.getUsedSize(); j++) {
				p = this.getUsed(j);
				Analysis.getCell(sheet, 4 + i, 0)
						.setCellValue(root.getIndex() + 1);

				cell = Analysis.getCell(sheet, 4 + i, 1 + j);
				data = p.getData();
				if ((data instanceof TNumeric) && ((TNumeric) data).isScalar()
						&& ((TNumeric) data).isReal()) {
					cell.setCellValue(((TNumeric) data).getReal(0, 0));
				} else if (data instanceof TBoolean) {
					cell.setCellValue(((TBoolean) data).getBoolean());
				} else {
					cell.setCellValue(data.toString());
				}

				if (root.getIncomplete()) {
					Analysis.getCell(sheet, 4 + i, 0)
							.setCellStyle(incompleteStyle);
					Analysis.getCell(sheet, 4 + i, 1 + j)
							.setCellStyle(incompleteStyle);
				}
			}
		}
		root.setIndex(bckIndex);

		// Save report file
		fileOut = new FileOutputStream(excel);
		wb.write(fileOut);
		fileOut.close();

		return excel;
	}

	/**
	 * Get cell
	 * 
	 * @param sheet
	 *            Sheet
	 * @param rowIndex
	 *            Row index
	 * @param columnIndex
	 *            Column index
	 * @return Cell
	 */
	private static Cell getCell(Sheet sheet, int rowIndex, int columnIndex) {
		Row row;
		Cell cell;

		row = sheet.getRow(rowIndex);
		if (row == null) {
			row = sheet.createRow(rowIndex);
		}

		cell = row.getCell(columnIndex);
		if (cell == null) {
			cell = row.createCell(columnIndex);
		}

		return cell;
	}

	/**
	 * Load analysis settings from file
	 * 
	 * @param file
	 *            Settings-file
	 * @throws XMLStreamException
	 * @throws FactoryConfigurationError
	 * @throws IOException
	 * @throws SAXException
	 */
	public void loadSettings(File file) throws XMLStreamException,
			FactoryConfigurationError, IOException, SAXException {
		ZipInputStream zipStream = null;
		ZipEntry entry;
		XMLStreamReader xmlReader = null;

		try {
			// validation of File
			zipStream = new ZipInputStream(new FileInputStream(file));
			entry = zipStream.getNextEntry();
			while ((entry != null)
					&& !entry.getName().equals("AnalysisParam.xml")) {
				entry = zipStream.getNextEntry();
			}
			if (entry == null) {
				throw new IOException(
						"Invalid settings file! " + file.getName());
			}

			// validieren
			new XMLValidator().validate(zipStream, "AnalysisParam.xsd");

			// open File for reading
			zipStream = new ZipInputStream(new FileInputStream(file));
			entry = zipStream.getNextEntry();
			while ((entry != null)
					&& !entry.getName().equals("AnalysisParam.xml")) {
				entry = zipStream.getNextEntry();
			}

			// read file
			xmlReader = XMLInputFactory.newInstance()
					.createXMLStreamReader(zipStream, "utf-8");
			while (xmlReader
					.getEventType() != XMLStreamConstants.START_ELEMENT) {
				xmlReader.next();
			}

			// read data
			this.reset();
			this.xml2obj(xmlReader);

			// end of document
			xmlReader.next();
			if (xmlReader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
				throw new IOException("Undefined error by reading file!");
			}

		} finally {
			if (xmlReader != null) {
				xmlReader.close();
			}
			if (zipStream != null) {
				zipStream.close();
			}
		}
	}

	/**
	 * Store analysis settings to file
	 * 
	 * @param file
	 *            Settings-file
	 * @throws XMLStreamException
	 * @throws FactoryConfigurationError
	 * @throws IOException
	 */
	public void storeSettings(File file)
			throws XMLStreamException, FactoryConfigurationError, IOException {
		ZipOutputStream zipStream = null;
		XMLStreamWriter xmlWriter = null;

		try {
			zipStream = new ZipOutputStream(new FileOutputStream(file));
			zipStream.putNextEntry(new ZipEntry("AnalysisParam.xml"));

			xmlWriter = XMLOutputFactory.newInstance()
					.createXMLStreamWriter(zipStream, "utf-8");
			xmlWriter.writeStartDocument("utf-8", "1.0");

			// write data
			this.obj2xml(xmlWriter);

			xmlWriter.writeEndDocument();
		} finally {
			if (xmlWriter != null) {
				xmlWriter.close();
			}
			if (zipStream != null) {
				zipStream.close();
			}
		}
	}

	/**
	 * Reset all data
	 * 
	 * @throws RemoteException
	 */
	public void reset() throws RemoteException {
		this.unused.clear();
		this.used.clear();

		this.unused.addAll(Arrays.asList(this.project.getAnalysisParams()));
	}

	/**
	 * Add parameter from to used list
	 * 
	 * @param param
	 *            Parameter to select
	 * @return Selection finished
	 */
	public boolean select(IParameter param) {
		boolean modified;

		modified = this.unused.remove(param);
		if (modified) {
			this.used.add(param);
		}

		return modified;
	}

	/**
	 * Remove parameter from used list
	 * 
	 * @param param
	 *            Parameter to deselect
	 * @return Deselection finished
	 */
	public boolean deselect(IParameter param) {
		boolean modified;

		modified = this.used.remove(param);
		if (modified) {
			this.unused.add(param);
		}

		return modified;
	}

	/**
	 * Get number of used parameter
	 * 
	 * @return Size
	 */
	public int getUsedSize() {
		return this.used.size();
	}

	/**
	 * Get used parameter
	 * 
	 * @param index
	 *            Index of list
	 * @return Parameter
	 */
	public IParameter getUsed(int index) {
		return this.used.get(index);
	}

	/**
	 * Get number of unused parameter
	 * 
	 * @return Size
	 */
	public int getUnusedSize() {
		return this.unused.size();
	}

	/**
	 * Get unused parameter
	 * 
	 * @param index
	 *            Index in list
	 * @return Parameter
	 */
	public IParameter getUnused(int index) {
		return this.unused.get(index);
	}

	/**
	 * Get short identifier without project name
	 * 
	 * @param param
	 *            Parameter
	 * @return Short identifier
	 * @throws RemoteException
	 */
	private String getShortIdentifier(IParameter param) throws RemoteException {
		String ident;
		int index;

		ident = param.getIdentifier();
		index = ident.indexOf('.');
		if (index == -1) {
			ident = ".";
		} else {
			ident = ident.substring(index + 1, ident.length());
		}

		return ident;
	}

	/**
	 * Write data to xml
	 * 
	 * @param xmlWriter
	 *            XMLWriter
	 * @throws XMLStreamException
	 * @throws RemoteException
	 */
	void obj2xml(XMLStreamWriter xmlWriter)
			throws XMLStreamException, RemoteException {

		xmlWriter.writeStartElement("AnalysisParam");
		xmlWriter.writeAttribute("object", "AnalysisParam");
		xmlWriter.writeStartElement("Used");
		for (IParameter p : this.used) {
			xmlWriter.writeStartElement("item");
			xmlWriter.writeCharacters(this.getShortIdentifier(p));
			xmlWriter.writeEndElement();
		}
		xmlWriter.writeEndElement();
		xmlWriter.writeEndElement();

	}

	/**
	 * Read data from xml
	 * 
	 * @param xmlReader
	 *            XMLReader
	 * @throws XMLStreamException
	 * @throws RemoteException
	 */
	void xml2obj(XMLStreamReader xmlReader)
			throws XMLStreamException, RemoteException {
		String str;
		IParameter param;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			if (xmlReader.getLocalName().equals("Used")) {
				xmlReader.next();
				while (xmlReader
						.getEventType() == XMLStreamConstants.START_ELEMENT) {
					xmlReader.next();
					str = "";
					while (xmlReader
							.getEventType() != XMLStreamConstants.END_ELEMENT) {
						str += xmlReader.getText();
						xmlReader.next();
					}
					param = this.getParameter(str);
					if (param != null) {
						this.select(param);
					}
					xmlReader.next();
				}
				xmlReader.next();
			} else {
				throw new XMLStreamException(
						"Unrecognized property: " + xmlReader.getLocalName());
			}
		}

	}

	/**
	 * Get parameter by name
	 * 
	 * @param shortIdentifier
	 *            Short identifier of parameter (without project name)
	 * @return Parameter from project. If not found null
	 * @throws RemoteException
	 */
	private IParameter getParameter(String shortIdentifier)
			throws RemoteException {
		IParameter p = null;
		int k = 0;

		while ((k < this.unused.size()) && (p == null)) {
			if (this.getShortIdentifier(this.unused.get(k))
					.equals(shortIdentifier)) {
				p = this.unused.get(k);
			}
			k++;
		}

		return p;
	}

}
