/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.utils;

import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.Objects;

/**
 * Class loader with child-first strategy (class and resources)
 *
 * @author Werner Koppelsttter
 */
public class ChildFirstURLClassLoader extends URLClassLoader {

	/** System class loader used for security reason */
//	private final ClassLoader sysCL;

	/**
	 * Create class loader for urls
	 * 
	 * @param urls
	 *            List of urls
	 * @param parent
	 *            Parent class loader
	 */
	public ChildFirstURLClassLoader(URL[] urls, ClassLoader parent) {
		super(urls, parent);
//		this.sysCL = ClassLoader.getSystemClassLoader();
	}

	@Override
	protected Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {

		// First, check if the class has already been loaded
		Class<?> c = this.findLoadedClass(name);
		if (c == null) {
//			if (this.sysCL != null) {
//				try {
//					// Find the class in system class loader (security reason)
//					c = this.sysCL.loadClass(name);
//				} catch (ClassNotFoundException e) {
//					// ClassNotFoundException thrown if class not found from system class loader
//				}
//			}

//			if (c == null) {
			try {
				// Find the class in current class loader
				c = this.findClass(name);
			} catch (ClassNotFoundException e) {
				// ClassNotFoundException thrown if class not found
				// from the current class loader

				// use super method
				c = super.loadClass(name, resolve);
			}
//			}
		}

		if (resolve) {
			this.resolveClass(c);
		}

		return c;
	}

	@Override
	public Enumeration<URL> getResources(String name) throws IOException {
		Objects.requireNonNull(name);
		@SuppressWarnings("unchecked")
		final Enumeration<URL>[] enums = (Enumeration<URL>[]) new Enumeration<?>[2];

//		if (this.sysCL != null) {
//			// find in system class loader (security reason)
//			enums[0] = this.sysCL.getResources(name);
//		}
		// find in current class loader
		enums[0] = this.findResources(name);
		// use super method
		enums[1] = super.getResources(name);

		return new Enumeration<URL>() {
			private int index = 0;

			@Override
			public boolean hasMoreElements() {
				while (this.index < enums.length) {
					if ((enums[this.index] != null) && enums[this.index].hasMoreElements()) {
						return true;
					}
					this.index++;
				}
				return false;
			}

			@Override
			public URL nextElement() {
				if (!this.hasMoreElements()) {
					throw new NoSuchElementException();
				}
				return enums[this.index].nextElement();
			}
		};

	}

	@Override
	public URL getResource(String name) {
		URL url = null;
		Objects.requireNonNull(name);

//		if (this.sysCL != null) {
//			// find in system class loader (security reason)
//			url = this.sysCL.getResource(name);
//		}
//		if (url == null) {
		// find in current class loader
		url = this.findResource(name);
//		}
		if (url == null) {
			// use super method
			super.getResource(name);
		}

		return url;
	}

}
