/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JSplitPane;
import javax.swing.WindowConstants;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.magopt.structure.IParameter;
import at.lcm.bu21.magopt.structure.IProject;
import at.lcm.bu21.magopt.structure.ParamContainer;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Project;
import magopt.gui.dataset.DataSetListener;
import magopt.gui.dataset.DataSetView;
import magopt.gui.event.EventView;
import magopt.gui.helpers.JProjectChooser;
import magopt.gui.tree.ProjectTreeView;
import magopt.gui.view.ParamListener;
import magopt.gui.view.View;

/**
 * MagOpt User Interface (UI).
 * 
 * @author Aigner Markus
 * 
 */
public class MagOptGui extends JFrame implements ClipboardOwner {
	private static final long serialVersionUID = 1L;

	protected static final String GUITITLE = "MagOpt UI";

	protected Project mProject;
	protected ProjectTreeView mProjectTree;
	protected View mParamView;
	protected EventView mEventView;
	protected DataSetView mDataSetView;

	private JSplitPane mJspMain;
	private JSplitPane mJspTop;
	private JSplitPane mJspBottom;

	private MenuBar mMenuBar;
	private ToolBar mToolBar;

	protected Clipboard mClipboard;

	protected static enum ParamType {
		FIELD, FUNCTION, FORMULA, FILE, DIRECTORY, CONTAINER, TEMPLATE
	};

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		/* gui must run in event dispatch thread */
		java.awt.EventQueue.invokeLater(new Runnable() {
			@Override
			public void run() {
				new MagOptGui();
			}
		});
	}

	protected MagOptGui() {
		this.initComponents();
	}

	/**
	 * Initialize GUI Components.
	 */
	protected void initComponents() {
		Container cp = this.getContentPane();

		this.setTitle(MagOptGui.GUITITLE);
		this.setSize(new Dimension(800, 600));
		this.setLocationRelativeTo(null);

		this.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				MagOptGui.this.quitGui();
			}
		});

		this.mClipboard = this.getToolkit().getSystemClipboard();

		/* menubar */
		this.mMenuBar = new MenuBar(this);
		this.setJMenuBar(this.mMenuBar);

		/* toolbar */
		this.mToolBar = new ToolBar(this);
		cp.add(this.mToolBar, BorderLayout.NORTH);

		this.mProjectTree = new ProjectTreeView();
		this.mProjectTree.tree.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseReleased(MouseEvent e) {
				MagOptGui.this.updateViewComponent();
				MagOptGui.this.updateBars();
			}
		});

		this.mParamView = new View();
		this.mParamView.addParamListener(new ParamListener() {
			@Override
			public void parameterChanged() {
				MagOptGui.this.refresh();
			}
		});

		this.mDataSetView = new DataSetView();
		this.mDataSetView.addDataSetListener(new DataSetListener() {
			@Override
			public void fireDataSetChanged(int index) {
				MagOptGui.this.mProject.setIndex(index);
				MagOptGui.this.refresh();
			}
		});

		this.mEventView = new EventView();

		this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);

		this.mJspMain = new JSplitPane(JSplitPane.VERTICAL_SPLIT, true);
		this.mJspTop = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, this.mProjectTree, this.mParamView);
		this.mJspTop.setContinuousLayout(true);
		this.mJspBottom = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, this.mDataSetView, this.mEventView);
		this.mJspBottom.setContinuousLayout(true);
		this.mJspMain.setLeftComponent(this.mJspTop);
		this.mJspMain.setRightComponent(this.mJspBottom);
		cp.add(this.mJspMain);

		this.setVisible(true);

		/* default divider locations */
		this.mJspMain.setDividerLocation(0.7);
		this.mJspBottom.setDividerLocation(0.3);
		this.mJspTop.setDividerLocation(0.3);
	}

	/**
	 * Creates new Project.
	 */
	protected void newProject() {
		try {
			boolean doIt = false;

			/* no dialog if no project is loaded */
			if (this.mProject == null) {
				doIt = true;
			} else {
				int response = JOptionPane.showConfirmDialog(this, "This will discard the loaded Project.\nContinue ?",
						"Warning", JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE);
				if (response == JOptionPane.YES_OPTION) {
					doIt = true;
				}
			}

			if (doIt == true) {
				JProjectChooser chooser = new JProjectChooser(JProjectChooser.ALL, new File("."));
				chooser.setDialogType(JFileChooser.SAVE_DIALOG);

				int response = chooser.showSaveDialog(this);
				if (response == JFileChooser.APPROVE_OPTION) {
					this.mProject = new Project("ProjectName", "ProjectDesc", chooser.getSelectedFile());
					this.mProjectTree.setProject(this.mProject);
					this.mDataSetView.setProject(this.mProject);
					this.mParamView.set(this.mProject);
					this.mEventView.setEventLog(this.mProject.getEventLog());
					this.updateBars();
					this.setTitle(MagOptGui.GUITITLE + " - " + Project.getVersion() + " - "
							+ this.mProject.getMopFile().getAbsolutePath());
				}
			}
		} catch (Exception myEx) {
			JOptionPane.showMessageDialog(this, myEx.getMessage(), "New Project", JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Opens Project (mop) or Template (mopt)
	 */
	protected void openProject() {
		new OpenThread(this).execute();
	}

	/**
	 * Saves current Project as Project (mop) or Template (mopt)
	 * 
	 * @param overwrite
	 *            - Save (true) or SaveAs (false)
	 */
	protected void saveProject(boolean overwrite) {
		if (this.mProject == null) {
			return;
		}
		new SaveThread(this, overwrite).execute();
	}

	/**
	 * Closes Window.
	 */
	protected void quitGui() {
		int response = JOptionPane.showConfirmDialog(this,
				"This will close the Application.\n" + "Changes in current Project will be lost.\n"
						+ "Do you really want to quit ?",
				"Warning", JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE);
		if (response == JOptionPane.YES_OPTION) {
			this.dispose();
		}
	}

	/**
	 * Refreshes GUI.
	 */
	protected void refresh() {
		this.mProject.refresh();
		this.updateTreeComponent();
		this.updateViewComponent();
		this.updateEventComponent();
		this.updateIndexComponent();
	}

	/**
	 * Updates View Component (north east component)
	 */
	protected void updateViewComponent() {
		try {
			this.mParamView.set(this.mProjectTree.getSelectedParameter());
		} catch (Exception myEx) {
			JOptionPane.showMessageDialog(this, myEx.getMessage(), "Update View Component", JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Updates Overview (Index List) Component (south west component)
	 */
	protected void updateIndexComponent() {
		this.mDataSetView.refresh();
	}

	/**
	 * Updates Tree Component (north west component)
	 */
	protected void updateTreeComponent() {
		this.mProjectTree.setProject(this.mProject);
	}

	/**
	 * Updates Menu Bar.
	 */
	protected void updateBars() {
		this.mMenuBar.check();
		this.mToolBar.check();
	}

	/**
	 * Inserts Parameter.
	 * 
	 * @param type
	 *            - Parameter Type
	 */
	protected void insertComponent(ParamType type) {
		if (this.mProject == null) {
			return;
		}
		new InsertThread(this, type).execute();
		// try {
		// ParamContainer parent;
		// Parameter selected = this.mProjectTree.getSelectedParameter();
		// if (selected instanceof ParamContainer || selected instanceof
		// Project) {
		// parent = (ParamContainer)selected;
		// } else {
		// parent = selected.getParent();
		// }
		//
		// String newName = JOptionPane.showInputDialog("Name:", "Default");
		// if (newName == null || newName.isEmpty()) {
		// return;
		// }
		//
		// switch (type) {
		// case CONTAINER:
		// parent.addParam(new ParamContainer(newName));
		// break;
		// case FUNCTION:
		// JLibraryChooser chooser = new JLibraryChooser();
		// if (chooser.showOpenDialog(this) == JLibraryChooser.APPROVE_OPTION) {
		// if (chooser.getSelectedType().equals("M")) {
		// parent.addParam(new ParamMatlabFunction(newName));
		// } else {
		// parent.addParam(new ParamJavaFunction(newName));
		// }
		// }
		// break;
		// case FILE:
		// parent.addParam(new ParamFile(newName));
		// break;
		// case FORMULA:
		// parent.addParam(new ParamFormula(newName));
		// break;
		// case DIRECTORY:
		// parent.addParam(new ParamDirectory(newName));
		// break;
		// case FIELD:
		// parent.addParam(new ParamField(newName));
		// break;
		// case TEMPLATE:
		// JProjectChooser pc = new
		// JProjectChooser(JProjectChooser.TEMPLATE_ONLY,
		// new File(Project.getLibraryPath()));
		// if (pc.showOpenDialog(this) == JProjectChooser.APPROVE_OPTION) {
		// ParamContainer pContainer =
		// Project.load(new
		// MopFile(pc.getSelectedFile())).convertToParamContainer();
		// /* adds each parameter of template to parent */
		// for (Parameter param:pContainer.getParam()) {
		// parent.addParam(param);
		// }
		// } else {
		// return;
		// }
		// break;
		// default:
		// System.err.println("insertNewComponent: Invalid Parameter Type");
		// return;
		// }
		// this.updateTreeComponent();
		// this.updateEventComponent();
		// this.updateBars();
		// } catch (Exception myEx) {
		// JOptionPane.showMessageDialog(this,
		// myEx.getMessage(),
		// "Insert New Component",
		// JOptionPane.ERROR_MESSAGE);
		// }
	}

	/**
	 * Deletes Selected Parameter.
	 */
	protected void deleteComponent() {
		if ((this.mProject == null) || (this.mProjectTree.getSelectedParameter() instanceof Project)) {
			return;
		}
		new DeleteThread(this).execute();
	}

	/**
	 * Updates Event Log.
	 */
	protected void updateEventComponent() {
		this.mEventView.refresh();
	}

	public void updateData(TData data) {
		this.updateData(this.mProjectTree.getSelectedParameter(), data);
	}

	public void updateData(IParameter param, TData data) {
		try {
			this.mProject.getUniqueObject(param.getIdentifier()).setData(data);
			this.updateTreeComponent();
			this.updateEventComponent();
		} catch (Exception myEx) {
			JOptionPane.showMessageDialog(this, myEx.getMessage(), "Update Data", JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Returns used Project.
	 * 
	 * @return Used Project
	 */
	public IProject getProject() {
		return this.mProject;
	}

	/**
	 * Updates only current or all datasets.
	 * 
	 * @param all
	 *            - true: update all, false: update current
	 */
	protected void regenerate(boolean all) {
		new RegenerateThread(this, all).execute();
	}

	/**
	 * Adds DataSet with current DataSets' Settings.
	 */
	protected void addDataSet() {
		new ConfigThread(this, true).execute();
	}

	/**
	 * Removes current DataSet.
	 */
	protected void removeDataSet() {
		new ConfigThread(this, false).execute();
	}

	/**
	 * Checks Project.
	 */
	protected void check() {
		new CheckThread(this).execute();
	}

	protected void copyParameter(boolean cut) {
		// TODO RMI
		ParameterSelection content = new ParameterSelection((Parameter) this.mProjectTree.getSelectedParameter(), cut);
		this.mClipboard.setContents(content, this);
		this.mMenuBar.check();
	}

	protected void pasteParameter() {
		try {
			ParamContainer parent;
			// TODO RMI
			Parameter curSelection = (Parameter) this.mProjectTree.getSelectedParameter();
			if (curSelection instanceof ParamContainer) {
				parent = (ParamContainer) curSelection;
			} else {
				parent = curSelection.getParent();
			}
			Transferable content = this.mClipboard.getContents(this);
			ClipboardData cbData = (ClipboardData) (content.getTransferData(ParameterSelection.mParamFlavor));

			Parameter param = cbData.getParameter();
			if (cbData.isCut() == true) {
				this.mProject.getUniqueObject(param.getIdentifier()).removeParam();
			}
			parent.addParam(param);

		} catch (UnsupportedFlavorException ex1) {
			JOptionPane.showMessageDialog(this, "Can't paste Clipboard Content as Parameter", "Paste",
					JOptionPane.ERROR_MESSAGE);
		} catch (IOException ex2) {
			ex2.printStackTrace();
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		this.mMenuBar.check();
		this.refresh();
		this.updateBars();
	}

	@Override
	public void lostOwnership(Clipboard arg0, Transferable arg1) {
		/* do nothing */
	}
}
