/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.event;

import java.awt.BorderLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.AbstractAction;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JToggleButton;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;

import at.lcm.bu21.magopt.structure.event.Event;
import at.lcm.bu21.magopt.structure.event.EventLog;
import at.lcm.bu21.magopt.structure.event.EventLogEvent;
import at.lcm.bu21.magopt.structure.event.EventLogListener;

/**
 * Display events of event-log
 * 
 * @author Werner Koppelsttter
 */
public class EventView extends JPanel implements EventLogListener {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;
	/** Project to display */
	protected EventLog log = null;
	/** JTable to display event-list */
	protected JTable table;
	/** Event-model */
	protected EventModel model;
	/** Filter info-event */
	protected JToggleButton info;
	/** Filter warning event */
	protected JToggleButton warning;
	/** Filter error-event */
	protected JToggleButton error;

	/**
	 * Create a panel to display a project tree
	 */
	public EventView() {
		super();
		ActionListener listener = new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				EventView.this.refresh();
			}
		};

		// this.setBorder(new EmptyBorder(5, 5, 5, 5));
		this.setLayout(new BorderLayout());

		// add button panel
		JPanel panel = new JPanel();
		panel.setBorder(new EmptyBorder(0, 0, 0, 0));
		panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
		// filter info
		this.info = new JToggleButton("Info", true);
		this.info.addActionListener(listener);
		panel.add(this.info);
		// filter warning
		this.warning = new JToggleButton("Warning", true);
		this.warning.addActionListener(listener);
		panel.add(this.warning);
		// filter error
		this.error = new JToggleButton("Error", true);
		this.error.addActionListener(listener);
		panel.add(this.error);
		// title
		panel.add(Box.createHorizontalGlue());
		JLabel label = new JLabel("E V E N T   L O G");
		label.setFont(label.getFont().deriveFont(Font.BOLD));
		panel.add(label);
		panel.add(Box.createHorizontalGlue());
		// clear button
		JButton button = new JButton("Clear Log");
		button.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				if (EventView.this.log != null) {
					EventView.this.log.clear();
				}
				EventView.this.refresh();
			}
		});
		panel.add(button);
		this.add(panel, BorderLayout.NORTH);

		// add table
		this.table = new JTable();
		this.table.getTableHeader().setReorderingAllowed(false);
		this.table.setShowGrid(false);
		this.model = new EventModel();
		this.table.setModel(this.model);

		// cell renderer
		this.table.setDefaultRenderer(Event.class, new EventRenderer());

		// set max width for checkbox column
		this.table.getColumnModel().getColumn(EventColumn.TYPE.getValue()).setPreferredWidth(30);
//		this.table.getColumnModel().getColumn(EventColumn.TYPE.getValue()).setMaxWidth(50);
		this.table.getColumnModel().getColumn(EventColumn.DATE.getValue()).setPreferredWidth(120);
//		this.table.getColumnModel().getColumn(EventColumn.DATE.getValue()).setMaxWidth(200);
		this.table.getColumnModel().getColumn(EventColumn.DESCRIPTION.getValue()).setPreferredWidth(600);

		// mouse-listener
		this.table.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				if (event.getClickCount() == 2) {
					EventView.this.openDialog();
				}
			}

			@Override
			public void mouseReleased(MouseEvent event) {
				this.showPopup(event);
			}

			@Override
			public void mousePressed(MouseEvent event) {
				this.showPopup(event);
			}

			private void showPopup(MouseEvent event) {
				if (event.isPopupTrigger()) {
					// select row
					int row = EventView.this.table.rowAtPoint(event.getPoint());
					EventView.this.table.getSelectionModel().setSelectionInterval(row, row);

					// menu
					JPopupMenu popup = new JPopupMenu();

					JMenuItem item = new JMenuItem("Open");
					item.addActionListener(new ActionListener() {
						@Override
						public void actionPerformed(ActionEvent event) {
							EventView.this.openDialog();
						}
					});
					popup.add(item);

					popup.show(event.getComponent(), event.getX(), event.getY());
				}
			}

		});

		// key-action
		this.table.getInputMap().put(KeyStroke.getKeyStroke("ENTER"), "enter");
		this.table.getActionMap().put("enter", new AbstractAction() {
			private static final long serialVersionUID = 1L;

			@Override
			public void actionPerformed(ActionEvent e) {
				EventView.this.openDialog();
			}
		});

		this.add(new JScrollPane(this.table), BorderLayout.CENTER);
	}

	/**
	 * Open Dialog
	 */
	private void openDialog() {
		new magopt.gui.event.EventDialog(EventView.this,
				(Event) EventView.this.model.getValueAt(EventView.this.table.getSelectedRow(), 0));
	}

	/**
	 * Set depending event-log
	 * 
	 * @param log
	 *            event-log
	 */
	public void setEventLog(EventLog log) {

		// remove listener from old log
		if (this.log != null) {
			this.log.removeEventLogListener(this);
		}

		this.log = log;

		// add listener to new log
		if (this.log != null) {
			this.log.addEventLogListener(this);
		}

		this.refresh();
	}

	/**
	 * Refresh panel with depending event-log
	 */
	public void refresh() {
		// update model;
		this.model.setEvents(this.log, this.info.isSelected(), this.warning.isSelected(), this.error.isSelected());
		// move to last line
		if (this.model.getRowCount() > 0) {
			this.table.scrollRectToVisible(this.table.getCellRect(this.model.getRowCount() - 1, 0, false));
		}
	}

	@Override
	public void eventLogChanged(final EventLogEvent event) {
		if (this.log != null) {
			SwingUtilities.invokeLater(new Runnable() {

				@Override
				public void run() {
					switch (event.getType()) {
					case EVENT_CHANGED:
						EventView.this.model.fireTableRowsUpdated(event.getFromIndex(), event.getToIndex());
						break;
					case EVENT_ADDED:
						EventView.this.model.fireTableRowsInserted(event.getFromIndex(), event.getToIndex());
						break;
					case EVENT_REMOVED:
						EventView.this.model.fireTableRowsDeleted(event.getFromIndex(), event.getToIndex());
						break;
					}
					// move to last line
					if (EventView.this.model.getRowCount() > 0) {
						EventView.this.table.scrollRectToVisible(
								EventView.this.table.getCellRect(EventView.this.model.getRowCount() - 1, 0, false));
					}
				}
			});
		}
	}
}
