/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.helpers;

import java.awt.BorderLayout;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import javax.swing.DefaultListModel;
import javax.swing.JFileChooser;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.UIManager;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.TitledBorder;
import javax.swing.filechooser.FileNameExtensionFilter;

import at.lcm.bu21.magopt.structure.ParamJavaFnc;
import at.lcm.bu21.magopt.structure.Project;

/**
 * Library-chooser for selecting library-files from file-system
 * 
 * @author Werner Koppelsttter
 */
public class JLibraryChooser extends JFileChooser {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Function type java in jar-library */
	public static final String JAVA_JAR = "JAR";
	/** Function type java class-file */
	public static final String JAVA_CLASS = "CLASS";
	/** Function type matlab function */
	public static final String MATLAB = "M";
	/** Function type pythons function */
	public static final String PYTHON = "PY";

	/** List for displaying classes in zip-file */
	protected JList<String> classList;

	/** Selected file or rather path */
	protected File selectedFile = null;

	/** Selected class */
	protected String selectedClass = null;

	/** type of selected file */
	protected String selectedType = null;

	/**
	 * Constructor of library-chooser
	 */
	public JLibraryChooser() {

		super();

		// Filefilter
		this.removeChoosableFileFilter(this.getAcceptAllFileFilter());
		FileNameExtensionFilter filter = new FileNameExtensionFilter("All Functions", "m", "py", "jar", "class");
		this.addChoosableFileFilter(filter);
		this.addChoosableFileFilter(new FileNameExtensionFilter("Matlab Function (*.m)", "m"));
		this.addChoosableFileFilter(new FileNameExtensionFilter("Python Function (*.py)", "py"));
		this.addChoosableFileFilter(new FileNameExtensionFilter("Java Library (*.jar)", "jar"));
		this.addChoosableFileFilter(new FileNameExtensionFilter("Class File (*.class)", "class"));
		this.setFileFilter(filter);

		// settings
		this.setDialogTitle("Choose library file");
		this.setDialogType(JFileChooser.OPEN_DIALOG);
		this.setFileSelectionMode(JFileChooser.FILES_ONLY);
		this.setMultiSelectionEnabled(false);
		this.setCurrentDirectory(new File(Project.getLibraryPath()));

		// class list
		this.classList = new JList<String>();
		JPanel classPanel = new JPanel(new BorderLayout());
		classPanel.setBorder(
				new CompoundBorder(new CompoundBorder(new EmptyBorder(5, 5, 5, 5), new TitledBorder("Classes:")),
						new EmptyBorder(5, 5, 5, 5)));
		JScrollPane ps = new JScrollPane(this.classList);
		classPanel.add(ps, BorderLayout.CENTER);
		this.add(classPanel, BorderLayout.SOUTH);
		this.addPropertyChangeListener(new PropertyChangeListener() {

			@Override
			public void propertyChange(PropertyChangeEvent evt) {
				String className = null, c;
				URLClassLoader cl;

				if (evt.getPropertyName() == JFileChooser.SELECTED_FILE_CHANGED_PROPERTY) {
					// file changed
					// update class-list
					File f = JLibraryChooser.super.getSelectedFile();

					if ((f != null) && f.exists()) {
						if (f.getName().toLowerCase().endsWith(".jar")) {
							try {
								// insert content of jar to list
								ZipFile zipFile = new ZipFile(f.getPath());
								DefaultListModel<String> model = new DefaultListModel<String>();
								Enumeration<? extends ZipEntry> en = zipFile.entries();
								while (en.hasMoreElements()) {
									ZipEntry zipEntr = en.nextElement();
									className = zipEntr.getName();
									if (className.toLowerCase().endsWith(".class")
											&& !className.toLowerCase().contains("$")) {
										c = className.substring(0, className.length() - 6).replaceAll("/", ".");
										try {
											cl = new URLClassLoader(new URL[] { f.toURI().toURL() },
													this.getClass().getClassLoader());
											if (ParamJavaFnc.class.isAssignableFrom(cl.loadClass(c))) {
												model.addElement(c);
											}
										} catch (Throwable ex) {
										}
									}
								}
								zipFile.close();

								JLibraryChooser.this.classList.setModel(model);
								JLibraryChooser.this.classList.setSelectedIndex(0);
							} catch (Exception ex) {
							}
						} else if (f.getName().toLowerCase().endsWith(".py")) {
							try {
								// insert content of module to list
								BufferedReader in = new BufferedReader(new FileReader(f));
								DefaultListModel<String> model = new DefaultListModel<String>();
								String line;
								while ((line = in.readLine()) != null) {
									line = line.trim();
									if (line.startsWith("class ")) {
										int index = line.indexOf("(");
										if (index > -1) {
											line = line.substring(6, index).trim();
											model.addElement(line);
										}
									}
								}
								in.close();
								JLibraryChooser.this.classList.setModel(model);
								JLibraryChooser.this.classList.setSelectedIndex(0);
							} catch (Exception ex) {
							}
						} else {
							JLibraryChooser.this.classList.setModel(new DefaultListModel<String>());
						}
					}
				} else if (evt.getPropertyName().compareTo("JFileChooserDialogIsClosingProperty") == 0) {
					// closing dialog
					// separate path
					JLibraryChooser.this.selectedFile = JLibraryChooser.super.getSelectedFile();
					JLibraryChooser.this.selectedClass = JLibraryChooser.this.classList.getSelectedValue();
					if ((JLibraryChooser.this.selectedFile != null) && JLibraryChooser.this.selectedFile.exists()) {
						String path = JLibraryChooser.this.selectedFile.getPath();
						if (path.endsWith(".jar")) {
							// jar-file
							JLibraryChooser.this.selectedType = JLibraryChooser.JAVA_JAR;
						} else if (path.endsWith(".class")) {
							// class-file
							JLibraryChooser.this.selectedType = JLibraryChooser.JAVA_CLASS;
						} else if (path.endsWith(".m")) {
							// m-file
							JLibraryChooser.this.selectedType = JLibraryChooser.MATLAB;
						} else if (path.endsWith(".py")) {
							// py-file
							JLibraryChooser.this.selectedType = JLibraryChooser.PYTHON;
						}
					} else {
						JLibraryChooser.this.selectedType = null;
						JLibraryChooser.this.selectedFile = null;
						JLibraryChooser.this.selectedClass = null;
					}
				} else {
					JLibraryChooser.this.classList.setModel(new DefaultListModel<String>());
				}
			}
		});
	}

	/**
	 * Get selected type
	 * 
	 * @return Selected type
	 */
	public String getSelectedType() {
		return this.selectedType;
	}

	/**
	 * Get selected file or rather path
	 * 
	 * @return Selected path
	 */
	@Override
	public File getSelectedFile() {
		return this.selectedFile;
	}

	/**
	 * Get selected class
	 * 
	 * @return Selected class
	 */
	public String getSelectedClass() {
		return this.selectedClass;
	}

	/**
	 * Test library-chooser
	 * 
	 * @param args
	 *            Default parameter
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());

		JLibraryChooser chooser = new JLibraryChooser();

		if (chooser.showDialog(null, null) == JFileChooser.APPROVE_OPTION) {
			System.out.println("TYPE: " + chooser.getSelectedType());
			System.out.println("PATH: " + chooser.getSelectedFile());
			System.out.println("CLASS: " + chooser.getSelectedClass());
		}
	}

}
