/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.structure.database;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.LinkedList;

import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.magopt.structure.event.EventLog;

/**
 * Local material-database
 * 
 * @author Werner Koppelsttter
 */
public class FileConnection implements MaterialConnection {

	/** File of local database */
	private File database;

	/** List of dataset in local database */
	private LinkedList<Dataset> dsList = new LinkedList<Dataset>();

	/**
	 * Structure of database dataset's
	 * 
	 * @author Werner Koppelsttter
	 */
	private class Dataset {
		/** ID of material-class */
		long materialClassID = -1;
		/** Name of material-class */
		String materialClassName = "";
		/** ID of material */
		long materialID = -1;
		/** Name of material */
		String materialName = "";
		/** ID of dataset */
		long datasetID = -1;
		/** Name of dataset */
		String datasetName = "";

		@Override
		public String toString() {
			return this.materialClassName + " (" + this.materialClassID + ")\n\t" + this.materialName + " ("
					+ this.materialID + ")\n\t\t" + this.datasetName + " (" + this.datasetID + ")";
		}
	}

	/**
	 * Create offline-database form material-connection
	 * 
	 * @param file
	 *            Path of local database
	 * @param con
	 *            Connection to material database
	 * @throws Exception
	 */
	public static void createFile(File file, MaterialConnection con) throws Exception {
		DBItem[] cl, mat, ds;
		XMLStreamWriter xmlWriter;
		BufferedOutputStream os;
		// FIXME
		EventLog log = new EventLog();
		LinkedList<MaterialProperty<Double>> properties;
		LinkedList<MaterialDiagram> diagrams;
		Double[] values;
		StringBuffer str;

		os = new BufferedOutputStream(new FileOutputStream(file));
		xmlWriter = XMLOutputFactory.newInstance().createXMLStreamWriter(os, "utf-8");
		xmlWriter.writeStartDocument("utf-8", "1.0");
		xmlWriter.writeStartElement("MaterialDatabase");

		cl = con.getMaterialClassList();
		for (DBItem c : cl) {
			if (c.getID() != -1) {
				xmlWriter.writeStartElement("MaterialClass");
				xmlWriter.writeAttribute("Name", c.getName());
				xmlWriter.writeAttribute("ID", Long.toString(c.getID()));
				mat = con.getMaterialList(c);
				for (DBItem m : mat) {
					if (m.getID() != -1) {
						xmlWriter.writeStartElement("Material");
						xmlWriter.writeAttribute("Name", m.getName());
						xmlWriter.writeAttribute("ID", Long.toString(m.getID()));
						ds = con.getDatasetList(m);
						for (DBItem d : ds) {
							if (d.getID() != -1) {
								xmlWriter.writeStartElement("Dataset");
								xmlWriter.writeAttribute("Name", d.getName());
								xmlWriter.writeAttribute("ID", Long.toString(d.getID()));

								properties = con.getProperties(d.getID(), log);
								xmlWriter.writeStartElement("Properties");
								for (MaterialProperty<Double> prop : properties) {
									xmlWriter.writeStartElement("item");

									xmlWriter.writeStartElement("Name");
									xmlWriter.writeCharacters(prop.getName());
									xmlWriter.writeEndElement();

									xmlWriter.writeStartElement("Description");
									xmlWriter.writeCharacters(prop.getDescription());
									xmlWriter.writeEndElement();

									xmlWriter.writeStartElement("Group");
									xmlWriter.writeCharacters(prop.getGroup());
									xmlWriter.writeEndElement();

									xmlWriter.writeStartElement("Value");
									xmlWriter.writeCharacters(Double.toString(prop.getValue()));
									xmlWriter.writeEndElement();

									xmlWriter.writeStartElement("Unit");
									xmlWriter.writeCharacters(prop.getUnit());
									xmlWriter.writeEndElement();

									xmlWriter.writeStartElement("Tex");
									xmlWriter.writeCharacters(prop.getTex());
									xmlWriter.writeEndElement();

									xmlWriter.writeStartElement("Locked");
									xmlWriter.writeCharacters(Boolean.toString(prop.getLocked()));
									xmlWriter.writeEndElement();

									xmlWriter.writeStartElement("Fixed");
									xmlWriter.writeCharacters(Boolean.toString(prop.getFixed()));
									xmlWriter.writeEndElement();

									xmlWriter.writeEndElement();
								}
								xmlWriter.writeEndElement();

								diagrams = con.getDiagrams(d.getID(), log);
								xmlWriter.writeStartElement("Diagrams");
								for (MaterialDiagram diag : diagrams) {
									xmlWriter.writeStartElement("item");
									xmlWriter.writeStartElement("Name");
									xmlWriter.writeCharacters(diag.getName());
									xmlWriter.writeEndElement();

									xmlWriter.writeStartElement("Description");
									xmlWriter.writeCharacters(diag.getDescription());
									xmlWriter.writeEndElement();

									xmlWriter.writeStartElement("Axis");
									for (MaterialProperty<Double[]> prop : diag.getAxis()) {
										xmlWriter.writeStartElement("item");

										xmlWriter.writeStartElement("Name");
										xmlWriter.writeCharacters(prop.getName());
										xmlWriter.writeEndElement();

										xmlWriter.writeStartElement("Description");
										xmlWriter.writeCharacters(prop.getDescription());
										xmlWriter.writeEndElement();

										xmlWriter.writeStartElement("Group");
										xmlWriter.writeCharacters(prop.getGroup());
										xmlWriter.writeEndElement();

										xmlWriter.writeStartElement("Values");
										values = prop.getValue();
										str = new StringBuffer();
										for (Double val : values) {
											if (str.length() > 0) {
												str.append(";");
											}
											str.append(val);
										}
										xmlWriter.writeCharacters(str.toString());
										xmlWriter.writeEndElement();

										xmlWriter.writeStartElement("Unit");
										xmlWriter.writeCharacters(prop.getUnit());
										xmlWriter.writeEndElement();

										xmlWriter.writeStartElement("Tex");
										xmlWriter.writeCharacters(prop.getTex());
										xmlWriter.writeEndElement();

										xmlWriter.writeStartElement("Locked");
										xmlWriter.writeCharacters(Boolean.toString(prop.getLocked()));
										xmlWriter.writeEndElement();

										xmlWriter.writeStartElement("Fixed");
										xmlWriter.writeCharacters(Boolean.toString(prop.getFixed()));
										xmlWriter.writeEndElement();

										xmlWriter.writeEndElement();
									}
									xmlWriter.writeEndElement();
									xmlWriter.writeEndElement();
								}
								xmlWriter.writeEndElement();

								xmlWriter.writeEndElement();
							}
						}
						xmlWriter.writeEndElement();
					}
				}
				xmlWriter.writeEndElement();
			}
		}

		xmlWriter.writeEndElement();
		xmlWriter.writeEndDocument();
		xmlWriter.close();
		os.close();

	}

	/**
	 * Read offline database file and create structure of dataset's
	 * 
	 * @param file
	 *            Offline database
	 * @throws Exception
	 */
	public FileConnection(File file) throws Exception {
		XMLStreamReader xmlReader;
		BufferedInputStream is = null;
		String materialClassName, materialName, datasetName;
		long materialClassID, materialID, datasetID;
		Dataset ds;

		this.database = file;
		if ((this.database == null) && !this.database.exists()) {
			throw new Exception("File doesn't exist!");
		}

		try {
			is = new BufferedInputStream(new FileInputStream(this.database));
			xmlReader = XMLInputFactory.newInstance().createXMLStreamReader(is, "utf-8");

			// read file
			while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
				xmlReader.next();
			}

			if ("MaterialDatabase".compareTo(xmlReader.getLocalName()) != 0) {
				throw new Exception("File is not a project-file");
			}
			xmlReader.next();

			while ((xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT)
					&& "MaterialClass".equals(xmlReader.getLocalName())) {
				materialClassName = xmlReader.getAttributeValue(null, "Name");
				materialClassID = Long.parseLong(xmlReader.getAttributeValue(null, "ID"));
				xmlReader.next();
				while ((xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT)
						&& "Material".equals(xmlReader.getLocalName())) {
					materialName = xmlReader.getAttributeValue(null, "Name");
					materialID = Long.parseLong(xmlReader.getAttributeValue(null, "ID"));
					xmlReader.next();
					while ((xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT)
							&& "Dataset".equals(xmlReader.getLocalName())) {
						datasetName = xmlReader.getAttributeValue(null, "Name");
						datasetID = Long.parseLong(xmlReader.getAttributeValue(null, "ID"));
						xmlReader.next();

						ds = new Dataset();
						ds.materialClassName = materialClassName;
						ds.materialClassID = materialClassID;
						ds.materialName = materialName;
						ds.materialID = materialID;
						ds.datasetName = datasetName;
						ds.datasetID = datasetID;
						this.dsList.add(ds);

						while ((xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT)
								|| !"Dataset".equals(xmlReader.getLocalName())) {
							xmlReader.next();
						}
						xmlReader.next();
					}
					xmlReader.next();
				}
				xmlReader.next();

			}
			xmlReader.next();

			if (xmlReader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
				throw new IOException("Undefined error by reading file!");
			}
			xmlReader.close();
		} finally {
			if (is != null) {
				is.close();
			}
		}
	}

	@Override
	public DBItem[] getMaterialItems(long datasetID) {
		DBItem[] items = new DBItem[] { new DBItem(), new DBItem(), new DBItem() };
		Dataset ds = null;

		for (Dataset d : this.dsList) {
			if (d.datasetID == datasetID) {
				ds = d;
			}
		}

		if (ds != null) {
			items[0] = new DBItem(ds.materialClassID, ds.materialClassName);
			items[1] = new DBItem(ds.materialID, ds.materialName);
			items[2] = new DBItem(ds.datasetID, ds.datasetName);
		}
		return items;
	}

	@Override
	public DBItem[] getMaterialClassList() {
		LinkedList<DBItem> list = new LinkedList<DBItem>();
		DBItem item;

		list.add(new DBItem());
		for (Dataset ds : this.dsList) {
			item = new DBItem(ds.materialClassID, ds.materialClassName);
			if (!list.contains(item)) {
				list.add(item);
			}
		}

		return list.toArray(new DBItem[0]);
	}

	@Override
	public DBItem[] getMaterialList(DBItem materialClass) {
		LinkedList<DBItem> list = new LinkedList<DBItem>();
		DBItem item;
		long id = -1;

		if ((materialClass != null) && (materialClass.getID() > 0)) {
			id = materialClass.getID();
		} else {
			materialClass = null;
		}

		list.add(new DBItem());
		for (Dataset ds : this.dsList) {
			if ((materialClass == null) || (id == ds.materialClassID)) {
				item = new DBItem(ds.materialID, ds.materialName);
				if (!list.contains(item)) {
					list.add(item);
				}
			}
		}

		return list.toArray(new DBItem[0]);
	}

	@Override
	public DBItem[] getDatasetList(DBItem material) {
		LinkedList<DBItem> list = new LinkedList<DBItem>();
		long id;

		list.add(new DBItem());
		if ((material != null) && (material.getID() >= 0)) {
			id = material.getID();
			for (Dataset ds : this.dsList) {
				if (id == ds.materialID) {
					list.add(new DBItem(ds.datasetID, ds.datasetName));
				}
			}
		}

		return list.toArray(new DBItem[0]);
	}

	@Override
	public LinkedList<MaterialProperty<Double>> getProperties(long datasetID, EventLog log) {
		LinkedList<MaterialProperty<Double>> properties;
		XMLStreamReader xmlReader = null;
		BufferedInputStream is = null;
		String element, str;
		String name, desc, group, unit, tex;
		Boolean locked, fixed;
		Double value;

		properties = new LinkedList<MaterialProperty<Double>>();

		try {
			is = new BufferedInputStream(new FileInputStream(this.database));
			xmlReader = XMLInputFactory.newInstance().createXMLStreamReader(is, "utf-8");

			// read file
			while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
				xmlReader.next();
			}

			xmlReader.next();

			// search dataset with id
			while ((xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT)
					&& "MaterialClass".equals(xmlReader.getLocalName())) {
				xmlReader.next();
				while ((xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT)
						&& "Material".equals(xmlReader.getLocalName())) {
					xmlReader.next();
					while ((xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT)
							&& "Dataset".equals(xmlReader.getLocalName())) {
						if (Long.parseLong(xmlReader.getAttributeValue(null, "ID")) == datasetID) {
							xmlReader.next();

							// search properties
							while ((xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT)
									|| !"Properties".equals(xmlReader.getLocalName())) {
								xmlReader.next();
							}
							xmlReader.next();

							// read propeties items
							while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
								xmlReader.next();
								name = "";
								desc = "";
								group = "";
								value = Double.NaN;
								unit = "";
								tex = "";
								locked = false;
								fixed = false;

								// read properties
								while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
									element = xmlReader.getLocalName();
									xmlReader.next();
									if ("Name".equals(element)) {
										while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
											name += xmlReader.getText();
											xmlReader.next();
										}
									} else if ("Description".equals(element)) {
										while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
											desc += xmlReader.getText();
											xmlReader.next();
										}
									} else if ("Group".equals(element)) {
										while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
											group += xmlReader.getText();
											xmlReader.next();
										}
									} else if ("Value".equals(element)) {
										str = "";
										while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
											str += xmlReader.getText();
											xmlReader.next();
										}
										value = Double.parseDouble(str);
									} else if ("Unit".equals(element)) {
										while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
											unit += xmlReader.getText();
											xmlReader.next();
										}
									} else if ("Tex".equals(element)) {
										while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
											tex += xmlReader.getText();
											xmlReader.next();
										}
									} else if ("Locked".equals(element)) {
										locked = Boolean.parseBoolean(xmlReader.getText());
										xmlReader.next();
									} else if ("Fixed".equals(element)) {
										fixed = Boolean.parseBoolean(xmlReader.getText());
										xmlReader.next();
									} else {
										throw new Exception("Undefined tag");
									}
									xmlReader.next();
								}
								properties.add(new MaterialProperty<Double>(name, desc, group, value, unit, tex, locked,
										fixed));
								xmlReader.next();
							}
							xmlReader.next();

							// skip rest of dataset
							while ((xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT)
									|| !"Dataset".equals(xmlReader.getLocalName())) {
								xmlReader.next();
							}

							xmlReader.next();
						} else {
							// skip dataset
							xmlReader.next();
							while ((xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT)
									|| !"Dataset".equals(xmlReader.getLocalName())) {
								xmlReader.next();
							}
							xmlReader.next();
						}
					}
					xmlReader.next();
				}
				xmlReader.next();
			}
			xmlReader.next();

			xmlReader.close();
			is.close();
		} catch (FileNotFoundException ex) {
			ex.printStackTrace();
		} catch (XMLStreamException ex) {
			ex.printStackTrace();
		} catch (FactoryConfigurationError ex) {
			ex.printStackTrace();
		} catch (IOException ex) {
			ex.printStackTrace();
		} catch (Exception ex) {
			ex.printStackTrace();
		} finally {
			if (is != null) {
				try {
					is.close();
				} catch (IOException e) {
				}
			}
		}

		return properties;
	}

	@Override
	public LinkedList<MaterialDiagram> getDiagrams(long datasetID, EventLog log) {
		LinkedList<MaterialDiagram> diagrams;
		XMLStreamReader xmlReader = null;
		BufferedInputStream is = null;
		String element, str;
		String diagName, diagDesc;
		LinkedList<MaterialProperty<Double[]>> axis;
		String name, desc, group, unit, tex;
		Boolean locked, fixed;
		Double[] values;
		String[] list;

		diagrams = new LinkedList<MaterialDiagram>();

		try {
			is = new BufferedInputStream(new FileInputStream(this.database));
			xmlReader = XMLInputFactory.newInstance().createXMLStreamReader(is, "utf-8");

			// read file
			while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
				xmlReader.next();
			}

			xmlReader.next();

			// search dataset with id
			while ((xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT)
					&& "MaterialClass".equals(xmlReader.getLocalName())) {
				xmlReader.next();
				while ((xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT)
						&& "Material".equals(xmlReader.getLocalName())) {
					xmlReader.next();
					while ((xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT)
							&& "Dataset".equals(xmlReader.getLocalName())) {
						if (Long.parseLong(xmlReader.getAttributeValue(null, "ID")) == datasetID) {
							xmlReader.next();

							// search diagrams
							while ((xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT)
									|| !"Diagrams".equals(xmlReader.getLocalName())) {
								xmlReader.next();
							}
							xmlReader.next();

							// read diagrams items
							while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
								xmlReader.next();
								diagName = "";
								diagDesc = "";
								axis = new LinkedList<MaterialProperty<Double[]>>();

								// read properties
								while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
									element = xmlReader.getLocalName();
									xmlReader.next();
									if ("Name".equals(element)) {
										while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
											diagName += xmlReader.getText();
											xmlReader.next();
										}
									} else if ("Description".equals(element)) {
										while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
											diagDesc += xmlReader.getText();
											xmlReader.next();
										}
									} else if ("Axis".equals(element)) {
										// read diagrams items
										while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
											xmlReader.next();
											name = "";
											desc = "";
											group = "";
											values = new Double[0];
											unit = "";
											tex = "";
											locked = false;
											fixed = false;

											// read properties
											while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
												element = xmlReader.getLocalName();
												xmlReader.next();
												if ("Name".equals(element)) {
													while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
														name += xmlReader.getText();
														xmlReader.next();
													}
												} else if ("Description".equals(element)) {
													while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
														desc += xmlReader.getText();
														xmlReader.next();
													}
												} else if ("Group".equals(element)) {
													while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
														group += xmlReader.getText();
														xmlReader.next();
													}
												} else if ("Values".equals(element)) {
													str = "";
													while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
														str += xmlReader.getText();
														xmlReader.next();
													}
													list = str.split(";");
													values = new Double[list.length];
													for (int k = 0; k < list.length; k++) {
														values[k] = Double.parseDouble(list[k]);
													}
												} else if ("Unit".equals(element)) {
													while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
														unit += xmlReader.getText();
														xmlReader.next();
													}
												} else if ("Tex".equals(element)) {
													while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
														tex += xmlReader.getText();
														xmlReader.next();
													}
												} else if ("Locked".equals(element)) {
													locked = Boolean.parseBoolean(xmlReader.getText());
													xmlReader.next();
												} else if ("Fixed".equals(element)) {
													fixed = Boolean.parseBoolean(xmlReader.getText());
													xmlReader.next();
												} else {
													throw new Exception("Undefined tag");
												}
												xmlReader.next();
											}
											axis.add(new MaterialProperty<Double[]>(name, desc, group, values, unit,
													tex, locked, fixed));
											xmlReader.next();
										}
									} else {
										throw new Exception("Undefined tag");
									}
									xmlReader.next();
								}
								diagrams.add(new MaterialDiagram(diagName, diagDesc, axis));
								xmlReader.next();
							}
							xmlReader.next();

							// skip rest of dataset
							while ((xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT)
									|| !"Dataset".equals(xmlReader.getLocalName())) {
								xmlReader.next();
							}

							xmlReader.next();
						} else {
							// skip dataset
							xmlReader.next();
							while ((xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT)
									|| !"Dataset".equals(xmlReader.getLocalName())) {
								xmlReader.next();
							}
							xmlReader.next();
						}
					}
					xmlReader.next();
				}
				xmlReader.next();
			}
			xmlReader.next();

			xmlReader.close();
			is.close();

		} catch (FileNotFoundException ex) {
			ex.printStackTrace();
		} catch (XMLStreamException ex) {
			ex.printStackTrace();
		} catch (FactoryConfigurationError ex) {
			ex.printStackTrace();
		} catch (IOException ex) {
			ex.printStackTrace();
		} catch (Exception ex) {
			ex.printStackTrace();
		} finally {
			if (is != null) {
				try {
					is.close();
				} catch (IOException e) {
				}
			}
		}

		return diagrams;
	}

	@Override
	public void close() {

	}
}
