/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.utils;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.LinkedList;
import java.util.List;

import com.jezhumble.javasysmon.JavaSysMon;
import com.jezhumble.javasysmon.OsProcess;

/**
 * This class execute external command with watchdog-observation
 * 
 * @author Werner Koppelsttter
 */
public class ExecuteWithoutStarter extends Execute {

	/**
	 * Construct a process with specified command
	 * 
	 * @param command
	 *            Command to execute
	 */
	public ExecuteWithoutStarter(String... command) {
		super(null, command);
	}

	/**
	 * Construct a process with specified command in specified directory
	 * 
	 * @param directory
	 *            Working directory
	 * @param command
	 *            Command to execute
	 */
	public ExecuteWithoutStarter(File directory, String... command) {
		super(directory, 0, command);
	}

	/**
	 * Construct a process with specified command in specified directory and a
	 * watchdog
	 * 
	 * @param directory
	 *            Working directory
	 * @param timeout
	 *            Timeout for Watchdog in millisecounds (0 if no watchdog needed)
	 * @param command
	 *            Command to execute
	 */
	public ExecuteWithoutStarter(File directory, long timeout, String... command) {
		super(directory, timeout, command);
	}

	/**
	 * Start execution of the external process
	 */
	@Override
	protected Process execute() throws IOException, URISyntaxException {
		ProcessBuilder builder;
		Process p;
		int currentPid;
		OsProcess before, after;
		List<?> listBefore, listAfter;
		LinkedList<Integer> possiblePids;
		File file;
		JavaSysMon sys;

		// append directory to program if local
		if ((ExecuteWithoutStarter.this.directory != null)
				&& !new File(ExecuteWithoutStarter.this.command[0]).isAbsolute()) {
			file = new File(ExecuteWithoutStarter.this.directory, ExecuteWithoutStarter.this.command[0]);
			if (file.exists()) {
				ExecuteWithoutStarter.this.command[0] = file.getAbsolutePath();
			} else {
				// exe
				file = new File(ExecuteWithoutStarter.this.directory, ExecuteWithoutStarter.this.command[0] + ".exe");
				if (file.exists()) {
					ExecuteWithoutStarter.this.command[0] = file.getAbsolutePath();
				} else {
					// bat
					file = new File(ExecuteWithoutStarter.this.directory,
							ExecuteWithoutStarter.this.command[0] + ".bat");
					if (file.exists()) {
						ExecuteWithoutStarter.this.command[0] = file.getAbsolutePath();
					}
				}
			}
		}
		builder = new ProcessBuilder(ExecuteWithoutStarter.this.command);
		builder.directory(ExecuteWithoutStarter.this.directory);

		// set environment
		if (ExecuteWithoutStarter.this.environment != null) {
			for (String key : ExecuteWithoutStarter.this.environment.keySet()) {
				builder.environment().put(key, ExecuteWithoutStarter.this.environment.get(key));
			}
		}

		// FIXME find PID of subprocess (since java 9 there is a method pid() of class
		// Process)

		// start and list sub processes
		sys = new JavaSysMon();
		currentPid = sys.currentPid();
		before = sys.processTree().find(currentPid);
		p = builder.start();
		after = sys.processTree().find(currentPid);

//					for (Object o : before.children()) {
//						ProcessInfo p = ((OsProcess) o).processInfo();
//						System.out.println("b: " + p.getPid() + "\t" + p.getName() + "\t" + p.getCommand());
//					}

		// compare sub-processes
		if ((before != null) && (after != null)) {
//						System.out.println(before.children().size() + "->" + after.children().size());
//						for (Object o : after.children()) {
//							ProcessInfo p = ((OsProcess) o).processInfo();
//							System.out.println("a: " + p.getPid() + "\t" + p.getName() + "\t" + p.getCommand());
//						}

			listBefore = before.children();
			listAfter = after.children();
			possiblePids = new LinkedList<Integer>();
			for (Object aObj : listAfter) {
				int aPid = ((OsProcess) aObj).processInfo().getPid();
				boolean found = false;
				for (Object bObj : listBefore) {
					int bPid = ((OsProcess) bObj).processInfo().getPid();
					found |= (bPid == aPid);
				}
				if (!found) {
					possiblePids.add(aPid);
				}
			}
			if (possiblePids.size() == 1) {
				ExecuteWithoutStarter.this.pid = possiblePids.get(0);
			}
		}

//					System.out.println("PDI: " + ExecuteWithoutStarter.this.pid);

		return p;
	}
}
