/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.algorithm;

import jmetal.core.Solution;
import jmetal.core.SolutionSet;
import jmetal.core.Variable;
import jmetal.problems.ME_StandardProblemReal;
import jmetal.problems.StandardProblemReal;
import at.lcm.bu21.optim.algorithm.me_nsga2.ME_Nsga2;
import at.lcm.bu21.optim.core.Archive;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.ObjectiveType;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.core.VariableDesc;

/**
 * JMetal solver-algorithm
 * 
 * @author Gnther Weidenholzer
 * @author Werner Koppelsttter
 */
public abstract class JMSolver extends Solver {

	/** JMetal problem object */
	protected StandardProblemReal jmProblem = null;

	/**
	 * Create optimizer
	 * 
	 * @param problem
	 *            Based problem with description of variables and results
	 */
	public JMSolver(Problem problem) {
		super(problem);

		VariableDesc[] vars;
		double[] lb, ub;
		String[] vn;
		int nVariables;

		if (this.getProblem() != null) {
			nVariables = this.getProblem().getNumVariedVariables();
			lb = new double[nVariables];
			ub = new double[nVariables];
			vn = new String[nVariables];

			vars = this.getProblem().getVariedVariables();
			for (int k = 0; k < vars.length; k++) {
				lb[k] = vars[k].getLB();
				ub[k] = vars[k].getUB();
				String nameTokens[] = vars[k].getName().split("\\.");
				vn[k] = nameTokens[nameTokens.length - 1];
			}

			if (this instanceof ME_Nsga2) {
				this.jmProblem = new ME_StandardProblemReal(nVariables, lb, ub,
						vn, this.getProblem().getNumObjectives());

			} else {
				this.jmProblem = new StandardProblemReal(nVariables, lb, ub,
						this.getProblem().getNumObjectives());
			}

		}
	}

	/**
	 * Get Individuals from solutionSet
	 * 
	 * @param solution
	 *            Solution from JMetal
	 * @return Individual
	 * @throws Exception
	 */
	protected Individual getIndividual(Solution solution) throws Exception {
		VariableDesc[] varDesc;
		Individual ind;
		Archive archive;
		double val;

		// return new parameters
		archive = this.getProblem().getArchive();
		varDesc = this.getProblem().getVariables();

		Variable[] vars = solution.getDecisionVariables();
		ind = new Individual(archive);
		int n = 0;
		for (int p = 0; p < varDesc.length; p++) {
			if (varDesc[p].isVaried()) {
				val = varDesc[p].round(vars[n++].getValue());
			} else {
				val = varDesc[p].getLB();
			}
			ind.setVariable(p, val);

		}

		return ind;
	}

	/**
	 * Get Individuals from solutionSet
	 * 
	 * @param solutionSet
	 *            Solution-set of JMetal
	 * @return Array of Individuals
	 * @throws Exception
	 */
	public Individual[] getIndividuals(SolutionSet solutionSet)
			throws Exception {
		Individual[] individuals;

		// return new parameters
		individuals = new Individual[solutionSet.size()];

		for (int k = 0; k < solutionSet.size(); k++) {
			individuals[k] = this.getIndividual(solutionSet.get(k));
		}

		return individuals;
	}

	/**
	 * Get JMetal solution of individual
	 * 
	 * @param individual
	 *            Individual
	 * @return Solution-set for JMetal
	 * @throws Exception
	 */
	protected Solution getSolution(Individual individual) throws Exception {
		Solution solution;
		Variable[] variables;
		VariableDesc[] varDesc;
		ResultDesc[] resDesc;
		double val;
		int n;

		varDesc = this.getProblem().getVariables();
		resDesc = this.getProblem().getResults();

		solution = new Solution(this.jmProblem);
		variables = solution.getDecisionVariables();

		// set variable values
		n = 0;
		for (int p = 0; p < varDesc.length; p++) {
			if (varDesc[p].isVaried()) {
				variables[n++].setValue(individual.getVariable(p));
			}
		}

		// set objective values
		n = 0;
		for (int p = 0; p < resDesc.length; p++) {
			if (resDesc[p].isObjective()) {
				val = individual.getResult(p);
				if (resDesc[p].getObjective() == ObjectiveType.MAX) {
					val *= -1.0;
				}
				solution.setObjective(n++, val);
			}
		}

		return solution;
	}

	/**
	 * Get JMetal solution-set of individual
	 * 
	 * @param individual
	 *            Individual
	 * @return Solution-set for JMetal
	 * @throws Exception
	 */
	protected SolutionSet getSolutionSet(Individual individual)
			throws Exception {
		SolutionSet solutionSet;

		solutionSet = new SolutionSet(1);

		// set objective values
		solutionSet.add(this.getSolution(individual));

		return solutionSet;

	}

	/**
	 * Get JMetal solution-set of individuals
	 * 
	 * @param individuals
	 *            Individuals
	 * @return Solution-set for JMetal
	 * @throws Exception
	 */
	public SolutionSet getSolutionSet(Individual[] individuals)
			throws Exception {
		SolutionSet solutionSet;

		solutionSet = new SolutionSet(individuals.length);

		// set objective values
		for (int k = 0; k < individuals.length; k++) {
			if (individuals[k] != null) {
				solutionSet.add(this.getSolution(individuals[k]));
			}
		}

		return solutionSet;
	}
}
