/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.algorithm.nsga2;

import java.util.LinkedList;

import jmetal.util.JMException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.optim.algorithm.GenerationBased;
import at.lcm.bu21.optim.algorithm.JPanelSolver;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;

/**
 * Generation-based NSGA-II implementation
 * 
 * @author Gnther Weidenholzer
 * @author Werner Koppelsttter
 */
public class GenNsga2 extends Nsga2 implements GenerationBased {

	/**
	 * Constructor of NSGA-II solver
	 * 
	 * @param problem
	 *            Based problem with description of variables and results
	 * @throws JMException
	 */
	public GenNsga2(Problem problem) throws JMException {
		super(problem);
	}

	/** Size of population */
	private int populationSize = 100;

	/** Generation timeout */
	private int generationTimeout = 120;

	/** Unsolved individuals of generation */
	private LinkedList<Individual> unsolved = new LinkedList<Individual>();

	/** Solved individuals of generation */
	private LinkedList<Individual> solved = new LinkedList<Individual>();

	@Override
	public TStruct getSettings() throws Exception {
		TStruct resumeData;

		resumeData = super.getSettings();

		// population size
		resumeData
				.put("populationSize", new TNumeric(this.getPopulationSize()));

		// generation timeout
		resumeData.put("generationTimeout",
				new TNumeric(this.getGenerationTimeout()));

		return resumeData;
	}

	@Override
	public void setSettings(TStruct resumeData) throws Exception {
		TNumeric num;

		super.setSettings(resumeData);

		// population size
		num = (TNumeric) resumeData.get("populationSize");
		if (num != null) {
			this.setPopulationSize((int) num.getReal(0, 0));
		}

		// generation timeout
		num = (TNumeric) resumeData.get("generationTimeout");
		if (num != null) {
			this.setGenerationTimeout((int) num.getReal(0, 0));
		}
	}

	@Override
	public String getOptimizerName() {
		return "Generational " + super.getOptimizerName();
	}

	@Override
	public String getDescription() {
		return "Generational " + super.getDescription();
	}

	@Override
	public JPanelSolver getPanel() {
		return new GenNsga2Panel(this);
	}

	@Override
	public void createPopulationIndividuals() throws Exception {
		this.unsolved.clear();
		while (this.unsolved.size() < this.populationSize) {
			this.unsolved.add(this.getIndividual(this.solver.getOffspring()
					.get(0)));
			if (this.unsolved.size() < this.populationSize) {
				this.unsolved.add(this.getIndividual(this.solver.getOffspring()
						.get(1)));
			}
		}
	}

	@Override
	public void takePopulationResults() throws Exception {

		this.solver.union(this.getSolutionSet(this.solved
				.toArray(new Individual[0])));
		this.solved.clear();

	}

	@Override
	public Individual getNextIndividual() {
		Individual ind = null;

		if (!this.unsolved.isEmpty()) {
			ind = this.unsolved.remove();
		}

		return ind;
	}

	@Override
	public void setResolvedIndividual(Individual ind) {
		if (this.getProblem().getArchive() == ind.getArchive()) {
			this.solved.add(ind);
		}

	}

	@Override
	public boolean hasNext() {
		return !this.unsolved.isEmpty();
	}

	@Override
	public int getPopulationSize() {
		return this.populationSize;
	}

	@Override
	public void setPopulationSize(int popSize) {
		this.populationSize = popSize;
	}

	@Override
	public int getGenerationTimeout() {
		return this.generationTimeout;
	}

	@Override
	public void setGenerationTimeout(int timeout) {
		this.generationTimeout = timeout;
	}

}
