/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core;

import java.io.Serializable;

import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;

/**
 * Description of variable. Range of value for the variable
 * 
 * @author Werner Koppelsttter
 */
public class VariableDesc implements Serializable {

	/** Name of variable */
	private String name;

	/** Lower boundary of value */
	private double lb = Double.NaN;

	/** Step size of discretisation */
	private double stepSize = 0.0;

	/** Upper boundary of value */
	private double ub = Double.NaN;

	/**
	 * Create variable from resumeData
	 * 
	 * @param resumeData
	 *            ResumeData
	 * @throws TDataException
	 */
	VariableDesc(TStruct resumeData) throws TDataException {
		this.setResumeData(resumeData);
	}

	/**
	 * Create variable
	 * 
	 * @param name
	 *            Name of variable
	 */
	public VariableDesc(String name) {
		this.name = name;
	}

	/**
	 * Create variable
	 * 
	 * @param name
	 *            Name of variable
	 * @param lb
	 *            Lower boundary of value
	 * @param stepSize
	 *            Step-size of discretisation
	 * @param ub
	 *            Upper boundary of value
	 */
	public VariableDesc(String name, double lb, double stepSize, double ub) {
		this(name);
		this.setBounds(lb, stepSize, ub);
	}

	/**
	 * Get name of varaible
	 * 
	 * @return Name
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Get lower boundary of value
	 * 
	 * @return Lower boundary
	 */
	public double getLB() {
		return this.lb;
	}

	/**
	 * Step-size of discretisation
	 * 
	 * @return Step-size
	 */
	public double getStepSize() {
		return this.stepSize;
	}

	/**
	 * Get number of steps
	 * 
	 * @return Step-size
	 */
	public double getNumSteps() {
		double num, tmp;

		if (!this.isValid()) {
			num = Double.NaN;
		} else if (Double.isInfinite(this.lb) || Double.isInfinite(this.ub)
				|| ((this.lb < this.ub) && (this.stepSize == 0.0))) {
			num = Double.POSITIVE_INFINITY;
		} else if (this.lb == this.ub) {
			num = 1;
		} else {
			tmp = ((this.ub - this.lb) / this.stepSize) + 1;
			num = (int) tmp;
			// check for numeric error
			if ((tmp - num) > (1 - 1e-14)) {
				num++;
			}
		}

		return num;
	}

	/**
	 * Get upper boundary of value
	 * 
	 * @return Upper boundary
	 */
	public double getUB() {
		return this.ub;
	}

	/**
	 * Set bounds of value
	 * 
	 * @param lb
	 *            Lower boundary
	 * @param step
	 *            Step-size
	 * @param ub
	 *            Upper boundary
	 */
	public void setBounds(double lb, double step, double ub) {
		this.lb = lb;
		this.stepSize = step;
		this.ub = ub;
	}

	/**
	 * Is value unvaried (value is fix, in case of lower-bound is equal upper-bound)
	 * 
	 * @return Is unvaried
	 */
	public boolean isVaried() {
		return this.lb != this.ub;
	}

	/**
	 * Is value valid. Is value between lower and upper boundary and on the
	 * step-size
	 * 
	 * @param val
	 *            Value to check
	 * @return Is value valid
	 */
	public boolean isValidValue(double val) {
		return (this.lb <= val) && (val <= this.ub) && (val == this.round(val));
	}

	/**
	 * Check if variable definition is valid (completely defined)
	 * 
	 * @return Is valid
	 */
	public boolean isValid() {
		return !Double.isNaN(this.lb) && !Double.isNaN(this.stepSize) && !Double.isNaN(this.ub) && (this.lb <= this.ub);
	}

	/**
	 * Round value to steps
	 * 
	 * @param val
	 *            Value to round
	 * @return Round value
	 */
	public double round(double val) {
		if (this.stepSize != 0.0) {
			// check limits;
			if (val < this.lb) {
				val = this.lb;
			} else if (val > this.ub) {
				val = this.ub;
			}

			// calc posible step
			val = this.lb + (this.stepSize * Math.round((val - this.lb) / this.stepSize));

			// check upper limit
			if (val > this.ub) {
				// check for numeric error
				if (val > (this.ub + (this.stepSize * 1e-14))) {
					val -= this.stepSize;
				} else {
					val = this.ub;
				}
			}

			// round
			val = Double.parseDouble(TNumeric.num2str(val, 14));
		}

		return val;
	}

	/**
	 * Get resume data
	 * 
	 * @return Resume data
	 * @throws TDataException
	 */
	TStruct getResumeData() throws TDataException {
		TStruct resumeData;

		resumeData = new TStruct();
		resumeData.put("name", new TString(this.name));
		resumeData.put("lb", new TNumeric(this.lb));
		resumeData.put("steps", new TNumeric(this.stepSize));
		resumeData.put("ub", new TNumeric(this.ub));

		return resumeData;
	}

	/**
	 * Set resume data
	 * 
	 * @param resumeData
	 *            Resume Data
	 * @throws TDataException
	 */
	void setResumeData(TStruct resumeData) throws TDataException {
		this.name = ((TString) resumeData.get("name")).getString();
		this.lb = ((TNumeric) resumeData.get("lb")).getReal(0, 0);
		this.stepSize = ((TNumeric) resumeData.get("steps")).getReal(0, 0);
		this.ub = ((TNumeric) resumeData.get("ub")).getReal(0, 0);
	}

	@Override
	public String toString() {
		StringBuffer str;

		str = new StringBuffer();
		str.append(this.name);
		str.append(":\t");
		str.append(this.lb);
		str.append(" : ");
		str.append(this.stepSize);
		str.append(" : ");
		str.append(this.ub);
		str.append("\t#");
		str.append(this.getNumSteps());
		str.append("\n");

		return str.toString();
	}
}
