/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.mapping.utils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.UUID;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;

import at.lcm.bu21.optim.algorithm.JMSolver;
import at.lcm.bu21.optim.algorithm.SolverList;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.core.VariableDesc;
import at.lcm.bu21.optim.core.mapping.MappingManager;
import at.lcm.bu21.optim.core.mapping.utils.comparators.MappingTaskAccMinComparator;
import at.lcm.bu21.optim.core.mapping.utils.comparators.MappingTaskCorrMinComparator;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants.THolderState;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingTask;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingTask.TRegressionModelType;
import jmetal.metaheuristics.nsgaII.Nsga2Real;
import main.ocvwrap.java.OpenCVMLPPredictor;

public class MappingHolder implements Serializable {
	private static final long serialVersionUID = 00000000000000315L; // opencv-based
																		// mapping
	private String targetName;

	private OpenCVMLPPredictor fastModel;
	private TMappingTask fastModelTask;
	private ConcurrentLinkedQueue<TMappingTask> completedFastCVEvaluationTasks = new ConcurrentLinkedQueue<TMappingTask>();
	private List<OpenCVMLPPredictor> ensModels = new ArrayList<OpenCVMLPPredictor>();
	private List<TMappingTask> ensModelTasks = new ArrayList<TMappingTask>();
	private ConcurrentLinkedQueue<TMappingTask> completedEnsCVEvaluationTasks = new ConcurrentLinkedQueue<TMappingTask>();
	private int nrOfRegressionSamples = -1;
	private int nrOfClassificationSamples = -1;
	private String fastModelQuality = "-";
	private String ensModelQuality = "-";

	private int nrOfVariables = -1;
	private int reducedTrainingSetSize = -1;
	private int maxTrainingDuration = -1;
	private int ensembleSize = 5;
	private double fastModelAccuracyThreshold = 0.95;

	// private int percentOfHighQualityIndividuals = 2;
	// private String mappingType;

	private transient OpenCVMLPPredictor prevFastModel;
	private transient TMappingTask prevFastModelTask;
	private transient List<OpenCVMLPPredictor> prevEnsModels = new ArrayList<OpenCVMLPPredictor>();
	private transient List<TMappingTask> prevEnsModelTasks = new ArrayList<TMappingTask>();
	private transient int prevNrOfRegressionSamples = 0;
	private transient int prevNrOfClassificationSamples = 0;
	private transient String prevFastModelQuality = "-";
	private transient String prevEnsModelQuality = "-";

	private transient boolean fastModelSelected = false;
	private transient boolean ensModelSelected = false;
	private transient int validEnsembleModels = 0;
	private transient int retrievedEnsembleModels = 0;
	private transient MappingManager parentMappingManager;
	private transient THolderState state;
	private transient MappingCreationThread creationThread;
	private transient AtomicBoolean stopMappingCreation;
	private transient AtomicBoolean hasUnremovedJobs;
	private transient boolean fastTrainingEnabled = false;
	private transient List<TMappingTask> candidateEnsModelTasks = new ArrayList<TMappingTask>();

	public MappingHolder(String targetName, MappingManager parentMappingManager, String mappingType) {
		super();
		this.targetName = targetName;
		// this.mappingType = mappingType;
		// this.percentOfHighQualityIndividuals = 2;
		this.parentMappingManager = parentMappingManager;
		if (this.reducedTrainingSetSize < 0) {
			this.reducedTrainingSetSize = TMappingConstants.DEFAULT_NR_OF_TRAINING_SAMPLES;
		}
		if (this.maxTrainingDuration < 0) {
			this.maxTrainingDuration = TMappingConstants.DEFAULT_MAPPING_CREATION_TIME_LIMIT;
		}

		this.nrOfVariables = this.parentMappingManager.getProblem().getVariedVariables().length;

		// this.nrOfRegressionSamples = this.parentMappingManager.getProblem()
		// .getArchive().getMemory()
		// .getValidIndividual(this.parentMappingManager.getProblem()).length;

		Individual[] validIndividuals = this.parentMappingManager.getProblem().getArchive().getMemory()
				.getValidIndividual(this.parentMappingManager.getProblem());

		this.nrOfRegressionSamples = validIndividuals.length;

		Individual[] errorIndividuals = this.parentMappingManager.getProblem().getArchive().getMemory()
				.getErrorIndividual();

		this.nrOfClassificationSamples = this.nrOfRegressionSamples + errorIndividuals.length;

		this.stopMappingCreation = new AtomicBoolean(false);
		this.hasUnremovedJobs = new AtomicBoolean(false);
		this.state = THolderState.IDLE;

		// TODO update build number
		this.parentMappingManager.addLog("Build " + MappingHolder.serialVersionUID
				+ "A - OpenCV Ensemble mapping holder created for: " + targetName);
	}

	protected void addLog(String message) {
		this.parentMappingManager.addLog(message);
	}

	private void backupExistingEnsModel() {
		/** back-up existing ensemble model */
		this.prevEnsModels.clear();
		for (OpenCVMLPPredictor cls : this.ensModels) {
			this.prevEnsModels.add(cls);
			this.parentMappingManager.addLog("Backed up one more ensemble model!");
		}
		this.prevEnsModelQuality = this.ensModelQuality;
		this.prevEnsModelTasks.clear();
		for (TMappingTask tmt : this.ensModelTasks) {
			this.prevEnsModelTasks.add(tmt);
			this.parentMappingManager.addLog("Backed up one more ensemble model task!");
		}
		this.prevNrOfClassificationSamples = this.nrOfClassificationSamples;
		this.prevNrOfRegressionSamples = this.nrOfRegressionSamples;
	}

	private void restorePreviousEnsembleModel() {
		/** restore previous ensemble model */
		this.ensModels.clear();
		for (OpenCVMLPPredictor cls : this.prevEnsModels) {
			this.ensModels.add(cls);
			this.parentMappingManager.addLog("Restored one more ensemble model!");
		}
		this.ensModelQuality = this.prevEnsModelQuality;
		this.ensModelTasks.clear();
		for (TMappingTask tmt : this.prevEnsModelTasks) {
			this.ensModelTasks.add(tmt);
			this.parentMappingManager.addLog("Restored one more ensemble model task!");
		}
		this.nrOfClassificationSamples = this.prevNrOfClassificationSamples;
		this.nrOfRegressionSamples = this.prevNrOfRegressionSamples;
	}

	private void clearFastModelInformation() {
		/** clear existing fast model data */
		this.fastModelTask = null;
		this.fastModelQuality = "-";
		this.completedFastCVEvaluationTasks.clear();
		this.fastModel = null;
		this.nrOfClassificationSamples = 0;
		this.nrOfRegressionSamples = 0;
	}

	private void clearEnsModelInformation() {
		/** clear existing ensemble model data */
		this.ensModelTasks.clear();
		this.ensModelQuality = "-";
		this.completedEnsCVEvaluationTasks.clear();
		this.ensModels.clear();
		this.nrOfClassificationSamples = 0;
		this.nrOfRegressionSamples = 0;
	}

	private boolean createBaseDatasetFile(String baseDatasetFileName, boolean createReducedDataset) {
		boolean outcome = true;
		try {
			Individual[] allValidIndividuals = this.parentMappingManager.getProblem().getArchive().getMemory()
					.getValidIndividual(this.parentMappingManager.getProblem());

			this.nrOfRegressionSamples = allValidIndividuals.length;

			Individual[] allErrorIndividuals = this.parentMappingManager.getProblem().getArchive().getMemory()
					.getErrorIndividual();

			this.nrOfClassificationSamples = this.nrOfRegressionSamples + allErrorIndividuals.length;

			Individual[] validIndividuals;
			Individual[] errorIndividuals;
			System.out.println(
					createReducedDataset + "," + this.nrOfClassificationSamples + "," + this.reducedTrainingSetSize);
			if ((!createReducedDataset) || (this.nrOfRegressionSamples < this.reducedTrainingSetSize)) {
				validIndividuals = new Individual[allValidIndividuals.length];
				for (int i = 0; i < allValidIndividuals.length; i++) {
					validIndividuals[i] = allValidIndividuals[i];
				}
				errorIndividuals = new Individual[allErrorIndividuals.length + 1];
				for (int i = 0; i < allErrorIndividuals.length; i++) {
					errorIndividuals[i] = allErrorIndividuals[i];
				}
				errorIndividuals[errorIndividuals.length - 1] = validIndividuals[0];
			} else {
				Random rndGen = new Random(System.currentTimeMillis());
				double redRatio = (this.nrOfRegressionSamples * 1.0) / this.reducedTrainingSetSize;
				this.parentMappingManager.addLog("redRatio=" + redRatio);
				int validSize = this.reducedTrainingSetSize;
				validIndividuals = new Individual[validSize];
				List<Individual> nonDominatedInds = new ArrayList<Individual>();
				nonDominatedInds.addAll(Arrays.asList(((JMSolver) SolverList
						.getSolverListEntry(this.parentMappingManager.getProblem().getSolver().getOptimizerName())
						.getSolver(this.parentMappingManager.getProblem()))
								.getIndividuals(Nsga2Real.refineSolutionSet(
										((JMSolver) SolverList
												.getSolverListEntry(this.parentMappingManager.getProblem().getSolver()
														.getOptimizerName())
												.getSolver(this.parentMappingManager.getProblem()))
														.getSolutionSet(allValidIndividuals),
										validSize / 2, this.parentMappingManager.getProblem().getNumObjectives()))));
				for (int i = 0; i < nonDominatedInds.size(); i++) {
					validIndividuals[i] = this.parentMappingManager.getProblem().getArchive().getMemory()
							.getIndividual(nonDominatedInds.get(i));
				}
				int i = nonDominatedInds.size();
				while (i < validSize) {
					int j = rndGen.nextInt(allValidIndividuals.length);
					validIndividuals[i] = allValidIndividuals[j];
					i++;
				}
				errorIndividuals = new Individual[allErrorIndividuals.length + 1];
				for (int j = 0; j < allErrorIndividuals.length; j++) {
					errorIndividuals[j] = allErrorIndividuals[j];
				}
				errorIndividuals[errorIndividuals.length - 1] = validIndividuals[0];
			}

			int samplesToSave = validIndividuals.length;
			if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
				samplesToSave = validIndividuals.length + errorIndividuals.length;
			}
			this.parentMappingManager.addLog("Saving " + (samplesToSave) + " samples in the base dataset file.");
			this.nrOfRegressionSamples = validIndividuals.length;
			this.nrOfClassificationSamples = validIndividuals.length + errorIndividuals.length;

			File f = new File(baseDatasetFileName);
			if (f.getParent() != null) {
				File dir = new File(f.getParent());
				if (!dir.exists() && !dir.mkdirs()) {
					throw new IOException("ERROR: could not create directory path: " + f.getParent());
				}
			}
			if (f.exists()) {
				f.delete();
			}

			boolean created = false;
			while (!created) {
				try {
					f.createNewFile();
					created = true;
				} catch (Exception e) {
					e.printStackTrace();
					created = false;
					Thread.sleep(1000);
				}
			}

			// create the .arff file and write the header data
			int numMappingTargets = this.parentMappingManager.getNumMappingTargets();
			if (!this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
				numMappingTargets--;
			}
			BufferedWriter bw = new BufferedWriter(new FileWriter(baseDatasetFileName));
			bw.write("@RELATION targetCount_" + numMappingTargets + "\n\n");
			bw.write("@ATTRIBUTE " + TMappingTask.TParameter.IND_ID + " NUMERIC\n");
			for (VariableDesc varDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
				String originalName = varDesc.getName();
				String opencvSafeName = originalName.replace(" ", "_");
				bw.write("@ATTRIBUTE " + opencvSafeName + " NUMERIC\n");

				// bw.write("@ATTRIBUTE " + varDesc.getName() + " NUMERIC\n");
			}
			for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
				if (rDesc.isObjective() || rDesc.isConstraint()) {

					String originalName = rDesc.getName();
					String opencvSafeName = originalName.replace(" ", "_");
					bw.write("@ATTRIBUTE " + opencvSafeName + " NUMERIC\n");

					// bw.write("@ATTRIBUTE " + rDesc.getName() + " NUMERIC\n");
				}
			}
			if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
				bw.write("@ATTRIBUTE " + TMappingConstants.ERRORS_CLASSIFIER + " NUMERIC\n");
			}
			bw.write("\n@DATA\n");

			// write the individual information (generation + variables +
			// targets)
			int count = 0;
			for (Individual ind : validIndividuals) {
				count++;
				String line = count + ",";
				for (VariableDesc vdesc : this.parentMappingManager.getProblem().getVariedVariables()) {
					line += ind.getVariable(vdesc.getName()) + ",";
				}

				// for (int i = 0; i < this.nrOfVariables; i++) {
				// Double val = ind.getVariable(i);
				// line += val + ", ";
				// }

				for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
					if (rDesc.isObjective() || rDesc.isConstraint()) {
						Double val = ind.getResult(rDesc.getName());
						line += val + ", ";
					}
				}

				if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
					line += TMappingConstants.NO_ERROR;
				} else {
					line = line.substring(0, line.length() - 2);
				}

				bw.write(line + "\n");
			}

			if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {

				for (Individual ind : errorIndividuals) {
					count++;
					String line = count + ",";
					for (VariableDesc vdesc : this.parentMappingManager.getProblem().getVariedVariables()) {
						line += ind.getVariable(vdesc.getName()) + ",";
					}

					// for (int i = 0; i < this.nrOfVariables; i++) {
					// Double val = ind.getVariable(i);
					// line += val + ", ";
					// }

					for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
						if (rDesc.isObjective() || rDesc.isConstraint()) {
							line += "0.0, ";
						}
					}

					line += TMappingConstants.ERROR;

					// line = line.substring(0, line.length() - 2);

					bw.write(line + "\n");
				}
			}

			bw.close();
		} catch (Exception e) {
			e.printStackTrace();
			outcome = false;
		}

		return outcome;
	}// end of createBaseDatasetFile method

	private boolean createCrossValidationEvaluationExperimentMasterFile(String masterFileName,
			String baseDatasetFileName, TRegressionModelType modelType) {
		boolean outcome = true;

		try {
			/** create directory structure if it does not exist */
			File f = new File(masterFileName);

			this.parentMappingManager.addLog("Temp master file absolute path:" + f.getAbsolutePath());
			if (f.getParent() != null) {
				File dir = new File(f.getParent());
				if (!dir.exists() && !dir.mkdirs()) {
					throw new IOException("ERROR: could not create directory path: " + f.getParent());
				}
			}

			BufferedWriter bw = new BufferedWriter(new FileWriter(masterFileName, false));
			if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
				bw.write("experimentType = cv_evaluation_cls\n");
			} else {
				bw.write("experimentType = cv_evaluation_reg\n");
			}

			/** if we are creating LINEAR regression/classification models */
			if (modelType.equals(TRegressionModelType.LINEAR)) {

				// int batchCount = 1;
				//
				// bw.write("nrOfCrossValidationFolds = 10\n");
				// bw.write("nrOfTaskBatches = " + batchCount + "\n");
				// if
				// (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER))
				// {
				// bw.write("trainingSetSize = "
				// + this.nrOfClassificationSamples + "\n");
				// } else {
				// bw.write("trainingSetSize = " + this.nrOfRegressionSamples
				// + "\n");
				// }
				// bw.write("baseDataset = " + baseDatasetFileName + "\n");
				// // bw.write("target = " + this.targetName + "\n");
				// bw.write("target = " + this.targetName.replace(" ", "_") +
				// "\n");
				// if
				// (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER))
				// {
				// bw.write("modelType = LOGISTIC\n");
				//
				// } else {
				// bw.write("modelType = LINEAR\n");
				// }
				// bw.write("ridgeParam = 0.00000001\n");

				bw.write("nrOfCrossValidationFolds = 10\n");
				bw.write("nrOfTaskBatches = " + 3 + "\n");
				if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
					bw.write("trainingSetSize = " + this.nrOfClassificationSamples + "\n");
				} else {
					bw.write("trainingSetSize = " + this.nrOfRegressionSamples + "\n");
				}
				bw.write("baseDataset = " + baseDatasetFileName + "\n");
				bw.write("target = " + this.targetName + "\n");
				bw.write("modelType = ANN\n");

				int maxNodes = Math.min(15, (2 * this.nrOfVariables));
				bw.write("hiddenUnits = " + maxNodes + "\n");
				bw.write("learnRate = 0.3, 0.2, 0.1\n");
				bw.write("momentum = 0.1\n");
				bw.write("trainingTime = 500\n");

			}

			if (modelType.equals(TRegressionModelType.ANN)) {

				int maxNodes = Math.min(31, (2 * this.nrOfVariables));
				int batchCount = (maxNodes - 1) * 10;
				if (this.fastTrainingEnabled) {
					maxNodes = Math.min(6, (2 * this.nrOfVariables));
					batchCount = (maxNodes - 1) * 9;
				}

				bw.write("nrOfCrossValidationFolds = 10\n");
				bw.write("nrOfTaskBatches = " + batchCount + "\n");
				if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
					bw.write("trainingSetSize = " + this.nrOfClassificationSamples + "\n");
				} else {
					bw.write("trainingSetSize = " + this.nrOfRegressionSamples + "\n");
				}
				bw.write("baseDataset = " + baseDatasetFileName + "\n");
				bw.write("target = " + this.targetName + "\n");
				bw.write("modelType = ANN\n");

				if (!this.fastTrainingEnabled) {
					// max 900 grid points
					bw.write("hiddenUnits = ");
					for (int i = 2; i <= maxNodes; i++) {
						bw.write("" + i);
						if (i < maxNodes) {
							bw.write(", ");
						}
					}
					/** real */
					// bw.write("\n");
					// bw.write("learnRate = 0.4, 0.3, 0.2, 0.1, 0.05\n");
					// bw.write("momentum = 0.3, 0.2, 0.1, 0.05\n");
					// bw.write("trainingTime = 1200\n");

					/** test */
					bw.write("\n");
					bw.write("learnRate = 0.4, 0.3, 0.2, 0.1, 0.05\n");
					bw.write("momentum = 0.3, 0.2, 0.1, 0.05\n");
					bw.write("trainingTime = 144\n");
				} else {
					// max 25 grid points
					bw.write("hiddenUnits = ");
					for (int i = 2; i <= maxNodes; i++) {
						bw.write("" + i);
						if (i < maxNodes) {
							bw.write(", ");
						}
					}
					bw.write("\n");
					bw.write("learnRate = 0.4, 0.3, 0.2, 0.1\n");
					bw.write("momentum = 0.3, 0.2, 0.1\n");
					bw.write("trainingTime = 300\n");
				}

			}

			bw.close();
		} catch (Exception e) {
			this.parentMappingManager.addLog("ERROR: could not create Experiment master file at: " + masterFileName);
			e.printStackTrace();
			outcome = false;

		}

		return outcome;
	} // end of createCrossValidationEvaluationExperimentMasterFile method

	private boolean createModelTrainingExperimentMasterFile(String masterFileName, String baseDatasetFileName,
			TMappingTask trainingTask) {
		boolean outcome = true;

		try {
			/** create directory structure if it does not exist */
			File f = new File(masterFileName);
			if (f.getParent() != null) {
				File dir = new File(f.getParent());
				if (!dir.exists() && !dir.mkdirs()) {
					throw new IOException("ERROR: could not create directory path: " + f.getParent());
				}
			}

			BufferedWriter bw = new BufferedWriter(new FileWriter(masterFileName, false));
			if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
				bw.write("experimentType = model_training_cls\n");
			} else {
				bw.write("experimentType = model_training_reg\n");
			}
			bw.write("nrOfCrossValidationFolds = 10\n");
			bw.write("nrOfTaskBatches = 1\n");
			if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
				bw.write("trainingSetSize = " + this.nrOfClassificationSamples + "\n");
			} else {
				bw.write("trainingSetSize = " + this.nrOfRegressionSamples + "\n");
			}
			bw.write("baseDataset = " + baseDatasetFileName + "\n");
			bw.write("target = " + this.targetName.replace(" ", "_") + "\n");
			// bw.write("target = " + this.targetName + "\n");

			/**
			 * In the current version, it is assumed that the current model type is ANN or
			 * LINEAR/LOGISTIC
			 */

			if (trainingTask.getParameters().get(TMappingTask.TParameter.MOD_TYPE.getValue()).toUpperCase()
					.equals(TMappingConstants.TRegressionModelType.ANN.getValue())) {
				bw.write("modelType = ANN\n");
				bw.write("hiddenUnits = "
						+ trainingTask.getParameters().get(TMappingTask.TParameter.HIDDEN_UNITS.getValue()) + "\n");
				bw.write("learnRate = "
						+ trainingTask.getParameters().get(TMappingTask.TParameter.LEARN_RATE.getValue()) + "\n");
				bw.write("momentum = " + trainingTask.getParameters().get(TMappingTask.TParameter.MOMENTUM.getValue())
						+ "\n");
				bw.write("trainingTime = "
						+ trainingTask.getParameters().get(TMappingTask.TParameter.TRAINING_TIME.getValue()) + "\n");
			}

			if (trainingTask.getParameters().get(TMappingTask.TParameter.MOD_TYPE.getValue()).toUpperCase()
					.equals(TMappingConstants.TRegressionModelType.LINEAR.getValue())) {
				bw.write("modelType = LINEAR\n");
				bw.write("ridgeParam = "
						+ trainingTask.getParameters().get(TMappingTask.TParameter.RIDGE_PARAM.getValue()) + "\n");
			}

			if (trainingTask.getParameters().get(TMappingTask.TParameter.MOD_TYPE.getValue()).toUpperCase()
					.equals(TMappingConstants.TRegressionModelType.LOGISTIC.getValue())) {
				bw.write("modelType = LOGISTIC\n");
				bw.write("ridgeParam = "
						+ trainingTask.getParameters().get(TMappingTask.TParameter.RIDGE_PARAM.getValue()) + "\n");
			}

			bw.close();
		} catch (Exception e) {
			this.parentMappingManager.addLog("ERROR: could not create Experiment master file at: " + masterFileName);
			e.printStackTrace();
			outcome = false;
		}

		return outcome;
	} // end of createModelTrainingExperimentMasterFile

	public String getNrOfUsedTrainingSamples() {
		if (this.targetName.equalsIgnoreCase(TMappingConstants.ERRORS_CLASSIFIER)) {
			return "" + this.nrOfClassificationSamples;
		} else {
			return "" + this.nrOfRegressionSamples;
		}
	}

	public int getNrOfCurrentTrainingSamples() {
		return this.parentMappingManager.getProblem().getArchive().getMemory()
				.getValidIndividual(this.parentMappingManager.getProblem()).length;
	}

	public String getNrOfUsedClassificationSamples() {
		return "" + this.nrOfClassificationSamples;
	}

	public double getMappingPrediction(List<Double> dataSample) throws Exception {
		double result = 0.0;
		if (this.targetName.equalsIgnoreCase(TMappingConstants.ERRORS_CLASSIFIER)) {
			/** ensemble 2-class classification prediction */
			int cls1Count = 0; // negative values
			int cls2Count = 0; // positive values
			for (int i = 0; i < this.ensModels.size(); i++) {
				double prediction = this.ensModels.get(i).predictSingleSample(dataSample);
				if (prediction < 0) {
					cls1Count++;
				} else {
					cls2Count++;
				}
			}
			if (cls1Count >= cls2Count) {
				result = Double.parseDouble(TMappingConstants.ERROR);
			} else {
				result = Double.parseDouble(TMappingConstants.NO_ERROR);
			}

		} else {
			/** ensemble regression prediction */

			double sum = 0.0;
			for (int i = 0; i < this.ensModels.size(); i++) {
				sum += this.ensModels.get(i).predictSingleSample(dataSample);
			}
			sum /= this.ensModels.size();
			result = sum;

		}

		return result;
	}

	public String getModelQuality() {

		String modelQuality = "-";

		if (this.ensModels.size() != 0) {
			modelQuality = this.ensModelQuality;
		}

		return modelQuality;
		// this.fastModelQuality = "-";
		// try {
		// if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
		// this.fastModelQuality = this.fastModelTask.getResults()
		// .get(TMappingTask.TResult.CV_ACC_MIN.getValue())
		// .substring(0, 6);
		// } else {
		// this.fastModelQuality = this.fastModelTask.getResults()
		// .get(TMappingTask.TResult.CV_CORR_MIN.getValue())
		// .substring(0, 6);
		// }
		// } catch (Exception e) {
		// e.printStackTrace();
		// }
		// return this.fastModelQuality;
	}

	public THolderState getState() {
		return this.state;
	}

	public String getTargetName() {
		return this.targetName;
	}

	void handleEndOfFastCVExperiment() {
		this.parentMappingManager.addLog("Fast CV Experiment - collected " + this.completedFastCVEvaluationTasks.size()
				+ " regression models candidates for target: " + this.targetName);

		/**
		 * Select the best candidate using the "high CV-adjusted correlation" method
		 */
		if (!this.stopMappingCreation.get()) {
			this.fastModelTask = this.selectFastModelTaskUsingCVAdjustedCorrelation(this.targetName);
		} else {
			this.fastModelTask = null;
		}

		if (this.fastModelTask != null) {
			/** Launch a fast model training experiment for the best candidate */
			this.fastModelSelected = true;
			this.performFastModelTrainingExperiment();
		} else {
			/** Restore previous models */
			// this.restorePreviousFastModel();
			this.restorePreviousEnsembleModel();
			if (this.ensModels.size() == 0) {
				this.state = THolderState.IDLE;
			} else {
				this.state = THolderState.READY;
			}

			this.parentMappingManager
					.addLog("ERROR: The desired fast model could not be created! :  handleEndOfFastCVExperiment()");

			this.parentMappingManager.handleHolderStateChange(this);
		}

		/** Clear the temp directory */
	}

	void handleEndOfEnsCVExperiment() {
		this.parentMappingManager
				.addLog("Ensemble CV Experiment - collected " + this.completedEnsCVEvaluationTasks.size()
						+ " regression models candidates for target: " + this.targetName);

		/**
		 * Select the best candidate using the "high CV-adjusted correlation" method
		 */
		if (!this.stopMappingCreation.get() && (this.completedEnsCVEvaluationTasks.size() != 0)) {
			this.candidateEnsModelTasks = this.selectEnsModelTasksUsingCVAdjustedCorrelation(this.targetName);
		} else {
			this.candidateEnsModelTasks = new ArrayList<TMappingTask>();
		}

		if (this.candidateEnsModelTasks.size() != 0) {
			/**
			 * Launch an ensemble model training experiment for the best candidate
			 */
			this.ensModelSelected = true;
			this.performEnsModelTrainingExperiment();
		} else {
			/** Restore previous models */
			// this.restorePreviousFastModel();
			this.restorePreviousEnsembleModel();
			if (this.ensModels.size() == 0) {
				this.state = THolderState.IDLE;
			} else {
				this.state = THolderState.READY;
			}

			this.parentMappingManager
					.addLog("ERROR: The desired ensemble model could not be created! :  handleEndOfEnsCVExperiment()");

			this.parentMappingManager.handleHolderStateChange(this);
		}

		/** Clear the temp directory */
	}

	void handleEndOfCrossValidationExperiment() {
		if (!this.fastModelSelected) {
			this.handleEndOfFastCVExperiment();
		} else {
			this.handleEndOfEnsCVExperiment();
		}
	}

	synchronized void handleEndOfFastTrainingExperiment(OpenCVMLPPredictor regressionModel, TMappingTask trainingTask) {
		if ((regressionModel != null) && (trainingTask != null)) {
			// this.state = THolderState.READY;
			this.fastModel = regressionModel;
			this.fastModelTask = trainingTask;
			if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
				this.fastModelQuality = this.fastModelTask.getResults().get(TMappingTask.TResult.CV_ACC_MIN.getValue())
						.substring(0, 6);
			} else {
				this.fastModelQuality = this.fastModelTask.getResults().get(TMappingTask.TResult.CV_CORR_MIN.getValue())
						.substring(0, 6);
			}
			this.parentMappingManager.addLog("Created the fast model for target: " + this.targetName);

			double accEstimation = Double.parseDouble(this.fastModelQuality);
			this.parentMappingManager
					.addLog("Estimed fast model accuracy of " + accEstimation + " for target: " + this.targetName);
			if (accEstimation < this.fastModelAccuracyThreshold) {
				this.parentMappingManager.addLog("Non-linear modeling IS required for target: " + this.targetName);
				// ** a non-linear ensemble model must be trained */
				this.startEnsCrossValidationExperiment();
			} else {
				this.parentMappingManager.addLog("Non-linear modeling IS NOT required for target: " + this.targetName);
				this.ensModels.clear();
				this.ensModelTasks.clear();
				this.ensModels.add(this.fastModel);
				this.ensModelTasks.add(this.fastModelTask);
				this.ensModelQuality = this.fastModelQuality;
				this.state = THolderState.READY;

			}
		} else {
			// this.restorePreviousFastModel();
			this.restorePreviousEnsembleModel();

			if (this.ensModels.size() == 0) {
				this.state = THolderState.IDLE;
			} else {
				this.state = THolderState.READY;
			}

			this.parentMappingManager.addLog(
					"ERROR: The desired fast model could not be created! : handleEndOfModelTrainingExperiment()");
		}
		this.parentMappingManager.handleHolderStateChange(this);
	}

	synchronized void handleEndOfEnsTrainingExperiment(OpenCVMLPPredictor regressionModel, TMappingTask trainingTask) {
		this.retrievedEnsembleModels++;
		if ((regressionModel != null) && (trainingTask != null)) {
			this.validEnsembleModels++;
			this.ensModels.add(regressionModel);
			this.ensModelTasks.add(trainingTask);
		}
		/** after all model training threads have finished */
		if (this.retrievedEnsembleModels == this.candidateEnsModelTasks.size()) {

			if (this.validEnsembleModels > 0) {
				double sum = 0;
				int i = 1;
				for (TMappingTask tmt : this.ensModelTasks) {
					String performanceInd = "";
					if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
						performanceInd = tmt.getResults().get(TMappingTask.TResult.CV_ACC_MIN.getValue()).substring(0,
								6);
					} else {
						performanceInd = tmt.getResults().get(TMappingTask.TResult.CV_CORR_MIN.getValue()).substring(0,
								6);
					}
					double accEstimation = Double.parseDouble(performanceInd);
					this.parentMappingManager.addLog("Estimed accuracy of " + accEstimation + " for model nr." + i
							+ " target: " + this.targetName);
					sum += accEstimation;
				}
				sum /= this.validEnsembleModels;
				this.parentMappingManager
						.addLog("Estimed ensemble accuracy of " + sum + " for target: " + this.targetName);
				this.ensModelQuality = String.format("%.6f", sum);
				this.state = THolderState.READY;
			} else {
				// this.restorePreviousFastModel();
				this.restorePreviousEnsembleModel();

				if (this.ensModels.size() == 0) {
					this.state = THolderState.IDLE;
				} else {
					this.state = THolderState.READY;
				}

				this.parentMappingManager.addLog(
						"ERROR: The desired ensemble model could not be created! : handleEndOfEnsTrainingExperiment");
			}
			this.parentMappingManager.handleHolderStateChange(this);
		}
	}

	synchronized void handleEndOfModelTrainingExperiment(OpenCVMLPPredictor regressionModel,
			TMappingTask trainingTask) {

		if (!this.ensModelSelected) {
			this.handleEndOfFastTrainingExperiment(regressionModel, trainingTask);
		} else {
			this.handleEndOfEnsTrainingExperiment(regressionModel, trainingTask);
		}
	}

	public boolean hasSufficientSamples() {
		Individual[] validIndividuals = this.parentMappingManager.getProblem().getArchive().getMemory()
				.getValidIndividual(this.parentMappingManager.getProblem());
		return validIndividuals.length >= TMappingConstants.MIN_NR_OF_TRAINING_SAMPLES;
	}

	public boolean hasUnremovedJobs() {
		return this.hasUnremovedJobs.get();
	}

	public boolean isMappingCreated() {
		boolean result = true;
		if (this.ensModels.size() < 1) {
			result = false;
		}
		return result;
	}

	private void performEnsModelTrainingExperiment() {

		this.validEnsembleModels = 0;
		this.retrievedEnsembleModels = 0;
		this.clearEnsModelInformation();
		for (TMappingTask tmt : this.candidateEnsModelTasks) {
			/** clear existing data */
			this.completedEnsCVEvaluationTasks.clear();

			/** the unique ID of the experiment */
			String experimentName = UUID.randomUUID().toString();

			/**
			 * the temporary directory where all experiment related files are stored
			 */
			String experimentFolder = System.getenv("TEMP") + "/ModelTraining_" + experimentName;

			/** create the base .arff file of the experiment */
			this.createBaseDatasetFile(experimentFolder + "/" + TMappingConstants.BASE_DATASET_FILE, true);

			/**
			 * Create the master file that describes the model training experiment.
			 */
			this.createModelTrainingExperimentMasterFile(
					experimentFolder + "/" + TMappingConstants.EXPERIMENT_MASTER_FILE,
					experimentFolder + "/" + TMappingConstants.BASE_DATASET_FILE, tmt);

			this.parentMappingManager.addFolderToDelete(experimentFolder);

			/** Start the experiment via a MappingCreatorThread */
			this.stopMappingCreation.set(false);
			MappingCreationThread mct = new MappingCreationThread(experimentFolder, this.completedFastCVEvaluationTasks,
					this, this.stopMappingCreation, this.hasUnremovedJobs, this.maxTrainingDuration);
			mct.start();
		}
	}

	private void performFastModelTrainingExperiment() {
		/** clear existing data */
		this.completedFastCVEvaluationTasks.clear();

		/** the unique ID of the experiment */
		String experimentName = UUID.randomUUID().toString();

		/**
		 * the temporary directory where all experiment related files are stored
		 */
		String experimentFolder = System.getenv("TEMP") + "/ModelTraining_" + experimentName;

		/** create the base .arff file of the experiment */
		this.createBaseDatasetFile(experimentFolder + "/" + TMappingConstants.BASE_DATASET_FILE, false);

		/** Create the master file that describes the model training experiment. */
		this.createModelTrainingExperimentMasterFile(experimentFolder + "/" + TMappingConstants.EXPERIMENT_MASTER_FILE,
				experimentFolder + "/" + TMappingConstants.BASE_DATASET_FILE, this.fastModelTask);

		this.parentMappingManager.addFolderToDelete(experimentFolder);

		/** Start the experiment via a MappingCreatorThread */
		MappingCreationThread mct = new MappingCreationThread(experimentFolder, this.completedFastCVEvaluationTasks,
				this, this.stopMappingCreation, this.hasUnremovedJobs, this.maxTrainingDuration);
		mct.start();
	}

	public void performPostDeserializationActions(MappingManager parentManager) {

		if (this.fastModel != null) {
			this.fastModel.performPostDeserializationActions();
		}
		for (OpenCVMLPPredictor predictor : this.ensModels) {
			predictor.performPostDeserializationActions();
		}

		this.parentMappingManager = parentManager;
		this.getModelQuality();
		this.state = THolderState.READY;
		this.stopMappingCreation = new AtomicBoolean(false);
		this.hasUnremovedJobs = new AtomicBoolean(false);
		this.prevEnsModels = new ArrayList<OpenCVMLPPredictor>();
		this.prevEnsModelTasks = new ArrayList<TMappingTask>();
		this.candidateEnsModelTasks = new ArrayList<TMappingTask>();

		this.parentMappingManager
				.addLog(this.targetName + " " + this.ensModels.size() + " " + this.ensModelTasks.size());
	}

	private TMappingTask selectFastModelTaskUsingCVAdjustedCorrelation(String targetName) {

		TMappingTask.TResult qualityDiscriminant = TMappingTask.TResult.CV_CORR_MIN;
		if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
			qualityDiscriminant = TMappingTask.TResult.CV_ACC_MIN;
		}

		List<TMappingTask> allTasks = new ArrayList<TMappingTask>();
		allTasks.addAll(this.completedFastCVEvaluationTasks);

		/**
		 * By default, the best performing task is the one with the best value of the
		 * quality discriminant parameter
		 */
		TMappingTask chosenTask = null;
		double bestDiscriminantValue = -2;
		for (TMappingTask mappingTask : allTasks) {
			double currentDiscriminantValue = Double
					.parseDouble(mappingTask.getResults().get(qualityDiscriminant.getValue()));
			if (bestDiscriminantValue < currentDiscriminantValue) {
				bestDiscriminantValue = currentDiscriminantValue;
				chosenTask = new TMappingTask(mappingTask);
			}
		}

		return chosenTask;
	}

	private List<TMappingTask> selectEnsModelTasksUsingCVAdjustedCorrelation(String targetName) {

		List<TMappingTask> allTasks = new ArrayList<TMappingTask>();
		allTasks.addAll(this.completedEnsCVEvaluationTasks);

		List<TMappingTask> chosenTasks = new ArrayList<TMappingTask>();

		if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
			Collections.sort(allTasks, new MappingTaskAccMinComparator());
			for (int i = 0; i < allTasks.size(); i++) {
				this.parentMappingManager.addLog("Task " + i + " - quality:"
						+ allTasks.get(i).getResults().get(TMappingTask.TResult.CV_ACC_MIN.getValue()));
			}
		} else {
			Collections.sort(allTasks, new MappingTaskCorrMinComparator());
			for (int i = 0; i < allTasks.size(); i++) {
				this.parentMappingManager.addLog("Task " + i + " - quality:"
						+ allTasks.get(i).getResults().get(TMappingTask.TResult.CV_CORR_MIN.getValue()));
			}
		}

		chosenTasks.add(allTasks.get(0));
		for (int i = 1; i < Math.min(this.ensembleSize, allTasks.size() - 1); i++) {
			if (this.targetName.equals(TMappingConstants.ERRORS_CLASSIFIER)) {
				double bAcc = Double
						.parseDouble(allTasks.get(0).getResults().get(TMappingTask.TResult.CV_ACC_MIN.getValue()));
				double cAcc = Double
						.parseDouble(allTasks.get(i).getResults().get(TMappingTask.TResult.CV_ACC_MIN.getValue()));
				this.parentMappingManager.addLog("Ens. candidate " + i + ": " + cAcc + " "
						+ (TMappingConstants.ENSEMBLE_QUALITY_THRESHOLD * bAcc));
				if (cAcc >= (TMappingConstants.ENSEMBLE_QUALITY_THRESHOLD * bAcc)) {
					chosenTasks.add(allTasks.get(i));
				}
			} else {
				double bCorr = Double
						.parseDouble(allTasks.get(0).getResults().get(TMappingTask.TResult.CV_CORR_MIN.getValue()));
				double cCorr = Double
						.parseDouble(allTasks.get(i).getResults().get(TMappingTask.TResult.CV_CORR_MIN.getValue()));
				this.parentMappingManager.addLog("Ens. candidate " + i + ": " + cCorr + " "
						+ (TMappingConstants.ENSEMBLE_QUALITY_THRESHOLD * bCorr));
				if (cCorr >= (TMappingConstants.ENSEMBLE_QUALITY_THRESHOLD * bCorr)) {
					chosenTasks.add(allTasks.get(i));
				}
			}
		}

		return chosenTasks;
	}

	public void setParentMappingManager(MappingManager parentMappingManager) {
		this.parentMappingManager = parentMappingManager;
	}

	public void setState(THolderState state) {
		this.state = state;
	}

	/**
	 * Create a new ensemble mapping by performing a full grid-search cross
	 * validation on a reduced data set
	 */
	public void startEnsCrossValidationExperiment() {

		/** back-up existing models */
		// this.backupExistingFastModel();
		this.backupExistingEnsModel();

		/** clear existing data */
		this.clearFastModelInformation();
		this.clearEnsModelInformation();

		// /** set the state of the holder to BUSY */
		// this.state = THolderState.BUSY;
		// this.parentMappingManager.handleHolderStateChange(this);

		/** the unique ID of the experiment */
		String experimentName = UUID.randomUUID().toString();

		/**
		 * the temporary directory where all experiment related files are stored
		 */
		String experimentFolder = System.getenv("TEMP") + "/CrossValidation_" + experimentName;

		/**
		 * Create the base .arff file of the experiment. A reduced dataset is required
		 * since the non-linear model creation method is quite slow.
		 */
		this.createBaseDatasetFile(experimentFolder + "/" + TMappingConstants.BASE_DATASET_FILE, true);

		/**
		 * Create the master file that describes the cross validation experiment.
		 */

		this.createCrossValidationEvaluationExperimentMasterFile(
				experimentFolder + "/" + TMappingConstants.EXPERIMENT_MASTER_FILE,
				experimentFolder + "/" + TMappingConstants.BASE_DATASET_FILE, TMappingTask.TRegressionModelType.ANN);

		/** Delete the experiment folder at the end of the experiment */
		this.parentMappingManager.addFolderToDelete(experimentFolder);

		/** Start the experiment via a MappingCreatorThread */
		this.completedEnsCVEvaluationTasks.clear();
		this.stopMappingCreation.set(false);
		this.creationThread = new MappingCreationThread(experimentFolder, this.completedEnsCVEvaluationTasks, this,
				this.stopMappingCreation, this.hasUnremovedJobs, this.maxTrainingDuration);
		this.creationThread.start();
	}

	/**
	 * Create a new fast mapping by performing a very fast cross validation
	 * experiment on the full data set
	 */
	public void startFastCrossValidationExperiment(boolean fastTrainingEnabled) {

		/** mark the fact that no models are currently available */
		this.fastModelSelected = false;
		this.ensModelSelected = false;

		/**
		 * if ANN non-linear modeling is necessary, a 90 or 900 points grid search will
		 * be applied
		 */
		this.fastTrainingEnabled = fastTrainingEnabled;

		/** back-up existing models */
		// this.backupExistingFastModel();
		this.backupExistingEnsModel();

		/** clear existing data */
		this.clearFastModelInformation();
		this.clearEnsModelInformation();

		/** set the state of the holder to BUSY */
		this.state = THolderState.BUSY;
		this.parentMappingManager.handleHolderStateChange(this);

		/** the unique ID of the experiment */
		String experimentName = UUID.randomUUID().toString();

		/**
		 * the temporary directory where all experiment related files are stored
		 */
		String experimentFolder = System.getenv("TEMP") + "/CrossValidation_" + experimentName;

		/**
		 * Create the base .arff file of the experiment. A reduced dataset is not
		 * required since the model creation methods is assumed to be very fast.
		 */
		this.createBaseDatasetFile(experimentFolder + "/" + TMappingConstants.BASE_DATASET_FILE, false);

		/**
		 * Create the master file that describes the cross validation experiment.
		 */

		this.createCrossValidationEvaluationExperimentMasterFile(
				experimentFolder + "/" + TMappingConstants.EXPERIMENT_MASTER_FILE,
				experimentFolder + "/" + TMappingConstants.BASE_DATASET_FILE, TMappingTask.TRegressionModelType.LINEAR);

		/** Delete the experiment folder at the end of the experiment */
		this.parentMappingManager.addFolderToDelete(experimentFolder);

		/** Start the experiment via a MappingCreatorThread */
		this.completedFastCVEvaluationTasks.clear();
		this.stopMappingCreation.set(false);
		this.creationThread = new MappingCreationThread(experimentFolder, this.completedFastCVEvaluationTasks, this,
				this.stopMappingCreation, this.hasUnremovedJobs, this.maxTrainingDuration);
		this.creationThread.start();
	}

	/** Start a stand-alone retraining of the current mapping */
	public void startEnsRetrainingExperiment() {
		/** back-up existing model */
		// this.backupExistingFastModel();
		this.backupExistingEnsModel();

		/** set the state of the holder to BUSY */
		this.state = THolderState.BUSY;
		this.parentMappingManager.handleHolderStateChange(this);

		this.fastModelSelected = true;
		this.ensModelSelected = true;
		this.candidateEnsModelTasks.clear();
		for (TMappingTask tmt : this.ensModelTasks) {
			this.candidateEnsModelTasks.add(tmt);
		}
		this.ensModels.clear();
		this.ensModelTasks.clear();

		this.parentMappingManager.addLog("Retraining: " + this.candidateEnsModelTasks.size() + " models!");

		/** Perform the actual model training experiment */
		this.performEnsModelTrainingExperiment();
	}

	public void stopMappingCreationThread() {
		this.stopMappingCreation.set(true);
	}

	public int getReducedTrainingSetSize() {
		return this.reducedTrainingSetSize;
	}

	public void setReducedTrainingSetSize(int reducedTrainingSetSize) {
		this.reducedTrainingSetSize = reducedTrainingSetSize;
	}

	public int getMaxTrainingDuration() {
		return this.maxTrainingDuration;
	}

	public void setMaxTrainingDuration(int maxTrainingDuration) {
		this.maxTrainingDuration = maxTrainingDuration;
	}

}
