/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.mapping.utils.dstructures;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;

/**
 * Contains the definition of a opencv test
 * 
 * @author Ciprian Zavoianu
 */
public class TMappingTask implements Serializable {

	public enum TParameter {
		IND_ID("IND_ID"), EX_TYPE("experimentType"), MOD_TYPE("modelType"), HIDDEN_UNITS("hiddenUnits"),
		TRAINING_TIME("trainingTime"), LEARN_RATE("learnRate"), MOMENTUM("momentum"), COST("cost"), EPSILON("epsilon"),
		EXPONENT("exponent"), GAMMA("gamma"), RIDGE_PARAM("ridgeParam"), CLUSTERS("nrOfClusters"),
		ITERATIONS("nrOfIterations"), MIN_CLUSTER_STDEV("minStdevPerCluster"), TARGET("target"),
		TRAIN_START("trainStart"), TRAIN_END("trainEnd"), TEST_START("testStart"), TEST_END("testEnd");

		public static boolean contains(String value) {
			for (TParameter r : TParameter.values()) {
				if (r.getValue().toUpperCase().equals(value.toUpperCase())) {
					return true;
				}
			}
			return false;
		}

		private final String value;

		TParameter(String value) {
			this.value = value;
		}

		public TParameter getTParameter(String value) {
			for (TParameter r : TParameter.values()) {
				if (r.getValue().toUpperCase().equals(value.toUpperCase())) {
					return r;
				}
			}
			return null;
		}

		public String getValue() {
			return this.value;
		}
	};

	public enum TRegressionModelType {
		ANN("ANN"), SMO_POLY("SMO_POLY"), SMO_RBF("SMO_RBF"), LINEAR("LINEAR"), ISOTONIC("ISOTONIC"),
		RBF_NET("RBF_NET"), LOGISTIC("LOGISTIC");

		public static boolean contains(String value) {
			for (TRegressionModelType r : TRegressionModelType.values()) {
				if (r.getValue().toUpperCase().equals(value.toUpperCase())) {
					return true;
				}
			}
			return false;
		}

		private final String value;

		TRegressionModelType(String value) {
			this.value = value;
		}

		public String getValue() {
			return this.value;
		}
	}

	public enum TResult {
		TRAIN_CORR("trainingCorrelation"), TRAIN_RMSE("trainingRmse"), TRAIN_RRSE("trainingRrse"),
		TEST_CORR("testCorrelation"), TEST_RMSE("testRmse"), TEST_RRSE("testRrse"), CV_CORR_MEAN("cvCorrelationMean"),
		CV_RMSE_MEAN("cvRmseMean"), CV_RRSE_MEAN("cvRrseMean"), CV_CORR_STDEV("cvCorrelationStdev"),
		CV_RMSE_STDEV("cvRMSEStdev"), CV_RRSE_STDEV("cvRRSEStdev"), CV_CORR_MIN("cvCorrelationMin"),
		CV_RMSE_MIN("cvRMSEMin"), CV_RRSE_MIN("cvRRSEMin"),
		CORRECTLY_CLASSIFIED_INSTANCES("correctlyClassifiedInstances"), CV_ACC_MEAN("cvAccuracyMean"),
		CV_ACC_STDEV("cvAccuracyStdev"), CV_ACC_MIN("cvAccuracyMin");

		public static boolean contains(String value) {
			for (TResult r : TResult.values()) {
				if (r.getValue().toUpperCase().equals(value.toUpperCase())) {
					return true;
				}
			}
			return false;
		}

		private final String value;

		TResult(String value) {
			this.value = value;
		}

		public TResult getTResult(String value) {
			for (TResult r : TResult.values()) {
				if (r.getValue().toUpperCase().equals(value.toUpperCase())) {
					return r;
				}
			}
			return null;
		}

		public String getValue() {
			return this.value;
		}

	}

	public enum TRunInfo {
		EVAL_TIME("evaluationTime"), EVAL_SEC("evaluationSeconds"), TRAIN_SIZE("trainingSetSize");

		public static boolean contains(String value) {
			for (TRunInfo r : TRunInfo.values()) {
				if (r.getValue().toUpperCase().equals(value.toUpperCase())) {
					return true;
				}
			}
			return false;
		}

		private final String value;

		TRunInfo(String value) {
			this.value = value;
		}

		public TRunInfo getTRunInfo(String value) {
			for (TRunInfo r : TRunInfo.values()) {
				if (r.getValue().toUpperCase().equals(value.toUpperCase())) {
					return r;
				}
			}
			return null;
		}

		public String getValue() {
			return this.value;
		}
	}

	private static final long serialVersionUID = 2454686474380562405L;

	private final Map<String, String> parameters = new Hashtable<String, String>();
	private final Map<String, String> results = new Hashtable<String, String>();
	private final Map<String, String> runInfo = new Hashtable<String, String>();

	public TMappingTask() {

	}

	public TMappingTask(TMappingTask otherTest) {
		List<String> parameterKeys = new ArrayList<String>(otherTest.getParameters().keySet());
		List<String> runInfoKeys = new ArrayList<String>(otherTest.getRunInfo().keySet());
		List<String> resultKeys = new ArrayList<String>(otherTest.getResults().keySet());
		for (String paramKey : parameterKeys) {
			this.parameters.put(paramKey, otherTest.getParameters().get(paramKey));
		}
		for (String riKey : runInfoKeys) {
			this.runInfo.put(riKey, otherTest.getRunInfo().get(riKey));
		}
		for (String resKey : resultKeys) {
			this.results.put(resKey, otherTest.getResults().get(resKey));
		}
	}

	public Map<String, String> getParameters() {
		return this.parameters;
	}

	public Map<String, String> getResults() {
		return this.results;
	}

	public Map<String, String> getRunInfo() {
		return this.runInfo;
	}

	public boolean lessComplex(TMappingTask otherTest, boolean useRuntimeEvaluation) {
		boolean lessComplex = false;
		if ((this.getParameters().get(TParameter.MOD_TYPE.getValue()) == null)
				|| (!this.getParameters().get(TParameter.MOD_TYPE.getValue())
						.equals(otherTest.getParameters().get(TParameter.MOD_TYPE.getValue())))) {
			throw new RuntimeException("ERROR: Trying to compare models of different type!");
		}

		if (this.getParameters().get(TParameter.MOD_TYPE.getValue()).toUpperCase()
				.equals(TRegressionModelType.ANN.getValue())) {
			int thisH = (int) Double.parseDouble(this.getParameters().get(TParameter.HIDDEN_UNITS.getValue()));
			int otherH = (int) Double.parseDouble(otherTest.getParameters().get(TParameter.HIDDEN_UNITS.getValue()));
			double thisESec = Double.parseDouble(this.getRunInfo().get(TRunInfo.EVAL_SEC.getValue()));
			double otherESec = Double.parseDouble(otherTest.getRunInfo().get(TRunInfo.EVAL_SEC.getValue()));
			if (thisH < otherH) {
				lessComplex = true;
			} else {
				if ((thisH == otherH) && useRuntimeEvaluation) {
					if (thisESec < otherESec) {
						lessComplex = true;
					}
				}
			}
		}

		if (this.getParameters().get(TParameter.MOD_TYPE.getValue()).toUpperCase()
				.equals(TRegressionModelType.SMO_POLY.getValue())) {
			double thisC = Double.parseDouble(this.getParameters().get(TParameter.COST.getValue()));
			double otherC = Double.parseDouble(otherTest.getParameters().get(TParameter.COST.getValue()));
			double thisE = Double.parseDouble(this.getParameters().get(TParameter.EXPONENT.getValue()));
			double otherE = Double.parseDouble(otherTest.getParameters().get(TParameter.EXPONENT.getValue()));
			double thisESec = Double.parseDouble(this.getRunInfo().get(TRunInfo.EVAL_SEC.getValue()));
			double otherESec = Double.parseDouble(otherTest.getRunInfo().get(TRunInfo.EVAL_SEC.getValue()));
			if ((thisC <= otherC) && (thisE <= otherE)) {
				lessComplex = true;
			} else {
				if (((thisC <= otherC) || (thisE <= otherE)) && (useRuntimeEvaluation)) {
					if (thisESec < otherESec) {
						lessComplex = true;
					}
				}
			}
		}

		if (this.getParameters().get(TParameter.MOD_TYPE.getValue()).toUpperCase()
				.equals(TRegressionModelType.SMO_RBF.getValue())) {
			double thisC = Double.parseDouble(this.getParameters().get(TParameter.COST.getValue()));
			double otherC = Double.parseDouble(otherTest.getParameters().get(TParameter.COST.getValue()));
			double thisG = Double.parseDouble(this.getParameters().get(TParameter.GAMMA.getValue()));
			double otherG = Double.parseDouble(otherTest.getParameters().get(TParameter.GAMMA.getValue()));
			double thisESec = Double.parseDouble(this.getRunInfo().get(TRunInfo.EVAL_SEC.getValue()));
			double otherESec = Double.parseDouble(otherTest.getRunInfo().get(TRunInfo.EVAL_SEC.getValue()));
			if ((thisC <= otherC) && (thisG >= otherG)) {
				lessComplex = true;
			} else {
				if (((thisC <= otherC) || (thisG >= otherG)) && (useRuntimeEvaluation)) {
					if (thisESec < otherESec) {
						lessComplex = true;
					}
				}
			}
		}

		if (this.getParameters().get(TParameter.MOD_TYPE.getValue()).toUpperCase()
				.equals(TRegressionModelType.RBF_NET.getValue())) {
			double thisC = Double.parseDouble(this.getParameters().get(TParameter.CLUSTERS.getValue()));
			double otherC = Double.parseDouble(otherTest.getParameters().get(TParameter.CLUSTERS.getValue()));
			double thisMinStdev = Double.parseDouble(this.getParameters().get(TParameter.MIN_CLUSTER_STDEV.getValue()));
			double otherMinStdev = Double
					.parseDouble(otherTest.getParameters().get(TParameter.MIN_CLUSTER_STDEV.getValue()));
			double thisESec = Double.parseDouble(this.getRunInfo().get(TRunInfo.EVAL_SEC.getValue()));
			double otherESec = Double.parseDouble(otherTest.getRunInfo().get(TRunInfo.EVAL_SEC.getValue()));
			if ((thisC <= otherC) && (thisMinStdev >= otherMinStdev)) {
				lessComplex = true;
			} else {
				if (((thisC <= otherC) || (thisMinStdev >= otherMinStdev)) && (useRuntimeEvaluation)) {
					if (thisESec < otherESec) {
						lessComplex = true;
					}
				}
			}
		}

		return lessComplex;
	}

	@Override
	public String toString() {
		String line = "";
		for (String paramKey : new TreeSet<String>(this.parameters.keySet())) {
			line += paramKey + "=" + this.parameters.get(paramKey) + ", ";
		}
		for (String riKey : new TreeSet<String>(this.runInfo.keySet())) {
			line += riKey + "=" + this.runInfo.get(riKey) + ", ";
		}
		for (String resKey : new TreeSet<String>(this.results.keySet())) {
			line += resKey + "=" + this.results.get(resKey) + ", ";
		}
		return line;
	}
}
