/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.gui.setting;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.text.NumberFormat;

import javax.swing.DefaultCellEditor;
import javax.swing.GroupLayout;
import javax.swing.GroupLayout.Alignment;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.LayoutStyle.ComponentPlacement;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;
import javax.swing.border.LineBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.VariableDesc;

/**
 * Panel to change variables descriptions
 * 
 * @author Gnther Weidenholzer
 * @author Werner Koppelsttter
 */
public class ChooseVariables extends SettingsPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Number-formatter */
	private static NumberFormat formatter = NumberFormat.getInstance();

	/** Problem where variables are defined */
	private Problem problem = null;

	/** Table model for variables */
	private VariableTableModel model;

	/** List of available variable parameters */
	private Parameter[] vars;

	/** Refresh panel */
	private PanelRefresh refresh = null;

	/** Panel enabled */
	private boolean enable = true;

	/** Select based project-index */
	private JSpinner curIndex;

	/** Number of possible individuals */
	private JLabel labNumberIndividuals;

	/** Table to configure parameters */
	private JTable tableParameters;

	/**
	 * Definition of table-columns
	 * 
	 * @author Werner Koppelsttter
	 */
	private enum VarColumn {
		/** Name */
		NAME(0, "Name", String.class),
		/** Description */
		DESCRIPTION(1, "Description", String.class),
		/** Current value */
		VALUE(2, "Value", Double.class),
		/** Selected */
		SELECT(3, "Select", Boolean.class),
		/** Min limit */
		MIN(4, "Min", Double.class),
		/** Stepsize */
		STEPSIZE(5, "Stepsize", Double.class),
		/** Max limit */
		MAX(6, "Max", Double.class),
		/** Number of steps */
		STEPS(7, "Steps", Double.class);

		/** Column number */
		private final int column;

		/** Column name */
		private final String name;

		/** Type of column object */
		private final Class<?> type;

		/**
		 * Create enumeration
		 * 
		 * @param column
		 *            Column number
		 * @param name
		 *            Column name
		 * @param tpye
		 *            Column type
		 */
		private VarColumn(int column, String name, Class<?> tpye) {
			this.column = column;
			this.name = name;
			this.type = tpye;
		}

		/**
		 * Get column number
		 * 
		 * @return Number
		 */
		int getValue() {
			return this.column;
		}

		/**
		 * Get column name
		 * 
		 * @return Name
		 */
		String getName() {
			return this.name;
		}

		/**
		 * Get column type
		 * 
		 * @return Type
		 */
		Class<?> getType() {
			return this.type;
		}

		/**
		 * Get element of enumeration with to the index
		 * 
		 * @param index
		 *            Index of column
		 * @return Enumeration element
		 */
		static VarColumn getVarColumn(int index) {
			VarColumn var = null;

			for (VarColumn v : VarColumn.values()) {
				if (v.getValue() == index) {
					var = v;
				}
			}

			return var;
		}
	}

	/**
	 * Table model for choosing variables
	 * 
	 * @author Werner Koppelsttter
	 */
	private class VariableTableModel extends AbstractTableModel {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public String getColumnName(int column) {
			return VarColumn.getVarColumn(column).getName();
		}

		@Override
		public Class<?> getColumnClass(int column) {
			return VarColumn.getVarColumn(column).getType();
		}

		@Override
		public int getColumnCount() {
			return VarColumn.values().length;
		}

		@Override
		public int getRowCount() {
			return ChooseVariables.this.vars.length;
		}

		@Override
		public boolean isCellEditable(int row, int column) {
			boolean edit;

			switch (VarColumn.getVarColumn(column)) {
			case SELECT:
				edit = !ChooseVariables.this.problem.getArchive().isLocked();
				break;
			case MIN:
			case STEPSIZE:
			case MAX:
				edit = (ChooseVariables.this.getVarDesc(ChooseVariables.this.vars[row].getIdentifier()) != null);
				break;
			default:
				edit = false;
				break;
			}

			edit &= ChooseVariables.this.enable;

			return edit;
		}

		@Override
		public Object getValueAt(int row, int column) {
			Object obj = null;
			Parameter param;
			VariableDesc varDesc = null;

			param = ChooseVariables.this.vars[row];
			varDesc = ChooseVariables.this.getVarDesc(param.getIdentifier());

			switch (VarColumn.getVarColumn(column)) {
			case NAME:
				obj = param.getIdentifier();
				break;
			case DESCRIPTION:
				obj = param.getDescription();
				break;
			case VALUE:
				obj = ((TNumeric) param.getData()).getReal(0, 0);
				break;
			case SELECT:
				obj = (varDesc != null);
				break;
			case MIN:
				if ((varDesc != null) && !Double.isNaN(varDesc.getLB())) {
					obj = varDesc.getLB();
				}
				break;
			case STEPSIZE:
				if ((varDesc != null) && !Double.isNaN(varDesc.getStepSize())) {
					obj = varDesc.getStepSize();
				}
				break;
			case MAX:
				if ((varDesc != null) && !Double.isNaN(varDesc.getUB())) {
					obj = varDesc.getUB();
				}
				break;
			case STEPS:
				if (varDesc != null) {
					double val = varDesc.getNumSteps();
					if (!Double.isNaN(val)) {
						obj = val;
					}
				}
				break;
			default:
				obj = param;
				break;
			}

			return obj;
		}

		@Override
		public void setValueAt(Object aValue, int row, int column) {
			VariableDesc varDesc;
			String name;
			double value;

			name = ChooseVariables.this.vars[row].getIdentifier();
			varDesc = ChooseVariables.this.getVarDesc(name);

			switch (VarColumn.getVarColumn(column)) {
			case SELECT:
				if (!ChooseVariables.this.problem.getArchive().isLocked()) {
					if ((Boolean) aValue) {
						ChooseVariables.this.problem.addVariable(name);
					} else {
						ChooseVariables.this.problem.rmVariable(name);
					}
				}
				break;
			case MIN:
				if (varDesc != null) {
					if (aValue == null) {
						value = Double.NaN;
					} else {
						value = (Double) aValue;
					}
					varDesc.setBounds(value, varDesc.getStepSize(), varDesc.getUB());
				}
				break;
			case STEPSIZE:
				if (varDesc != null) {
					if (aValue == null) {
						value = Double.NaN;
					} else {
						value = (Double) aValue;
					}
					varDesc.setBounds(varDesc.getLB(), value, varDesc.getUB());
				}
				break;
			case MAX:
				if (varDesc != null) {
					if (aValue == null) {
						value = Double.NaN;
					} else {
						value = (Double) aValue;
					}
					varDesc.setBounds(varDesc.getLB(), varDesc.getStepSize(), value);
				}
				break;
			default:
				break;
			}

			this.fireTableDataChanged();
			ChooseVariables.this.refreshIndividuals();

			// Select cell for checkbox editor
			if (VarColumn.getVarColumn(column) == VarColumn.SELECT) {
				ChooseVariables.this.tableParameters.setRowSelectionInterval(row, row);
				ChooseVariables.this.tableParameters.setColumnSelectionInterval(column, column);
			}
		}
	}

	/**
	 * Validate problem
	 * 
	 * @param prob
	 *            Problem
	 * 
	 * @return Message if not valid
	 */
	public static String validate(Problem prob) {
		String message = null;

		if (prob.getNumVariedVariables() < 1) {
			message = "Select at least one varied parameter!";
		} else {
			for (VariableDesc var : prob.getVariables()) {
				if (!var.isValid()) {
					message = "Please complete settings";
				}
			}
		}

		return message;
	}

	/**
	 * Create panel. Used by WindowsBuilder
	 */
	private ChooseVariables() {
		this.model = new VariableTableModel();

		this.initComponents();
	}

	/**
	 * Creates new form choose-variables
	 * 
	 * @param problem
	 *            Problem to set variables
	 * @param refresh
	 *            Refresh panel
	 */
	public ChooseVariables(Problem problem, PanelRefresh refresh) {
		this(problem);
		this.refresh = refresh;
	}

	/**
	 * Creates new form choose-variables
	 * 
	 * @param problem
	 *            Problem to set variables
	 */
	public ChooseVariables(Problem problem) {
		this();
		if (problem != null) {
			this.problem = problem;
		}

		Project project = problem.getArchive().getProject();
		if (project != null) {
			this.vars = project.getDynamicParams();
			project.setIndex(problem.getArchive().getBasedIndex());
		} else {
			this.vars = new Parameter[0];
		}
		this.curIndex.setValue(problem.getArchive().getBasedIndex() + 1);

		this.tableParameters.setDefaultRenderer(Double.class, new DefaultTableCellRenderer() {
			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			{
				this.setHorizontalAlignment(SwingConstants.RIGHT);
			}

			@Override
			protected void setValue(Object value) {
				String str = "";
				Double val;

				if (value != null) {
					val = (Double) value;
					if (val.isInfinite()) {
						str = ChooseVariables.formatter.format(val);
					} else {
						str = val.toString();
					}
				}
				this.setText(str);
			}

			@Override
			public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
					boolean hasFocus, int row, int column) {
				VariableDesc var;

				super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);

				this.setToolTipText(this.getText());

				if (isSelected) {
					table.getSelectionBackground();
				} else {
					switch (VarColumn.getVarColumn(column)) {
					case MIN:
					case STEPSIZE:
					case MAX:
						var = ChooseVariables.this.getVarDesc(ChooseVariables.this.vars[row].getIdentifier());
						if ((var != null) && !var.isValid()) {
							this.setBackground(Color.RED);
						} else {
							this.setBackground(table.getBackground());
						}
						break;
					default:
						this.setBackground(table.getBackground());
						break;
					}
				}

				return this;
			}
		});
		this.tableParameters.setDefaultRenderer(String.class, new DefaultTableCellRenderer() {
			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
					boolean hasFocus, int row, int column) {

				super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
				this.setToolTipText(this.getText());

				return this;
			}
		});

		JTextField tf = new JTextField();
		tf.setHorizontalAlignment(SwingConstants.RIGHT);
		this.tableParameters.setDefaultEditor(Double.class, new DefaultCellEditor(tf) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			/** Value of cell */
			Double value;

			@Override
			public boolean stopCellEditing() {
				String s = (String) super.getCellEditorValue();

				if ((s != null) && !s.isEmpty()) {
					try {
						this.value = Double.valueOf(s);
					} catch (Exception e) {
						((JComponent) this.getComponent()).setBorder(new LineBorder(Color.red));
						return false;
					}
				} else {
					this.value = null;
				}

				return super.stopCellEditing();
			}

			@Override
			public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row,
					int column) {
				JTextField tf;
				this.value = null;
				((JComponent) this.getComponent()).setBorder(new LineBorder(Color.black));
				tf = (JTextField) super.getTableCellEditorComponent(table, value, isSelected, row, column);
				tf.selectAll();

				return tf;
			}

			@Override
			public Object getCellEditorValue() {
				return this.value;
			}

		});
		this.tableParameters.getColumnModel().getColumn(0).setMinWidth(250);
		this.tableParameters.getColumnModel().getColumn(1).setMinWidth(150);
		this.tableParameters.getModel().addTableModelListener(new TableModelListener() {

			@Override
			public void tableChanged(TableModelEvent e) {
				ChooseVariables.this.refresh();
			}
		});
		this.refreshIndividuals();

		this.refreshEnabled();
	}

	/**
	 * Refresh panel
	 */
	protected void refresh() {
		if (this.refresh != null) {
			this.refresh.refresh();
		}
	}

	@Override
	public String validateContents() {
		return ChooseVariables.validate(this.getProblem());
	}

	/**
	 * Are settings enabled
	 * 
	 * @return Setting enable
	 */
	public boolean isSettingEnabled() {
		return this.enable;
	}

	/**
	 * Set settings enabled
	 * 
	 * @param enable
	 *            Settings enable
	 */
	public void setSettingEnabled(boolean enable) {
		this.enable = enable;
		this.refreshEnabled();
	}

	/**
	 * Refresh enable state of components
	 */
	private void refreshEnabled() {
		this.curIndex.setEnabled(this.enable && !this.problem.getArchive().isLocked());
		this.tableParameters.setEnabled(this.enable && !this.problem.isLocked());
	}

	/**
	 * Get variable-description to the name
	 * 
	 * @param name
	 *            Name of variable
	 * @return Variable-description
	 */
	private VariableDesc getVarDesc(String name) {
		VariableDesc varDesc = null;

		for (VariableDesc var : ChooseVariables.this.problem.getVariables()) {
			if (var.getName().compareTo(name) == 0) {
				varDesc = var;
			}
		}

		return varDesc;
	}

	/**
	 * Refresh label with number of defined individuals
	 */
	private void refreshIndividuals() {
		double num = 1.0;

		if (this.validateContents() == null) {
			for (VariableDesc var : this.problem.getVariables()) {
				num *= var.getNumSteps();
			}
			if (!Double.isInfinite(num) && (num >= 1000000.0)) {
				this.labNumberIndividuals.setText(String.format("%g", num));
			} else {
				this.labNumberIndividuals.setText(ChooseVariables.formatter.format(num));
			}
		} else {
			this.labNumberIndividuals.setText("");
		}
	}

	/**
	 * Get problem of panel
	 * 
	 * @return Problem
	 */
	public Problem getProblem() {
		return this.problem;
	}

	/**
	 * Initialize components
	 */
	private void initComponents() {
		JScrollPane jScrollParameters;
		JLabel labCurIndex, labIndividuals;

		// current index
		labCurIndex = new JLabel("Initial values from project index:");
		labCurIndex.setFont(new Font("Tahoma", 1, 12));

		this.curIndex = new JSpinner();
		this.curIndex.setFont(new Font("Tahoma", 1, 12));
		this.curIndex.setValue(1);
		this.curIndex.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent evt) {
				int val;
				Project project;

				val = (Integer) ChooseVariables.this.curIndex.getValue();
				project = ChooseVariables.this.problem.getArchive().getProject();
				if ((project == null) || (val <= 0)) {
					val = 1;
				} else if (val > project.getLength()) {
					val = project.getLength();
				}

				ChooseVariables.this.curIndex.setValue(val);
				project.setIndex(val - 1);
				ChooseVariables.this.problem.getArchive().setBasedIndex(val - 1);
				ChooseVariables.this.model.fireTableDataChanged();
			}
		});

		// number individuals
		labIndividuals = new JLabel("Possible Individuals:");
		labIndividuals.setFont(new Font("Tahoma", 1, 12));

		this.labNumberIndividuals = new JLabel("0");
		this.labNumberIndividuals.setFont(new Font("Tahoma", 1, 12));

		// table
		this.tableParameters = new JTable(this.model);
		this.tableParameters.setColumnSelectionAllowed(true);
		this.tableParameters.getTableHeader().setReorderingAllowed(false);
		jScrollParameters = new JScrollPane();
		jScrollParameters.setViewportView(this.tableParameters);
		this.tableParameters.getColumnModel().getSelectionModel().setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

		// layout
		GroupLayout layout = new GroupLayout(this);
		layout.setHorizontalGroup(layout.createSequentialGroup().addContainerGap().addGroup(layout
				.createParallelGroup(Alignment.LEADING)
				.addGroup(layout.createSequentialGroup().addComponent(labCurIndex).addGap(18)
						.addComponent(this.curIndex, GroupLayout.PREFERRED_SIZE, 56, GroupLayout.PREFERRED_SIZE)
						.addPreferredGap(ComponentPlacement.RELATED, 28, Short.MAX_VALUE).addComponent(labIndividuals)
						.addPreferredGap(ComponentPlacement.RELATED).addComponent(this.labNumberIndividuals,
								GroupLayout.PREFERRED_SIZE, 100, GroupLayout.PREFERRED_SIZE))
				.addComponent(jScrollParameters, Alignment.LEADING, GroupLayout.DEFAULT_SIZE, 580, Short.MAX_VALUE))
				.addContainerGap());
		layout.setVerticalGroup(layout.createSequentialGroup().addContainerGap()
				.addGroup(layout.createParallelGroup(Alignment.LEADING).addComponent(labCurIndex)
						.addComponent(this.curIndex, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE,
								GroupLayout.PREFERRED_SIZE)
						.addComponent(labIndividuals).addComponent(this.labNumberIndividuals))
				.addPreferredGap(ComponentPlacement.RELATED)
				.addComponent(jScrollParameters, GroupLayout.DEFAULT_SIZE, 250, Short.MAX_VALUE).addContainerGap());
		this.setLayout(layout);
	}
}
