'''
Created on 05.12.2016

@author: Werner Koppelstaetter
'''

import os
import sys
import re
import traceback

try:
    # noinspection PyUnresolvedReferences
    from java.lang import StackTraceElement
except Exception as e:
    print(e)

import importlib


def eval_fnc_method(structure, *args):
    val = None
    errMsg = None
    errStack = None
    cp = None
    eggs = []
    try:
        (cp, eggs) = _path_load(structure)

        # import class
        importlib.reload(importlib.import_module(structure.getModuleName()))
        exec('from ' + structure.getModuleName() + ' import ' + structure.getClassName())

        # load fnc
        fnc = eval(structure.getClassName() + '(structure)')

        # eval method
        if args[0] == 'refreshDependency':
            if len(fnc.getImports()) == 0 and len(fnc.getExports()) == 0:
                fnc.setImportVars(fnc.importVars())
                fnc.setExportVars(fnc.exportVars())

        method = eval(structure.getClassName() + '.' + args[0])
        if len(args) == 1:
            val = method(fnc)
        else:
            val = method(fnc, *(args[1:]))

    except:
        errStack = []
        exc_type, exc_value, exc_traceback = sys.exc_info()
        errMsg = str(exc_value)
        for trace in traceback.extract_tb(exc_traceback):
            errStack.insert(0, StackTraceElement('', trace.name, trace.filename, trace.lineno))

        _path_unload(cp, eggs)

    return errMsg, errStack, val


def _path_load(structure):
    # set dynamic system-path
    cp = structure.getClassPath()

    eggs = _load_latest_eggs(cp)

    if cp in sys.path:
        cp = None
    else:
        sys.path.append(cp)

    return cp, eggs


def _path_unload(cp, eggs):
    # restore system-path
    if cp is not None:
        sys.path.remove(cp)
    for egg in eggs:
        sys.path.remove(egg)


def _load_latest_eggs(cp: str) -> list:
    """
    Search for egg files in the classpath folder and add them to the python path.
    If multiple egg files of the same package are found, only the latest version will be added to the python path.

    Parameters
    ----------
    cp : str
     Classpath

    Returns
    -------
    eggs : list
        List of egg file paths, which have been added to the python path
    """
    eggs = []
    py_version = '{}.{}'.format(sys.version_info.major, sys.version_info.minor)

    if os.path.exists(cp):
        '''make sure the path exists'''
        pkgs = dict()
        for file in os.listdir(cp):
            '''look for egg files in the classpath 
            (do not use the glob functions for this e.g. 
            eggs = glob.glob(os.path.join(cp, '*.egg'), recursive=False)) 
            '''
            if file.endswith('.egg'):
                '''extract package name and versions of the egg'''
                match_obj = re.search(r"^(?P<eggName>\w+)-(?P<eggVersion>[\d\.]+)-py(?P<pyVersion>[\d\.]+).egg$", file)
                if match_obj is None:
                    raise Exception('Parsing filename of "{}" resulted in an error'.format(file))
                pkg_name = match_obj.group(1)
                pkg_version = match_obj.group(2)
                egg_pyc_version = match_obj.group(3)

                if egg_pyc_version != py_version:
                    '''compare if the egg file is compatible with the current python version
                       
                       if the egg contains bytecode (*.pyc) generated by another python version,
                       it might fail to load the packages inside
                    '''
                    msg = 'Package with incompatible python version detected: Python {} != {} ({})'
                    print(msg.format(py_version, egg_pyc_version, os.path.join(cp, file)))
                    raise ImportError(msg.format(py_version, egg_pyc_version, file))

                if pkg_name not in pkgs:
                    pkgs[pkg_name] = dict(version=pkg_version, file=file, unique=True)
                else:
                    pkgs[pkg_name]['unique'] = False
                    if parse_version(pkgs[pkg_name]['version']) < parse_version(pkg_version):
                        pkgs[pkg_name]['version'] = pkg_version
                        pkgs[pkg_name]['file'] = file

        for key, value in pkgs.items():
            egg = os.path.join(cp, value['file'])
            folder = os.path.join(cp, key)
            if os.path.isdir(folder):
                '''check if the source package folder is in the same location as the egg file'''
                print('Warning: Folder and egg file of the same package "{}" detected in {}!\n'.format(key, cp))
            if egg not in sys.path:
                '''add to sys path'''
                sys.path.append(egg)
                eggs.append(egg)
                if not value['unique']:
                    print('Warning: Multiple egg files of package "{}" detected.'
                          ' The latest package version "{}" will be loaded!\n'.format(key, value['version']))

    return eggs


def parse_version(v):
    """Simple version parser

    Alternative to:
    - from packaging.version import parse as parse_version  # recommended
    - from distutils.version import LooseVersion as parse_version # older

    """
    return tuple(map(int, (v.split("."))))


def getVariables(loc):
    import inspect

    variables = []
    for key in loc:
        if not key.startswith('_'):
            if not inspect.isfunction(loc[key]) and not inspect.ismethod(loc[key]) and not inspect.ismodule(
                    loc[key]) and not inspect.isbuiltin(loc[key]):
                variables.append(key)

    return variables


def redirect_output(output_printstream, output_encoding, error_printstream, error_encoding):
    sys.stdout = _RedirectOutputstream(output_printstream, output_encoding)
    sys.stderr = _RedirectErrorstream(error_printstream, error_encoding)


class _RedirectOutputstream():

    def __init__(self, printstream, encoding):
        self.printmethod = getattr(printstream, 'print')
        self.flushmethod = getattr(printstream, 'flush')
        self.encoding = encoding

    def write(self, msg):
        self.printmethod(msg)

    def flush(self):
        self.flushmethod()


class _RedirectErrorstream():

    def __init__(self, printstream, encoding):
        self.printmethod = getattr(printstream, 'print')
        self.flushmethod = getattr(printstream, 'flush')
        self.encoding = encoding

    def write(self, msg):
        self.printmethod(msg)

    def flush(self):
        self.flushmethod()
