#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 9643 $
$LastChangedDate:: 2020-02-14 15:31:06 +0100#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import warnings
from symspace.wrapper.symwrapper import SymProject, SymContainer
from symspace.generic_gui_qt import GUIMessageThread


class ScriptUtils:
    """Helper class to run scripts in the symspace console"""

    @classmethod
    def runWithParams(cls, fnc_handle, params, needs_project: bool = True) -> None:
        """
        Run a function in the symspace console with  standardized error handling.
        The function takes project and parameter handles.

        Parameters
        ----------
        fnc_handle : fnc
        params : dict
            parameter dictionary
        needs_project: bool
            Flag to indicate if the a loaded project is needed for execution

        Examples
        --------

        >>> from symspace.console.script_utils import ScriptUtils
        >>> def main(project, parameter): print(project, parameter)
        >>> ScriptUtils.runWithParams(main, locals())

        """
        try:
            if '__project__' in params:
                '''run macro'''
                project = params['__project__']  # type: SymProject
                parameter = params['__parameter__']  # type: SymContainer
                fnc_handle(project, parameter)
            elif 'CurProject' in params and 'CurObject' in params:
                '''manual assignment'''
                project = params['CurProject']  # type: SymProject
                parameter = params['CurObject']  # type: SymContainer
                fnc_handle(project, parameter)
            else:
                if needs_project:
                    cls.showNoProjectWarning()
                else:
                    fnc_handle(None, None)

        except Exception as ex:
            print(ex)
            cls.showError(str(ex))

    @classmethod
    def runWithoutParams(cls, fnc_handle) -> None:
        """
        Run a function in the symspace console with  standardized error handling.

        Examples
        --------

        >>> from symspace.console.script_utils import ScriptUtils
        >>> def main(): print('Do sth')
        >>> ScriptUtils.runWithoutParams(main)

        """
        try:
            fnc_handle()

        except Exception as ex:
            print(ex)
            cls.showError(str(ex))

    @staticmethod
    def showNoProjectWarning():
        """Show the standard warning in console and gui"""
        warnings.warn('Run this script via "Run macro (Ctrl+M)"!')
        msg_elements = {'icon': 'warning', 'text': 'No SyMSpace project found!\n\nOpen a SyMSpace project and '
                                                   'run this script via Run macro (Ctrl+M)!', 'buttons': 'ok'}
        msg = GUIMessageThread('Warning', msg_elements)
        msg.start()
        print('Script finished with error')

    @staticmethod
    def showError(text: str):
        """
        Show a specific warning in console and gui

        Parameters
        ----------
        text : str
            error message
        """
        warnings.warn(text)
        msg_elements = {'icon': 'critical', 'text': text, 'buttons': 'ok'}
        msg = GUIMessageThread('Error', msg_elements)
        msg.start()
        print('Script finished with error')
