#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision$
$LastChangedDate::                           $

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import stat
import shutil
import json
import fileinput
from symspace.osutils.pathutils import cs_home


def writeJSON(prj: 'SymProject', fname_comp: str):
    """
    Write JSON File

    Parameters
    ----------
    prj : SymProject
    fname_comp : string
        File name of the Component.

    """
    inc_lst = ['Info', 'Files', 'Docu', 'Filter']
    prj_dict = prj.toDict()
    json_dict = {}

    for item in inc_lst:
        if item in prj_dict:
            json_dict[item] = prj_dict[item]

    jsonstr = json.dumps(json_dict)
    fname = fname_comp.replace('.mopc', '.json')
    fh = open(fname, "w+")
    fh.write(jsonstr)
    fh.close()


def moveCompfile(fname: str, userGroup: str):
    """
    Move component file to correct directory

    Parameters
    ----------
    fname : string
        File name
    userGroup: string
        UserGroup directory

    """

    basename = os.path.basename(fname)

    if (cs_home is not None) and fname.startswith(compspace):
        path = os.path.join(cs_home, '_UserGroups', userGroup, 'Components')
        if os.path.exists(path):
            dst_file = os.path.join(path, basename)
            print('Component directory: ' + dst_file)
        else:
            raise FileNotFoundError('Path does not exist' + path)
    else:
        pos = fname.find('Components')
        if pos > 0:
            dst_file = fname[0:pos] + 'Components\\' + basename
        else:
            dst_file = fname

    if os.path.exists(dst_file) and (fname != dst_file):
        os.remove(dst_file)
    shutil.move(fname, dst_file)
    json_src = fname.replace('.mopc', '.json')
    json_dst = dst_file.replace('.mopc', '.json')
    if os.path.exists(json_dst) and (json_src != json_dst):
        os.remove(json_dst)
    shutil.move(json_src, json_dst)
    return dst_file


def readInterface(params: list) -> list:
    """
    Read parameters of the Component function

    Parameters
    ----------
    params : list
        List of SymDependItem
        SyMSpace import or export

    Returns
    -------
    par_lst : list
        List of parameters
    """
    par_lst = []
    for param in params:
        item = {}
        # if not (param.getParamName() == ''):
        #     ''' get parameters from the fields linked to the function interface'''
        #     item['name'] = param.getParamName()
        #     if param.getHandle() is None:
        #         item['description'] = param.javaObj.getVarDesc().getDescription()
        #         item['info'] = 'external'
        #         item['comment'] = ''
        #     else:
        #         item['description'] = param.getHandle().getDescription()
        #         item['info'] = ''
        #         if param.getHandle().getComment().replace("'''", "").strip().find('#') == 0:
        #             item['comment'] = param.getHandle().getComment().replace("'''", "").replace("#", "").strip()
        # else:
        #     item['name'] = param.getVarName()
        #     item['description'] = param.javaObj.getVarDesc().getDescription()
        #     item['info'] = 'optional'
        #     item['comment'] = ''

        ''' get parameters from the function interface'''
        ports_desc = param.javaObj.getVarDesc()
        name = ports_desc.getName()
        description = ports_desc.getDescription()
        unit = ports_desc.getUnit()
        optional = ports_desc.getOptional()
        dimension = ports_desc.getDimension()
        if param.getHandle() is None:
            info = 'external'
        else:
            info = ''

        item['name'] = name
        item['description'] = description
        item['unit'] = unit
        item['optional'] = optional
        item['dimension'] = dimension
        # ToDo: do we need info/comment?
        item['info'] = info
        item['comment'] = ''

        par_lst.append(item)
    return par_lst


def copytree(src: str, dst: str, symlinks: bool = False, ignore=None):
    """
    Copies directory with files recursively

    Parameters
    ----------
    src : str
        Source path
    dst : str
        Destination path
    symlinks : bool
    ignore

    """

    if not os.path.exists(dst):
        os.makedirs(dst)
    for item in os.listdir(src):
        # Skip Subversion directories
        if item == '.svn':
            continue
        # Skip __pycache__ directory
        if item == '__pycache__':
            continue
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            copytree(s, d, symlinks, ignore)
        else:
            if not os.path.exists(d) or os.stat(s).st_mtime - os.stat(d).st_mtime > 1:
                shutil.copy2(s, d)
                os.chmod(d, stat.S_IWRITE)
                # print(d, os.access(d, os.W_OK))


def copyFile(src: str, dst: str) -> str:
    """
    Copies files and removes readonly

    Parameters
    ----------
    src : str
        Source path
    dst : str
        Destination path

    Returns
    -------
    dst_file : str
        Destination file path
    """

    if not os.path.exists(dst):
        os.makedirs(dst)

    dst_file = shutil.copy2(src, dst)
    os.chmod(dst_file, stat.S_IWRITE)
    return dst_file


def copyFileAndReplace(infile: str, outfile: str, replace: list):
    """
    Copy text file and replace content

    Parameters
    ----------
    infile : string
    outfile : string
    replace : list of tuples
    """
    py_src = ''
    with open(infile, 'r') as file:
        for line in file:
            py_src = py_src + line
    for r in replace:
        py_src = py_src.replace(r[0], r[1])

    with open(outfile, 'w+') as file:
        file.write(py_src)


def replaceInFile(file: str, old: str, new: str):
    """Replace text in file

    Parameters
    ----------
    file : str
        File name
    old : str
        Text to be replaced
    new : str
        Text replaced
    """
    with fileinput.FileInput(file, inplace=True) as f:
        for line in f:
            print(line.replace(old, new), end='')
