#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 9898 $
$LastChangedDate:: 2020-06-30 14:21:09 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import sys
import subprocess
import symspace.osutils.pathutils as pathutils


def scriptRun(executable: str, filename: str, timeout: float = 60, hide_console: bool = True,
              stdout=subprocess.PIPE, stderr=subprocess.PIPE, 
              cmd_args: list = None, cmd_kwargs: dict = None) -> subprocess.CompletedProcess:
    """
    Run a Python Script in the defined executable

    Parameters
    ----------
    executable :str
        Path to the python interpreter executable
    filename : str
        Name of the python-file to execute
    timeout : float
        Process timeout in seconds
    hide_console : bool
        If True the subprocess console is during the build process is hidden / does not pop up, Default: True
    stdout :
        Output stream, subprocess.PIPE (Default) or None
    stderr :
        Error stream, subprocess.PIPE (Default) or None
    cmd_args: list
        list of strings, with additional commandline arguments
    cmd_kwargs: dict
        dict with additional commandline arguments, key and value have to be of type string!

    Returns
    -------
    process : subprocess.CompletedProcess
    """
    if not os.path.exists(filename):
        raise FileNotFoundError('File not found: {}!'.format(filename))

    si = subprocess.STARTUPINFO()
    if hide_console:
        si.dwFlags |= subprocess.STARTF_USESHOWWINDOW

    run_args = [executable, filename]

    if cmd_args is not None:
        run_args.extend(cmd_args)
    if cmd_kwargs is not None:
        for key, value in cmd_kwargs.items():
            run_args.append(key)
            run_args.append(value)
    try:
        process = subprocess.run(run_args, timeout=timeout, startupinfo=si, check=True,
                                 stdout=stdout, stderr=stderr)
    except subprocess.CalledProcessError as ex:
        if ex.stderr is None:
            raise Exception('Process faild with exit status {}'.format(ex.returncode))
        else:
            raise Exception(ex.stderr.decode('utf-8'))
    except subprocess.TimeoutExpired as ex:
        raise TimeoutError('Timeout after {} s'.format(ex.timeout))

    # if process.returncode != 0:
    #     raise Exception(process.stderr)
    # else:
    #     return process
    return process


def csymspaceRun(filename: str, timeout: float = 60, hide_console: bool = True,
                 stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                 cmd_args: list = None, cmd_kwargs: dict = None) -> subprocess.CompletedProcess:
    """
    Run a python script in the SyMSpace Console

    Parameters
    ----------
    filename : str
        Name of the python-file to execute
    timeout : float
        Process timeout in seconds
    hide_console : bool
        If True the subprocess console is during the build process is hidden / does not pop up, Default: True
    stdout :
        Output stream, subprocess.PIPE (Default) or None
    stderr :
        Error stream, subprocess.PIPE (Default) or None
    cmd_args: list
        list of strings, with additional commandline arguments
    cmd_kwargs: dict
        dict with additional commandline arguments, key and value have to be of type string!
        
    Returns
    -------
    process : subprocess.CompletedProcess
    """
    process = scriptRun(pathutils.csymspace_exe, filename,
                        timeout=timeout, hide_console=hide_console,
                        stdout=stdout, stderr=stderr, cmd_args=cmd_args, cmd_kwargs=cmd_kwargs)
    return process


def pythonRun(filename: str, timeout: float = 60, hide_console: bool = True,
              stdout=subprocess.PIPE, stderr=subprocess.PIPE, 
              cmd_args: list = None, cmd_kwargs: dict = None) -> subprocess.CompletedProcess:
    """
    Run a python script in the python.exe

    Parameters
    ----------
    filename :  str
        Name of the pythonfile to execute
    timeout : float
        Process timeout in seconds
    hide_console : bool
        If True the subprocess console is during the build process is hidden / does not pop up, Default: True
    stdout :
        Output stream, subprocess.PIPE (Default) or None
    stderr :
        Error stream, subprocess.PIPE (Default) or None
    cmd_args: list
        list of strings, with additional commandline arguments
    cmd_kwargs: dict
        dict with additional commandline arguments, key and value have to be of type string!
        
    Returns
    -------
    process : subprocess.CompletedProcess
    """
    process = scriptRun(pathutils.python_exe, filename,
                        timeout=timeout, hide_console=hide_console,
                        stdout=stdout, stderr=stderr, cmd_args=cmd_args, cmd_kwargs=cmd_kwargs)
    return process


def getSVNRevision(dirpath):
    # ToDo: check if subprocess.run should replace subprocess.Popen
    svn_cmd = 'svn info "{}" --show-item last-changed-revision'.format(dirpath)
    svn_co = subprocess.Popen(svn_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                              universal_newlines=True)
    svn_revision, error = svn_co.communicate()
    svn_revision = svn_revision.strip()
    return svn_revision, error


def getInterpreter():
    """
    Get the name of the interpreter

    e.g.

    if the executable is python.exe return python
    if the executable is java.exe return java
    if the executable is CSyMSpace.exe return CSyMSpace

    Returns
    -------
    interpreter name : str
    """
    interpPath = os.path.normpath(sys.executable)
    return os.path.splitext(os.path.basename(interpPath))[0]


def needsJepInterpreter():
    """
    Raise an error if the interpreter name is python

    """
    if getInterpreter() == 'python':
        msg = 'This must be run in a jep interpreter / SyMSpace console!'
        print(msg)
        raise Exception(msg)


def needsJepDecorator(func):
    """
    Decorator which raises an error if the interpreter name is python

    Parameters
    ----------
    func : function

    Returns
    -------
    function_wrapper : function
    """

    def function_wrapper(*args, **kwargs):
        needsJepInterpreter()
        return func(*args, **kwargs)

    return function_wrapper
