# -*- coding: utf-8 -*-
"""
Created on Fr Juli 21 09:00:02 2017

@author: mkitzberger

Helper classes that allow for fast formatted print inside MagOpt jep/python-console
by shortening the standard representation (repr-method) of numpy arrays and
provide formatted indentation for lists and dictionarys.

This class is required because Jep uses the repr string for printing
inside the console which can lead to blocks due to the long strings returned
by large numpy arrays.

"""
import numpy as np
import copy


class SymList(list):
    """Pretty print version of python list"""

    def __repr__(self):
        """override list __repr__() magic method"""
        nind = 5
        string = '[\n'
        i = 0
        for item in super().__iter__():
            i += 1
            if i != super().__len__() - 1:
                string = string + nind * ' ' + str(item) + ',\n'
            else:
                string = string + nind * ' ' + str(item) + '\n'
        string = string + ']'
        return string

    def __str__(self):
        return 'list with length {}'.format(super().__len__())


class SymDict(dict):
    """Pretty print version of python dict"""

    def __repr__(self):
        """override dict __repr__() magic method"""
        nind = 5
        keys = super().keys()
        string = '{\n'
        for i, key in enumerate(keys):
            if i != len(keys) - 1:
                string = string + nind * ' ' + "'{}' : {}".format(key, str(super().__getitem__(key))) + ',\n'
            else:
                string = string + nind * ' ' + "'{}' : {}".format(key, str(super().__getitem__(key))) + '\n'
        string = string + '}'
        return string

    def __str__(self):
        return 'dict with length {}'.format(len(super().keys()))


class SymNdArray(np.ndarray):
    """Pretty print version of numpy array
       increases speed when printing in MagOpt Jep python console"""

    def __new__(cls, input_array):
        obj = np.asarray(input_array).view(cls)
        cls.obj = obj
        return obj

    def __repr__(self):
        """override numpy np.ndarray __str__() magic method"""
        return "np.ndarray with np.ndarray.shape = {} and np.ndarray.dtype = '{}'".format(super().shape,
                                                                                          str(super().dtype))

    def __str__(self):
        return "np.ndarray with np.ndarray.shape = {} and np.ndarray.dtype = '{}'".format(super().shape,
                                                                                          str(super().dtype))


class SymPrettyPrintConverter():
    """Static conversion class to recursively replace dict, list and np.ndarray
     by their inherited pretty print versions PyDict, PyList, PyNdArray which overwrite
     the __repr__ method of their respective superclass.
     This allows a nice fast print in MagOpt Jep console"""

    @staticmethod
    def pretty2ugly(tmp):
        raise NotImplementedError('Method not Implemented')

    @staticmethod
    def ugly2pretty(tmp):
        if isinstance(tmp, list) and not isinstance(tmp, SymList):
            for i, item in enumerate(tmp):
                tmp[i] = SymPrettyPrintConverter.ugly2pretty(tmp[i])
            return SymList(tmp)

        elif isinstance(tmp, dict) and not isinstance(tmp, SymDict):
            for i, key in enumerate(tmp.keys()):
                tmp[key] = SymPrettyPrintConverter.ugly2pretty(tmp[key])
            return SymDict(tmp)

        elif isinstance(tmp, np.ndarray) and not isinstance(tmp, SymNdArray):
            return SymNdArray(tmp)

        else:
            return tmp
