# -*- coding: utf-8 -*-
"""
Created on Thu May 04 16:39:02 2017

@author: ssilber, mkitzberger

"""

import base64
import traceback
import sys
import os
import numbers
import json
import pickle
import shutil
from io import IOBase
from abc import ABCMeta, abstractmethod
import numpy as np
from symspace.wrapper.prettyprint import SymPrettyPrintConverter
from lcm.utils.pytperiodic import PyTPeriodicMatrix
from symspace.osutils.pathutils import symspace_home
from symspace.fileutils import copyFileAndReplace

try:
    # Start of java imports
    # noinspection PyUnresolvedReferences
    import java
    # noinspection PyUnresolvedReferences
    from jep import jarray
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import Parameter as JParameter
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import Project as JProject
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamField as JParamField
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamComboBox as JParamComboBox
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamContainer as JParamContainer
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import PortDesc as JPortDesc
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamLinkedField as JParamLinkedField
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamDirectory as JParamDirectory
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamFile as JParamFile
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamMaterial as JParamMaterial
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamCollector as JParamCollector
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamFunction as JParamFunction
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamJavaFunction as JParamJavaFunction
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamPythonFunction as JParamPythonFunction
    # noinspection PyUnresolvedReferences
    # from at.lcm.bu21.magopt.structure import Component as JComponent
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import DependItem as JDependItem
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure.simulate import SimulateAll as JSimulateAll
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.general.utils import FileUtils as JFileUtils
    # noinspection PyUnresolvedReferences
    from magopt.utils import ComponentInfo
    # noinspection PyUnresolvedReferences
    from java.lang import StackTraceElement as JStackTraceElement
    # noinspection PyUnresolvedReferences
    from java.lang import String as JString
    # noinspection PyUnresolvedReferences
    from java.lang import Integer as JInteger
    # noinspection PyUnresolvedReferences
    from java.lang import Long as JLong
    # noinspection PyUnresolvedReferences
    from java.lang.reflect import Array as JArray
except Exception as e:
    print('Unable to load Java Libraries!\n{}'.format(e))

try:
    '''java-matlab dependent imports'''
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamFormula as JParamFormula
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import ParamMatlabFunction as JParamMatlabFunction
except Exception as e:
    print('Unable to load Java Libraries for Matlab!\n{}'.format(e))

# java wrapper/conversion imports
from lcm.tdata_interface.pyjdata_interface import PyJDataInterface
from symspace.wrapper.symplot import SymPlot2Line, SymPlot2Bar

"""
Additional methods not implemented in Java:
    SymParameter.getParameters: 'SymComponent' included
    SymContainer.addComponent: 
    SymContainer.isComponent: 
"""


class SymFileUtils:
    """
    Additional File utilities
    """

    @staticmethod
    def checkSum(file):
        if not isinstance(file, IOBase):
            # Function is not a python file handle therefore java functions can be used.
            if isinstance(file, str):
                normpath = os.path.normpath(file)
                return JFileUtils.checkSum(normpath)
            else:
                return JFileUtils.checkSum(file)
        else:
            # Function is a python file handle -> convert to java file handle
            filename = os.path.realpath(file)
            return JFileUtils.checkSum(filename)


class SymSimulateAll:
    """
    Class for simulating entire projects
    """

    def __init__(self, project):
        if not isinstance(project, SymProject):
            raise ValueError('project must be a SymProject instance')

        self.__jSimulateAllInstance = JSimulateAll(project.javaObj)
        # set defaults
        self.setOptimize(False)
        self.setRemoteCompute(False)
        self.setRunSimulation(False)

    def start(self):
        """
        Start to regenerate all indices (similar to regenerate all button in GUI).
        """
        self.__jSimulateAllInstance.start()

    def stop(self):
        """
        Stop regenerate.
        """
        self.__jSimulateAllInstance.stop()

    def getNumError(self):
        """
        Get current number of errors.
        """
        return self.__jSimulateAllInstance.getNumError()

    def getNumFinished(self):
        """
        Get number of currently finished indices.
        """
        return self.__jSimulateAllInstance.getNumFinished()

    def getNumIncomplete(self):
        """
        Get number of current incomplete indices.
        """
        return self.__jSimulateAllInstance.getNumIncomplete()

    def getNumNotInitialized(self):
        """
        Get number of not initialized indices.
        """
        return self.__jSimulateAllInstance.getNumNotInitialized()

    def getOptimize(self):
        """
        Get optimize flag value.
        """
        status = self.__jSimulateAllInstance.getOptimize()
        if status == 0:
            return False
        else:
            return True

    def getRemoteCompute(self):
        """
        Get currently set value of remote compute flag.
        """
        status = self.__jSimulateAllInstance.getRemoteCompute()
        if status == 0:
            return False
        else:
            return True

    def getRunSimulation(self):
        """
        Get currently set value of run simulation flag.
        """
        status = self.__jSimulateAllInstance.getRunSimulation()
        if status == 0:
            return False
        else:
            return True

    def isStart(self):
        """
        Check if start was called.
        """
        status = self.__jSimulateAllInstance.isStart()
        if status == 0:
            return False
        else:
            return True

    def isStop(self):
        """
        Check if not running.
        """
        status = self.__jSimulateAllInstance.isStop()
        if status == 0:
            return False
        else:
            return True

    def setOptimize(self, optimize):
        """
        Set the optimize flag.
        """
        if not isinstance(optimize, bool):
            raise ValueError('Optimize flag must be a boolean instance')
        self.__jSimulateAllInstance.setOptimize(optimize)

    def setRemoteCompute(self, remoteCompute):
        """
        Set the remote compute flag.
        """
        if not isinstance(remoteCompute, bool):
            raise ValueError('Remote Compute flag must be boolean instance')
        self.__jSimulateAllInstance.setRemoteCompute(remoteCompute)

    def setRunSimulation(self, runSimulation):
        """
        Set the run simulation flag.
        """
        if not isinstance(runSimulation, bool):
            raise ValueError('Run simulation flag must be boolean instance')
        self.__jSimulateAllInstance.setRunSimulation(runSimulation)

    def setTimeout(self, timeout):
        """
        Set the simulation timeout.
        """
        if not isinstance(timeout, numbers.Number):
            raise ValueError('Timeout needs to be a number instance')
        else:
            timeout = int(timeout)

        self.__jSimulateAllInstance.setTimeout(timeout)


class SymComponentInfo:
    """
    Create ComponentInfo for SyMSpace-file
    """

    def __init__(self, javaObj):
        self.javaObj = javaObj

    @staticmethod
    def listComponents():
        """
        List info about all component in the library

        Returns
        -------
        ret : list
            List of SymComponentInfo
        """
        ret = []
        for comp in ComponentInfo.listComponents():
            ret.append(SymComponentInfo(comp))
        return ret

    @staticmethod
    def getComponents():
        """
        Dict of components in library

        Returns
        -------
        ret : dict
            Dictionary of all components with file
        """
        ret = {}
        for comp in SymComponentInfo.listComponents():
            ret[comp.getName()] = comp.getCompFile()
        return ret

    def getCompFile(self):
        """
        Get absolute file of Component

        Returns
        -------
        ret : str
        """
        return self.javaObj.getMop().getAbsolutePath()

    def getName(self):
        """
        Get name of component

        Returns
        -------
        ret : str
        """
        return self.javaObj.getName()

    def getPackageName(self):
        """
        Get name of package

        Returns
        -------
        ret : str
        """
        return self.javaObj.getPackageName()

    def getProperties(self):
        """
        Get list of properties

        Returns
        -------
        ret : dict
            Dict of component properties
        """
        return PyJDataInterface.convert_J2Py(self.javaObj.getProperties())

    def getProperty(self, name):
        """
        Get property-value for name

        Parameters
        ----------
        name : str
            Name of property

        Returns
        -------
        ret : str
        """
        return self.javaObj.getProperty(name)


class SymParameter(metaclass=ABCMeta):
    """
    Python wrapper base class
    For some basic usage examples of the wrapper refer to files in:
    System/Python/src/symspace/wrapper/scripting_examples

    All given examples for SymParameter
    are based on the following basis project setups:

    >>> CurProject1 = SymProject('test','test','test')
    >>> param1 = SymField('param1')
    >>> param2 = SymField('param2')
    >>> param3 = SymField('param3')
    >>> param4 = SymField('param4')
    >>> param5 = SymField('param5')
    >>> param6 = SymField('param6')
    >>> CurProject1.addParam(param1)
    >>> CurProject1.addParam(param2)
    >>> CurProject1.addParam(param3)
    >>> CurProject1.addParam(param4)
    >>> CurProject1.addParam(param5)
    >>> CurProject1.addParam(param6)

    >>> CurProject2 = SymProject('test','test','test')
    >>> formula = SymFormula('test', 'test')
    >>> CurProject2.addParam(param1)
    >>> CurProject2.addParam(param2)
    >>> CurProject2.addParam(param3)
    >>> CurProject2.addParam(formula)

    """

    @staticmethod
    def convertToPython(obj):
        """
        Convert javaObject to python object
        >>> # noinspection PyUnresolvedReferences
        >>> from at.lcm.bu21.magopt.structure import Project as JProject
        >>> # create a java project
        >>> javaProj = JProject('test','test','test')
        >>> # convert to python version SymProject
        >>> pythonProject = SymParameter.convertToPython(javaProj)

        Parameters
        ----------
        obj : JavaParameterObject
            Parameter object to convert

        Returns
        -------
        ret : SymParameter
            Matching Python Parameter object

        """
        return SymParameter._convertObject(obj)

    @staticmethod
    def _convertObject(obj):
        # @Todo: change to isinstance(???)
        if str(type(obj)) == "<class 'jep.PyJArray'>":
            ret = [SymParameter._convertObject(o) for o in obj]
        elif obj is None:
            return None
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.Project':
            ret = SymProject(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamMaterial':
            ret = SymMaterial(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamCollector':
            ret = SymCollector(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamMatlabFunction':
            ret = SymMatlabFunction(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamPythonFunction':
            ret = SymPythonFunction(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamJavaFunction':
            ret = SymJavaFunction(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamContainer':
            ret = SymContainer(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamComboBox':
            ret = SymComboBox(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamFormula':
            ret = SymFormula(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamLinkedField':
            ret = SymLinkedField(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamField':
            ret = SymField(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamFile':
            ret = SymFile(javaObj=obj)
        elif obj.java_name == 'at.lcm.bu21.magopt.structure.ParamDirectory':
            ret = SymDirectory(javaObj=obj)
        else:
            ret = None

        return SymPrettyPrintConverter.ugly2pretty(ret)

    def __init__(self, javaObj):
        self.javaObj = javaObj

    def __str__(self):
        return 'SymParameter'

    def __repr__(self):
        return 'SymParameter'

    def __eq__(self, other):
        """
        For equality check equals() method of the
        underlying java object is called.
        """
        return self.javaObj.equals(other.javaObj)

    @property
    def javaObj(self):
        """
        Property attribute that returns the underlying java object

        Returns
        -------
        ret : JavaParameterInstance
        """
        return self._javaObj

    @javaObj.setter
    def javaObj(self, obj):
        """
        Property setter for the wrappers java object

        Parameters
        ----------
        obj : JavaParameterInstance

        """
        self._javaObj = obj

    # methods
    def listTabNames(self):
        """
        List name of tabs

        Returns
        -------
        tabNames : list
            List of tab names

        """
        tabs = str(self.javaObj.listTabNames())
        return tabs

    def toString(self):
        """
        To string for interface

        Returns
        -------
        val : string
            String representation of parameter
        """
        str_val = self.javaObj.toString()
        print(str_val)

    def refresh(self):
        """
        Refresh structure and flags
        """
        self.javaObj.refresh()

    def refreshFlags(self):
        """
        Refresh flags
        """
        self.javaObj.refreshFlags()

    def removeParam(self):
        """
        Remove this parameter from parent parameter

        Remove a parameter from project
        >>> CurProject = locals()['CurProject']
        >>> CurProject.param1.removeParam()
        """
        self.javaObj = self.javaObj.removeParam()

    def moveUp(self):
        """
        Move this parameter up in list by one position
        """
        self.javaObj.moveUp()

    def moveDown(self):
        """
        Move this parameter down in list by one position
        """
        self.javaObj.moveDown()

    def cloneParam(self, *args):
        """
        Clone parameter

        Parameters
        ----------
        *args : tuple

            clone : SymParameter, optional
                Existing object to clone to
            index : int, optional
                Index to clone (-1 for all datasets)
            docFiles : bool, optional
                Clone doc-files

        Returns
        -------
        clone : SymParameter
            Clone of SymParameter

        """
        nargs = len(args)
        if nargs == 0:
            py_obj = SymParameter._convertObject(self.javaObj.cloneParam())
            return py_obj

        elif nargs == 1:
            obj = args[0]
            py_obj = SymParameter._convertObject(self.javaObj.cloneParam(obj.javaObj, -1))
            return py_obj

        elif nargs == 2:
            obj = args[0]
            index = args[1]
            py_obj = SymParameter._convertObject(self.javaObj.cloneParam(obj.javaObj, index))
            return py_obj

        elif nargs == 3:
            obj = args[0]
            index = args[1]
            docFiles = args[2]
            py_obj = SymParameter._convertObject(self.javaObj.cloneParam(obj.javaObj, index, docFiles))
            return py_obj

        else:
            raise TypeError('cloneParam(): max. three positional arguments')

    def info(self, strInfo):
        """
        Display infos
        """
        try:
            raise Exception('Test')

        except Exception as e:
            # get stack traceback
            exc_type, exc_value, exc_traceback = sys.exc_info()
            trace_list = traceback.extract_tb(exc_traceback, limit=2)
            stack = JArray.newInstance(JStackTraceElement, len(trace_list))
            for i, item in enumerate(trace_list):
                (filename, lineno, locals, name) = item
                stack[i] = JStackTraceElement('', name, filename, lineno)

        self.javaObj.info(strInfo, stack)

    def warning(self, strWarning):
        """
        Display warnings
        """
        try:
            raise Exception('Test')

        except Exception as e:
            # get stack traceback
            exc_type, exc_value, exc_traceback = sys.exc_info()
            trace_list = traceback.extract_tb(exc_traceback, limit=2)
            stack = JArray.newInstance(JStackTraceElement, len(trace_list))
            for i, item in enumerate(trace_list):
                (filename, lineno, _, name) = item
                stack[i] = JStackTraceElement('', name, filename, lineno)

        self.javaObj.warning(strWarning, stack)

    def error(self, varError, addMessage=''):
        """
        Display error
        """

        if isinstance(varError, Exception):
            # process a given Exception
            strError = str(varError)
            exc_traceback = varError.__traceback__
            trace_list = traceback.extract_tb(exc_traceback, limit=2)
            stack = JArray.newInstance(JStackTraceElement, len(trace_list))
            for i, item in enumerate(trace_list):
                (filename, lineno, _, name) = item
                stack[i] = JStackTraceElement('', name, filename, lineno)

        elif isinstance(varError, str):
            # produce Exception
            try:
                raise Exception('Test')

            except Exception:
                # get stack traceback
                strError = varError
                exc_type, exc_value, exc_traceback = sys.exc_info()
                trace_list = traceback.extract_tb(exc_traceback, limit=2)
                stack = JArray.newInstance(JStackTraceElement, len(trace_list))
                for i, item in enumerate(trace_list):
                    (filename, lineno, _, name) = item
                    stack[i] = JStackTraceElement('', name, filename, lineno)

        else:
            raise ValueError('error(): parameter varError must be a Exception object or string')

        strError = addMessage + strError
        self.javaObj.error(self.getIdentifier(), strError, stack)

    # getter methods
    def getName(self):
        """
        Get name of parameter

        Returns
        -------
        name : string
            Name of parameter
        """
        return self.javaObj.getName()

    def getDynamicOutputs(self):
        """
        Get list of dynamic output parameters.
        Dynamic output parameter are parameter which can diversify.
        ParamField which are used, not constant, not writable and has one numeric value.

        Returns
        -------
        ret : list
            List of parameter
        """
        return self._convertObject(self.javaObj.getDynamicOutputs())

    def getDynamicParams(self):
        """
        Get list of dynamic parameters.
        Dynamic parameter are parameter which can diversify.
        ParamField which are used, not constant, writable and has one numeric value.

        Returns
        -------
        ret : list
            List of parameter
        """
        return self._convertObject(self.javaObj.getDynamicParams())

    def getAnalysisParams(self):
        """
        Get list of analysis parameter.
        Analysis parameters are ParamField which are used.

        Returns
        -------
        ret : list
            List of parameter

        """
        return self._convertObject(self.javaObj.getAnalysisParams())

    def getObject(self, identString):
        """
        Get first object which identifier correlate with given identifier.

        Parameters
        ----------
        identString : string
            Identifier

        Returns
        -------
        ret : SymParameter
            Found parameter

        """
        pyobj = self._convertObject(self.javaObj.getObject(identString))
        # if not pyobj: print('None')
        return pyobj

    def getUniqueObject(self, uniqueIdentString):
        """
        Get object which unique identifier
        returns a SymParameter Object

        Parameters
        ----------
        uniqueIdentString : string

        Returns
        -------
        ret : SymParameter
            Found parameter
        """
        pyobj = self._convertObject(self.javaObj.getUniqueObject(uniqueIdentString))
        if not pyobj:
            print('None')
        return pyobj

    def getObjectPath(self, identString):
        """Get first object which identifier correlate with given identifier"""
        # TODO: implement return type Parameter.ParameterPath
        return self._convertObject(self.javaObj.getObjectPath(identString))

    def getObjectPaths(self, identString):
        """Get list of objects which identifier correlate with given identifier"""
        # TODO: implement return type Parameter.ParameterPath
        return self._convertObject(self.javaObj.getObjectPaths(identString))

    def getParamByIdentifier(self, identString):
        """Return a list of Parameter which identifier correlate with given identifier"""
        # TODO: implement return type Parameter.ParameterPath
        return self._convertObject(self.javaObj.getParamByIdentifier(identString))

    def getParamByPid(self, pid):
        """Get parameter by parameter-identifier"""
        return self._convertObject(self.javaObj.getParamByPid(pid))

    def getParamByUniqueIdentifier(self, uIdName):
        """Return Parameter that identifier match with given identifier"""
        # TODO: implement return type Parameter.ParameterPath
        return self._convertObject(self.javaObj.getParamByUniqueIdentifier(uIdName))

    def getParameters(self, val=None):
        """
        Get list of sub- parameters that are instances of class provided
        in val.

        If val is a SymParameter instance all parameters
        corresponding to the type of val are returned as a list.

        If val is a string which equals to 'SymParameter', 'SymProject',
        'SymMaterial', 'SymCollector', 'SymFunction', 'SymMatlabFunction',
        'SymPythonFunction', 'SymJavaFunction', 'SymContainer',
        'SymComboBox', 'SymFormula', 'SymLinkedField', 'SymField',
        'SymFile', 'SymDirectory', 'SymComponent' all sub- parameters of the respective
        type are returned as a list.

        Parameters
        ----------
        val : SymParameter or string

        Returns
        -------
        found : list
            List of found parameters

        Examples
        --------
        Get a list of all parameters of type SymField in a project
        >>> CurProject = locals()['CurProject']
        >>> CurProject.getParameters('SymField')
        >>>
        >>> CurProject.getParameters('SymContainer')

        Get a list of all parameters in a project which are of same type as a given parameter

        >>> param1 = SymField('default', 'default')
        >>> CurProject.getParameters(param1)


        """
        if val is None:
            # Todo: Is this extension of the java implementation necessary?
            if isinstance(self, SymContainer):
                return self.getParam()
            else:
                raise TypeError('Instance need to be a SymContainer (sub) class, got {} instead'.format(type(self)))

        if isinstance(val, str):
            if val == 'SymParameter':
                tmp = JParameter
            elif val == 'SymProject':
                tmp = JProject
            elif val == 'SymMaterial':
                tmp = JParamMaterial
            elif val == 'SymCollector':
                tmp = JParamCollector
            elif val == 'SymFunction':
                tmp = JParamFunction
            elif val == 'SymMatlabFunction':
                tmp = JParamMatlabFunction
            elif val == 'SymPythonFunction':
                tmp = JParamPythonFunction
            elif val == 'SymJavaFunction':
                tmp = JParamJavaFunction
            elif val == 'SymContainer':
                tmp = JParamContainer
            elif val == 'SymComboBox':
                tmp = JParamComboBox
            elif val == 'SymFormula':
                tmp = JParamFormula
            elif val == 'SymLinkedField':
                tmp = JParamLinkedField
            elif val == 'SymField':
                tmp = JParamField
            elif val == 'SymFile':
                tmp = JParamFile
            elif val == 'SymDirectory':
                tmp = JParamDirectory
            elif val == 'SymComponent':
                tmp = JParamContainer
            else:
                raise ValueError('Unknown python parameter')
            if val == 'SymComponent':
                ret = []
                for par in self._convertObject(self.javaObj.getParameters(tmp)):
                    if par.isComponent():
                        ret.append(par)
                return ret
            else:
                return self._convertObject(self.javaObj.getParameters(tmp))
        else:
            return self._convertObject(self.javaObj.getParameters(val.javaObj))

    def getParent(self):
        """
        Get parent-parameter
        The parent parameter is also a container because
        only containers can have child parameters.
        For parameters in a container this method gives the container.

        Returns
        -------
        ret : SymContainer
           Parent parameter
        """
        return self._convertObject(self.javaObj.getParent())

    def getPId(self):
        """
        Get ID of parameter

        Returns
        -------
        ret : string
            Parameter ID
        """
        return self.javaObj.getPId()

    def getDSId(self):
        """
        Get ID of current dataset ( = current index)

        Returns
        -------
        ret : string
            Dataset ID
        """
        return self.javaObj.getDsId()

    def getDescription(self):
        """
        Get decription of parameter

        Returns
        -------
        ret : string
            Description of parameter
        """
        return self.javaObj.getDescription()

    def getModified(self):
        """
        Get modified-status

        Returns
        -------
        status : bool
            Get modified-status

        """
        return self.javaObj.getModified() == 1

    def getUncertain(self, index=None):
        """
        Get uncertain-status

        Parameters
        ----------
        index : int, optional

        Returns
        -------
        status : bool
            Get uncertain status
        """
        if index is None:
            return self.javaObj.getUncertain() == 1
        else:
            return self.javaObj.getUncertain(index) == 1

    def getPosition(self):
        """
        Get position of parameter

        Returns
        -------
        pos : int
            Position starts with 0
        """
        lst = [p.getName() for p in self.getParent().getParam()]
        return lst.index(self.getName())

    def getPrint(self):
        """
        Get print-flag

        Returns
        -------
        status : bool
            Get print flag
        """
        return self.javaObj.getPrint() == 1

    def getReadOnly(self):
        """
        Get read-only-status

        Returns
        -------
        status : bool
            Get read-only status


        """
        return self.javaObj.getReadOnly() == 1

    def getUsed(self):
        """
        Get used-status

        Returns
        -------
        status : bool
            Get used status
        """
        return self.javaObj.getUsed() == 1

    def getUsedBy(self):
        """
        Get list of parameter which use this parameter

        Returns
        -------
        usedBy : list
        """
        return self._convertObject(self.javaObj.getUsedBy())

    def getSetBy(self):
        """
        Get list of parameter which export to this parameter

        Returns
        -------
        setBy : list
        """
        return self._convertObject(self.javaObj.getSetBy())

    def getUsedInputParameter(self):
        """
        Get list of used input parameter

        Returns
        -------
        used : list
        """
        return self._convertObject(self.javaObj.getUsedInputParameter())

    def getError(self):
        """
        Get error-status of current dataset

        Returns
        -------
        error : bool
        """
        return self.javaObj.getError() == 1

    def getIndex(self):
        """
        Get current dataset index

        Returns
        -------
        index : int
            Index of current dataset (zero-based)
        """
        return self.javaObj.getIndex()

    def getConstant(self):
        """
        Get constant-flag

        Returns
        -------
        flag : bool
            Constant-flag

        """
        return self.javaObj.getConstant() == 1

    def getRoot(self):
        """
        Get root-parameter

        Todo: Should typehints/docstrings be changed SymParameter -> SymProject?
              Java implementation: public Parameter getRoot()
              But can the root-parameter be different type than SymProject?

        Returns
        -------
        root : SymParameter
            Root parameter

        """
        return self._convertObject(self.javaObj.getRoot())

    def getIdentifier(self):
        """
        Get identifier string of this parameter

        Returns
        -------
        identifier : string
        """
        return self.javaObj.getIdentifier()

    def getLength(self):
        """
        Get size of datasets (number of indexes)

        Returns
        -------
        size : int
            Size of dataset

        """
        return self.javaObj.getLength()

    def getData(self, index=None):
        """
        Get data of SymParameter of defined dataset index.

        Parameters
        ----------
        index : int (zero-based)
            defines dataset index. If not defined, current index is used

        Returns
        -------
        data : PyTData
            Python version of TData
        """
        if index is not None:
            tmp = PyJDataInterface.convert_J2Py(self.javaObj.getData(index))
        else:
            tmp = PyJDataInterface.convert_J2Py(self.javaObj.getData())

        return SymPrettyPrintConverter.ugly2pretty(tmp)

    def getJavaData(self):
        """
        Get Java Data

        Returns
        -------
        data : TData
            Java version of TData

        """
        tmp = self.javaObj.getData()
        return tmp

    def getDataString(self):
        """
        Get String representation of java data

        Returns
        -------
        val : string
            String representation of java data
        """
        return self.javaObj.getData().toString()

    def getDocFile(self, name):
        """
        Get doc-files with the given name

        Parameters
        ----------
        name : str
            Name of Doc file

        Returns
        -------
        ret : dict
            Doc file dictionary
        """
        doc = self.javaObj.getDocFile(name)
        docfile = {'name': doc.getName(),
                   'description': doc.getDescription(),
                   'file': os.path.join(self.getRoot().getFilesDirectory(), doc.getFile())}
        return docfile

    def getDocFiles(self):
        """
        Get doc-files

        Returns
        -------
        ret : list
            List of doc files
        """
        ret = []
        for doc in PyJDataInterface.convert_J2Py(self.javaObj.getDocFiles()):
            ret.append({'name': doc.getName(),
                        'description': doc.getDescription(),
                        'file': os.path.join(self.getRoot().getFilesDirectory(), doc.getFile())})
        return ret

    def getComment(self):
        """
        Get comment of parameter

        Returns
        -------
        comment : string
        """
        return self.javaObj.getComment()

    def getJava(self):
        """
        Get Java Object of Parameter

        Returns
        -------
        obj : Java parameter
        """
        return self.javaObj

    # setter methods
    def setName(self, val):
        """
        Set name of parameter and change in Java Dependency and Java Interface

        Parameters
        ----------
        val : string
            Name to set
        """
        self.javaObj.setName(val)
        self.refresh()

    def setDescription(self, val):
        """
        Set description of parameter

        Parameters
        ----------
        val : string
            Description to set
        """
        self.javaObj.setDescription(val)

    def setPrint(self, val):
        """
        Set print-flag

        Parameters
        ----------
        val : bool
            Print-flag
        """
        self.javaObj.setPrint(val)

    def setData(self, val):
        """
        Set data of parameter

        Parameters
        ----------
        val : PyTData
            Python data to set
        """
        if self.getReadOnly():
            raise ValueError('Cannot write to read only parameters')
        else:
            tmp = PyJDataInterface.convert_Py2J(val)
            self.javaObj.setData(tmp)

    def setDataString(self, val):
        """
        Set string of data

        Parameters
        ----------
        val : string

        """
        self.javaObj.setDataString(val)

    def setModified(self, val):
        """
        Set modified-status

        Parameters
        ----------
        val : bool

        """
        self.javaObj.setModified(val)

    def setConstant(self, val):
        """
        Set constant flag of parameter

        Parameters
        ----------
        val : bool

        """
        self.javaObj.setConstant(val)

    def setComment(self, val):
        """
        Set comment of parameter

        Parameters
        ----------
        val : str

        """
        self.javaObj.setComment(val)

    def setError(self, val):
        """
        Set error flag

        Parameters
        ----------
        val : bool
            Errror-flag
        """
        self.javaObj.setError(val)

    def toDict(self, detailed=False):
        """
        Convert data to Python dictionary

        Returns
        -------
        val : dict
        """

        if not self.getUncertain():
            data = self.getData()
            if detailed and (
                    isinstance(data, np.ndarray) or isinstance(data, dict) or isinstance(data, PyTPeriodicMatrix)):
                pkl = pickle.dumps(data)
                ret = base64.b64encode(pkl).decode('ascii')
            elif isinstance(data, str):
                ret = data
            else:
                ret = self.getDataString()

        else:
            # ret = float('nan')
            # This is not supported by JSON
            ret = 'NaN'
        return ret

    def _getInfo(self):
        ret = dict(identifier=self.getIdentifier(),
                   name=self.getName(),
                   description=self.getDescription(),
                   comment=self.getComment(),
                   constant=self.getConstant(),
                   used=self.getUsed(),
                   modified=self.getModified(),
                   uncertain=self.getUncertain(),
                   error=self.getError(),
                   readonly=self.getReadOnly())
        return ret


class SymContainer(SymParameter):
    """
    Python wrapper class for |sym|'s java class ParamContainer
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamContainer(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamContainer(name, description)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, key):
        """
        Indexing of parameter
        """
        if not isinstance(key, str):
            raise ValueError('key has to be of type string')
        if key == ':':
            return self.getParam()
        else:
            childs = self.getParameters()
            childNames = [x.getName() for x in childs]
            for i, val in enumerate(childNames):
                if val == key:
                    return (childs[i])

            return self.getObject(key)

    # methods
    def addParam(self, pyParam, pos=None):
        """
        Add Parameter to structure

        Parameters
        ----------
        pyParam : SymParameter
            Parameter to add
        pos : int, optional
            Position to add parameter default end
        """
        if pos is not None:
            # add param at certain position in container
            self.javaObj.addParam(pyParam.javaObj, pos)
        else:
            # add param at end of container
            self.javaObj.addParam(pyParam.javaObj)

    def addComponent(self, comp, pos=None):
        """
        Add Component to structure

        Parameters
        ----------
        comp : str
            Name of Component, or filename of component
        pos : int, optional
            Position to add Component. First parameter is at pos = 0.
        """
        if os.path.isfile(comp):
            fname = comp
        else:
            lib_comp = SymComponentInfo.getComponents()
            if comp in lib_comp:
                fname = lib_comp[comp]
            else:
                raise ValueError('Component ' + comp + ' not found')

        ident = self.getIdentifier()
        self.getRoot().importTemplate(fname, ident)
        ret = self.getParam()[-1]

        if pos is not None:
            diff = len(self.getParam()) - (pos + 1)
            if 0 < diff < len(self.getParam()):
                for i in range(diff):
                    ret.moveUp()
        return ret

    def getComponent(self):
        """
        Get component property of container

        Returns
        -------
        pyobj : SymComponent
        """
        javaObj = self.javaObj.getComponent()
        if javaObj.java_name == 'at.lcm.bu21.magopt.structure.Component':
            pyobj = SymComponent(javaObj=javaObj)
        else:
            pyobj = None

        return pyobj

    def addComponentProperty(self):
        """
        Add Component property of container

        Java equivalent : public void addComponent()

        """
        self.javaObj.addComponent()

    def removeComponentProperty(self):
        """
        Remove Component property of container

        Java equivalent : public void removeComponent()

        """
        self.javaObj.removeComponent()

    # def updateComponent(self, comp):
    #     """
    #     Replace Component with a newer version or a different Component
    #     Warning: this method is not finished yet
    #
    #     Parameters
    #     ----------
    #     comp : str
    #         Component that will be used
    #     """
    #     CurObject = locals()['CurObject']
    #     files = []
    #     for f in self.javaObj.getComponent().listLinkedFiles():
    #         files.append(f)
    #
    #     # Remove files of Component
    #     for f in files:
    #         if os.path.isfile(f):
    #             os.remove(f)
    #
    #     new_comp = self.getParent().addComponent(comp)
    #
    #     newid = new_comp.getIdentifier()
    #     lstin = []
    #     for inpar in new_comp.getUsedInputParameter():
    #         lstin.append(inpar.getIdentifier().replace(newid + '.', ''))
    #
    #     oldid = CurObject.getIdentifier()
    #     for inpar in CurObject.getUsedInputParameter():
    #         parid = inpar.getIdentifier().replace(oldid + '.', '')
    #         if parid in lstin:
    #             # print(parid)
    #             new_comp[parid].setData(CurObject[parid].getData())
    #
    #     name = self.getName()
    #     self.removeParam()
    #     new_comp.setName(name)

    def updateComponent(self):
        """
        Replace Component with a newer version or a different Component
        Warning: this method is not finished yet

        Parameters
        ----------
        comp : str
            Component that will be used
        """

        prj = self.getRoot()
        # Name of component is currently coded in comment field
        name = self.getComment()

        src_id = self.getIdentifier()
        dst_comp = self.getParent().addComponent(name, self.getPosition())
        dst_id = dst_comp.getIdentifier()
        print('SRC ID: {}'.format(src_id))
        print('DST ID: {}'.format(dst_id))

        # Compare containers
        src = [p.getIdentifier().replace(src_id + '.', '') for p in self.getParameters('SymContainer')[1:]]
        dst = [p.getIdentifier().replace(dst_id + '.', '') for p in dst_comp.getParameters('SymContainer')[1:]]

        # Add containers if required
        print('\n ---------- Add container:')
        for ident in src:
            if ident not in dst:
                fullid = dst_id + '.' + ident
                parentid = '.'.join(fullid.split('.')[:-1])
                src_param = prj.getObject(src_id + '.' + ident)
                pos = src_param.getPosition()
                print('{}, fullid: {}, parentid: {}'.format(ident, fullid, parentid))
                prj.getObject(parentid).addParam(src_param.cloneParam())

        # Compare remaining parameters
        src = [p.getIdentifier().replace(src_id + '.', '') for p in self.getParameters()]
        dst = [p.getIdentifier().replace(dst_id + '.', '') for p in dst_comp.getParameters()]

        # Add parameters if required
        for ident in src:
            if ident not in dst:
                fullid = dst_id + '.' + ident
                parentid = '.'.join(fullid.split('.')[:-1])
                src_param = prj.getObject(src_id + '.' + ident)
                pos = src_param.getPosition()
                print('id: {}, fullid: {}, parentid: {}'.format(ident, fullid, parentid))
                prj.getObject(parentid).addParam(src_param.cloneParam())

        # Convert parameters if required
        for par in self.getParameters('SymParameter'):
            ident = '{}.{}'.format(dst_id, par.getIdentifier().replace(src_id + '.', ''))
            dst_par = prj[ident]
            if dst_par is not None:
                if type(par) != type(dst_par):
                    if isinstance(dst_par, SymField) and isinstance(par, SymLinkedField):
                        dst_par = dst_par.toSymLinkedField()
                        dst_par.setImportAssignments({'link': par.getImports()[0].getParamName()})

        # Set interface parameter
        print('\n ---------- Set interface assignments:')
        fnc_param = prj[src_id].getParameters('SymFunction')
        fnc_param.extend(prj[src_id].getParameters('SymFormula'))
        fnc_param.extend(prj[src_id].getParameters('SymLinkedField'))

        for param in fnc_param:
            src_par = param
            dst_par = prj[param.getIdentifier().replace(src_id, dst_id)]

            ipar_src = src_par.getImportAssignments()
            try:
                ipar_dst = dst_par.getImportAssignments()
            except:
                continue

            iassign = dict()
            for var in ipar_src:
                if var in ipar_dst:
                    if ipar_src[var] != ipar_dst[var]:
                        iassign[var] = ipar_src[var]
                        print('{} Import {}: "{}" changed to "{}"'.format(dst_par.getIdentifier(), var, ipar_dst[var],
                                                                          ipar_src[var]))
            if len(iassign) > 0:
                dst_par.setImportAssignments(iassign)

            if hasattr(src_par, 'getExportAssignments'):
                epar_src = src_par.getExportAssignments()
                epar_dst = dst_par.getExportAssignments()
                eassign = dict()
                for var in epar_src:
                    if var in epar_dst:
                        if epar_src[var] != epar_dst[var]:
                            eassign[var] = epar_src[var]
                            try:
                                print('{} Export {}: "{}" changed to "{}"'.format(dst_par.getIdentifier(), var,
                                                                                  ipar_dst[var], ipar_src[var]))
                            except:
                                print('Variable {} not in list'.format(var))
                dst_par.setExportAssignments(eassign)

        # Set data
        curindex = prj.getIndex()
        for idx in range(prj.getLength()):
            prj.setIndex(idx + 1)
            for par in self.getUsedInputParameter():
                ident = '{}.{}'.format(dst_id, par.getIdentifier().replace(src_id + '.', ''))
                dst_par = prj[ident]
                if dst_par is not None:
                    if not dst_par.getReadOnly():
                        dst_par.setData(par.getData())
                    else:
                        print('Warning: Parameter "{}" is read only, set value manually {}'.format(dst_par.getName(),
                                                                                                   str(par.getData())))

        prj.setIndex(curindex)

        comp_name = self.getName()
        # self.setName(comp_name + '_old')
        self.removeParam()
        dst_comp.setName(comp_name)

    def getData(self, index=None):
        """
        Overrides getData from SymParameter because
        containers cannot have data.
        """
        print('Containers have no data returning None')
        return None

    def convertToSymContainer(self):
        """
        Convert to class SymContainer

        Returns
        -------
        ret : SymContainer
        """
        return SymContainer(javaObj=self.javaObj.convertToParamContainer())

    def convertToSymMaterial(self):
        """
        Convert to class SymMaterial

        Returns
        -------
        ret : SymMaterial
        """
        return SymMaterial(javaObj=self.javaObj.convertToParamMaterial())

    def convertToSymCollector(self):
        """
        Convert to class SymCollector

        Returns
        -------
        ret : SymCollector
        """
        return SymCollector(javaObj=self.javaObj.convertToParamCollector())

    # getter methods
    def getParam(self):
        """
        Get List of child-parameters.

        Returns
        -------
        ret : list
            List of child parameters

        Examples
        --------
        Get a list of sub - parameters of a project
        >>> CurProject = locals()['CurProject']
        >>> param_list = CurProject.getParam()

        Get Python list of all Sub-Parameters included in Sub-Container of a project

        >>> param_list = CurProject['containerName'].getParam()

        """
        return super()._convertObject(self.javaObj.getParam())

    def getParamSize(self):
        """
        Get size of number of params in container

        Returns
        -------
        size : int
        """
        return self.javaObj.getParamSize()

    def getIncomplete(self, index=None):
        """
        Get given incomplete status

        Parameters
        ----------
        index : int, optional

        Returns
        -------
        status : bool
            Incomplete status

        """
        if index is not None:
            return self.javaObj.getIncomplete(index) == 1
        else:
            return self.javaObj.getIncomplete() == 1

    def convertToPythonFunction(self, classPath, className):
        """
        Convert parameter to Python function
        Parameters
        ----------
        classPath : string

        className : string

        Returns
        -------

        """
        self.javaObj.convertToParamPythonFunction(classPath, className)

    def convertToMatlabFunction(self, classPath, className):
        """
        Convert parameter to Matlab function
        Parameters
        ----------
        classPath : string

        className : string

        Returns
        -------

        """
        self.javaObj.convertToParamMatlabFunction(classPath, className)

    def convertToJavaFunction(self, classPath, className):
        """
        Convert parameter to Java function
        Parameters
        ----------
        classPath : string

        className : string

        Returns
        -------

        """
        self.javaObj.convertToParamJavaFunction(classPath, className)

    def addPublicInterface(self, name):
        """
        Add public-interface if not exists

        Parameters
        ----------
        name : str

        """
        self.javaObj.addPublicInterface(name)

    def getPublicInterfaceParam(self, name):
        """
        Get parameter-name of public-interface
        Parameters
        ----------
        name : str

        Returns
        -------
        name : str
        """
        return PyJDataInterface.convert_J2Py(self.javaObj.getPublicInterfaceParam(name))

    def getPublicInterfaceParameter(self, name):
        """
        Get parameter of public-interface
        Parameters
        ----------
        name : str

        Returns
        -------
        parm : SymParameter
        """
        obj = self.javaObj.getPublicInterfaceParameter(name)
        par = SymParameter._convertObject(obj)
        return par

    def setPublicInterface(self, name, param):
        """
        Set public-interface
        Parameters
        ----------
        name : str

        param : str
        """
        self.javaObj.setPublicInterface(name, param)

    def listPublicInterfaceName(self):
        """
        List all names of public-interfaces

        Returns
        -------
        ret : list
        """
        ret = PyJDataInterface.convert_J2Py(self.javaObj.listPublicInterfaceName())
        return ret

    def removePublicInterface(self, name):
        """
        Remove public-interface
        Parameters
        ----------
        name : str

        Returns
        -------

        """
        self.javaObj.removePublicInterface(name)

    def removeUnused(self):
        """
        Remove unused Parameters from the container
        """
        params = self.getParameters('SymParameter')
        for param in params:
            if not param.getUsed():
                print('Removing: {}'.format(param.getIdentifier()))
                param.removeParam()
        self.refresh()

    def isComponent(self):
        """
        Test if Container is a Component
        """
        if self.javaObj.getComponent() is None:
            ret = False
        else:
            ret = True
        return ret

    def getAccess(self):
        """
        Get access of a container
        """
        return self._javaObj.getAccess().toString()

    def getComponentDirectory(self):
        """
        Get path to Component directory
        """
        if self.isComponent():
            return self.javaObj.getComponent().getFolder().getAbsolutePath()
        else:
            return None

    def getAllImports(self):
        """
        Gets all imports of this Container and all subelements

        Returns
        -------
        ret : tuple
            (param: SymParameter, imp: SymDependItem)
        """
        imports = []
        if hasattr(self, 'getImports'):
            for imp in self.getImports():
                imports.append((self, imp))
        for par in self.getParam():
            if hasattr(par, 'getAllImports'):
                imports.extend(par.getAllImports())
            elif hasattr(par, 'getImports'):
                try:
                    for imp in par.getImports():
                        imports.append((par, imp))
                except Exception as e:
                    print(par.getName())
        return imports

    def getExternalImports(self):
        """
        Gets all imports that are linked to a parameter outside of this container

        Returns
        -------
        ret : tuple
            (param: SymParameter, imp: SymDependItem)
        """
        ret = []
        ident = self.getIdentifier()
        for imp in self.getAllImports():
            lnk = imp[1].getHandle()
            if lnk is not None:
                if not lnk.getIdentifier().startswith(ident):
                    ret.append(imp)
        return ret

    def getUnresolvedImports(self):
        """
        Gets all imports of this container and all subelements that are unresolved

        Returns
        -------
        ret : tuple
            (param: SymParameter, imp: SymDependItem)
        """
        ret = []
        for imp in self.getAllImports():
            if imp[1].getHandle() is None and imp[1].getParamName() != '':
                ret.append(imp)
        return ret

    def getAllExports(self):
        """
        Gets all Exports of this Container and all subelements

        Returns
        -------
        ret : tuple
            (param: SymParameter, imp: SymDependItem)
        """
        exports = []
        if hasattr(self, 'getExports'):
            for imp in self.getExports():
                exports.append((self, imp))
        for par in self.getParam():
            if hasattr(par, 'getAllExports'):
                exports.extend(par.getAllExports())
            elif hasattr(par, 'getExports'):
                try:
                    for imp in par.getExports():
                        exports.append((par, imp))
                except Exception as e:
                    print(par.getIdentifier())

        return exports

    def getExternalExports(self):
        """
        Gets all exports that are linked to a parameter outside of this container

        Returns
        -------
        ret : tuple
            (param: SymParameter, imp: SymDependItem)
        """
        ret = []
        ident = self.getIdentifier()
        for imp in self.getAllExports():
            lnk = imp[1].getHandle()
            if lnk is not None:
                if not lnk.getIdentifier().startswith(ident):
                    ret.append(imp)
        return ret

    def getUnresolvedExports(self):
        """
        Gets all exports of this container and all subelements that are unresolved

        Returns
        -------
        ret : tuple
            (param: SymParameter, imp: SymDependItem)
        """
        ret = []
        for imp in self.getAllExports():
            if imp[1].getHandle() is None and imp[1].getParamName() != '':
                ret.append(imp)
        return ret

    def toDict(self, detailed=False):
        """
        Convert data to Python dictionary

        Returns
        -------
        ret : dictionary
        """
        res = {'__error__': self.getError()}
        if hasattr(self, 'getUpdateRequired'):
            res['__updaterequired__'] = self.getUpdateRequired()
        for p in self.getParam():
            res[p.getName()] = p.toDict(detailed)
        return res

    """
    def _getTreeInfo(self, info=None):
        if info is None:
            info = {}
        for par in self.getParam():
            if isinstance(par, SymContainer):
                info[self.getName()] = par._getTreeInfo()
        info['name'] = self.getName()
    """

    def _getInfo(self):
        ret = super()._getInfo()

        ret['param'] = []
        for par in self.getParam():
            if not isinstance(par, SymContainer):
                ret['param'].append(par._getInfo())
        ret['docfiles'] = self.getDocFiles()
        return ret


class SymField(SymParameter):
    """
    Python wrapper class for |sym|'s java class ParamField
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamField(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamField(name, description)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    # methods
    def toSymField(self):
        """
        Convert to class SymField

        Returns
        -------
        param : SymField

        """
        return SymField(javaObj=self.javaObj.convertToParamField())

    def toSymFormula(self):
        """
        Convert to class SymFormula

        Returns
        -------
        param : SymFormula
        """
        return SymFormula(javaObj=self.javaObj.convertToParamFormula())

    def toSymLinkedField(self):
        """
        Convert to class SymLinkedField

        Returns
        -------
        param : SymLinkedField
        """
        return SymLinkedField(javaObj=self.javaObj.convertToParamLinkedField())

    def toSymComboBox(self):
        """
        Convert to class SymComboBox

        Returns
        -------
        param : SymComboBox
        """
        return SymComboBox(javaObj=self.javaObj.convertToParamComboBox())

    def clearDatabaseLink(self):
        """
        Clear the database link
        """
        self.javaObj.clearDbId()

    # getter
    def getTexString(self):
        """
        Get Tex-string of Parameter

        Returns
        -------
        texString : string
        """
        return self.javaObj.getTexString()

    def getUnit(self):
        """
        Get unit of Parameter

        Returns
        -------
        unit : string
        """
        return self.javaObj.getUnit()

    def getType(self):
        """
        Get type of Parameter

        Returns
        -------
        type : string
        """
        return self.javaObj.getType()

    def getLock(self):
        """
        Get lock status

        Returns
        -------
        status : bool
             Lock status
        """
        return self.javaObj.getLock() == 1

    # setter
    def setLock(self, lock=True):
        """
        Set lock status

        Parameters
        ----------
        lock : bool, optional
            Lock status defaults to True

        """
        self.javaObj.setLock(lock)

    def setTexString(self, val):
        """
        Set Tex-string of Parameter

        Parameters
        ----------
        val : string
            Tex-string
        """
        self.javaObj.setTexString(val)

    def setUnit(self, val):
        """
        Set unit of Parameter

        Parameters
        ----------
        val : string
            Unit-string
        """
        self.javaObj.setUnit(val)

    def setType(self, val):
        """
        Set type of Parameter

        Parameters
        ----------
        val : string
            Type-string
        """
        self.javaObj.setType(val)

    def _getInfo(self):
        ret = super()._getInfo()
        ret['value'] = self.getDataString()
        ret['unit'] = self.getUnit()
        ret['texstring'] = self.getTexString()
        return ret

    # def plot(self, *args):
    #     """
    #     Plots data of a SymField-Object with bokeh plotting library. The figure is opened in the standard web browser. At the moment, only array-type data is supported -> type(SymField.getData()) = SymNdArray. Possible plotting types implemented so far are:
    #        - self.getData() over index if *args is empty
    #        - self.getData() over args[0].getData() where args[0] = SymField-Object with type(args[0].getData()) = SymNdArray
    #     Further implementations are planned
    #
    #     Parameters
    #     ----------
    #     *args : SymField object where type(SymField.getData()) = SymNdArray
    #
    #     Returns
    #     -------
    #     fig : bokeh.plotting.figure object, which can be used for modifying the plot
    #     """
    #     import bokeh.plotting as bp
    # 
    #     bp.reset_output()
    #     fig = bp.figure(plot_width=600, plot_height=400, title=self.getDescription())
    #
    #     if len(args)==0:
    #         """self.array data over self.index"""
    #         if self.getData().shape[0]==1 or self.getData().shape[1]==1:
    #             data = self.getData().flatten()
    #             fig.line(np.arange(data.shape[0]),data)
    #             fig.xaxis.axis_label = "data index"
    #             fig.yaxis.axis_label = "{} in {}".format(self.getName(),self.getUnit())
    #             bp.show(fig)
    #         else:
    #             """self.matrix data over self.index"""
    #             """self.matrix data over self.matrix.column"""
    #             print('Dimension of Data has to be 1x?')
    #     elif len(args)==1:
    #         """self.array data over other.array"""
    #         if isinstance(args[0], self.__class__) and self.getData().shape[0]==1 or self.getData().shape[1]==1 and args[0].getData().shape[0]==1 or args[0].getData().shape[1]==1:
    #             y_data = self.getData().flatten()
    #             x_data = args[0].getData().flatten()
    #             fig.line(x_data, y_data)
    #             fig.xaxis.axis_label = "{} in {}".format(args[0].getName(), args[0].getUnit())
    #             fig.yaxis.axis_label = "{} in {}".format(self.getName(),self.getUnit())
    #             bp.show(fig)
    #         else:
    #             print('Dimension of Data has to be 1x?')
    #     else:
    #         print('only one additional argument possible')
    #
    #     return fig


class SymComboBox(SymField):
    """
    Python wrapper class for |sym|'s java class ParamComboBox
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    self.javaObj = JParamComboBox(name)
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    self.javaObj = JParamComboBox(name, description)
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    # getter
    def getComboList(self):
        """Get list of combobox"""
        return self.javaObj.getComboList()

    def getComboIndex(self):
        """Get index of combobox"""
        return self.javaObj.getComboIndex()

    # setter
    def setComboList(self, val):
        """Set list of combobox"""
        self.javaObj.setComboList(val)

    def setComboIndex(self, val):
        """Set index of combobox"""
        self.javaObj.setComboIndex(val)


class SymProject(SymContainer):
    """
    Python wrapper class for |sym|'s java class Project
    """

    @classmethod
    def load(cls, strFile):
        """
        load project from path

        Parameters
        ----------
        strFile : string
            Full path to .mop file
        """
        norm_path = os.path.normpath(strFile)
        javaObj = JProject.load(norm_path)

        if javaObj is None:
            raise Exception('Unable to load file: {}'.format(norm_path))
        return cls(javaObj=javaObj)

    @staticmethod
    def getDataOverIndices(pyParam):
        """
        Get indices and corresponding data of parameter as lists.

        Parameters
        ----------
        pyParam : SymParameter

        Returns
        -------
        values : list
            List containing data over indices

        """
        numInd = pyParam.getLength()
        values = []
        indices = []
        for index in range(numInd):
            values.append(pyParam.getData(index))
            indices.append(index)

        return indices, values

    @staticmethod
    def plotParameter(*args, **kwargs):
        """
        Static method to plot data of one or two parameter over indices as line or bar plot

        Parameters
        ----------
        kwargs:
            paramX : SymParameter, optional
                Parameter to use for the x-axis
            paramY : SymParameter, list of SymParameter
                Parameter(s) to use for the y-axis

            mode : {None, 'line', 'bar'}
                Mode of plotting the data


        Examples
        --------

        Plot parameter in Python Console

        >>> p = locals()['CurProject']
        >>> p.plotParameter(p['Motor_Results.P1']).show()

        """
        lineWidth = 2
        if len(args) == 1:
            if kwargs.get('mode'):
                mode = kwargs['mode']
            else:
                mode = 'line'
            pyParam = args[0]
            if pyParam is None:
                raise ValueError("nonexistent parameter")
            indices, values = SymProject.getDataOverIndices(pyParam)
            for value in values:
                if value is None:
                    raise ValueError('{} contains data of None type'.format(pyParam.getName()))
                elif not isinstance(value, numbers.Number):
                    raise ValueError('{} data must be numeric'.format(pyParam.getName()))
            if mode == 'line':
                plot = SymPlot2Line()
            elif mode == 'bar':
                plot = SymPlot2Bar()
            else:
                raise Exception('mode must be "line" or "bar"!')
            plot.addSignal('{}'.format(pyParam.getName()), values)
            plot.setTitle('{} vs. indices'.format(pyParam.getName()))
            plot.setXLabel('indices')
            plot.setYLabel('{} in {}'.format(pyParam.getName(), pyParam.getUnit()))
            plot.hfig.getSignal(0).getStyle().setLineWidth(lineWidth)
            plot.reset()
            return plot
        elif len(args) == 2:
            pyParam1 = args[0]
            if pyParam1 is None:
                raise ValueError('nonexistent parameter x')
            pyParam2 = args[1]
            if pyParam2 is None:
                raise ValueError('nonexistent parameter y')
            if isinstance(pyParam2, list):
                # plot multiple signals
                signalNames = None
                if kwargs.get('SignalNames'):
                    signalNames = kwargs['SignalNames']
                    if len(signalNames) != len(pyParam2):
                        raise ValueError('number of signal names must be the same as number of parameters')
                plot = SymPlot2Line()
                indices, values1 = SymProject.getDataOverIndices(pyParam1)
                for i, param in enumerate(pyParam2):
                    if param is None:
                        raise ValueError('nonexistent parameter in list position {}'.format(i))
                    indices, values2 = SymProject.getDataOverIndices(param)
                    for value1, value2 in zip(values1, values2):
                        if value1 is None:
                            raise ValueError('{} contains data of None type'.format(pyParam1.getName()))
                        elif value2 is None:
                            raise ValueError('{} contains data of None type'.format(param.getName()))
                        elif not isinstance(value1, numbers.Number):
                            raise ValueError('{} data is not numeric'.format(pyParam1.getName()))
                        elif not isinstance(value2, numbers.Number):
                            raise ValueError('{} data is not numeric'.format(param.getName()))
                    points = zip(values1, values2)
                    sorted_points = sorted(points)
                    v1 = [point[0] for point in sorted_points]
                    v2 = [point[1] for point in sorted_points]
                    if signalNames is not None:
                        plot.addSignal('{}'.format(signalNames[i]), v1, v2)
                    else:
                        plot.addSignal('{} in {}'.format(param.getName(), param.getUnit()), v1, v2)
                    plot.hfig.getSignal(i).getStyle().setLineWidth(lineWidth)
                plot.setXLabel('{} in {}'.format(pyParam1.getName(), pyParam1.getUnit()))
                plot.setLegend(True)
                return plot
            else:
                if pyParam2 is None:
                    raise ValueError('Parameter y is None')
                indices, values1 = SymProject.getDataOverIndices(pyParam1)
                indices, values2 = SymProject.getDataOverIndices(pyParam2)
                for value1, value2 in zip(values1, values2):
                    if value1 is None:
                        raise ValueError('{} contains data of None type'.format(pyParam1.getName()))
                    elif value2 is None:
                        raise ValueError('{} contains data of None type'.format(pyParam2.getName()))
                    elif not isinstance(value1, numbers.Number):
                        raise ValueError('{} data is not numeric'.format(pyParam1.getName()))
                    elif not isinstance(value2, numbers.Number):
                        raise ValueError('{} data is not numeric'.format(pyParam2.getName()))
                plot = SymPlot2Line()
                points = zip(values1, values2)
                sorted_points = sorted(points)
                values1 = [point[0] for point in sorted_points]
                values2 = [point[1] for point in sorted_points]
                plot.addSignal('{} in {}'.format(pyParam2.getName(), pyParam2.getUnit()), values1, values2)
                plot.setTitle('{} vs. {}'.format(pyParam2.getName(), pyParam1.getName()))
                plot.setXLabel('{} in {}'.format(pyParam1.getName(), pyParam1.getUnit()))
                plot.setYLabel('{} in {}'.format(pyParam2.getName(), pyParam2.getUnit()))
                plot.hfig.getSignal(0).getStyle().setLineWidth(lineWidth)
                plot.reset()
                return plot
        else:
            raise ValueError('Function takes one or two positional arguments')

    @staticmethod
    def plotParamHistogramm(*args, **kwargs):
        """
        Static method to plot histogramm of parameter over indices

        Parameters
        ----------
        args : list
        kwargs :
        """
        if len(args) == 1:
            pyParam = args[0]
            nbin = pyParam.getLength()
            if pyParam is None:
                raise ValueError('Parameter is None')
            indices, values = SymProject.getDataOverIndices(pyParam)
            for value in values:
                if value is None:
                    raise ValueError('Parameter contains data of None type')
                elif not isinstance(value, numbers.Number):
                    raise ValueError('Compare requires numeric data')
            y, bin_edges = np.histogram(values, bins=len(np.unique(values)))
            y_scaled = y / nbin
            plot = SymPlot2Bar()
            plot.addSignal('{}'.format(pyParam.getName()), np.unique(values), y_scaled.tolist())
            plot.setTitle('Histogramm of {}'.format(pyParam.getName()))
            plot.setXLabel('{} in {}'.format(pyParam.getName(), pyParam.getUnit()))
            plot.setYLabel('density')
            return plot

        else:
            raise ValueError('Function takes one positional arguments')

    @classmethod
    def new(cls, name, description, file):
        """
        Create a new project

        Parameters
        ----------
        name : string
            Project name
        description: string
            Project description
        file: string
            Project file

        Returns
        -------
        SymProject
        """

        obj = cls(name, description, file)
        return obj

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            self.javaObj = kwargs['javaObj']
        else:
            if args:
                nargs = len(args)
                if nargs == 3:
                    name = args[0]
                    description = args[1]
                    file = args[2]
                    self.javaObj = JProject(name, description, file)
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('missing parameter')

    def __str__(self):
        return 'Sym' + self.javaObj.toString()

    def __repr__(self):
        return self.__str__()

    # methods
    def close(self):
        """
        Close connections of project
        """
        self.javaObj.close()

    def regenerate(self, **kwargs):
        """
        Regenerate project and update required functions

        Parameters
        ----------
        kwargs
            RunSim : bool
                Run simulation if true
            RemoteCompute : bool
                Remote compute if true
            Optimize : bool
                Optimize if true
        """
        return self.javaObj.regenerate(kwargs)

    def resetAllFunctions(self):
        """
        Reset all functions in project
        """
        self.javaObj.resetAllFunctions()

    def check(self):
        """
        Check structure which updates are required
        """
        self.javaObj.check()

    def save(self):
        """
        Save Project to project-file
        """
        status = self.javaObj.save()
        return status

    def saveAs(self, strFile=None, override=None):
        """
        Save Project to given project-file

        Parameters
        ----------
        strFile : string
            Full path with .mop extension
        override : bool
            Override flag

        """
        if (strFile is not None) and (override is not None):
            status = self.javaObj.saveAs(strFile, override)
        elif strFile is not None:
            status = self.javaObj.saveAs(strFile)
        else:
            raise Exception('saveAs(): Wrong number of Parameters')
        return status

    def duplicate(self, index=None):
        """ Duplicate current dataset to the end of list

        Parameters
        ----------
        index : int (zero-based)
            defines dataset index. If not defined, current index is used

        Returns
        -------

        """
        if index is None:
            self.javaObj.duplicate()
        else:
            self.javaObj.duplicate(index)

    def reduce(self):
        """
        Reduce current dataset of list.
        (Removes one index at the end)
        """
        self.javaObj.reduce()

    def importTemplate(self, strFile, strObject):
        """
        Import template into project at given parameter
        """
        status = self.javaObj.importTemplate(strFile, strObject)
        return status

    def exportTemplate(self, strFile, strObjects):
        """
        Export template from given parameter
        """
        jarr = jarray(len(strObjects), java.lang.String)
        for i in range(len(strObjects)):
            if not isinstance(strObjects[i], str):
                raise ValueError('All args must be strings')
            jarr[i] = java.lang.String(strObjects[i])

        status = self.javaObj.exportTemplate(strFile, jarr)
        return status

    def moveDataset(self, from_index, to_index):
        """
        Move dataset in project
        """
        self.javaObj.moveDataset(from_index, to_index)

    def generateDocumentationFiles(self):
        """
        Generate documentationsfiles of current index
        """
        return self.javaObj.generateDocumentationFiles()

    # getter
    def getAllDSComment(self):
        """
        Get comment-list of datasets
        """
        tmp = self.javaObj.getAllDSComment()
        return tmp

    def getDSComment(self):
        """
        Get comment of current dataset
        """
        tmp = self.javaObj.getDSComment()
        return tmp

    def getEventLog(self):
        """
        Get event-log of current dataset
        """
        tmp = self.javaObj.getEventLog().toString()
        return tmp

    def getDirectory(self):
        """
        Get Project directory
        """
        full_path = self.javaObj.getFile()
        return os.path.dirname(full_path)

    def getFilesDirectory(self):
        """
        Get directory of dataset-files
        """
        return self.javaObj.getFilesDirectory().getAbsolutePath()

    def getAllFilesDirectory(self):
        """
        Get path to Project files directory
        """
        return self.javaObj.getAllFilesDirectory().getAbsolutePath()

    def getFile(self):
        """
        Get Project filename
        """
        # ToDo: Why is the implemented like this?
        #  Different behavior between python and java implementation:
        #  Java returns full path
        #  Python Wrapper returns filename

        full_path = self.javaObj.getFile()
        return os.path.basename(full_path)

    def getGuiHandle(self):
        """
        Get gui-handle from matlab
        """
        return self.javaObj.getGuiHandle()

    @staticmethod
    def getInstallPath():
        """
        Get install-path

        Returns
        -------
        str
        """
        return JProject.getInstallPath()

    @staticmethod
    def getLibraryPath():
        """
        Get path of library

        Returns
        -------
        str
        """
        return JProject.getLibraryPath()

    def getPrintEvent(self):
        """
        Get status of event-print
        """
        return self.javaObj.getPrintEvent()

    @staticmethod
    def getProperty(name):
        """
        Get property

        Returns
        -------
        str
        """
        return JProject.getProperty(name)

    @staticmethod
    def getVersion():
        """
        Get current program version

        Returns
        -------
        str
        """
        return JProject.getVersion()

    # setter
    def setIndex(self, val):
        """
        Set current dataset index zero-based

        Parameters
        ----------
        val : int
            Index to set (zero-based)
        """
        self.javaObj.setIndex(val)

    def setDSComment(self, val):
        """
        Set comment of current dataset
        """
        self.javaObj.setDSComment(val)

    def setGuiHandle(self, val):
        """Set gui-handle from Matlab"""
        self.javaObj.setGuiHandle(val)

    def setPrintEvent(self, val):
        """Set status of event-print"""
        self.javaObj.setPrintEvent(val)

    def _toJSON(self, param):
        """
        Only for test purpose - will be removed in future!!!
        """
        fname = param.getIdentifier() + '.json'
        with open(os.path.join(self.getDirectory(), fname), 'w+') as fh:
            fh.write(json.dumps(param._getInfo()))


class SymDependItem:
    """
    Python wrapper class for |sym|'s java class DependItem
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    var = args[0]
                    javaObj = JDependItem(var)
                    self.javaObj = javaObj
                elif nargs == 2:
                    var = args[0]
                    param = args[1]
                    javaObj = JDependItem(var, param)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        return 'Sym' + self.javaObj.toString()

    def __repr__(self):
        return self.__str__()

    def getHandle(self):
        """Get handle of DependItem"""
        jParam = self.javaObj.getHandle()
        return SymParameter._convertObject(jParam)

    def getParamName(self):
        """Get parameter (string) of DependItem"""
        tmp = self.javaObj.getParam()
        return tmp

    def getVarName(self):
        """Get Name of DependItem"""
        return self.javaObj.getVarName()


class SymDependency:
    """
    Python wrapper class for |sym|'s java class Dependency
    """

    # getter methods
    @staticmethod
    def getImports(obj):
        val = obj.javaObj.getDepend().getImports()
        # val contains Java Array of type Java class DependItem
        tmp = []
        for i, item in enumerate(val):
            tmp.append(SymDependItem(javaObj=item))
        return SymPrettyPrintConverter.ugly2pretty(tmp)

    @staticmethod
    def getExports(obj):
        val = obj.javaObj.getDepend().getExports()
        # val contains Java Array of type Java class DependItem
        tmp = []
        for i, item in enumerate(val):
            tmp.append(SymDependItem(javaObj=item))
        return SymPrettyPrintConverter.ugly2pretty(tmp)

    @staticmethod
    def getImportVars(obj):
        imp = obj.getJava().getDepend().getImports()
        # imp contains list of DependItems
        names = []
        data = []
        for k, item in enumerate(imp):
            if (item.getHandle() and item.getVarName()) is not None:
                names.append(str(item.getVarName()))
                data.append(PyJDataInterface.convert_J2Py(item.getHandle().getData()))

        return SymPrettyPrintConverter.ugly2pretty(names), SymPrettyPrintConverter.ugly2pretty(data)

    @staticmethod
    def getExportVars(obj):
        exp = obj.getJava().getDepend().getExports()
        # exp contains list of DependItems
        names = []
        data = []
        for k, item in enumerate(exp):
            if (item.getHandle() and item.getVarName()) is not None:
                names.append(str(item.getVarName()))
                data.append(PyJDataInterface.convert_J2Py(item.getHandle().getData()))

        return SymPrettyPrintConverter.ugly2pretty(names), SymPrettyPrintConverter.ugly2pretty(data)


class SymTopFunction(SymDependency, SymParameter):
    """
    Python wrapper class for |sym|'s java class Function
    """

    @staticmethod
    def getArgumentMap(varargin):
        map = JParamJavaFunction.getEmptyMap()
        for i, item in enumerate(varargin):
            if varargin[i] == 'Timeout':
                map.put(varargin[i], JLong(varargin[i + 1]))
            else:
                map.put(varargin[i], varargin[i + 1])

        return map

    def __str__(self):
        return 'Sym' + self.javaObj.toString()

    def getUpdateRequired(self):
        tmp = self.javaObj.getUpdateRequired()
        return tmp


class SymMaterial(SymContainer, SymTopFunction):
    """
    Python wrapper class for |sym|'s java class ParamMaterial
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamMaterial(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamMaterial(name, description)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    # methods
    def update(self, **kwargs):
        status = self.javaObj.update(kwargs)
        return status

    def reset(self):
        """Reset Function"""
        self.javaObj.reset()

    # getter
    def getMaterialClassName(self):
        """Get material-class-name"""
        return self.javaObj.getMaterialClassName()

    def getMaterialName(self):
        return self.javaObj.getMaterialName()

    def getDatasetName(self):
        return self.javaObj.getDatasetName()

    def getData(self, index=None):
        return SymParameter.getData(self, index)

    # getter
    def getImports(self):
        val = SymDependency.getImports(self)
        return val

    def getExports(self):
        val = SymDependency.getExports(self)
        return val

    # setter
    def setDatasetID(self, val):
        status = self.javaObj.setDatasetID(val, JParamMaterial.getConnection())
        return status


class SymCollector(SymContainer, SymTopFunction):
    """
    Python wrapper class for |sym|'s java class ParamCollector
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamCollector(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamCollector(name, description)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    # methods
    def update(self, **kwargs):
        status = self.javaObj.update(kwargs)
        return status

    def reset(self):
        """Reset Function"""
        self.javaObj.reset()

    # getter
    def getData(self, index=None):
        return SymParameter.getData(self, index)


class SymLinkedField(SymField, SymTopFunction):
    """
    Python wrapper class for |sym|'s java class ParamLinkedField
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamLinkedField(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamLinkedField(name, description)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    # methods
    def reset(self):
        self.javaObj.reset()

    # getter
    def getImports(self):
        val = SymDependency.getImports(self)
        return val

    def getExports(self):
        val = SymDependency.getExports(self)
        return val

    def getImportAssignments(self):
        val = dict()
        for imp in self.getImports():
            val[imp.getVarName()] = imp.getParamName()
        return val

    def setUpdateRequired(self, val):
        self.javaObj.setUpdateRequired(val)

    def setLink(self, val):
        self.javaObj.getDepend().setImportAssignmentsMap({'link': val})

    def setImportAssignments(self, val):
        if 'link' in val:
            self.javaObj.getDepend().setImportAssignmentsMap(val)
        else:
            raise Exception("Assignment only possible to 'link'")

    def update(self, **kwargs):
        status = self.javaObj.update(kwargs)
        return status


class SymFormula(SymField, SymTopFunction):
    """
    Python wrapper class for |sym|'s java class ParamFormula
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamFormula(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamFormula(name, description)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    # methods
    def reset(self):
        self.javaObj.reset()

    # getter
    def getImports(self):
        return SymDependency.getImports(self)

    def getExports(self):
        return SymDependency.getExports(self)

    def getFormulaType(self):
        return self.javaObj.getFormulaType().toString()

    def getExpression(self):
        return self.javaObj.getExpression()

    def getImportAssignments(self):
        val = dict()
        for imp in self.getImports():
            val[imp.getVarName()] = imp.getParamName()
        return val

    def getUnit(self):
        """
        Get unit of Parameter

        Returns
        -------
        unit : string
        """
        return self.javaObj.getUnit()

    def setImportAssignments(self, val):
        self.javaObj.getDepend().setImportAssignmentsMap(val)

    def setUpdateRequired(self, val):
        self.javaObj.setUpdateRequired(val)

    def setFormulaType(self, val):
        self.javaObj.setFormulaType(val)

    def setExpression(self, val):
        self.javaObj.setExpression(val)

    def setUnit(self, val):
        """
        Set unit of Parameter

        Parameters
        ----------
        val : string
            Unit-string
        """
        self.javaObj.setUnit(val)


class SymLocation:
    """
    Python wrapper class for |sym|'s java class Location
    """

    def __init__(self, javaLocation):
        self.javaObj = javaLocation

    def fileValues(self):
        javaArray = self.javaObj.fileValues()
        val = []
        for i, item in enumerate(javaArray):
            val.append(SymLocation(item))
        return val

    def functionValues(self):
        javaArray = self.javaObj.functionValues()
        val = []
        for i, item in enumerate(javaArray):
            val.append(SymLocation(item))
        return val

    def isConstantPossible(self):
        return self.javaObj.isConstantPossible() == 1

    def toString(self):
        return self.javaObj.toString()

    def valueOf(self, name):
        return self.javaObj.valueOf(name)

    def values(self):
        javaArray = self.javaObj.values()
        val = []
        for i, item in enumerate(javaArray):
            val.append(SymLocation(item))
        return val


class SymDirectory(SymDependency, SymParameter):
    """
    Python wrapper class for |sym|'s java class ParamDirectory
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamDirectory(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamDirectory(name, description)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    # methods
    def show(self):
        self.javaObj.show()

    def getAbsoluteDirectoryPath(self, index=None):
        if index is not None:
            val = str(self.javaObj.getAbsoluteDirectoryPath(index))
        else:
            val = str(self.javaObj.getAbsoluteDirectoryPath())
        return val

    def getRelDirPath(self, index=None):
        if index is not None:
            val = str(self.javaObj.getRelDirPath(index))
        else:
            val = str(self.javaObj.getRelDirPath())
        return val

    def getLocation(self):
        return SymLocation(self.javaObj.getLocation())

    # setter
    def setDirectory(self, val):
        self.javaObj.setRelDirPath(val)

    def setLocation(self, val):
        self.javaObj.changeLocation(val)

    def updateLocation(self):
        self.javaObj.updateLocation()


class SymFile(SymDirectory):
    """
    Python wrapper class for |sym|'s java class ParamFile
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamFile(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamFile(name, description)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    # getter
    def getCurCheckSum(self):
        return self.javaObj.getCheckSum()

    def getFile(self):
        return str(self.javaObj.getFileName())

    def getFileName(self):
        return str(self.javaObj.getFileName())

    # setter
    def setFile(self, val):
        self.javaObj.setFileName(val)

    def setFileName(self, val):
        self.javaObj.setFileName(val)

    def setAbsoluteFilePath(self, val):
        self.javaObj.setAbsoluteFilePath(val)

    def setAbsoluteFile(self, javaIoFile):
        self.javaObj.setAbsoluteFile(javaIoFile)


class SymFunction(SymContainer, SymTopFunction, metaclass=ABCMeta):
    """
    Python wrapper class for |sym|'s java class ParamFunction
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamFunction(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamFunction(name, description)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    def equals(self, other):
        return self.javaObj.equals(other.javaObj)

    # getter
    def getClassName(self):
        return self.javaObj.getClassName()

    def getClassPath(self):
        return self.javaObj.getClassPath()

    def setClassPath(self, name):
        self.javaObj.setClassPath(name)

    def getImports(self):
        val = SymDependency.getImports(self)
        return val

    def getExports(self):
        val = SymDependency.getExports(self)
        return val

    def getImportValue(self, name):
        return self.javaObj.getImportValue(name)

    def getExportValue(self, name):
        return self.javaObj.getExportValue(name)

    def getImportUnit(self, name):
        return self.javaObj.getImportUnit(name)

    def getExportUnit(self, name):
        return self.javaObj.getExportUnit(name)

    def getParamDirectory(self):
        return self.javaObj.getParamDirectory().toString()

    # setter
    @abstractmethod
    def setClassName(self, className):
        # has to be implemented in subclasses
        pass

    def getImportAssignments(self):
        val = dict()
        for imp in self.getImports():
            val[imp.getVarName()] = imp.getParamName()
        return val

    def getExportAssignments(self):
        val = dict()
        for imp in self.getExports():
            val[imp.getVarName()] = imp.getParamName()
        return val

    def setImportAssignments(self, val):
        self.javaObj.getDepend().setImportAssignmentsMap(val)

    def setExportAssignments(self, val):
        self.javaObj.getDepend().setExportAssignmentsMap(val)

    def setUpdateRequired(self, val):
        self.javaObj.setUpdateRequired(val)

    def generateDocumentationFiles(self):
        """
        Generate documentationsfiles of current index
        """
        return self.getRoot().javaObj.generateDocumentationFiles(self.javaObj)

    def generateInterfaceFields(self, replace: bool = False):
        self.generateImportFields(replace)
        self.generateExportFields(replace)

    def generateImportFields(self, replace):
        imports = self.getImports()
        self.createFunctionParameters(imports, 'imports', replace)

    def generateExportFields(self, replace):
        imports = self.getExports()
        self.createFunctionParameters(imports, 'exports', replace)

    def createFunctionParameters(self, params: list, direction: str, replace: bool):
        """
        Creates and links fields for imports or exports of a function

        Parameters
        ----------
        params
        direction : str
            depend direction 'import' or 'export'
        replace : bool
            replace existing links if True

        """
        javaObj = self.javaObj
        for param in params:
            ports_desc = param.javaObj.getVarDesc()
            name = ports_desc.getName()
            if param.getHandle() is None or replace:

                description = ports_desc.getDescription()
                unit = ports_desc.getUnit()
                # optional = ports_desc.getOptional()
                dimension = ports_desc.getDimension()

                new_param = SymField(name, description)
                new_param.setUnit(unit)
                new_param.setComment(dimension)

                self.addParam(new_param)

                if direction == 'imports':
                    javaObj.getDepend().setImportAssignmentsMap({name: name})

                    '''set default value'''
                    if 'Boolean' in dimension:
                        val = True
                    elif 'String' in dimension:
                        val = ''
                    elif 'Numeric' in dimension:
                        val = 1.0
                    elif 'Struct' in dimension:
                        val = {}
                    elif 'Array' in dimension:
                        val = np.array([1, 2, 3], dtype=np.object)
                    elif 'Periodic' in dimension:
                        val = None
                    else:
                        val = None

                    new_param.setData(val)

                else:
                    javaObj.getDepend().setExportAssignmentsMap({name: name})

            else:
                print('skipped {}'.format(name))

    def hasPreview(self):
        return self.javaObj.hasPreview() == 1

    def hasShow(self):
        return self.javaObj.hasShow() == 1

    def preview(self, panel):
        """
        Draw preview to panel

        Parameters
        ----------
        panel : javax.swing.JPanel

        """
        self.javaObj.preview(panel)


class SymJavaFunction(SymFunction):
    """
    Python wrapper class for |sym|'s java class ParamJavaFunction
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamJavaFunction(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamJavaFunction(name, description)
                    self.javaObj = javaObj
                elif nargs == 3:
                    name = args[0]
                    description = args[1]
                    classPath = args[2]
                    javaObj = JParamJavaFunction(name, description, classPath)
                    self.javaObj = javaObj
                elif nargs == 4:
                    name = args[0]
                    description = args[1]
                    classPath = args[2]
                    className = args[3]
                    javaObj = JParamJavaFunction(name, description, classPath, className)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    def update(self, **kwargs):
        status = self.javaObj.update(kwargs)
        return status

    def reset(self):
        self.javaObj.reset()

    def getFunctionObject(self):
        val = self.javaObj.getFunctionObject()
        return val

    def setClassName(self, className):
        self.javaObj.setClassName(className)


class SymPythonFunction(SymFunction):
    """
    Python wrapper class for |sym|'s java class ParamPythonFunction
    """

    @classmethod
    def new(cls, param: SymContainer, name: str, description: str, dirpath: str, create_test: bool = True):
        """
        This function creates a new python function and adds it to the project tree

        Parameters
        ----------

        param : SymContainer
            SyMSpace Container or Project where the function is added
        name : string
            Name of the python function
        description : string
            Function description
        dirpath: str
             Path to the directory, where the function and additional files will be saved.
        create_test : bool
            Create a test SyMSpace project
        """

        className = 'Fnc{}'.format(name)
        fncName = 'fnc_{}'.format(name.lower())

        if dirpath == '':
            dirpath = os.path.join(param.getRoot().getDirectory(), 'src')

        '''create directory if it does not exist'''
        os.makedirs(dirpath, exist_ok=True)

        res_dir = os.path.join(symspace_home, 'System', 'Python', 'resources', 'NewPythonFunction')
        template_file = os.path.join(res_dir, 'fnc_pytemplate.py')
        file = os.path.join(dirpath, fncName + '.py')

        copyFileAndReplace(template_file, file, [('PyTemplate', name)])

        fnc = cls(name, description, file, className)

        '''Insert python function'''
        param.addParam(fnc)

        if create_test:
            template_test_file = os.path.join(res_dir, 'test_pytemplate.mop')
            test_dir = os.path.join(dirpath, 'test')
            if not os.path.exists(test_dir):
                os.makedirs(test_dir)

            '''Create test SymProject'''
            test_prj = os.path.join(test_dir, 'test_' + name + '.mop')
            if os.path.isfile(test_prj):
                raise Exception('Test SymProject already exists')

            shutil.copy2(template_test_file, test_prj)
            prj = SymProject.load(test_prj)
            prj.setName('Test' + name)
            prj.setDescription('Test project for Python function {}'.format(className))
            prj.addParam(fnc.cloneParam())
            prj.save()

        return fnc

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamPythonFunction(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamPythonFunction(name, description)
                    self.javaObj = javaObj
                elif nargs == 4:
                    name = args[0]
                    description = args[1]
                    classPath = args[2]
                    className = args[3]
                    javaObj = JParamPythonFunction(name, description, classPath, className)
                    self.javaObj = javaObj
                elif nargs == 5:
                    name = args[0]
                    description = args[1]
                    classPath = args[2]
                    className = args[3]
                    moduleName = args[4]
                    javaObj = JParamPythonFunction(name, description, classPath, className, moduleName)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    def update(self, **kwargs):
        status = self.javaObj.update(kwargs)
        return status

    def reset(self):
        self.javaObj.reset()

    def getModuleName(self):
        return self.javaObj.getModuleName()

    def setModuleName(self, name):
        self.javaObj.setModuleName(name)

    def setClassName(self, className):
        self.javaObj.setClassName(className)

    def edit(self):
        self.javaObj.edit()


class SymMatlabFunction(SymFunction):
    """
    Python wrapper class for |sym|'s java class ParamMatlabFunction
    """

    def __init__(self, *args, **kwargs):
        if kwargs.get('javaObj'):
            javaObj = kwargs['javaObj']
            self.javaObj = javaObj
        else:
            if args:
                nargs = len(args)
                if nargs == 1:
                    name = args[0]
                    javaObj = JParamMatlabFunction(name)
                    self.javaObj = javaObj
                elif nargs == 2:
                    name = args[0]
                    description = args[1]
                    javaObj = JParamMatlabFunction(name, description)
                    self.javaObj = javaObj
                elif nargs == 3:
                    name = args[0]
                    description = args[1]
                    file = args[2]
                    javaObj = JParamMatlabFunction(name, description, file)
                    self.javaObj = javaObj
                elif nargs == 4:
                    name = args[0]
                    description = args[1]
                    classPath = args[2]
                    className = args[3]
                    javaObj = JParamMatlabFunction(name, description, classPath, className)
                    self.javaObj = javaObj
                else:
                    raise ValueError('wrong number of parameters')
            else:
                raise ValueError('no arguments supplied')

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val.replace('Param', '')

    def __repr__(self):
        return self.__str__()

    # methods
    def edit(self):
        self.javaObj.edit()

    def refreshDependency(self):
        self.javaObj.refreshDependency()

    def update(self, **kwargs):
        status = self.javaObj.update(kwargs)
        return status

    def reset(self):
        self.javaObj.reset()

    def setClassName(self, className):
        self.javaObj.setClassName(className)


class SymComponent:
    """
    Python wrapper class for |sym|'s java class Container
    """

    # @classmethod
    # def new(cls, container: SymContainer):
    #     obj = container.getComponent()
    #     if not isinstance(obj, SymComponent):
    #         # container.addComponentProperty()
    #         obj = cls(JComponent(container.javaObj))
    #
    #     return obj

    def __init__(self, javaObj):
        if javaObj.java_name == 'at.lcm.bu21.magopt.structure.Component':
            self.javaObj = javaObj
        else:
            raise TypeError('Wrong type of javaObj!')

    def getContainer(self) -> SymContainer:
        """
        Get parent container of component

        Returns
        -------
        SymContainer
        """
        return SymContainer(javaObj=self.javaObj.getContainer())

    def listLinkedFiles(self):
        """
        List linked-files as absolute path

        Returns
        -------
        list
        """
        return PyJDataInterface.convert_J2Py(self.javaObj.listLinkedFiles())

    def addLinkedFile(self, file: str):
        """
        Add file to linked-files

        Parameters
        ----------
        file : str
            File to add
        """
        self.javaObj.addLinkedFile(file)

    def getHelpFolder(self) -> str:
        """
        Get help-folder as absolute path

        Returns
        -------
        str
        """
        return self.javaObj.getHelpFolder().toString()

    def getFolder(self) -> str:
        """
        Get folder of component-files

        Returns
        -------
        str
        """
        return self.javaObj.getFolder().toString()

    def refresh(self):
        """
        Update project-path. Copy linked files and help-files if path is changed.
        """
        self.javaObj.refresh()

    def __str__(self):
        val = self.javaObj.toString()
        return 'Sym' + val
